//------------------------------------------------------------------------------
// AnimationMaster class.
//------------------------------------------------------------------------------

#include <Inventor/SoDB.h>
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include "AnimationMaster.h"


SO_NODE_SOURCE( AnimationMaster ) ;



//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
AnimationMaster::initClass()
{
	SO_NODE_INIT_CLASS( AnimationMaster, SoGroup, "Group" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
AnimationMaster::AnimationMaster()
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
AnimationMaster::AnimationMaster(
	int	numChildren
	) :
	SoGroup( numChildren )
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common construction steps.
//------------------------------------------------------------------------------
void
AnimationMaster::buildCommon()
{
	SO_NODE_CONSTRUCTOR( AnimationMaster ) ;
	SO_NODE_ADD_FIELD( timeIn, ( 0.0 ) ) ;
	SO_NODE_ADD_FIELD( speed, ( 1.0f ) ) ;

	timeSensor = new SoFieldSensor( timeChangedCB, (void *)this ) ;

	SoField *rt = SoDB::getGlobalField( "realTime" ) ;
	if( rt != NULL && rt->isOfType( SoSFTime::getClassTypeId() ) )
	{
		timeIn.connectFrom( rt ) ;
		currentTime = ((SoSFTime *)rt)->getValue().getValue() ;
	}
	else
	{
#ifdef	VERBOSE
		SoDebugError::post( this->getName().getString(), "`realTime' "
			"global field not found or not a SoSFTime field" ) ;
#endif	/* VERBOSE */
		currentTime = SbTime::getTimeOfDay().getValue() ;
	}
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
AnimationMaster::~AnimationMaster()
{
	timeSensor->detach() ;
	delete timeSensor ;
}



//------------------------------------------------------------------------------
// Check to see if a node is a SoSwitch node
//------------------------------------------------------------------------------
SbBool
AnimationMaster::typeOk(
	SoNode *child
	)
{
	if( child->isOfType( SoSwitch::getClassTypeId() ) )
	{
		return( TRUE ) ;
	}
	else
	{
#ifdef	VERBOSE
		SoDebugError::post( this->getName().getString(),
			"non-SoSwitch child encountered!" ) ;
#endif	/* VERBOSE */
		return( FALSE ) ;
	}
}



//------------------------------------------------------------------------------
// Check to attach the timer sensor (and update the time if necessary).
//------------------------------------------------------------------------------
void
AnimationMaster::checkSensor()
{
	if( timeSensor->getAttachedField() == NULL )
	{
		timeSensor->attach( &timeIn ) ;
		currentTime = (double)timeIn.getValue().getValue() ;
	}
}



//------------------------------------------------------------------------------
// Add a child if the type is okay.
//------------------------------------------------------------------------------
void
AnimationMaster::addChild(
	SoNode *child
	)
{
	if( typeOk( child ) )
	{
		checkSensor() ;
		childBirthTime.append( currentTime ) ;
		SoGroup::addChild( child ) ;
		( (SoSwitch *)child )->whichChild = 0 ;
	}
}



//------------------------------------------------------------------------------
// Insert a child if the type is okay.
//------------------------------------------------------------------------------
void
AnimationMaster::insertChild(
	SoNode	*child,
	int	newChildIndex
	)
{
	if( typeOk( child ) )
	{
		checkSensor() ;
		childBirthTime.insert( currentTime, newChildIndex ) ;
		SoGroup::insertChild( child, newChildIndex ) ;
	}
}



//------------------------------------------------------------------------------
// Replace a child if the type is okay.
//------------------------------------------------------------------------------
void
AnimationMaster::replaceChild(
	int	index,
	SoNode	*child
	)
{
	if( typeOk( child ) )
	{
		checkSensor() ;
		childBirthTime.remove( index ) ;
		childBirthTime.insert( currentTime, index ) ;
		SoGroup::replaceChild( index, child ) ;
	}
}



//------------------------------------------------------------------------------
// Remove a child.
//------------------------------------------------------------------------------
void
AnimationMaster::removeChild(
	int	index
	)
{
	if( index >= 0 && index < getNumChildren() )
	{
		if( getNumChildren() == 1 )
		{
			timeSensor->detach() ;
		}
		childBirthTime.remove( index ) ;
		SoGroup::removeChild( index ) ;
	}
}



//------------------------------------------------------------------------------
// Static callback for when the timeIn field changes.
//------------------------------------------------------------------------------
void
AnimationMaster::timeChangedCB(
	void		*data,
	SoSensor	*
	)
{
	( (AnimationMaster *)data )->updateChildren() ;
}



//------------------------------------------------------------------------------
// Update all of the children.
//------------------------------------------------------------------------------
void
AnimationMaster::updateChildren()
{
	currentTime = (double)timeIn.getValue().getValue() ;
	float speedFactor = speed.getValue() ;
	if( speedFactor <= 0.0f )
	{
		speedFactor = 0.0f ;
	}

	int i0 = getNumChildren() - 1 ;
	if( i0 < 0 )
	{
		timeSensor->detach() ;
		return ;
	}

	for( int i = i0 ; i >= 0 ; i-- )
	{
		float age = (float)( currentTime - childBirthTime[i] ) *
				speedFactor ;
		SoSwitch *sw = (SoSwitch *)getChild( i ) ;
		int n = (int)((float)sw->getNumChildren() * age ) ;
		if( n >= sw->getNumChildren() )
		{
			removeChild( i ) ;
		}
		else
		{
			sw->whichChild = n ;
		}
	}
}



//------------------------------------------------------------------------------
// Read in the node.
//------------------------------------------------------------------------------
SbBool
AnimationMaster::readInstance(
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoInput		*in,
	unsigned short
#else
	SoInput	*in
#endif
	)
{

	timeSensor->detach() ;

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SbBool status = SoGroup::readInstance( in, IS_GROUP ) ;
#else
	SbBool status = SoGroup::readInstance( in ) ;
#endif

	// updateChildren() ;

	// Reattach the sensor.
	if( getNumChildren() > 0 )
	{
		timeSensor->attach( &timeIn ) ;
	}

	return( status ) ;
}



//------------------------------------------------------------------------------
// Read in the children.
//------------------------------------------------------------------------------
SbBool
AnimationMaster::readChildren(
	SoInput *in
	)
{
	// Code straight from SoGroup.
	SoBase	*base ;
	SbBool	ret = TRUE ;

	// If reading binary, read number of children first
	if( in->isBinary() )
	{
		int numToRead, i ;

		if( !in->read( numToRead ) )
		{
			ret = FALSE;
		}
		else
		{
			for( i = 0 ; i < numToRead ; i++ )
			{
				if( SoBase::read( in, base,
					SoNode::getClassTypeId() ) &&
					base != NULL)
				{
					addChild( (SoNode *)base ) ;
				}
				// Running out of children is now an error,
				// since the number of children in the file
				// must be exact
				else
				{
					ret = FALSE ;
					break ;
				}
			}
			// If we are reading a 1.0 file, read the
			// GROUP_END_MARKER
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
			if( ret && in->getIVVersion() == 1.0 )
#else
			if( ret && in->getVersion() == 1.0 )
#endif
			{
				const int GROUP_END_MARKER = -1;
				int marker ;

				// Read end marker if it is there. If not,
				// some sort of error occurred.
				if( !in->read( marker ) ||
					marker != GROUP_END_MARKER )
				{
					ret = FALSE ;
				}
			}
		}
	}
	// ASCII: Read children until none left. Deal with children
	// causing errors by adding them as is.
	else
	{
		while( TRUE )
		{
			ret = SoBase::read( in, base, SoNode::getClassTypeId() )
				&& ret ;

			// Add child, even if error occurred, unless there
			// is no child to add.
			if( base != NULL )
			{
				addChild( (SoNode *)base ) ;
			}
			// Stop when we run out of valid children
			else
			{
				break ;
			}
		}
	}

	return ret ;
}



//------------------------------------------------------------------------------
// DoubleList helper class functions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Append a double value onto the list.
//------------------------------------------------------------------------------
void
DoubleList::append(
	double	value
	)
{
	double *dp = new double ;
	*dp = value ;
	SbPList::append( (void *)dp ) ;
}



//------------------------------------------------------------------------------
// Insert a double value into the list.
//------------------------------------------------------------------------------
void
DoubleList::insert(
	double	value,
	int	addBefore
	)
{
	double *dp = new double ;
	*dp = value ;
	SbPList::insert( (void *)dp, addBefore ) ;
}



//------------------------------------------------------------------------------
// Remove (and free) an entry.
//------------------------------------------------------------------------------
void
DoubleList::remove(
	int	which
	)
{
	double *dp = (double *)SbPList::operator[]( which ) ;

	if( dp != NULL )
	{
		delete dp ;
	}
	SbPList::remove( which ) ;
}



//------------------------------------------------------------------------------
// Remove all entries from the given index.
//------------------------------------------------------------------------------
void
DoubleList::truncate(
	int	start
	)
{
	double	*dp ;

	for( int i = start ; i < getLength() ; i++ )
	{
		dp = (double *)SbPList::operator[]( i ) ;

		if( dp != NULL )
		{
			delete dp ;
		}
	}

	SbPList::truncate( start ) ;
}
