//------------------------------------------------------------------------------
// Abstract base class for computer-controlled enemies in Pointblank.
//------------------------------------------------------------------------------

#include "GuidedMissile.h"
#include "Utilities.h"
#include "Vehicle.h"
#include "Automaton.h"


SO_NODE_ABSTRACT_SOURCE( Automaton ) ;


AutomatonStrikeHandler *Automaton::strikeHandler =
				Automaton::defaultStrikeHandler ;
void *Automaton::strikeData = NULL ;
AutomatonLaunchHandler *Automaton::launchHandler =
				Automaton::defaultLaunchHandler ;
void *Automaton::launchData = NULL ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Automaton::initClass()
{
	if( Obstacle::getClassTypeId() == SoType::badType() )
	{
		Obstacle::initClass() ;
	}
	SO_NODE_INIT_ABSTRACT_CLASS( Automaton, Obstacle, "Obstacle" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Automaton::Automaton()
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate Constructor.
//------------------------------------------------------------------------------
Automaton::Automaton(
	const char	*name
	) :
	Obstacle( name )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
Automaton::createIt()
{
	SO_NODE_CONSTRUCTOR( Automaton ) ;

	SO_NODE_ADD_FIELD( skill, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( armor, ( 1.0f ) ) ;
	SO_NODE_ADD_FIELD( value, ( 10 ) ) ;
	SO_NODE_ADD_FIELD( rangeOfInterest, ( 20.0f ) ) ;
	SO_NODE_ADD_FIELD( alive, ( TRUE ) ) ;

	damage = 0.0f ;
	missileList.truncate( 0 ) ;

	status = ALIVE ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Automaton::~Automaton()
{
}



//------------------------------------------------------------------------------ // Store the missile strike handler function.
//------------------------------------------------------------------------------
void
Automaton::setStrikeHandler(
	AutomatonStrikeHandler	*func,
	void			*userData
	)
{
	strikeHandler = func ;
	strikeData = userData ;
}



//------------------------------------------------------------------------------ // Store the missile launch handler function.
//------------------------------------------------------------------------------
void
Automaton::setLaunchHandler(
	AutomatonLaunchHandler	*func,
	void			*userData
	)
{
	launchHandler = func ;
	launchData = userData ;
}



//------------------------------------------------------------------------------
// Default handler for automaton's missile strikes.
//------------------------------------------------------------------------------
void
Automaton::defaultStrikeHandler(
	Missile *,
	SoNode *,
	void *
	)
{
#ifdef VERBOSE
	printf( "Need to set AutomatonStrikeHandler!!!!\n" ) ;
#endif // VERBOSE
}



//------------------------------------------------------------------------------
// Default handler for automaton's missile launches.
//------------------------------------------------------------------------------
Missile *
Automaton::defaultLaunchHandler(
	SoType,
	void *
	)
{
#ifdef VERBOSE
	printf( "Need to set AutomatonLaunchHandler!!!!\n" ) ;
#endif // VERBOSE

	return( new Missile ) ;
}



//------------------------------------------------------------------------------
// Suffer damage.
//------------------------------------------------------------------------------
int32_t
Automaton::sufferDamage(
	float	amount
	)
{
	int	pts = 0 ;

	damage += amount ;
	if( damage > armor.getValue() )
	{
		status = DYING ;
		die() ;
		pts = value.getValue() ;
	}

	return( pts ) ;
}



//------------------------------------------------------------------------------
// Update the object and all of its missiles.
//------------------------------------------------------------------------------
void
Automaton::update(
	float	deltaTime,
	SoNode	*scene,
	Vehicle	*player
	)
{
	if( status == ALIVE && player )
	{
		// Check if vehicle within range.
		SbVec3f &range = player->position.getValue() -
					position.getValue() ;
		if( rangeOfInterest.getValue() * rangeOfInterest.getValue() >=
			range.dot( range ) )
		{
			updateSelf( deltaTime, scene, player ) ;
		}
	}

	// Update all the missiles.
	if( status != DEAD )
	{
		for( int m = missileList.getLength() - 1 ; m >= 0 ; m-- )
		{
			Missile *mis = (Missile *)missileList[m] ;
			Missile::MissileStatus result ;

			if( mis->isOfType( GuidedMissile::getClassTypeId() ) &&
				player )
			{
				result = guideMissile( mis, deltaTime, scene,
							player ) ;
			}
			else
			{
				result = mis->update( 0.0f, 0.0f, deltaTime,
							scene ) ;
			}

			if( result != Missile::ACTIVE )
			{
				// Check to see if missile struck the player.
				const SoPath *obs = mis->getObstructor() ;
				if( result == Missile::STRUCK && obs != NULL )
				{
					strikeHandler( mis, obs->getTail(),
						strikeData ) ;
				}
				else
				{
					strikeHandler( mis, NULL, strikeData ) ;
				}

				// Remove from the list.
				missileList.remove( m ) ;
			}
		}

		if( status == DYING && missileList.getLength() == 0 )
		{
			status = DEAD ;
			alive.setValue( FALSE ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Steer a guided missile.
//------------------------------------------------------------------------------
Missile::MissileStatus
Automaton::guideMissile(
	Missile	*mis,
	float	deltaTime,
	SoNode	*scene,
	Vehicle	*target
	)
{
	float mx = 0.0f ;
	float my = 0.0f ;

	SbVec3f &delta = target->position.getValue() - mis->position.getValue();

	float t = skill.getValue() * delta.length() / mis->getMaxSpeed() ;
	SbVec3f tVel = target->getVelocity() ;
	tVel[2] = 0.0f ;		// Ignore gravity pull on vehicle.
	delta += t * tVel ;

	SbVec3f ahead ;
	const SbRotation &rot = mis->rotation.getValue() ;
	Utilities::setForwardVector( rot, ahead ) ;
	float da = delta.dot( ahead ) ;

	if( da > 0.0f )
	{
		SbVec3f up ;
		SbVec3f right ;
		Utilities::setUpVector( rot, up ) ;
		Utilities::setRightVector( rot, right ) ;
		mx = delta.dot( right ) ;
		my = delta.dot( up ) ;
		if( mx > 1.0f ) mx = 1.0f ;
		else if( mx < -1.0f ) mx = -1.0f ;
		if( my > 1.0f ) my = 1.0f ;
		else if( my < -1.0f ) my = -1.0f ;
	}

	return( mis->update( mx, my, deltaTime, scene ) ) ;
}



