//------------------------------------------------------------------------------
// Non-mobile automaton.
//------------------------------------------------------------------------------

#include <math.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include "Bomb.h"
#include "CollisionAction.h"
#include "CruiseMissile.h"
#include "ExplosionControl.h"
#include "GuidedMissile.h"
#include "LosAction.h"
#include "Mortar.h"
#include "Utilities.h"
#include "Vehicle.h"
#include "Battery.h"

SO_NODE_SOURCE( Battery ) ;


static const float turnRate = M_PI ;
static const float twoPi = 2.0f * (float)M_PI ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Battery::initClass()
{
	if( Automaton::getClassTypeId() == SoType::badType() )
	{
		Automaton::initClass() ;
	}
	SO_NODE_INIT_CLASS( Battery, Automaton, "Automaton" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Battery::Battery()
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
Battery::Battery(
	const char *name
	) :
	Automaton( name )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
Battery::createIt()
{
	SO_NODE_CONSTRUCTOR( Battery ) ;

	SO_NODE_ADD_FIELD( minAzim, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( maxAzim, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( minIncl, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( maxIncl, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( typeOfMissile, ( "Missile" ) ) ;
	SO_NODE_ADD_FIELD( reloadTime, ( 10.0f ) ) ;

	timeToFire = 0.0f ;
	missileSpeed = 1.0f / 10.0f ;

	incl = minIncl.getValue() ;
	azim = minAzim.getValue() ;

	minAzimSensor = new SoFieldSensor( checkAnglesCB, this ) ;
	minAzimSensor->setPriority( 0 ) ;
	maxAzimSensor = new SoFieldSensor( checkAnglesCB, this ) ;
	maxAzimSensor->setPriority( 0 ) ;
	minInclSensor = new SoFieldSensor( checkAnglesCB, this ) ;
	minInclSensor->setPriority( 0 ) ;
	maxInclSensor = new SoFieldSensor( checkAnglesCB, this ) ;
	maxInclSensor->setPriority( 0 ) ;

	enableSensors() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Battery::~Battery()
{
	if( view )
	{
		view->unref() ;
	}

	delete minAzimSensor ;
	delete maxAzimSensor ;
	delete minInclSensor ;
	delete maxInclSensor ;
}




//------------------------------------------------------------------------------
// Process the geometry after it has been read in.
//------------------------------------------------------------------------------
void
Battery::processGeometry()
{
	if( root->getNumChildren() != 1 )
	{
		SoDebugError::post( "Battery", "geometry file, %s, results in"
			"wrong number of children (%d instead of 1)\nfor the"
			"object's geometry root", name.getValue().getString(),
			root->getNumChildren() ) ;
		exit( -1 ) ;
	}

	SoNode *newCopy = root->getChild( 0 )->copy( TRUE ) ;
	root->replaceChild( 0, newCopy ) ;

	SoSearchAction sa ;
	sa.setName( "StatusSwitch" ) ;
	sa.setInterest( SoSearchAction::FIRST ) ;
	sa.setSearchingAll( FALSE ) ;

	sa.apply( root ) ;

	SoPath *path = sa.getPath() ;

	//
	// If geometry does not contain a switch named "StatusSwitch" for
	// displaying a live and a destroyed representation of the battery,
	// then put the root under a switch that can be turned off.
	//
	if( path == NULL ||
		!path->getTail()->isOfType( SoSwitch::getClassTypeId() ) )
	{
#ifdef VERBOSE
		printf( "Battery file %s does not contain a StatusSwitch "
			"node\n", name.getValue().getString() ) ;
#endif // VERBOSE
		view = new SoSwitch ;
		view->ref() ;
		view->addChild( root->getChild( 0 ) ) ;
		root->removeAllChildren() ;
		root->addChild( view ) ;
	}
	else
	{
		view = (SoSwitch *)path->getTail() ;
		view->ref() ;
	}

	//
	// Now search for the inclination rotation node.
	//
	sa.setName( "TurretInclRotation" ) ;
	sa.apply( root ) ;
	path = sa.getPath() ;
	if( path == NULL ||
		!path->getTail()->isOfType( SoRotation::getClassTypeId() ) )
	{
#ifdef VERBOSE
		printf( "Battery file %s does not contain a TurretInclRotation "
			"node\n", name.getValue().getString() ) ;
#endif // VERBOSE
		inclRot = new SoRotation ;
		inclRot->ref() ;
		minIncl.setValue( 0.0f ) ;
		maxIncl.setValue( 0.0f ) ;
	}
	else
	{
		inclRot = (SoRotation *)path->getTail() ;
		inclRot->ref() ;
	}

	//
	// Now search for the azimuth rotation node.
	//
	sa.setName( "TurretAzimRotation" ) ;
	sa.apply( root ) ;
	path = sa.getPath() ;
	if( path == NULL ||
		!path->getTail()->isOfType( SoRotation::getClassTypeId() ) )
	{
#ifdef VERBOSE
		printf( "Battery file %s does not contain a TurretAzimRotation "
			"node\n", name.getValue().getString() ) ;
#endif // VERBOSE
		azimRot = new SoRotation ;
		azimRot->ref() ;
		minAzim.setValue( 0.0f ) ;
		maxAzim.setValue( 0.0f ) ;
	}
	else
	{
		azimRot = (SoRotation *)path->getTail() ;
		azimRot->ref() ;
	}

	//
	// Now find the launch point.
	//
	sa.setName( "LaunchPoint" ) ;
	sa.setSearchingAll( TRUE ) ;
	sa.apply( root ) ;
	path = sa.getPath() ;
	if( path == NULL ||
		!path->getTail()->isOfType( SoCoordinate3::getClassTypeId() ) )
	{
#ifdef VERBOSE
		printf( "Battery file %s does not contain a LaunchPoint "
			"node\n", name.getValue().getString() ) ;
#endif // VERBOSE
		launchPoint.setValue( 0.0f, 0.0f, 0.0f ) ;
	}
	else
	{
		SoCoordinate3 *lp = (SoCoordinate3 *)path->getTail() ;
		if( lp->point.getNum() > 0 )
		{
			launchPoint = lp->point[0] ;
		}
		else
		{
			launchPoint.setValue( 0.0f, 0.0f, 0.0f ) ;
		}
	}

	incl = minIncl.getValue() ;
	azim = minAzim.getValue() ;

	setRotations() ;
}



//------------------------------------------------------------------------------
// Update the automaton.
//------------------------------------------------------------------------------
void
Battery::updateSelf(
	float	deltaTime,
	SoNode	* ,// scene,
	Vehicle	*player
	)
{
	timeToFire += deltaTime ;

	// Just fire like crazy for now.
	if( aim( deltaTime, player ) && timeToFire >= reloadTime.getValue() )
	{
		fire() ;
	}
}



//------------------------------------------------------------------------------
// Fire a missile
//------------------------------------------------------------------------------
void
Battery::fire()
{
	Missile *missile = launchMissile( missileType ) ;

	missile->setHeight( 0.2f ) ;
	missile->setOwner( this ) ;
	missileList.append( missile ) ;

	SbVec3f lp = launchPoint + getCenterOffset() ;
	SbVec3f pos = position.getValue() + lp[0] * aimRight +
			lp[1] * aimAhead + lp[2] * aimUp ;

	missile->position.setValue( pos ) ;
	missile->rotation.setValue( aimAngle ) ;
	missile->setVelocity( SbVec3f( 0.0f, 0.0f, 0.0f ) ) ;
	timeToFire = 0.0f ;
}



//------------------------------------------------------------------------------
// Check for collisions.
//------------------------------------------------------------------------------
void
Battery::collide(
	CollisionAction *action
	)
{
	if( status != ALIVE )
	{
		return ;
	}

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

 	action->checkCollision( boundingBox, rotation.getValue(),
		position.getValue(), collisionRadius ) ;
}



//------------------------------------------------------------------------------
// LosAction handler.
//------------------------------------------------------------------------------
void
Battery::obstruct(
	LosAction *action
	)
{
	if( status != ALIVE ||
		action->isSelf( this ) || action->isTarget( this ) )
	{
		return ;
	}

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	//
	// Check if line intersects bounding box.
	//
	action->checkBoundingBox( boundingBox, rotation.getValue(),
					position.getValue(), collisionRadius ) ;
}



//------------------------------------------------------------------------------
// Aim for the player, returning true if close enough to fire.
//------------------------------------------------------------------------------
SbBool
Battery::aim(
	float	deltaTime,
	Vehicle	*target
	)
{
	SbBool	fireOk ;
	float	minVal = 0.1f + ( 1.0f - skill.getValue() ) * 0.2f ;

	SbVec3f lp = launchPoint + getCenterOffset() ;
	SbVec3f pos = position.getValue() + lp[0] * aimRight +
			lp[1] * aimAhead + lp[2] * aimUp ;
	SbVec3f &delta = target->position.getValue() - pos ;

	//
	// Add lead.
	//
	float t = skill.getValue() * delta.length() * missileSpeed ;
	SbVec3f tVel = target->getVelocity() ;
	tVel[2] = 0.0f ;		// Ignore gravity pull on vehicle.
	delta += t * tVel ;

	float dr = delta.dot( aimRight ) ;
	float du = delta.dot( aimUp ) ;
	float da = delta.dot( aimAhead ) ;

	switch( tracking )
	{
		default :
			adjustAzim( deltaTime, da, dr ) ;
			adjustIncl( deltaTime, da, du ) ;
			setRotations() ;
			dr = delta.dot( aimRight ) ;
			du = delta.dot( aimUp ) ;
			fireOk = -minVal <= dr && dr <= minVal &&
				 -minVal <= du && du <= minVal &&
				 checkLimits() ;
			break ;

		//
		// GUIDE has a looser aim
		//
		case GUIDE :
			adjustAzim( deltaTime, da, dr ) ;
			adjustIncl( deltaTime, da, du ) ;
			setRotations() ;
			da = delta.dot( aimAhead ) ;
			fireOk = da > 0.0f && checkLimits() ;
			break ;

		//
		// AZIM_ONLY, so don't adjust the inclination
		//
		case AZIM_ONLY :
			adjustAzim( deltaTime, da, dr ) ;
			incl = minIncl.getValue() ;
			setRotations() ;
			dr = delta.dot( aimRight ) ;
			fireOk = -minVal <= dr && dr <= minVal &&
				 checkLimits() ;
			break ;

		//
		// LOB gets tricky
		//
		case LOB :
			float z = delta[2] ;
			float x = sqrtf( delta[0] * delta[0] +
					delta[1] * delta[1] ) ;
			//
			// Theta is the true inclination angle and is
			// equal to asinf( aimAhead[2] )
			//
			aimAhead.normalize() ;
			float sinT = aimAhead[2] ;
			float cosT = sqrtf( 1.0f - sinT * sinT ) ;

			if( cosT == 0.0f )
			{
				fireOk = FALSE ;
				adjustIncl( deltaTime, da, du ) ;
				adjustAzim( deltaTime, da, dr ) ;
				setRotations() ;
				break ;
			}

			//
			// Knowing the horizontal distance, x, to the object,
			// calculate the value of z when the mortar reaches
			// x in terms of the current elevation angle.
			//
			float V = Mortar::muzzleVelocity ;
			float A = -0.125f * Constants::Gravity ;
			float f = 1.0f / ( V * V * cosT ) ;
			float zCur = x * ( sinT + 0.5f * A * x * f ) / cosT ;

			//
			// Now calculate the change in Z for change in the
			// elevation angle.
			//
			float dzda = ( 1.0f + A * x * sinT * f ) * x /
					( cosT * cosT ) ;

			float deltaZ = z - zCur ;
			if( dzda != 0.0f )
			{
				float dIncl = deltaZ / dzda ;
				float maxDi = turnRate * deltaTime *
						( skill.getValue() + 0.05f ) ;
				if( dIncl > maxDi )
				{
					incl += maxDi ;
				}
				else if( dIncl < -maxDi )
				{
					incl -= maxDi ;
				}
				else
				{
					incl += dIncl ;
				}
				if( incl > maxIncl.getValue() )
				{
					incl = maxIncl.getValue() ;
				}
				else if( incl < minIncl.getValue() )
				{
					incl = minIncl.getValue() ;
				}
			}

			// Adjust the azimuth.
			SbVec3f right( aimAhead[1], -aimAhead[0], 0.0f ) ;
			right.normalize() ;
			dr = delta.dot( right ) ;
			adjustAzim( deltaTime, da, dr ) ;

			setRotations() ;
			right.setValue( aimAhead[1], -aimAhead[0], 0.0f ) ;
			right.normalize() ;
			dr = delta.dot( right ) ;

			float minZ = .5f + ( 1.0f - skill.getValue() ) * 1.5f ;
			fireOk = -minVal <= dr && dr <= minVal &&
				 -minZ <= deltaZ && deltaZ <= minZ &&
				 checkLimits() ;
			break ;
	}


	//
	// Now check the aim.
	//
	return( fireOk ) ;
}



//------------------------------------------------------------------------------
// Set the rotations.
//------------------------------------------------------------------------------
void
Battery::setRotations()
{
	inclRot->rotation.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ), incl ) ;
	azimRot->rotation.setValue( SbVec3f( 0.0f, 0.0f, 1.0f ), azim ) ;
	aimAngle = inclRot->rotation.getValue() ;
	aimAngle *= azimRot->rotation.getValue() ;
	aimAngle *= rotation.getValue() ;

	Utilities::setUpVector( aimAngle, aimUp ) ;
	Utilities::setForwardVector( aimAngle, aimAhead ) ;
	Utilities::setRightVector( aimAngle, aimRight ) ;
}



//------------------------------------------------------------------------------
// Die.
//------------------------------------------------------------------------------
void
Battery::die()
{
	if( view->whichChild.getValue() == 0 )
	{
		if( view->getNumChildren() > 1 )
		{
			view->whichChild = 1 ;
		}
		else
		{
			view->whichChild = -1 ;
		}
	}
}



//------------------------------------------------------------------------------
// Reset for new level.
//------------------------------------------------------------------------------
void
Battery::reset()
{
	if( status == ALIVE )
	{
		timeToFire = 0.0f ;
	}
}



//------------------------------------------------------------------------------
// Adjust the azimuth.
//------------------------------------------------------------------------------
void
Battery::adjustAzim(
	float	deltaTime,
	float	da,
	float	dr
	)
{
	//
	// Adjust the azimuth.
	//
	float dAzim = turnRate * deltaTime * ( skill.getValue() + 0.05f ) ;
	if( dr < 0.0f )
	{
		if( dAzim * da > -dr )
		{
			azim += -dr / da ;
		}
		else
		{
			azim += dAzim ;
		}
	}
	else if( dr > 0.0f )
	{
		if( dAzim * da > dr )
		{
			azim += dr / da ;
		}
		else
		{
			azim -= dAzim ;
		}
	}
	azim = fmodf( azim + twoPi, twoPi ) ;
}



//------------------------------------------------------------------------------
// Adjust the inclination.
//------------------------------------------------------------------------------
void
Battery::adjustIncl(
	float	deltaTime,
	float	da,
	float	du
	)
{
	float dIncl = turnRate * deltaTime *
			( skill.getValue() + 0.05f ) ;
	if( du > 0.0f )
	{
		if( dIncl * da > du )
		{
			incl += du / da ;
		}
		else
		{
			incl += dIncl ;
		}
		if( incl > maxIncl.getValue() )
		{
			incl = maxIncl.getValue() ;
		}
	}
	else if( du < 0.0f )
	{
		if( dIncl * da > -du )
		{
			incl += du / da ;
		}
		else
		{
			incl -= dIncl ;
		}
		if( incl < minIncl.getValue() )
		{
			incl = minIncl.getValue() ;
		}
	}
}




//------------------------------------------------------------------------------
// Check the missile type after a read.
//------------------------------------------------------------------------------
SbBool
Battery::postRead(
	SbBool status
	)
{
	status = Automaton::postRead( status ) ;

	if( status )
	{
		missileType = SoType::fromName( typeOfMissile.getValue() ) ;
		if( missileType == SoType::badType() )
		{
#ifdef VERBOSE
			SoDebugError::post(
				getClassTypeId().getName().getString(),
				"Bad 'typeOfMissile' value (%s)",
				typeOfMissile.getValue().getString() ) ;
#endif // VERBOSE
			missileType = Missile::getClassTypeId() ;
		}

		Missile *testMissile = (Missile *)missileType.createInstance() ;
		if( testMissile == NULL )
		{
#ifdef VERBOSE
			SoDebugError::post(
				getClassTypeId().getName().getString(),
				"Difficulty instancing %s (using Missile "
				"instead)",
				typeOfMissile.getValue().getString() ) ;
#endif // VERBOSE
			missileType = Missile::getClassTypeId() ;
			testMissile = new Missile ;
		}
		testMissile->ref() ;
		missileSpeed = 1.0f / testMissile->getMaxSpeed() ;
		testMissile->unref() ;

		if( missileType.isDerivedFrom( Bomb::getClassTypeId() ) )
		{
			tracking = LOB ;
		}
		else if( missileType.isDerivedFrom(
			CruiseMissile::getClassTypeId() ) )
		{
			tracking = AZIM_ONLY ;
		}
		else if( missileType.isDerivedFrom(
			GuidedMissile::getClassTypeId() ) )
		{
			tracking = GUIDE ;
		}
		else
		{
			tracking = DEAD_ON ;
		}

		incl = minIncl.getValue() ;
		azim = minAzim.getValue() ;
	}

	checkAngles() ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Enable the field sensors.
//------------------------------------------------------------------------------
void
Battery::enableSensors()
{
	Automaton::enableSensors() ;

	minAzimSensor->attach( &minAzim ) ;
	maxAzimSensor->attach( &maxAzim ) ;
	minInclSensor->attach( &minIncl ) ;
	maxInclSensor->attach( &maxIncl ) ;
}



//------------------------------------------------------------------------------
// Disable the field sensors.
//------------------------------------------------------------------------------
void
Battery::disableSensors()
{
	Automaton::disableSensors() ;

	minAzimSensor->detach() ;
	maxAzimSensor->detach() ;
	minInclSensor->detach() ;
	maxInclSensor->detach() ;
}



//------------------------------------------------------------------------------
// Callback for when angle limits change.
//------------------------------------------------------------------------------
void
Battery::checkAnglesCB(
	void		*userData,
	SoSensor	*
	)
{
	( (Battery *)userData )->checkAngles() ;
}



//------------------------------------------------------------------------------
// Check new angle limits.
//------------------------------------------------------------------------------
void
Battery::checkAngles()
{
	maxAzimuth = maxAzim.getValue() - minAzim.getValue() ;
	if( maxAzimuth > twoPi )
	{
		maxAzimuth = twoPi ;
	}
	else
	{
		maxAzimuth = fmodf( maxAzimuth, twoPi ) ;
	}
	maxInclination = fmodf( maxIncl.getValue() - minIncl.getValue(), twoPi);
	azim = minAzim.getValue() ;
	incl = minIncl.getValue() ;
	setRotations() ;
}



//------------------------------------------------------------------------------
// Check if aim is within limits.
//------------------------------------------------------------------------------
SbBool
Battery::checkLimits()
{
	float a = fmodf( azim - minAzim.getValue(), twoPi ) ;
	float i = fmodf( incl - minIncl.getValue(), twoPi ) ;

	return( 0.0f <= a && a <= maxAzimuth &&
		0.0f <= i && i <= maxInclination ) ;
}
