//------------------------------------------------------------------------------
// Class for bombs.
//------------------------------------------------------------------------------

#include <Inventor/SbTime.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoSeparator.h>

#include "LosAction.h"
#include "Bomb.h"
#include "Utilities.h"
#include "Constants.h"

#include "BombGeom.h"

SO_NODE_SOURCE( Bomb ) ;

SoSeparator		*Bomb::geometry = NULL ;
SoSeparator		*Bomb::icon = NULL ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Bomb::initClass()
{
	if( Missile::getClassTypeId() == SoType::badType() )
	{
		Missile::initClass() ;
	}
	SO_NODE_INIT_CLASS( Bomb, Missile, "Missile" ) ;

	geometry = new SoSeparator ;
	geometry->ref() ;

	SoComplexity *comp = new SoComplexity ;
	geometry->addChild( comp ) ;
	comp->type = SoComplexity::OBJECT_SPACE ;
	comp->value = 0.2f ;

	SoCylinder *cone = new SoCylinder ;
	geometry->addChild( cone ) ;
	cone->height = Constants::MissileLength ;
	cone->radius = Constants::MissileRadius ;

	icon = Utilities::readAll( "icons/bomb.iv", FALSE ) ;
	if( icon == NULL )
	{
		SoInput input ;
		input.setBuffer( (void *)defaultGeom, sizeof( defaultGeom ) ) ;

		icon = SoDB::readAll( &input ) ;

		if( icon == NULL )
		{
			SoDebugError::post( "Bomb", "can not create icon "
				"geometry" ) ;
			exit( 1 ) ;
		}
	}
	icon->ref() ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Bomb::Bomb()
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Alternate constructor specifying the desired height.
//------------------------------------------------------------------------------
Bomb::Bomb(
	float	height
	) :
	Missile( height )
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common code for constructors.
//------------------------------------------------------------------------------
void
Bomb::buildCommon()
{
	SO_NODE_CONSTRUCTOR( Bomb ) ;

	children.truncate( 0 ) ;
	children.append( geometry ) ;

	los->setTarget( NULL ) ;

	velocity.setValue( 0.0f, 0.0f, 0.0f ) ;

	damageAmount = 10.0f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Bomb::~Bomb()
{
}



//------------------------------------------------------------------------------
// Move the missile and check for collisions.
//------------------------------------------------------------------------------
Missile::MissileStatus
Bomb::update(
	float	,
	float	,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( age > maxAge )
	{
		status = Missile::EXPIRED ;
		blowUp() ;
		return status ;
	}

	SbRotation	rot = rotation.getValue() ;

	// -Z vector
	SbVec3f ahead ;
	Utilities::setForwardVector( rot, ahead ) ;

	age += deltaTime ;

	SbVec3f	startPos( position.getValue() ) ;

	SbVec3f	&forward = velocity * deltaTime ;
	velocity[2] -= 0.1f * Constants::Gravity * deltaTime ;

	SbVec3f &curPos = startPos + forward ;

	los->setRay( startPos, curPos ) ;
	los->apply( scene ) ;

	if( los->isObstructed() )
	{
		blowUp() ;
		curPos = los->getEndPoint() ;
		status = Bomb::STRUCK ;
	}

	position.setValue( curPos ) ;
	rotation.setValue( rot ) ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Return whether or not the missile is a guided type.
//------------------------------------------------------------------------------
SbBool
Bomb::isGuided() const
{
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Return the missile type (this is a virtual method wrapper around the static
// GetClassTypeId).
//------------------------------------------------------------------------------
Missile::MissileType
Bomb::getMissileType() const
{
	return( Missile::BOMB ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the icon scene graph.
//------------------------------------------------------------------------------
SoNode *
Bomb::getIconGraph()
{
	return( Bomb::icon ) ;
}
