//------------------------------------------------------------------------------
// Class for cannons
//------------------------------------------------------------------------------

#include <Inventor/SbTime.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTranslation.h>

#include "LosAction.h"
#include "Cannon.h"
#include "Utilities.h"
#include "Constants.h"

#include "CannonGeom.h"

SO_NODE_SOURCE( Cannon ) ;

SoSeparator		*Cannon::geometry = NULL ;
SoSeparator		*Cannon::icon = NULL ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Cannon::initClass()
{
	if( Missile::getClassTypeId() == SoType::badType() )
	{
		Missile::initClass() ;
	}
	SO_NODE_INIT_CLASS( Cannon, Missile, "Missile" ) ;

	geometry = new SoSeparator ;
	geometry->ref() ;

	SoLightModel *lm = new SoLightModel ;
	geometry->addChild( lm ) ;
	lm->model = SoLightModel::BASE_COLOR ;

	SoBaseColor *bc = new SoBaseColor ;
	geometry->addChild( bc ) ;
	bc->rgb.setValue( 1.0f, 0.0f, 0.0f ) ;

	SoComplexity *comp = new SoComplexity ;
	geometry->addChild( comp ) ;
	comp->value = 0.0f ;

	SoTranslation *t = new SoTranslation ;
	geometry->addChild( t ) ;
	t->translation.setValue( 0.0f, -5.0f * Constants::MissileLength, 0.0f );

	SoCylinder *cyl = new SoCylinder ;
	geometry->addChild( cyl ) ;
	cyl->height = 10.0f * Constants::MissileLength ;
	cyl->radius = 0.1f * Constants::MissileRadius ;

	icon = Utilities::readAll( "icons/cannon.iv", FALSE ) ;
	if( icon == NULL )
	{
		SoInput input ;
		input.setBuffer( (void *)defaultGeom, sizeof( defaultGeom ) ) ;

		icon = SoDB::readAll( &input ) ;

		if( icon == NULL )
		{
			SoDebugError::post( "Cannon", "can not create icon "
				"geometry" ) ;
			exit( 1 ) ;
		}
	}
	icon->ref() ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Cannon::Cannon()
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Alternate constructor specifying the desired height.
//------------------------------------------------------------------------------
Cannon::Cannon(
	float	height
	) :
	Missile( height )
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common code for constructors.
//------------------------------------------------------------------------------
void
Cannon::buildCommon()
{
	SO_NODE_CONSTRUCTOR( Cannon ) ;

	status = Missile::ACTIVE ;

	children.truncate( 0 ) ;
	children.append( geometry ) ;

	maxSpeed = 50.0f ;
	maxAge = 1.0f ;

	buildTime = 0.0f ;

	damageAmount = 0.101f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Cannon::~Cannon()
{
}



//------------------------------------------------------------------------------
// Move the missile and check for collisions.
//------------------------------------------------------------------------------
Missile::MissileStatus
Cannon::update(
	float	,
	float	,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( age > maxAge )
	{
		status = Missile::EXPIRED ;
		blowUp() ;
		return status ;
	}

	SbRotation	rot = rotation.getValue() ;

	// -Z vector
	SbVec3f ahead ;
	Utilities::setForwardVector( rot, ahead ) ;

	SbVec3f curPos ;
	SbVec3f	startPos( position.getValue() ) ;

	if( age == 0.0f )
	{
		velocity = maxSpeed * ahead ;
		curPos = startPos + 10.0f * Constants::MissileLength * ahead ;
	}
	else
	{
		curPos = startPos + velocity * deltaTime ;
	}

	los->setRay( startPos, curPos ) ;
	los->apply( scene ) ;

	if( los->isObstructed() )
	{
		blowUp() ;
		curPos = los->getEndPoint() ;
		status = Missile::STRUCK ;
	}

	position.setValue( curPos ) ;
	age += deltaTime ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Return whether or not the missile is a guided type.
//------------------------------------------------------------------------------
SbBool
Cannon::isGuided() const
{
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Return the missile type (this is a virtual method wrapper around the static
// GetClassTypeId).
//------------------------------------------------------------------------------
Missile::MissileType
Cannon::getMissileType() const
{
	return( Missile::CANNON ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the icon scene graph.
//------------------------------------------------------------------------------
SoNode *
Cannon::getIconGraph()
{
	return( Cannon::icon ) ;
}
