//------------------------------------------------------------------------------
// Pointblank render area class.
//------------------------------------------------------------------------------

#include <stdlib.h>
#include <strings.h>
#include <Xm/Xm.h>
#include <Inventor/SbColor.h>
#include <Inventor/SoSceneManager.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/devices/SoXtMouse.h>
#include <Inventor/Xt/devices/SoXtKeyboard.h>
#include <GL/gl.h>
#include "CockPit.h"
#include "RedrawGroup.h"


class DisplayRegion {
    public :
	DisplayRegion( SoNode *newScene, SbVec2s &windowSize, float l, float r,
			float b, float t, SoGLRenderAction *renderAction ) ;
	~DisplayRegion() ;
	void	windowSizeChanged( const SbVec2s &windowSize ) ;
	void	redraw() ;
	void	setBackgroundColor( const SbColor &c )
			{ clearColor = c ; }
	void	setClearWindow( SbBool flag ) { clearWindow = flag ; }
	const SbVec2s &getCenter() const { return( center ) ; }
	const SbVec2s &getSize() const
		{ return( vpReg.getViewportSizePixels() ) ; }
	const SbViewportRegion &getViewportRegion() const { return( vpReg ) ; }

    private :
	DisplayRegion() ;
	DisplayRegion( const DisplayRegion &dr ) ;
	SoGLRenderAction	*ra ;
	SoNode			*sceneGraph ;
	float			left ;
	float			right ;
	float			bottom ;
	float			top ;
	SbColor			clearColor ;
	SbBool			clearWindow ;
	SbViewportRegion	vpReg ;
	SbVec2s			center ;
} ;



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
CockPit::CockPit(
	Widget		parent,
	const char	*name
	) :
	SoXtGLWidget(
		parent,
		name,
		TRUE,
		SO_GLX_RGB | SO_GLX_DOUBLE | SO_GLX_ZBUFFER | SO_GLX_OVERLAY,
		TRUE
		)
{
	ra = new SoGLRenderAction( SbViewportRegion( 1, 1 ) ) ;
	appHandler = NULL ;
	mouseDevice = new SoXtMouse ;
	keyboardDevice = new SoXtKeyboard ;
	needWindowClear = 2 ;
	overlayManager = new SoSceneManager() ;
	overlayManager->setRenderCallback( renderOverlayCB, this ) ;
	overlayMapColors = NULL ;
	useTexture = TRUE ;
	draining = FALSE ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
CockPit::~CockPit()
{
	for( int i = 0 ; i < sceneList.getLength() ; i++ )
	{
		DisplayRegion *dr = (DisplayRegion *)sceneList[i] ;
		delete dr ;
	}

	if( overlayMapColors != NULL )
	{
		free( overlayMapColors ) ;
	}

	delete overlayManager ;
	delete ra ;
	delete mouseDevice ;
	delete keyboardDevice ;
}



//------------------------------------------------------------------------------
// Initialize the input devices.
//------------------------------------------------------------------------------
void
CockPit::initOverlayGraphic()
{
	overlayManager->reinitialize() ;
	overlayManager->setRGBMode( FALSE ) ;
	overlayManager->setBackgroundIndex( 0 ) ;

	//
	// Load the overlay color map.
	//
	if( overlayMapColors != NULL )
	{
		free( overlayMapColors ) ;
	}
	int num = 3 ;
	SbColor	colors[3] ;
	colors[0].setValue( 1.0f, 1.0f, 0.0f ) ;
	colors[1].setValue( 1.0f, 0.0f, 0.0f ) ;
	colors[2].setValue( 0.0f, 0.0f, 0.0f ) ;

	overlayMapColors = (XColor *)malloc( sizeof(XColor) * num ) ;
	XColor *xcol = overlayMapColors ;
	for( int i = 0 ; i < num ; i++, xcol++ )
	{
		xcol->red   = (unsigned short)( colors[i][0] * 65535 ) ;
		xcol->green = (unsigned short)( colors[i][1] * 65535 ) ;
		xcol->blue  = (unsigned short)( colors[i][2] * 65535 ) ;
		xcol->flags = DoRed | DoGreen | DoBlue ;
		xcol->pixel = 1 + i ;
	}

	XStoreColors( getDisplay(), overlayColorMap, overlayMapColors, num ) ;

	Widget	w = getOverlayWidget() ;

	mouseDevice->enable( w, (XtEventHandler)SoXtGLWidget::eventHandler,
				(XtPointer)this, XtWindow( w ) ) ;
	keyboardDevice->enable( w, (XtEventHandler)SoXtGLWidget::eventHandler,
				(XtPointer)this, XtWindow( w ) ) ;
	mouseDevice->setWindowSize( getGlxSize() ) ;
	keyboardDevice->setWindowSize( getGlxSize() ) ;

	// Add an event handler to process EnterNotify events to force the
	// widget to have the keyboard focus.  Otherwise, keyboard events will
	// go to whichever text field has focus.  Otherwise, the keyboardFocus
	// would have to be POINTER.
	XtAddEventHandler( w, ( EnterWindowMask | LeaveWindowMask ), FALSE,
		(XtEventHandler)CockPit::windowEventCB, (XtPointer)this ) ;
}



//------------------------------------------------------------------------------
// Process EnterNotify events so that keyboard events are received properly.
//------------------------------------------------------------------------------
void
CockPit::windowEventCB(
	Widget		w,
	CockPit		*cp,
	XAnyEvent	*event,
	Boolean		*
	)
{
	//
	// Get keyboard focus.
	//
	if( event->type == EnterNotify )
	{
		XmProcessTraversal( cp->getOverlayWidget(),
					XmTRAVERSE_CURRENT ) ;
	}
	//
	// Lose focus.
	//
	else if( event->type == LeaveNotify )
	{
		XmProcessTraversal( SoXt::getShellWidget( w ),
					XmTRAVERSE_CURRENT ) ;
	}
}



//------------------------------------------------------------------------------
// Initialize the graphics.  Called when the window is created.
//------------------------------------------------------------------------------
void
CockPit::initGraphic()
{
	glDepthFunc( GL_LESS ) ;
	glClearDepth( 1.0f ) ;
}



//------------------------------------------------------------------------------
// Add a scene to the window.
//------------------------------------------------------------------------------
int
CockPit::addScene(
	SoNode	*scene,
	float	l,
	float	r,
	float	b,
	float	t,
	SbBool	clearWindow
	)
{
	DisplayRegion *dr = new DisplayRegion( scene, getSize(), l, r, b, t,
						ra ) ;
	dr->setClearWindow( clearWindow ) ;

	sceneList.append( dr ) ;

	return( sceneList.getLength() - 1 ) ;
}



//------------------------------------------------------------------------------
// Remove all scenes from the list.
//------------------------------------------------------------------------------
void
CockPit::clearScenes()
{
	sceneList.truncate( 0 ) ;
}



//------------------------------------------------------------------------------
// Specify the background color for one of the scenes.
//------------------------------------------------------------------------------
void
CockPit::setSceneBackgroundColor(
	const SbColor	&col,
	int		sceneNumber
	)
{
	if( sceneNumber >= 0 && sceneNumber < sceneList.getLength() )
	{
		DisplayRegion *dr = (DisplayRegion *)sceneList[sceneNumber] ;

		dr->setBackgroundColor( col ) ;
	}
}



//------------------------------------------------------------------------------
// Return the center pixel of a display region.
//------------------------------------------------------------------------------
const SbVec2s &
CockPit::getSceneCenter(
	int	sceneNo
	) const
{
	return( ((DisplayRegion *)sceneList[sceneNo])->getCenter() ) ;
}



//------------------------------------------------------------------------------
// Return the size of a display region.
//------------------------------------------------------------------------------
const SbVec2s &
CockPit::getSceneSize(
	int	sceneNo
	) const
{
	return( ((DisplayRegion *)sceneList[sceneNo])->getSize() ) ;
}



//------------------------------------------------------------------------------
// Redraw all of the scenes, forcing a screen clear first.  Normally only
// called by expose events.
//------------------------------------------------------------------------------
void
CockPit::redraw()
{
	RedrawGroup::forceRedraw() ;
	needWindowClear = 2 ;

	if( draining )
	{
		render() ;
	}
}



//------------------------------------------------------------------------------
// Force a redraw of the screen.
//------------------------------------------------------------------------------
void
CockPit::forceRedraw()
{
	RedrawGroup::forceRedraw() ;
	needWindowClear = 2 ;
	render() ;
}



//------------------------------------------------------------------------------
// Redraw all the scenes listed.
//------------------------------------------------------------------------------
void
CockPit::render()
{
	glXMakeCurrent( getDisplay(), getNormalWindow(), getNormalContext() ) ;

	if( needWindowClear > 0 )
	{
		needWindowClear-- ;
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
	}

	for( int i = 0 ; i < sceneList.getLength() ; i++ )
	{
		DisplayRegion *dr = (DisplayRegion *)sceneList[i] ;

		dr->redraw() ;
	}
	glXSwapBuffers( getDisplay(), getNormalWindow() ) ;
	RedrawGroup::clearRedraw() ;
}



//------------------------------------------------------------------------------
// Redraw all the scenes listed but don't swap.
//------------------------------------------------------------------------------
void
CockPit::renderNoSwap()
{
	glXMakeCurrent( getDisplay(), getNormalWindow(), getNormalContext() ) ;

	if( needWindowClear > 0 )
	{
		needWindowClear-- ;
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
	}

	for( int i = 0 ; i < sceneList.getLength() ; i++ )
	{
		DisplayRegion *dr = (DisplayRegion *)sceneList[i] ;

		dr->redraw() ;
	}
}



//------------------------------------------------------------------------------
// Adjust all the viewports when the window changes sizes.
//------------------------------------------------------------------------------
void
CockPit::sizeChanged(
	const SbVec2s	&newSize
	)
{
	SbViewportRegion vp( ra->getViewportRegion() ) ;
	vp.setWindowSize( newSize ) ;
	ra->setViewportRegion( vp ) ;

	mouseDevice->setWindowSize( getGlxSize() ) ;
	keyboardDevice->setWindowSize( getGlxSize() ) ;

	for( int i = 0 ; i < sceneList.getLength() ; i++ )
	{
		DisplayRegion *dr = (DisplayRegion *)sceneList[i] ;
		dr->windowSizeChanged( newSize ) ;
	}

	RedrawGroup::forceRedraw() ;
	needWindowClear = 2 ;

	overlayManager->setWindowSize( newSize ) ;
}



//------------------------------------------------------------------------------
// Handle an event.
//------------------------------------------------------------------------------
void
CockPit::processEvent(
	XAnyEvent	*anyevent
	)
{
	if( appHandler != NULL )
	{
		(*appHandler)( appHandlerData, anyevent ) ;
	}
}



//------------------------------------------------------------------------------
// Callback for rendering the overlay window.
//------------------------------------------------------------------------------
void
CockPit::renderOverlayCB(
	void		*data,
	SoSceneManager	*
	)
{
	( (CockPit *)data )->redrawOverlay() ;
}



//------------------------------------------------------------------------------
// Redraw the overlay scene if visible.
//------------------------------------------------------------------------------
void
CockPit::redrawOverlay()
{
	if( !isVisible() || getOverlayWindow() == NULL )
	{
		return ;
	}

	// set the window
	glXMakeCurrent( getDisplay(), getOverlayWindow(),
			getOverlayContext() ) ;

	// draw the scene
	overlayManager->render( TRUE, FALSE ) ;

	glFlush();
}



//------------------------------------------------------------------------------
// Set the overlay scene graph.
//------------------------------------------------------------------------------
void
CockPit::setOverlaySceneGraph(
	SoNode	*node
	)
{
	overlayManager->deactivate() ;

	overlayManager->setSceneGraph( node ) ;

	overlayManager->activate() ;
	overlayManager->scheduleRedraw() ;
}



//------------------------------------------------------------------------------
// Return the viewport region of a scene.
//------------------------------------------------------------------------------
const SbViewportRegion &
CockPit::getSceneViewportRegion(
	int sceneNo
	)
{
	return( ((DisplayRegion *)sceneList[sceneNo])->getViewportRegion() ) ;
}



//------------------------------------------------------------------------------
// Deterimine if textures should be used.
//------------------------------------------------------------------------------
void
CockPit::afterRealizeHook()
{
	// call the base class
	SoXtGLWidget::afterRealizeHook();

	const char *renderer = (const char *)glGetString( GL_RENDERER ) ;

	//
	// On the following SGI machines we don't want texture rendering 
	// to be turned on by default (for speed reason) - machines not listed
	// below are assumed to be fast enough to do textures.
	//
    
	if( strncmp( (const char *)glGetString(GL_VENDOR), "SGI", 3 ) == 0 )
	{

		if( !strncmp( renderer, "NEWPORT", 7 ) ||	// Indy, XL
			!strncmp( renderer, "GR1", 3 ) ||	// P.I.
			!strncmp( renderer, "VGX", 3 ) ||	// VGX(T)
			!strncmp( renderer, "LIGHT", 5 ) ||	// Starter
			!strncmp( renderer, "GR2", 3 ) ||	// XZ
			!strncmp( renderer, "GR3", 3 ) ||	// Extreme
			!strncmp( renderer, "GU1", 3 ) )
		{
			useTexture = FALSE;
		}
	}
}



//------------------------------------------------------------------------------
// Drain events from the Xevent queue and return.
//------------------------------------------------------------------------------
void
CockPit::drainEvents()
{
	draining = TRUE ;
	XtInputMask mask ;

	while( !isVisible() ||
		( ( mask = XtAppPending( SoXt::getAppContext() ) ) &
		XtIMXEvent ) != 0 )
	{
		XEvent event ;
		SoXt::nextEvent( SoXt::getAppContext(), &event ) ;
		SoXt::dispatchEvent( &event ) ;
	}

	draining = FALSE ;
}



//------------------------------------------------------------------------------
// DisplayRegion helper class.
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
DisplayRegion::DisplayRegion(
	SoNode	*newScene,
	SbVec2s	&windowSize,
	float	l,
	float	r,
	float	b,
	float	t,
	SoGLRenderAction	*renderAction
	)
{
	ra = renderAction ;

	sceneGraph = newScene ;
	sceneGraph->ref() ;

	left = l ;
	right = r ;
	bottom = b ;
	top = t ;

	clearWindow = TRUE ;

	clearColor.setValue( 0.0f, 0.0f, 0.0f ) ;

	windowSizeChanged( windowSize ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
DisplayRegion::~DisplayRegion()
{
	sceneGraph->unref() ;
}



//------------------------------------------------------------------------------
// Adjust the window and viewport size.
//------------------------------------------------------------------------------
void
DisplayRegion::windowSizeChanged(
	const SbVec2s &windowSize
	)
{
	vpReg.setWindowSize( windowSize ) ;
	vpReg.setViewport( left, bottom, ( right - left ), ( top - bottom ) ) ;

	center.setValue(
		(short)( 0.5f * ( left + right ) * (float)windowSize[0] ),
		(short)( 0.5f * ( bottom + top ) * (float)windowSize[1] ) ) ;
}



//------------------------------------------------------------------------------
// Render the scene.
//------------------------------------------------------------------------------
void
DisplayRegion::redraw()
{
	ra->setViewportRegion( vpReg ) ;

	if( clearWindow )
	{
		const SbViewportRegion &reg = ra->getViewportRegion() ;

		const SbVec2s &size = reg.getViewportSizePixels() ;
		const SbVec2s &orig = reg.getViewportOriginPixels() ;

		glScissor( orig[0], orig[1], size[0], size[1] ) ;
		glEnable( GL_SCISSOR_TEST ) ;
		glClearColor( clearColor[0], clearColor[1], clearColor[2],
				0.0f ) ;
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
		glDisable( GL_SCISSOR_TEST ) ;
	}

	ra->apply( sceneGraph ) ;
}
