//------------------------------------------------------------------------------
// CollisionAction class.
//------------------------------------------------------------------------------

#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoComplexityElement.h>
#include <Inventor/elements/SoComplexityTypeElement.h>
#include <Inventor/elements/SoCoordinateElement.h>
#include <Inventor/elements/SoFontNameElement.h>
#include <Inventor/elements/SoFontSizeElement.h>
#include <Inventor/elements/SoProfileCoordinateElement.h>
#include <Inventor/elements/SoProfileElement.h>
#include <Inventor/elements/SoSwitchElement.h>
#include <Inventor/elements/SoUnitsElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/elements/SoViewingMatrixElement.h>
#include <Inventor/elements/SoViewportRegionElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoGroup.h>
#include <math.h>
#include "CollisionAction.h"
#include "CollisionBox.h"
#include "Obstacle.h"


// #define COLLISION_ACTION_DEBUG 1

#define CL(i,j,k)	(((i)<<2)|((j)<<1)|k)

static int32_t firstIndex[] = {
		CL(0,0,0),CL(0,1,0),CL(1,0,0),CL(1,1,0),
		CL(0,0,0),CL(0,0,1),CL(1,0,0),CL(1,0,1),
		CL(0,0,0),CL(0,0,1),CL(0,1,0),CL(0,1,1),
		} ;
static int32_t secondIndex[] = {
		CL(0,0,1),CL(0,1,1),CL(1,0,1),CL(1,1,1),
		CL(0,1,0),CL(0,1,1),CL(1,1,0),CL(1,1,1),
		CL(1,0,0),CL(1,0,1),CL(1,1,0),CL(1,1,1),
		} ;


#define PRINT_VEC(s,v) \
	printf("%-20s: %7.4f %7.4f %7.4f\n",(s),(v)[0],(v)[1],(v)[2])


SO_ACTION_SOURCE( CollisionAction ) ;


//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
CollisionAction::initClass()
{
	SO_ACTION_INIT_CLASS( CollisionAction, SoAction ) ;

	// Enable the elements that are involved.
	SO_ENABLE( CollisionAction, SoModelMatrixElement ) ;
	SO_ENABLE( CollisionAction, SoComplexityElement ) ;
	SO_ENABLE( CollisionAction, SoComplexityTypeElement ) ;
	SO_ENABLE( CollisionAction, SoCoordinateElement ) ;
	SO_ENABLE( CollisionAction, SoFontNameElement ) ;
	SO_ENABLE( CollisionAction, SoFontSizeElement ) ;
	SO_ENABLE( CollisionAction, SoProfileCoordinateElement ) ;
	SO_ENABLE( CollisionAction, SoProfileElement ) ;
	SO_ENABLE( CollisionAction, SoSwitchElement ) ;
	SO_ENABLE( CollisionAction, SoUnitsElement ) ;
	SO_ENABLE( CollisionAction, SoViewVolumeElement ) ;
	SO_ENABLE( CollisionAction, SoViewingMatrixElement ) ;
	SO_ENABLE( CollisionAction, SoViewportRegionElement ) ;
	SO_ENABLE( CollisionAction, SoNormalBindingElement ) ;

	// Register static methods fo the action.
	SO_ACTION_ADD_METHOD( SoNode, doNada ) ;

	SO_ACTION_ADD_METHOD( SoGroup, callDoAction ) ;
	SO_ACTION_ADD_METHOD( Obstacle, collideS ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
CollisionAction::CollisionAction() :
	self( NULL )
{
	SO_ACTION_CONSTRUCTOR( CollisionAction ) ;
	boundingBox.makeEmpty() ;
	collisionRadius = 0.0f ;
	centerOffset.setValue( 0.0f, 0.0f, 0.0f ) ;
	setLocation( SbVec3f( 0.0f, 0.0f, 0.0f ),
			SbRotation( 0.0f, 0.0f, 1.0f, 0.0f ) ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
CollisionAction::~CollisionAction()
{
}



//------------------------------------------------------------------------------
// Initiate the action.
//------------------------------------------------------------------------------
void
CollisionAction::beginTraversal(
	SoNode *node
	)
{
	cList.truncate( 0 ) ;
	traverse( node ) ;
}



//------------------------------------------------------------------------------
// Generic action.
//------------------------------------------------------------------------------
void
CollisionAction::callDoAction(
	SoAction	*action,
	SoNode		*node
	)
{
	node->doAction( action ) ;
}



//------------------------------------------------------------------------------
// Initialize the bounding box to check collisions against.
//------------------------------------------------------------------------------
void
CollisionAction::setBoundingBox(
	const SbBox3f &box
	)
{
	boundingBox.setBounds( box.getMin(), box.getMax() ) ;
	centerOffset = boundingBox.getCenter() ;

	float w, h, l ;
	boundingBox.getSize( w, h, l ) ;
	collisionRadius = sqrtf( 0.25f * ( w * w + h * h + l * l ) ) ;
}



//------------------------------------------------------------------------------
// Initialize the location matrix of the bounding box and calculate the inverse.
//------------------------------------------------------------------------------
void
CollisionAction::setLocation(
	const SbVec3f &location,
	const SbRotation &rot
	)
{
	collisionCenter = location + centerOffset ;

	rot.getValue( localToWorldMatrix ) ;
	// Stuff in translation values (quicker than building translation
	// matrix and multiplying the two matrices together).
	localToWorldMatrix[3][0] = location[0] ;
	localToWorldMatrix[3][1] = location[1] ;
	localToWorldMatrix[3][2] = location[2] ;

	worldToLocalMatrix = localToWorldMatrix.inverse() ;
}



//------------------------------------------------------------------------------
// Given the 8 local-transformed points, save the collision point and normal
// for the case when an obstacle point is inside the local bounding box.
//------------------------------------------------------------------------------
Collision *
CollisionAction::savePointCollisionInfo(
	const SbRotation &rot,		// Obstacle rotation
	const SbVec3f	 &normal,	// Obstacle/World space normal
	SbBool		 normalInObjectSpace, // False means in world space
	const SbVec3f	 &localPt,	// Local space collision point
	const SbVec3f	 &worldPt,	// World space collision point
	SbBool		 findWorldPt	// Flag
	)
{
	// Add collision point and set world point and normal.
	Collision *collision = addCollision( localPt ) ;


	// Transform obstacle space normal to world space normal.
	if( normalInObjectSpace )
	{
		SbVec3f worldNorm ;
		// Create obstacle-to-world transformation matrix.
		SbMatrix obstacleToWorld ;
		rot.getValue( obstacleToWorld ) ;
		obstacleToWorld.multDirMatrix( normal, worldNorm ) ;
		worldNorm.normalize() ;
		collision->setWorldNormal( worldNorm ) ;
	}
	else
	{
		collision->setWorldNormal( normal ) ;
	}

	// Find the world point if necessary.
	if( findWorldPt )
	{
		SbVec3f wp ;
		localToWorldMatrix.multVecMatrix( localPt, wp ) ;
		collision->setWorldPoint( wp ) ;
	}
	else
	{
		collision->setWorldPoint( worldPt ) ;
	}

	return( collision ) ;
}



//------------------------------------------------------------------------------
// Trivial rejection test based on the distance between the center and
// collision radii.
//------------------------------------------------------------------------------
SbBool
CollisionAction::checkDistance(
	const SbVec3f	&center,
	float		radius
	)
{
	//
	// Check for trivial rejection based on distance between centers of
	// objects and sum of collision radii.
	//
	SbVec3f dVec( center[0]-collisionCenter[0],
		center[1]-collisionCenter[1], center[2]-collisionCenter[2] ) ;
	radius += collisionRadius ;
	return( dVec.dot( dVec ) > ( radius * radius ) ) ;
}



//------------------------------------------------------------------------------
// Perform the collision check given the 8 corner points of the obstacle.
//------------------------------------------------------------------------------
SbBool
CollisionAction::checkCollision(
	const SbVec3f corner[2][2][2],	// Obstacle's world space corner points
	const SbRotation &obstacleRot,	// Obstacle's rotation matrix
	const SbVec3f &center,		// Obstacle's center
	float radius			// Obstacle's collision radius
	)
{
	//
	// Check for trivial rejection based on distance between centers of
	// objects and sum of collision radii.
	//
	SbVec3f dVec( center[0]-collisionCenter[0],
		center[1]-collisionCenter[1], center[2]-collisionCenter[2] ) ;
	radius += collisionRadius ;
	if( dVec.dot( dVec ) > radius * radius )
	{
		return FALSE ;
	}

	//
	// Transform points and check status relative to local bounding box.
	// Save status for point to help reduce the number of checks made for
	// edges intersecting bounding box (made later if no point is found
	// inside the bounding box).
	//
	int n, i, j, k ;
	SbVec3f pt[8] ;
	unsigned int pStatus[8] ;
	float xMin, xMax, yMin, yMax, zMin, zMax ;
	boundingBox.getBounds( xMin, yMin, zMin, xMax, yMax, zMax ) ;
	for( n = 0 ; n < 8 ; n++ )
	{
		i = ( n & 0x04 ) >> 2 ;
		j = ( n & 0x02 ) >> 1 ;
		k = n & 0x01 ;
		pStatus[n] = 0 ;
		worldToLocalMatrix.multVecMatrix( corner[i][j][k], pt[n] );
		if( pt[n][0] >= xMin ) pStatus[n] |= XMIN_INNER ;
		if( pt[n][0] <= xMax ) pStatus[n] |= XMAX_INNER ;
		if( pt[n][1] >= yMin ) pStatus[n] |= YMIN_INNER ;
		if( pt[n][1] <= yMax ) pStatus[n] |= YMAX_INNER ;
		if( pt[n][2] >= zMin ) pStatus[n] |= ZMIN_INNER ;
		if( pt[n][2] <= zMax ) pStatus[n] |= ZMAX_INNER ;
		//
		// If point was inside, stop and return true.
		//
		if( pStatus[n] == ALL_INNER )
		{
#ifdef COLLISION_ACTION_DEBUG
printf( "point %d was inside\n", n ) ;
#endif
			// Object space normal
			SbVec3f oDir( i == 0 ? -1.0f : 1.0f,
				j == 0 ? -1.0f : 1.0f, k == 0 ? -1.0f : 1.0f ) ;

			savePointCollisionInfo( obstacleRot, oDir, TRUE, pt[n],
						corner[i][j][k], FALSE ) ;

			return TRUE ;
		}
	}

	//
	// No point was inside, so check lines and then surfaces.
	//
	return( detailedCheckCollision( pt, pStatus, obstacleRot ) ) ;
}



//------------------------------------------------------------------------------
// Perform the collision check given the bounding box of the obstacle and its
// position and rotation.
//------------------------------------------------------------------------------
SbBool
CollisionAction::checkCollision(
	const SbBox3f &box,
	const SbRotation &rotation,
	const SbVec3f &center,
	float radius
	)
{
	//
	// Check for trivial rejection based on distance between centers of
	// objects and sum of collision radii.
	//
	SbVec3f dVec( center[0]-collisionCenter[0],
		center[1]-collisionCenter[1], center[2]-collisionCenter[2] ) ;
	radius += collisionRadius ;
	if( dVec.dot( dVec ) > radius * radius )
	{
		return FALSE ;
	}

	//
	// Transform points and check status relative to local bounding box.
	// Save status for point to help reduce the number of checks made for
	// edges intersecting bounding box (made later if no point is found
	// inside the bounding box).
	//
	int n, i, j, k ;
	SbVec3f pt[8] ;
	SbVec3f boxExt[2] ;
	boxExt[0] = box.getMin() ;
	boxExt[1] = box.getMax() ;
	unsigned int pStatus[8] ;
	float xMin, xMax, yMin, yMax, zMin, zMax ;
	boundingBox.getBounds( xMin, yMin, zMin, xMax, yMax, zMax ) ;

	SbMatrix obstacleToLocal ;
	rotation.getValue( obstacleToLocal ) ;
	// Stuff in translation values (quicker than building translation
	// matrix and multiplying the two matrices together).
	obstacleToLocal[3][0] = center[0] ;
	obstacleToLocal[3][1] = center[1] ;
	obstacleToLocal[3][2] = center[2] ;
	obstacleToLocal.multRight( worldToLocalMatrix ) ;

	for( n = 0 ; n < 8 ; n++ )
	{
		i = ( n & 0x04 ) >> 2 ;
		j = ( n & 0x02 ) >> 1 ;
		k = n & 0x01 ;
		pStatus[n] = 0 ;
		SbVec3f A( boxExt[i][0], boxExt[j][1], boxExt[k][2] ) ;
		obstacleToLocal.multVecMatrix( A, pt[n] );
		if( pt[n][0] >= xMin ) pStatus[n] |= XMIN_INNER ;
		if( pt[n][0] <= xMax ) pStatus[n] |= XMAX_INNER ;
		if( pt[n][1] >= yMin ) pStatus[n] |= YMIN_INNER ;
		if( pt[n][1] <= yMax ) pStatus[n] |= YMAX_INNER ;
		if( pt[n][2] >= zMin ) pStatus[n] |= ZMIN_INNER ;
		if( pt[n][2] <= zMax ) pStatus[n] |= ZMAX_INNER ;
		//
		// If point was inside, stop and return true.
		//
		if( pStatus[n] == ALL_INNER )
		{
#ifdef COLLISION_ACTION_DEBUG
printf( "point %d was inside\n", n ) ;
#endif
			// Object space normal
			SbVec3f oDir( i == 0 ? -1.0f : 1.0f,
				j == 0 ? -1.0f : 1.0f, k == 0 ? -1.0f : 1.0f ) ;

			SbVec3f worldPt ;
			savePointCollisionInfo( rotation, oDir, TRUE, pt[n],
						worldPt, TRUE ) ;

			return TRUE ;
		}
	}

	//
	// No point was inside, so check lines.
	//
	return( detailedCheckCollision( pt, pStatus, rotation ) ) ;
}



//------------------------------------------------------------------------------
// Common code for more detailed collision check.
//------------------------------------------------------------------------------
SbBool
CollisionAction::detailedCheckCollision(
	const SbVec3f		pt[8],		// Obstacle world space corners
	const unsigned int	pStatus[8],	// Status of point checks
	const SbRotation	&obstacleRot	// Obstacle rotation
	)
{
	int		n ;
	SbVec3f		intersection ;
	SbVec3f		faceNormal ;
	static float	lookUp[] = { -1.0f, 1.0f, 0.0f, 0.0f } ;

	CollisionBox *b = NULL ;
	for( n = 0 ; n < 12 ; n++ )
	{
		//
		// If the following test fails, the line can't intersect
		// the bounding box, so don't bother testing.
		//
		if( ( pStatus[firstIndex[n]] | pStatus[secondIndex[n]] ) ==
			ALL_INNER )
		{
#ifdef COLLISION_ACTION_DEBUG
			printf( "detailed line collision check\n" ) ;
#endif
			if( b == NULL ) b = new CollisionBox( boundingBox ) ;
			if( b->checkLine( pt[firstIndex[n]],
				pt[secondIndex[n]], intersection, faceNormal ) )
			{
#ifdef COLLISION_ACTION_DEBUG
				printf( "line collision\n" ) ;
#endif
				//
				// Found a line of the obstacle that
				// intersects our local bounding box.
				// Now find world point and world-space normal.
				//

				// Calculate obstacle space normal based on
				// horid logic ops.  This is dependent upon
				// the matchup between i,j,k and n and the
				// x, y, and z minimums and maximums of the
				// obstacle's bounding box.
				//
				// XorFlags should contain just 1 bit set
				// indicating which dimension (x, y, or z) is
				// changing in the obstacle's space on the
				// line that intersects the local bounding box.
				//
				// The rest of the logic produces a value with
				// the lower two bits possible set.
				//
				// Bit 0: indicates min (0) or max (1) edge iff
				//	  bit 1 is clear
				// Bit 1: set indicates that normal is
				//	  perpendicular to this dimension, hence
				//	  this component is zero
				int xorFlags = firstIndex[n] ^ secondIndex[n] ;
				int i = ( ( xorFlags & 0x04 ) >> 1 ) |
					( firstIndex[n] & 0x04 ) >> 2 ;
				int j = ( ( xorFlags & 0x02 ) ) |
					( firstIndex[n] & 0x02 ) >> 1 ;
				int k = ( ( xorFlags & 0x01 ) << 1 ) |
					( firstIndex[n] & 0x01 )  ;

				// Finally, the obstacle space normal.
				SbVec3f oDir( lookUp[i], lookUp[j], lookUp[k] );

				SbVec3f worldPt ;
				savePointCollisionInfo( obstacleRot, oDir, TRUE,
							intersection, worldPt,
							TRUE ) ;
				return TRUE ;
			}
		}
	}
	if( b != NULL ) delete b ;

	//
	// No collision yet, so check if the edges of the action's bounding
	// box intersect with the planes of the obstacles's bounding box.
	//
	CollisionBox objectBox( pt ) ;
	if( objectBox.checkBox( boundingBox, intersection, faceNormal ) )
	{
#ifdef COLLISION_ACTION_DEBUG
		printf( "surface collision\n" ) ;
#endif
		// The local normal is in faceNormal.  Transform to world space.
		SbVec3f worldNormal ;
		localToWorldMatrix.multDirMatrix( faceNormal, worldNormal ) ;
		worldNormal.normalize() ;
// PRINT_VEC( "worldNormal", worldNormal ) ;
		SbVec3f worldPt ;
		savePointCollisionInfo( obstacleRot, worldNormal, FALSE,
					intersection, worldPt, TRUE ) ;
		return TRUE ;
	}

	//
	// No collision, so return.
	//
	return FALSE ;
}



//------------------------------------------------------------------------------
// Static method for all checking collisions with all obstacles.
//------------------------------------------------------------------------------
void
CollisionAction::collideS(
	SoAction *action,
	SoNode	*node
	)
{
	( (Obstacle *)node )->collide( (CollisionAction *)action ) ;
}



//------------------------------------------------------------------------------
// Static method for ignoring the collision action.
//------------------------------------------------------------------------------
void
CollisionAction::doNada(
	SoAction *,
	SoNode	 *
	)
{
}



//------------------------------------------------------------------------------
// Set the self node.
//------------------------------------------------------------------------------
void
CollisionAction::setSelf(
	SoNode	*node
	)
{
	if( self != NULL )
	{
		self->unref() ;
	}
	self = node ;
	if( self != NULL )
	{
		self->ref() ;
	}
}



//------------------------------------------------------------------------------
// Add a collision to the collision list.
//------------------------------------------------------------------------------
Collision *
CollisionAction::addCollision(
	const SbVec3f	&pt
	)
{
	Collision *c = new Collision( getCurPath(), pt ) ;
	cList.append( c ) ;

	return( c ) ;
}



//------------------------------------------------------------------------------
// Collision class definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Collision::Collision(
	const SoPath	*collisionPath,
	const SbVec3f	&lPoint
	) :
	worldPoint( 0.0f, 0.0f, 0.0f ),
	worldNormal( 0.0f, 0.0f, 0.0f )
{
	path = collisionPath->copy() ;
	path->ref() ;

	localPoint.setValue( lPoint[0], lPoint[1], lPoint[2] ) ;
}



//------------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------------
Collision::Collision(
	const Collision &c
	)
{
	path = c.path ;
	path->ref() ;

	localPoint = c.localPoint ;
	worldPoint = c.worldPoint ;
	worldNormal = c.worldNormal ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Collision::~Collision()
{
	path->unref() ;
}



//------------------------------------------------------------------------------
// CollisionList class definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Truncate the list starting at a givent entry.
//------------------------------------------------------------------------------
void
CollisionList::truncate(
	int	start
	)
{
	int	i ;

	for( i = start ; i < getLength() ; i++ )
	{
		if( (*this)[i] != NULL )
		{
			delete (*this)[i] ;
		}
	}

	SbPList::truncate( start ) ;
}



//------------------------------------------------------------------------------
// Set the element of the list, deleting the old one.
//------------------------------------------------------------------------------
void
CollisionList::set(
	int		i,
	Collision	*c
	)
{
	if( (*this)[i] != NULL )
	{	
		delete (*this)[i] ;
	}

	(* (const SbPList *)this )[i] = (void *)c ;
}

