//------------------------------------------------------------------------------
// Class header for CollisionAction
//------------------------------------------------------------------------------

#ifndef _COLLISIONACTION_H_
#define _COLLISIONACTION_H_

#include <Inventor/SbLinear.h>
#include <Inventor/SbBox.h>
#include <Inventor/actions/SoSubAction.h>


class SoNode ;

class Collision {
    public :
	Collision( const Collision &c ) ;
	~Collision() ;
	SoPath		*getPath() const { return path ; }
	const SbVec3f	&getLocalPoint() const { return localPoint ; }
	const SbVec3f	&getWorldPoint() const { return worldPoint ; }
	const SbVec3f	&getWorldNormal() const { return worldNormal ; }
	void	setWorldPoint( const SbVec3f &w ) { worldPoint = w ; }
	void	setWorldNormal( const SbVec3f &n ) { worldNormal = n ; }

    SoINTERNAL public :
	Collision( const SoPath *collisionPath, const SbVec3f &cPoint ) ;

    private :
	Collision() ;
	SoPath	*path ;
	SbVec3f localPoint ;
	SbVec3f	worldPoint ;
	SbVec3f	worldNormal ;
} ;



class CollisionList : public SbPList {
    public :
	CollisionList() : SbPList() {}
	CollisionList( int size ) : SbPList(size) {}
	CollisionList( const CollisionList &l ) : SbPList(l)  {} ;
	~CollisionList() { truncate( 0 ) ; }

	void append( Collision *c ) { SbPList::append( (void *)c ) ; }
	void insert( Collision *c, int addBefore )
		{ SbPList::insert( (void *)c, addBefore ) ; }
	void truncate( int start ) ;
	Collision *operator [](int i) const
		{ return ( (Collision *)( (* (const SbPList *) this )[i] ) ) ; }
	void set( int i, Collision *c ) ;
    private :
	CollisionList &operator =( const CollisionList & ) ;
} ;



class CollisionAction : public SoAction {

	SO_ACTION_HEADER( CollisionAction ) ;

    public :
	enum IntersectionStatus {
		XMIN_INNER = ( 0x01 << 0 ),
		XMAX_INNER = ( 0x01 << 1 ),
		YMIN_INNER = ( 0x01 << 2 ),
		YMAX_INNER = ( 0x01 << 3 ),
		ZMIN_INNER = ( 0x01 << 4 ),
		ZMAX_INNER = ( 0x01 << 5 ),
		ALL_INNER = XMIN_INNER | XMAX_INNER | YMIN_INNER |
			    YMAX_INNER | ZMIN_INNER | ZMAX_INNER
		} ;

	static void initClass() ;
	CollisionAction() ;
	virtual ~CollisionAction() ;
	void setBoundingBox( const SbBox3f &box ) ;
	void setLocation( const SbVec3f &location, const SbRotation &rot ) ;
	void setSelf( SoNode *n ) ;
	SbBool isSelf( SoNode *n ) { return( n == self ) ; }
	Collision *addCollision( const SbVec3f &point ) ;
	const CollisionList &getCollisionList() const
		{ return cList ; }
	SbBool	checkDistance( const SbVec3f &center, float radius ) ;

    SoEXTENDER public :
	SbBool checkCollision( const SbVec3f corner[2][2][2],
			const SbRotation &rot, const SbVec3f &center,
			float collisionRadius ) ;
	SbBool checkCollision( const SbBox3f &box, const SbRotation &rot,
			const SbVec3f &center, float collisionRadius ) ;
	void xformWorldToLocalPt( const SbVec3f &src, SbVec3f &dst ) 
		{ worldToLocalMatrix.multVecMatrix( src, dst ) ; }
	void xformLocalToWorldPt( const SbVec3f &src, SbVec3f &dst ) 
		{ localToWorldMatrix.multVecMatrix( src, dst ) ; }
	Collision *savePointCollisionInfo( const SbRotation &rot,
		const SbVec3f &normal, SbBool normalInObjectSpace,
		const SbVec3f &localPt, const SbVec3f &worldPt,
		SbBool findWorldPt ) ;
	SbBool detailedCheckCollision( const SbVec3f pt[8],
		const unsigned int pStatus[8], const SbRotation &rot ) ;
	void getBounds( float &xMin, float &yMin, float &zMin,
			float &xMax, float &yMax, float &zMax )
		{ boundingBox.getBounds( xMin, yMin, zMin, xMax, yMax, zMax ); }
	const SbVec3f &getCenter() { return collisionCenter ; }
	float getRadius() { return collisionRadius ; }

    protected :
	virtual void beginTraversal( SoNode *node ) ;

    private :
	static void callDoAction( SoAction *, SoNode * ) ;
	static void	collideS( SoAction *, SoNode * ) ;
	static void	doNada( SoAction *, SoNode * ) ;

	CollisionList	cList ;
	SbBox3f		boundingBox ;
	SbVec3f		collisionCenter ;
	SbVec3f		centerOffset ;
	float		collisionRadius ;
	SbMatrix	localToWorldMatrix ;
	SbMatrix	worldToLocalMatrix ;
	SoNode		*self ;		// Node instigating the collision check
} ;

#endif /* !_COLLISIONACTION_H_ */
