//------------------------------------------------------------------------------
// CollisionBox class.
//------------------------------------------------------------------------------

#include "CollisionBox.h"
#include "Utilities.h"


#ifdef VERBOSE
static Utilities	Util( "DEBUG_COLLISIONBOX" ) ;
#endif /* VERBOSE */

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
CollisionBox::CollisionBox()
{
	for( int i = 0 ; i < 8 ; i++ )
	{
		points[i].setValue( 0.0f, 0.0f, 0.0f ) ;
	}
	updateFromPoints() ;
}



//------------------------------------------------------------------------------
// Update the data fields from the points array.
//------------------------------------------------------------------------------
void
CollisionBox::updateFromPoints()
{
	origin[0][0] = points[0] ;
	origin[0][1] = points[1] ;
	origin[1][0] = points[0] ;
	origin[1][1] = points[2] ;
	origin[2][0] = points[0] ;
	origin[2][1] = points[4] ;

	N[0] = origin[0][1] - origin[0][0] ;
	N[1] = origin[1][1] - origin[1][0] ;
	N[2] = origin[2][1] - origin[2][0] ;

	n2[0] = N[0].dot( N[0] ) ;
	n2[1] = N[1].dot( N[1] ) ;
	n2[2] = N[2].dot( N[2] ) ;

	d[0][0] = N[0].dot( origin[0][0] ) ;
	d[0][1] = N[0].dot( origin[0][1] ) ;
	d[1][0] = N[1].dot( origin[1][0] ) ;
	d[1][1] = N[1].dot( origin[1][1] ) ;
	d[2][0] = N[2].dot( origin[2][0] ) ;
	d[2][1] = N[2].dot( origin[2][1] ) ;
}



//------------------------------------------------------------------------------
// Set the points array.
//------------------------------------------------------------------------------
void
CollisionBox::setPoints(
	const SbVec3f corner[8]
	)
{
	points[0] = corner[0] ;
	points[1] = corner[1] ;
	points[2] = corner[2] ;
	points[3] = corner[3] ;
	points[4] = corner[4] ;
	points[5] = corner[5] ;
	points[6] = corner[6] ;
	points[7] = corner[7] ;

	updateFromPoints() ;
}



//------------------------------------------------------------------------------
// Set the points array.
//------------------------------------------------------------------------------
void
CollisionBox::setPoints(
	const SbVec3f	corner[2][2][2]
	)
{
	points[0] = corner[0][0][0] ;
	points[1] = corner[0][0][1] ;
	points[2] = corner[0][1][0] ;
	points[3] = corner[0][1][1] ;
	points[4] = corner[1][0][0] ;
	points[5] = corner[1][0][1] ;
	points[6] = corner[1][1][0] ;
	points[7] = corner[1][1][1] ;

	updateFromPoints() ;
}



//------------------------------------------------------------------------------
// Set the points array.
//------------------------------------------------------------------------------
void
CollisionBox::setPoints(
	const SbBox3f	&box
	)
{
	const SbVec3f &min = box.getMin() ;
	const SbVec3f &max = box.getMax() ;

	points[0].setValue( min[0], min[1], min[2] ) ;
	points[1].setValue( min[0], min[1], max[2] ) ;
	points[2].setValue( min[0], max[1], min[2] ) ;
	points[3].setValue( min[0], max[1], max[2] ) ;
	points[4].setValue( max[0], min[1], min[2] ) ;
	points[5].setValue( max[0], min[1], max[2] ) ;
	points[6].setValue( max[0], max[1], min[2] ) ;
	points[7].setValue( max[0], max[1], max[2] ) ;

	updateFromPoints() ;
}



//------------------------------------------------------------------------------
// Check if a line given by 2 points, A and B, intersects the surface of the
// box.
//------------------------------------------------------------------------------
SbBool
CollisionBox::checkLine(
	const SbVec3f	&A,
	const SbVec3f	&B,
	SbVec3f		&intersection,
	SbVec3f		&faceNormal
	) const 
{
	SbVec3f BA( B[0]-A[0], B[1]-A[1], B[2]-A[2] ) ;
	const float Epsilon = 1.e-10 ;
#ifdef VERBOSE
Util.print( 4, "------------checkLine\n" ) ;
Util.printVec( 4, "A", A ) ;
Util.printVec( 4, "B", B ) ;
Util.printVec( 4, "BA", BA ) ;
#endif /* VERBOSE */
	for( int dir = 0 ; dir < 3 ; dir++ )
	{
#ifdef VERBOSE
Util.print( 4, "+++ dir = %d\n", dir ) ;
Util.printVec( 4, "N", N[dir] ) ;
#endif /* VERBOSE */
		float tDenom = N[dir].dot( BA ) ;
		if( -Epsilon < tDenom && tDenom < Epsilon )
		{
#ifdef VERBOSE
Util.print( 4, "bailing due to tDenom = %f\n", tDenom ) ;
#endif /* VERBOSE */
			continue ;
		}
		float nDotA = N[dir].dot( A ) ;
		tDenom = 1.0f / tDenom ;
#ifdef VERBOSE
Util.print( 4, "nDotA = %f, tDenom = %f\n", nDotA, tDenom ) ;
#endif /* VERBOSE */
		for( int side = 0 ; side < 2 ; side++ )
		{
#ifdef VERBOSE
Util.print( 4, "*** side = %d\n", side ) ;
Util.print( 4, "%-20s: %f\n", "d", d[dir][side] ) ;
#endif /* VERBOSE */
			float t = ( d[dir][side] - nDotA ) * tDenom ;
			if( t < 0.0f || t > 1.0f )
			{
#ifdef VERBOSE
Util.print( 4, "bailing due to t = %f\n", d ) ;
#endif /* VERBOSE */
				continue ;
			}
#ifdef VERBOSE
Util.print( 4, "t = %f\n", t ) ;
Util.printVec( 4, "origin", origin[dir][side] ) ;
#endif /* VERBOSE */
			SbVec3f P(
				A[0]+t*BA[0]-origin[dir][side][0],
				A[1]+t*BA[1]-origin[dir][side][1],
				A[2]+t*BA[2]-origin[dir][side][2]
				) ;
			int dirJ = ( dir + 1 ) % 3 ;
			int dirK = ( dir + 2 ) % 3 ;
			float s = N[dirJ].dot( P ) ;
			float u = N[dirK].dot( P ) ;
#ifdef VERBOSE
Util.printVec( 4, "P", P ) ;
Util.print( 4, "s = %f, u = %f, n2[%d] = %f, ns[%d] = %f\n", s, u, dirJ, n2[dirJ], dirK, n2[dirK] ) ;
Util.printVec( 4, "N[dirJ]", N[dirJ] ) ;
Util.printVec( 4, "N[dirK]", N[dirK] ) ;
#endif /* VERBOSE */

			if( s < 0.0f || u < 0.0f ||
				s > n2[dirJ] || u > n2[dirK] )
			{
				continue ;
			}

			intersection.setValue( A[0]+t*BA[0], A[1]+t*BA[1],
						A[2]+t*BA[2] ) ;
			faceNormal = ( side == 1 ) ? N[dir] : -N[dir] ;
#ifdef VERBOSE
Util.print( 4, "intersecting\n" ) ;
Util.printVec( 4, "faceNormal", faceNormal ) ;
#endif /* VERBOSE */

			return TRUE ;
		}
	}

	return FALSE ;
}



//------------------------------------------------------------------------------
// Check to see if the edges of a bounding box intersect the collision box.
//------------------------------------------------------------------------------
SbBool
CollisionBox::checkBox(
	SbBox3f	&box,
	SbVec3f	&intersection,
	SbVec3f &faceNormal
	)
{
	SbVec3f boxExt[2] ;
	int	n ;
	static int index[12][6] = {
		{ 0, 1, 0, 1, 1, 0 }, // bottom front edge
		{ 1, 1, 1, 0, 1, 1 }, // top front edge
		{ 1, 0, 0, 0, 0, 0 }, // bottom rear edge
		{ 0, 0, 1, 1, 0, 1 }, // top rear edge
		{ 0, 0, 1, 0, 0, 0 }, // left rear edge
		{ 0, 1, 0, 0, 0, 0 }, // left bottom edge
		{ 0, 1, 0, 0, 1, 1 }, // left front edge
		{ 0, 0, 1, 0, 1, 1 }, // left top edge
		{ 1, 0, 0, 1, 0, 1 }, // right rear edge
		{ 1, 1, 1, 1, 0, 1 }, // right top edge
		{ 1, 1, 1, 1, 1, 0 }, // right front edge
		{ 1, 0, 0, 1, 1, 0 }, // right bottom edge
		} ;

	boxExt[0] = box.getMin() ;
	boxExt[1] = box.getMax() ;

	for( n = 0 ; n < 12 ; n++ )
	{
		int *k = index[n] ;
		if( checkLine(
		    SbVec3f(boxExt[k[0]][0],boxExt[k[1]][1],boxExt[k[2]][2]),
		    SbVec3f(boxExt[k[3]][0],boxExt[k[4]][1],boxExt[k[5]][2]),
		    intersection, faceNormal ) )
		{
#ifdef COLLISION_ACTION_DEBUG
			printf( "checkBox hit on line %d\n", n ) ;
#endif
			return TRUE ;
		}
	}

	return FALSE ;
}



//------------------------------------------------------------------------------
// Debug print function.
//------------------------------------------------------------------------------
void
CollisionBox::print(
	const char *label
	) const
{
#define PRINT_V(l,a) printf( "--%15s:  %7.4f %7.4f %7.4f\n",l,a[0],a[1],a[2])
	printf( "%s:\n", label ) ;
	PRINT_V( "points", points[0] ) ;
	for( int i = 1 ; i < 8 ; i++ )
	{
		PRINT_V( " ", points[i] ) ;
	}
	PRINT_V( "N[0]", N[0] ) ;
	printf( "--%10s:  %7.4f\n", "n2[0]", n2[0] ) ;
	PRINT_V( "origin[0][0]", origin[0][0] ) ;
	printf( "--%10s:  %7.4f\n", "d[0][0]", d[0][0] ) ;
	PRINT_V( "origin[0][1]", origin[0][1] ) ;
	printf( "--%10s:  %7.4f\n", "d[0][1]", d[0][1] ) ;
	PRINT_V( "N[1]", N[1] ) ;
	printf( "--%10s:  %7.4f\n", "n2[1]", n2[1] ) ;
	PRINT_V( "origin[1][0]", origin[1][0] ) ;
	printf( "--%10s:  %7.4f\n", "d[1][0]", d[1][0] ) ;
	PRINT_V( "origin[1][1]", origin[1][1] ) ;
	printf( "--%10s:  %7.4f\n", "d[1][1]", d[1][1] ) ;
	PRINT_V( "N[2]", N[2] ) ;
	printf( "--%10s:  %7.4f\n", "n2[2]", n2[2] ) ;
	PRINT_V( "origin[2][0]", origin[2][0] ) ;
	printf( "--%10s:  %7.4f\n", "d[2][0]", d[2][0] ) ;
	PRINT_V( "origin[2][1]", origin[2][1] ) ;
	printf( "--%10s:  %7.4f\n", "d[2][1]", d[2][1] ) ;
#undef PRINT_V
}
