//------------------------------------------------------------------------------
// Class for cruise missiles.
//------------------------------------------------------------------------------

#include <Inventor/SbTime.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoSeparator.h>

#include "FollowTerrainAction.h"
#include "LosAction.h"
#include "CruiseMissile.h"
#include "Utilities.h"
#include "Constants.h"

#include "CruiseMissileGeom.h"

SO_NODE_SOURCE( CruiseMissile ) ;

SoSeparator		*CruiseMissile::geometry = NULL ;
SoSeparator		*CruiseMissile::icon = NULL ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
CruiseMissile::initClass()
{
	if( Missile::getClassTypeId() == SoType::badType() )
	{
		Missile::initClass() ;
	}
	SO_NODE_INIT_CLASS( CruiseMissile, Missile, "Missile" ) ;

	geometry = new SoSeparator ;
	geometry->ref() ;

	SoComplexity *comp = new SoComplexity ;
	geometry->addChild( comp ) ;
	comp->type = SoComplexity::OBJECT_SPACE ;
	comp->value = 0.2f ;

	SoCylinder *cyl = new SoCylinder ;
	geometry->addChild( cyl ) ;
	cyl->height = Constants::MissileLength ;
	cyl->radius = Constants::MissileRadius ;

	icon = Utilities::readAll( "icons/cruise_missile.iv", FALSE ) ;
	if( icon == NULL )
	{
		SoInput input ;
		input.setBuffer( (void *)defaultGeom, sizeof( defaultGeom ) ) ;

		icon = SoDB::readAll( &input ) ;

		if( icon == NULL )
		{
			SoDebugError::post( "CruiseMissile", "can not create "
				"icon geometry" ) ;
			exit( 1 ) ;
		}
	}
	icon->ref() ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
CruiseMissile::CruiseMissile()
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Alternate constructor specifying the desired height.
//------------------------------------------------------------------------------
CruiseMissile::CruiseMissile(
	float	height
	) :
	Missile( height )
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common code for constructors.
//------------------------------------------------------------------------------
void
CruiseMissile::buildCommon()
{
	SO_NODE_CONSTRUCTOR( CruiseMissile ) ;
	SO_NODE_ADD_FIELD( maxPitchRate, ( 2.5f ) ) ;

	status = Missile::ACTIVE ;
	fta = new FollowTerrainAction ;

	children.truncate( 0 ) ;
	children.append( geometry ) ;

	maxSpeed = maxSpeed * 0.75f ;

	buildTime = 8.0f ;

	damageAmount = 6.0f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
CruiseMissile::~CruiseMissile()
{
	delete fta ;
}



//------------------------------------------------------------------------------
// Move the missile and check for collisions.
//------------------------------------------------------------------------------
Missile::MissileStatus
CruiseMissile::update(
	float	,
	float	,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( age > maxAge )
	{
		status = Missile::EXPIRED ;
		blowUp() ;
		return status ;
	}

	SbVec3f	startPos( position.getValue() ) ;

	SbRotation	rot = rotation.getValue() ;

	// -Z vector
	SbVec3f ahead ;
	Utilities::setForwardVector( rot, ahead ) ;

	float	forward = maxSpeed * deltaTime ;

	SbVec3f &curPos = startPos + forward * ahead ;

	fta->setSearchPoint( curPos ) ;
	fta->apply( scene ) ;

	const GroundPointList &gpl = fta->getGroundPointList() ;

	int nHits = gpl.getLength() ;

	if( nHits > 0 )
	{
		int n = 0 ;

		while( ( n < nHits - 1 ) &&
			gpl[n]->getHeight() > Constants::MissileRadius )
		{
			n++ ;
		}

		float h = gpl[n]->getHeight() ;

		// Want height to be desiredHeight.  Adjust rotation so that
		// it will be that height in 1/20 of a second.
		float x = forward ;
		float y = desiredHeight - h ;

		float theta = atan2f( y, x ) ;
		float maxTheta = maxPitchRate.getValue() * deltaTime ;
		if( theta > maxTheta )
		{
			theta = maxTheta ;
		}
		else if( theta < -maxTheta )
		{
			theta = -maxTheta ;
		}

		// Try to dampen out the turn if going against the normal.
		float dp = gpl[n]->getNormal().dot( ahead ) ;
		if( y > 0.0f && dp < 0.0f )
		{
			theta /= 1.0f - 2.0f * dp ;
		}

		if( ahead[2] > 0.999f || ahead[2] < -0.999f )
		{
			rot = SbRotation( SbVec3f( -1.0f, 0.0, 0.0f ), theta )
				* rot ;
		}
		else
		{
			SbVec3f &axis = SbVec3f( 0.0f, 0.0f, 1.0f ).
						cross( ahead ) ;
			rot = rot * SbRotation( axis, theta ) ;
		}

		Utilities::setForwardVector( rot, ahead ) ;

		forward = maxSpeed * deltaTime ;

		curPos = startPos + forward * ahead ;
	}

	age += deltaTime ;

	los->setRay( startPos, curPos ) ;
	los->apply( scene ) ;

	if( los->isObstructed() )
	{
		blowUp() ;
		curPos = los->getEndPoint() ;
		status = Missile::STRUCK ;
	}

	position.setValue( curPos ) ;
	rotation.setValue( rot ) ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Return the missile type (this is a virtual method wrapper around the static
// GetClassTypeId).
//------------------------------------------------------------------------------
Missile::MissileType
CruiseMissile::getMissileType() const
{
	return( Missile::CRUISE_MISSILE ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the icon scene graph.
//------------------------------------------------------------------------------
SoNode *
CruiseMissile::getIconGraph()
{
	return( CruiseMissile::icon ) ;
}
