//------------------------------------------------------------------------------
// DamageControl class.
//------------------------------------------------------------------------------

#include <Inventor/SoDB.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include "Utilities.h"
#include "DamageControl.h"



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
DamageControl::DamageControl()
{
	createDamageDisplay() ;
	restoreArmor( 1.0f ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
DamageControl::~DamageControl()
{
	damageDisplay->unref() ;
}



//------------------------------------------------------------------------------
// Create the damage panel display.
//------------------------------------------------------------------------------
void
DamageControl::createDamageDisplay()
{
	SoInput	*in = Utilities::createInput() ;

	static const char *filename = "graphs/damage_display.iv" ;

	if( !in->openFile( filename, TRUE ) )
	{
		SoReadError::post( in, "Can't open %s", filename ) ;
		exit( -1 ) ;
	}

	damageDisplay = SoDB::readAll( in ) ;
	in->closeFile() ;

	if( damageDisplay == NULL )
	{
		SoReadError::post( in, "Can't read %s", filename ) ;
		exit( -1 ) ;
	}

	damageDisplay->ref() ;

	SbString name( "ArmorLeftIndicator" ) ;
	armorLeft = (SoTriangleStripSet *)Utilities::getNamedNode(
				name.getString(),
				SoTriangleStripSet::getClassTypeId() ) ;
	if( armorLeft == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"TriangleStripSet node named `%s'\n",
			filename, name.getString() ) ;
		exit( -1 ) ;
	}

	name = "ArmorGoneIndicator" ;
	armorGone = (SoTriangleStripSet *)Utilities::getNamedNode(
				name.getString(),
				SoTriangleStripSet::getClassTypeId() ) ;
	if( armorGone == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"TriangleStripSet node named `%s'\n",
			filename, name.getString() ) ;
		exit( -1 ) ;
	}

	name = "ArmorStatusColor" ;
	armorColor = (SoBaseColor *)Utilities::getNamedNode(
				name.getString(),
				SoBaseColor::getClassTypeId() ) ;
	if( armorColor == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"BaseColor node named `%s'\n", filename,
			name.getString() ) ;
		exit( -1 ) ;
	}

	delete in ;
}



//------------------------------------------------------------------------------
// Take away armor and update the display.  Return the amount of armor left.
//------------------------------------------------------------------------------
float
DamageControl::sufferDamage(
	float	amount
	)
{
	armor -= amount ;

	updateDisplay() ;

	return( armor ) ;
}



//------------------------------------------------------------------------------
// Repair armor and update the display.
//------------------------------------------------------------------------------
void
DamageControl::repair(
	float	amount
	)
{
	armor += amount ;

	if( armor > maxArmor )
	{
		armor = maxArmor ;
	}

	updateDisplay() ;
}



//------------------------------------------------------------------------------
// Update the armor display.
//------------------------------------------------------------------------------
void
DamageControl::updateDisplay()
{
	int	newLevel ;
	static int32_t	fours[10] = { 4, 4, 4, 4, 4, 4, 4, 4, 4, 4 } ;

	if( armor <= 0.0f )
	{
		newLevel = 0 ;
	}
	else
	{
		newLevel = (int)( 10.0f * ( armor / maxArmor ) + 0.5f ) ;
	}

	if( newLevel != damageLevel )
	{
		if( newLevel <= 0 )
		{
			newLevel = 0 ;
			armorLeft->numVertices.setValue( 0 ) ;
			armorGone->numVertices.setNum( 10 ) ;
			armorGone->numVertices.setValues( 0, 10, fours ) ;
		}
		else if( newLevel >= 10 )
		{
			newLevel = 10 ;
			armorLeft->numVertices.setNum( 10 ) ;
			armorLeft->numVertices.setValues( 0, 10, fours ) ;
			armorGone->numVertices.setValue( 0 ) ;
		}
		else
		{
			armorLeft->numVertices.setNum( newLevel ) ;
			armorLeft->numVertices.setValues( 0, newLevel, fours ) ;
			armorGone->numVertices.setNum( 10 - newLevel ) ;
			armorGone->numVertices.setValues( 0, 10 - newLevel,
							fours ) ;
		}
		armorGone->startIndex.setValue( 4 * newLevel ) ;

		if( newLevel > 5 )
		{
			armorColor->rgb.setValue( SbColor( 0.0f, 0.5f, 0.0f ) );
		}
		else if( newLevel > 2 )
		{
			armorColor->rgb.setValue( SbColor( 0.5f, 0.5f, 0.0f ) );
		}
		else
		{
			armorColor->rgb.setValue( SbColor( 0.5f, 0.0f, 0.0f ) );
		}

		damageLevel = newLevel ;
	}
}



//------------------------------------------------------------------------------
// Restore the armor to its maximum value.
//------------------------------------------------------------------------------
void
DamageControl::restoreArmor(
	float	maxDamage
	)
{
	armor = maxDamage ;
	maxArmor = maxDamage ;
	damageLevel = -1 ;
	updateDisplay() ;
}
