//------------------------------------------------------------------------------ // Convenience class for managing the dashboard.
//------------------------------------------------------------------------------

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <Inventor/SoDB.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoText3.h>

#include "DamageControl.h"
#include "RadarControl.h"
#include "WeaponControl.h"
#include "Dashboard.h"
#include "Speedometer.h"
#include "Utilities.h"



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Dashboard::Dashboard()
{
	SoInput	*in = Utilities::createInput() ;

	if( !in->openFile( "graphs/cockpit.iv", TRUE ) )
	{
		SoReadError::post( in, "Can't open cockpit.iv" ) ;
		exit( -1 ) ;
	}

	root = SoDB::readAll( in ) ;

	if( root == NULL )
	{
		SoReadError::post( in, "Can't read cockpit.iv" ) ;
		exit( -1 ) ;
	}

	root->ref() ;

	//
	// Set up radar control
	//
	initRadarControl( in ) ;

	//
	// Set up weapon control
	//
	initWeaponControl( in ) ;

	//
	// Set up damage control
	//
	initDamageControl( in ) ;

	//
	// Set up the score display.
	//
	initScoreDisplay( in ) ;

	//
	// Set up the number of players display.
	//
	initNumPlayersDisplay( in ) ;

	//
	// Set up the message display.
	//
	initMessageDisplay( in ) ;

	//
	// Set up the speedometer.
	//
	initSpeedometer( in ) ;

	//
	// Set up the compass.
	//
	initCompass( in ) ;

	in->closeFile() ;

	delete in ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Dashboard::~Dashboard()
{
	root->unref() ;
	delete radarControl ;
}



//------------------------------------------------------------------------------
// Initialize the radar control.
//------------------------------------------------------------------------------
void
Dashboard::initRadarControl(
	SoInput	*in
	)
{
	const char *name = "RadarHolder" ;

	SoGroup *radarHolder = (SoGroup *)Utilities::getNamedNode( name,
						SoGroup::getClassTypeId() ) ;
	if( radarHolder == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a group node named `%s'", name ) ;
		exit( -1 ) ;
	}

	radarControl = new RadarControl ;
	radarControl->setRange( RadarControl::EXTRA_FAR ) ;

	radarHolder->addChild( radarControl->getSceneGraph() ) ;
}



//------------------------------------------------------------------------------
// Initialize the weapon control.
//------------------------------------------------------------------------------
void
Dashboard::initWeaponControl(
	SoInput	*in
	)
{
	const char *name = "WeaponPanelHolder" ;

	SoGroup *weaponHolder = (SoGroup *)Utilities::getNamedNode( name,
						SoGroup::getClassTypeId() ) ;
	if( weaponHolder == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a group node named `%s'", name ) ;
		exit( -1 ) ;
	}

	weaponControl = new WeaponControl ;
	weaponHolder->addChild( weaponControl->getWeaponDisplay() ) ;
}



//------------------------------------------------------------------------------
// Initialize the damage control.
//------------------------------------------------------------------------------
void
Dashboard::initDamageControl(
	SoInput	*in
	)
{
	const char *name = "DamagePanelHolder" ;

	SoGroup *damageHolder = (SoGroup *)Utilities::getNamedNode( name,
						SoGroup::getClassTypeId() ) ;
	if( damageHolder == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a group node named `%s'", name ) ;
		exit( -1 ) ;
	}

	damageControl = new DamageControl ;
	damageHolder->addChild( damageControl->getDamageDisplay() ) ;
}



//------------------------------------------------------------------------------
// Initialize the score indicator.
//------------------------------------------------------------------------------
void
Dashboard::initScoreDisplay(
	SoInput	*in
	)
{
	const char *name = "ScoreHolder" ;

	scoreHolder = (SoText3 *)Utilities::getNamedNode( name,
						SoText3::getClassTypeId() ) ;
	if( scoreHolder == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a text3 node named `%s'", name ) ;
		exit( -1 ) ;
	}
}



//------------------------------------------------------------------------------
// Update the score display.
//------------------------------------------------------------------------------
void
Dashboard::updateScoreDisplay(
	int	score
	)
{
	char buf[20] ;

	sprintf( buf, "%+05d", score ) ;

	scoreHolder->string.setValue( buf ) ;
}



//------------------------------------------------------------------------------
// Initialize the number of players indicator.
//------------------------------------------------------------------------------
void
Dashboard::initNumPlayersDisplay(
	SoInput	*in
	)
{
	const char *name = "NumPlayersHolder" ;

	numPlayersHolder = (SoText3 *)Utilities::getNamedNode( name,
						SoText3::getClassTypeId() ) ;
	if( numPlayersHolder == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a text3 node named `%s'", name ) ;
		exit( -1 ) ;
	}
}



//------------------------------------------------------------------------------
// Update the score display.
//------------------------------------------------------------------------------
void
Dashboard::updateNumPlayers(
	int	val
	)
{
	char buf[20] ;

	sprintf( buf, "%02d", val ) ;

	numPlayersHolder->string.setValue( buf ) ;
}



//------------------------------------------------------------------------------
// Initialize the speedometer.
//------------------------------------------------------------------------------
void
Dashboard::initSpeedometer(
	SoInput	*in
	)
{
	const char *name = "SpeedometerHolder" ;

	SoGroup *s = (SoGroup *)Utilities::getNamedNode( name,
						SoGroup::getClassTypeId() ) ;
	if( s == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a group node named `%s'", name ) ;
		exit( -1 ) ;
	}

	speedometer = new Speedometer ;
	s->addChild( speedometer->getDisplay() ) ;
}



//------------------------------------------------------------------------------
// Initialize the compass.
//------------------------------------------------------------------------------
void
Dashboard::initCompass(
	SoInput	*in
	)
{
	const char *name = "HeadingHolder" ;

	headingHolder = (SoText3 *)Utilities::getNamedNode( name,
						SoText3::getClassTypeId() ) ;
	if( headingHolder == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a text3 node named `%s'", name ) ;
		exit( -1 ) ;
	}

	heading = -1 ;
	updateHeading( 0 ) ;
}



//------------------------------------------------------------------------------
// Initialize the message display.
//------------------------------------------------------------------------------
void
Dashboard::initMessageDisplay(
	SoInput	*in
	)
{
	const char *name = "MessageHolder" ;

	messages = (SoText3 *)Utilities::getNamedNode( name,
						SoText3::getClassTypeId() ) ;
	if( messages == NULL )
	{
		SoReadError::post( in, "File `cockpit.iv' does not contain"
			" a text3 node named `%s'", name ) ;
		exit( -1 ) ;
	}

	maxMessages = messages->string.getNum() ;
}



//------------------------------------------------------------------------------
// Add a message to the message display.
//------------------------------------------------------------------------------
void
Dashboard::addMessage(
	char *fmt,
	...
	)
{
	va_list args ;
	static char buffer[512] ;

	va_start( args, fmt ) ;

	vsprintf( buffer, fmt, args ) ;

	va_end( args ) ;

	messages->string.setNum( maxMessages - 1 ) ;
	messages->string.insertSpace( 0, 1 ) ;
	messages->string.set1Value( 0, SbString( buffer ) ) ;
}



//------------------------------------------------------------------------------
// Update the heading display.
//------------------------------------------------------------------------------
void
Dashboard::updateHeading(
	int newHeading
	)
{
	if( heading == newHeading )
	{
		return ;
	}

	char buf[4] ;
	buf[0] = '0' + ( newHeading / 100 ) % 10 ;
	buf[1] = '0' + ( newHeading / 10 ) % 10 ;
	buf[2] = '0' + newHeading % 10 ;
	buf[3] = '\0' ;

	heading = newHeading ;
	headingHolder->string.setValue( buf ) ;
}
