//------------------------------------------------------------------------------
// A class to manage the debris.
//------------------------------------------------------------------------------

#include <Inventor/SoDB.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoShapeHints.h>
#include "AnimationMaster.h"
#include "TransformSwitch.h"
#include "Utilities.h"
#include "Debris.h"
#include "DebrisControl.h"



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
DebrisControl::DebrisControl()
{
	root = new SoSeparator ;
	root->ref() ;
        root->renderCaching = SoSeparator::OFF ;
        root->renderCulling = SoSeparator::OFF ;
        root->boundingBoxCaching = SoSeparator::OFF ;
        root->pickCulling = SoSeparator::OFF ;

	SoShapeHints *hints = new SoShapeHints ;
	root->addChild( hints ) ;
	hints->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE ;
	hints->shapeType = SoShapeHints::UNKNOWN_SHAPE_TYPE ;
	hints->faceType = SoShapeHints::CONVEX ;

	master = new AnimationMaster ;
	root->addChild( master ) ;
	master->speed.setValue( 1.0f / 20.0f ) ;

	animation = NULL ;
	SoNode *node ;
	SoInput *in = Utilities::createInput() ;
	if( in->openFile( "graphs/debris_group.iv" ) &&
		SoDB::read( in, node ) &&
		node->isOfType( SoGroup::getClassTypeId() ) )
	{
		animation = (SoGroup *)node ;
		animation->ref() ;
	}
	else
	{
		SoReadError::post( in, "Problem reading %s",
					in->getCurFileName() );
		exit( -1 ) ;
	}

	in->closeFile() ;
	delete in ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
DebrisControl::~DebrisControl()
{
	root->unref() ;
	if( animation )
	{
		animation->unref() ;
	}
}



//------------------------------------------------------------------------------
// Add an explosion.
//------------------------------------------------------------------------------
void
DebrisControl::addDebris(
	const SbVec3f	&location,
	float		armorValue
	)
{
	Debris *debris = new Debris ;

	master->addChild( debris ) ;

	for( int i = 0 ; i < animation->getNumChildren() ; i++ )
	{
		debris->addChild( animation->getChild( i ) ) ;
	}

	debris->value = armorValue ;
	debris->position.setValue( location ) ;
	debris->rotation.setValue( SbVec3f( 0.0f, 0.0f, 1.0f ), 0.0f ) ;
}



//------------------------------------------------------------------------------
// Update all of the debris so that it is laying on the ground.
//------------------------------------------------------------------------------
void
DebrisControl::update(
	float	dt,
	SoNode	*scene
	)
{
	for( int i = 0 ; i < master->getNumChildren() ; i++ )
	{
		Debris *debris = (Debris *)master->getChild( i ) ;

		debris->update( dt, scene ) ;
	}
}



//------------------------------------------------------------------------------
// Check to see if in range to scavenge.
//------------------------------------------------------------------------------
float
DebrisControl::scavenge(
	const SbVec3f	&pos
	) const
{
	float	amount = 0.0f ;

	for( int i = 0 ; i < master->getNumChildren() ; i++ )
	{
		Debris *debris = (Debris *)master->getChild( i ) ;

		SbVec3f &dis = pos - debris->position.getValue() ;

		if( dis.dot( dis ) < 1.0f )
		{
			float mult = debris->value.getValue() * 0.067f ;
			amount += ( 1.0f - (float)debris->whichChild.getValue()
				/ ( (float)debris->getNumChildren() ) ) * mult ;
		}
	}

	return( amount ) ;
}
