//------------------------------------------------------------------------------
// DestructibleObstacle class.
//------------------------------------------------------------------------------

#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoSeparator.h>
#include "CollisionAction.h"
#include "LosAction.h"
#include "DestructibleObstacle.h"

SO_NODE_SOURCE( DestructibleObstacle ) ;


//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
DestructibleObstacle::initClass()
{
	if( StaticObstacle::getClassTypeId() == SoType::badType() )
	{
		StaticObstacle::initClass() ;
	}
	SO_NODE_INIT_CLASS( DestructibleObstacle, StaticObstacle,
				"StaticObstacle" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
DestructibleObstacle::DestructibleObstacle()
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
DestructibleObstacle::DestructibleObstacle(
	const char *name
	) :
	StaticObstacle( name )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
DestructibleObstacle::createIt()
{
	SO_NODE_CONSTRUCTOR( DestructibleObstacle ) ;
	SO_NODE_ADD_FIELD( armor, ( 1.0f ) ) ;
	view = NULL ;
	totalDamage = 0 ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
DestructibleObstacle::~DestructibleObstacle()
{
	if( view )
	{
		view->unref() ;
	}
}



//------------------------------------------------------------------------------
// Process the geometry to add in a switch node.
//------------------------------------------------------------------------------
void
DestructibleObstacle::processGeometry()
{
	StaticObstacle::processGeometry() ;

	if( root->getNumChildren() != 1 )
	{
		SoDebugError::post( getClassTypeId().getName().getString(),
			"geometry file, %s,\nresults in wrong number of "
			"children (%d instead of 1)\n"
			"for the object's geometry root",
			name.getValue().getString(), root->getNumChildren() ) ;
		exit( -1 ) ;
	}

	SoSearchAction sa ;
	sa.setName( "StatusSwitch" ) ;
	sa.setInterest( SoSearchAction::FIRST ) ;
	sa.setSearchingAll( FALSE ) ;

	sa.apply( root ) ;

	SoPath *path = sa.getPath() ;

	//
	// If geometry does not contain a switch named "StatusSwitch" for
	// displaying a live and a destroyed representation of the battery,
	// then put the root under a switch that can be turned off.
	//
	if( path == NULL ||
		!path->getTail()->isOfType( SoSwitch::getClassTypeId() ) ||
		( (SoSwitch *)( path->getTail() ) )->getNumChildren() < 1 )
	{
		view = new SoSwitch ;
		view->ref() ;
		view->addChild( root->getChild( 0 ) ) ;
		root->removeAllChildren() ;
		root->addChild( view ) ;
		view->whichChild = 0 ;
	}
	else
	{
		view = (SoSwitch *)path->getTail() ;
		view->ref() ;
	}
}



//------------------------------------------------------------------------------
// Destroy the object.
//------------------------------------------------------------------------------
SbBool
DestructibleObstacle::destroy(
	float damage
	)
{
	if( view && view->whichChild.getValue() == 0 )
	{
		totalDamage += damage ;
		if( totalDamage > armor.getValue() )
		{
			if( view->getNumChildren() > 1 )
			{
				view->whichChild = 1 ;
				calculateCorners() ;
			}
			else
			{
				view->whichChild = -1 ;
			}
			return( TRUE ) ;
		}
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// CollisionAction handler.
//------------------------------------------------------------------------------
void
DestructibleObstacle::collide(
	CollisionAction *action
	)
{
	if( view && view->whichChild.getValue() != -1 )
	{
		StaticObstacle::collide( action ) ;
	}
}



//------------------------------------------------------------------------------
// LosAction handler.
//------------------------------------------------------------------------------
void
DestructibleObstacle::obstruct(
	LosAction *action
	)
{
	if( view && view->whichChild.getValue() != -1 )
	{
		StaticObstacle::obstruct( action ) ;
	}
}
