//------------------------------------------------------------------------------
// A class to manage the explosions.
//------------------------------------------------------------------------------

#include <GL/gl.h>
#include <Inventor/SoDB.h>
#include <Inventor/actions/SoAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoShapeHints.h>
#include "AnimationMaster.h"
#include "TransformSwitch.h"
#include "Vehicle.h"
#include "SfxPoint.h"
#include "Utilities.h"
#include "ExplosionControl.h"



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
ExplosionControl::ExplosionControl()
{
	root = new SoSeparator ;
	root->ref() ;
	root->renderCaching = SoSeparator::OFF ;
	root->renderCulling = SoSeparator::OFF ;
	root->boundingBoxCaching = SoSeparator::OFF ;
	root->pickCulling = SoSeparator::OFF ;

	SoShapeHints *hints = new SoShapeHints ;
	root->addChild( hints ) ;
	hints->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE ;
	hints->shapeType = SoShapeHints::SOLID ;
	hints->faceType = SoShapeHints::CONVEX ;

	SoLightModel *lm = new SoLightModel ;
	root->addChild( lm ) ;
	lm->model = SoLightModel::BASE_COLOR ;

	SoMaterialBinding *mb = new SoMaterialBinding ;
	root->addChild( mb ) ;
	mb->value = SoMaterialBinding::PER_VERTEX_INDEXED ;

	SoCallback *cb = new SoCallback() ;
	root->addChild( cb ) ;
	cb->setCallback( setDepthMask, 0 ) ;

	master = new AnimationMaster ;
	root->addChild( master ) ;
	master->speed.setValue( 1.0f ) ;

	cb = new SoCallback() ;
	root->addChild( cb ) ;
	cb->setCallback( setDepthMask, (void *)1 ) ;

	animation = NULL ;
	SoNode *node ;
	SoInput *in = Utilities::createInput() ;
	if( in->openFile( "graphs/explosion_group.iv" ) &&
		SoDB::read( in, node ) &&
		node->isOfType( SoGroup::getClassTypeId() ) )
	{
		animation = (SoGroup *)node ;
		animation->ref() ;
	}
	else
	{
		SoReadError::post( in, "Problem reading %s",
					in->getCurFileName() );
		exit( -1 ) ;
	}

	smallAnimation = new SoGroup ;
	smallAnimation->ref() ;

	SoScale *scale = new SoScale ;
	scale->ref() ;
	scale->scaleFactor.setValue( 0.5f, 0.5f, 0.5f ) ;

	for( int i = 0 ; i < animation->getNumChildren() ; i++ )
	{
		SoSeparator *sep = new SoSeparator ;
		smallAnimation->addChild( sep ) ;

		sep->addChild( scale ) ;
		sep->addChild( animation->getChild( i ) ) ;
	}

	scale->unref() ;

	in->closeFile() ;

	camera = NULL ;

	vehicle = NULL ;

	bang[0] = new SfxPoint( "explosion.aiff" ) ;
	bang[0]->setHalfVolumeDistance( 5.0f ) ;
	for( i = 1 ; i < MaxBangs ; i++ )
	{
		bang[i] = new SfxPoint( *bang[0] ) ;
	}
	nextBang = 0 ;

	delete in ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
ExplosionControl::~ExplosionControl()
{
	root->unref() ;
	animation->unref() ;
	smallAnimation->unref() ;

	if( camera )
	{
		camera->unref() ;
	}
	if( vehicle )
	{
		vehicle->unref() ;
	}
}



//------------------------------------------------------------------------------
// Set the camera location for semi-billboarding the explosions.
//------------------------------------------------------------------------------
void
ExplosionControl::setCamera(
	SoCamera *newCamera
	)
{
	newCamera->ref() ;

	if( camera )
	{
		camera->unref() ;
	}

	camera = newCamera ;
}



//------------------------------------------------------------------------------
// Set the vehicle the listener is in so the sounds will sound like they are
// coming from the correct location.
//------------------------------------------------------------------------------
void
ExplosionControl::setVehicle(
	Vehicle	*v
	)
{
	v->ref() ;

	if( vehicle )
	{
		vehicle->unref() ;
	}

	vehicle = v ;
}



//------------------------------------------------------------------------------
// Add an explosion.
//------------------------------------------------------------------------------
void
ExplosionControl::addExplosion(
	const SbVec3f	&location
	)
{
	TransformSwitch *ts = new TransformSwitch ;

	master->addChild( ts ) ;

	for( int i = 0 ; i < animation->getNumChildren() ; i++ )
	{
		ts->addChild( animation->getChild( i ) ) ;
	}

	ts->position.setValue( location ) ;
	ts->rotation.connectFrom( &camera->orientation ) ;

	SbVec3f d ;
	if( vehicle )
	{
		vehicle->rotation.getValue().inverse().multVec( 
			location - vehicle->position.getValue(), d ) ;
	}
	else
	{
		d = location ;
	}

	if( d.dot( d ) < 1000.0f )
	{
		bang[nextBang]->play( d[0], d[1], d[2] ) ;
		nextBang = ( nextBang + 1 ) % MaxBangs ;
	}
}



//------------------------------------------------------------------------------
// Add an small explosion.
//------------------------------------------------------------------------------
void
ExplosionControl::addBulletExplosion(
	const SbVec3f	&location
	)
{
	TransformSwitch *ts = new TransformSwitch ;

	master->addChild( ts ) ;

	for( int i = 0 ; i < smallAnimation->getNumChildren() ; i++ )
	{
		ts->addChild( smallAnimation->getChild( i ) ) ;
	}

	ts->position.setValue( location ) ;
	ts->rotation.connectFrom( &camera->orientation ) ;
}



//------------------------------------------------------------------------------
// Turn off the zbuffer.
//------------------------------------------------------------------------------
void
ExplosionControl::setDepthMask(
	void		*userData,
	SoAction	*action
	)
{
	if( action->isOfType( SoGLRenderAction::getClassTypeId() ) )
	{
		glDepthMask( userData != NULL ) ;
	}
}



