//------------------------------------------------------------------------------
// Flyer class.
//------------------------------------------------------------------------------

#include <Inventor/fields/SoMFString.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoInfo.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <X11/keysym.h>

#include "FollowTerrainAction.h"
#include "CollisionAction.h"
#include "ZbufferMode.h"
#include "Missile.h"
#include "Utilities.h"
#include "Flyer.h"


SO_NODE_SOURCE( Flyer ) ;


const float minThrust = 0.4f ;

//------------------------------------------------------------------------------
// Initialize the Flyer class.
//------------------------------------------------------------------------------
void
Flyer::initClass()
{
	if( Tank::getClassTypeId() == SoType::badType() )
	{
		Tank::initClass() ;
	}
	SO_NODE_INIT_CLASS( Flyer, Tank, "Tank" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Flyer::Flyer() :
	Tank( "models/flyer.iv" )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
Flyer::Flyer(
	const char *name
	) :
	Tank( name )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
Flyer::createIt()
{
	SO_NODE_CONSTRUCTOR( Flyer ) ;

	setParams() ;
	thrust = minThrust ;
	leftEngine = thrust ;
	rightEngine = thrust ;
	addFirstPersonGraphics() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Flyer::~Flyer()
{
}



//------------------------------------------------------------------------------
// Set up parameters.
//------------------------------------------------------------------------------
void
Flyer::setParams()
{
	maxSpeed = 5.0f ;
	maxTurnRate = 1.5f ;
	maxAcceleration = 4.0f ;
	maxDamage = 0.5f ;
	numBays = 3 ;

	numWeapons[WeaponControl::MAIN_POD] = 1 ;
	weaponList[WeaponControl::MAIN_POD].truncate( 0 ) ;
	weaponList[WeaponControl::MAIN_POD].append( Missile::CANNON );
	numWeapons[WeaponControl::AA_POD] = 1 ;
	weaponList[WeaponControl::AA_POD].truncate( 0 ) ;
	weaponList[WeaponControl::AA_POD].append( Missile::BOMB );

	bayPod = WeaponControl::AA_POD ;

	modelLabel = "Flyer" ;

	conView.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ), 0.5f * (float)M_PI ) ;

	points = 75 ;
}



//------------------------------------------------------------------------------
// Indicate that con view is available.
//------------------------------------------------------------------------------
SbBool
Flyer::hasConView() const
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Reset the flyer.
//------------------------------------------------------------------------------
void
Flyer::reset()
{
	SbVec3f pos = position.getValue() ;
	pos[2] += 7.5f ;
	position.setValue( pos ) ;
	onGround = 0 ;
	thrust = 0.5f * ( 1.0f + minThrust ) ;
	leftEngine = thrust ;
	rightEngine = thrust ;
}



//------------------------------------------------------------------------------
// Move the flyer.
//------------------------------------------------------------------------------
float
Flyer::move(
	float	xInput,
	float	yInput,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}
	xInput *= xInput > 0.0f ? xInput : -xInput ;
	yInput *= yInput > 0.0f ? yInput : -yInput ;

	float pitchRate = yInput * deltaTime * 1.25f * thrust ;
	float rollRate = xInput * deltaTime * 1.5f * thrust ;

	SbRotation roll( SbVec3f( 0.0f, 1.0f, 0.0f ), rollRate ) ;
	SbRotation pitch( SbVec3f( -1.0f, 0.0f, 0.0f ), pitchRate ) ;

	// Calculate the position of the tank without user input or environment
	// effects (ie, freefall).
	SbVec3f &newCenterPos = position.getValue() + velocity * deltaTime ;
	SbRotation &freeRot = roll * pitch * rotation.getValue() ;

	SbMatrix mat ;
	mat.setRotate( freeRot ) ;

	// Based on the current orientation, find the side that presents
	// the most space projected onto the XY plane.
	findBottomSide( mat ) ;

	SbVec3f	collisionNormal ;
	SbBool terrainCollision = followTerrain( scene, newCenterPos, mat,
					collisionNormal, deltaTime ) ;
	SbRotation newOrientation( mat ) ;

	ca->setLocation( newCenterPos, newOrientation ) ;
	ca->apply( scene ) ;

	const CollisionList &cList = ca->getCollisionList() ;

	// Handle a collision.
	if( onGround != 0 || terrainCollision || cList.getLength() > 0 )
	{
		position.setValue( newCenterPos[0], newCenterPos[1],
					newCenterPos[2] ) ;
		rotation.setValue( newOrientation ) ;
		radarRotation.setValue( newOrientation ) ;
		return( 100.0f ) ;
	}

	SbVec3f right( mat[0][0], mat[0][1], mat[0][2] ) ;
	SbVec3f ahead( mat[1][0], mat[1][1], mat[1][2] ) ;
	SbVec3f up( mat[2][0], mat[2][1], mat[2][2] ) ;

	// Adjust velocity based on user input.
	float desiredSpeed = thrust * maxSpeed -
			ahead[2] * 0.02f * Constants::Gravity ;
	float accel = maxAcceleration ;

	float forwardSpeed = velocity.dot( ahead ) ;
	float newSpeed = forwardSpeed ;
	if( forwardSpeed < desiredSpeed )
	{
		newSpeed += accel * deltaTime ;
		if( newSpeed > desiredSpeed )
		{
			newSpeed = desiredSpeed ;
		}
	}
	else if( newSpeed > desiredSpeed )
	{
		newSpeed -= accel * deltaTime ;
		if( newSpeed < desiredSpeed )
		{
			newSpeed = desiredSpeed ;
		}
	}
	velocity = newSpeed * ahead ;

	position.setValue( newCenterPos[0], newCenterPos[1], newCenterPos[2] ) ;
	rotation.setValue( newOrientation ) ;
	radarRotation.setValue( newOrientation ) ;

	// Get the height.
	float altitude ;
	char buf[20] ;
	fta->setSearchPoint( newCenterPos ) ;
	fta->apply( scene ) ;
	const GroundPointList &gpl = fta->getGroundPointList() ;
	if( gpl.getLength() > 0 )
	{
		altitude = -gpl[0]->getHeight() ;
		int	iAlt = (int)( altitude * 10.0f ) - 3 ;
		buf[0] = 'A' ;
		buf[1] = 'L' ;
		buf[2] = 'T' ;
		buf[3] = ':' ;
		buf[4] = ' ' ;
		if( iAlt < 0 )
		{
			buf[5] = '0' ;
			buf[6] = '\0' ;
		}
		else
		{
			buf[5] = '0' + ( iAlt / 100 ) % 10 ;
			buf[6] = '0' + ( iAlt / 10 ) % 10 ;
			buf[7] = '0' + ( iAlt ) % 10 ;
			buf[8] = '\0' ;
		}
		altText->string.setValue( buf ) ;
	}
	else
	{
		altitude = 0.0f ;
		altText->string.setValue( "ALT: XXX" ) ;
	}


	float A = -0.1f * Constants::Gravity ; // -2 * Bomb::Grav / 2
	float B = velocity[2] ;
	float C = altitude ;
	float D = B * B - 2.0f * A * C ; // A has factor of 2 built in
	float t ;
	if( D > 0.0f )
	{
		D = sqrtf( D ) ;
		t = ( -B - D ) / A ; // A has factor of 2 built in
	}
	else if( D < 0.0f )
	{
		t = 0.0f ;
	}
	else
	{
		t = -B / A ; // A has factor of 2 built in
	}
	SbVec3f groundZero( velocity[0] * t, velocity[1] * t, -altitude ) ;
	float gzX = right.dot( groundZero ) ;
	float gzY = ahead.dot( groundZero ) ;
	float gzZ = up.dot( groundZero ) ;

	if( gzY < 1.1f )
	{
		// Move off sight.
		bombSite->translation.setValue( 0.0f, -100.0f, 0.0f ) ;
	}
	else
	{
		gzY = 1.1f / gzY ;
		gzX *= gzY ;
		gzZ *= gzY ;
		bombSite->translation.setValue( gzX, 0.0f, gzZ ) ;
	}

	//
	// Determine the roll angle.
	//
	float rollAng = atan2f( right[2], up[2] ) ;
	rollIndicator->angle = rollAng ;

	//
	// Determine the pitch angle.
	//
	float pitchAng = asinf( ahead[2] ) * ( 180.0f / (float)M_PI ) ;
	int iPitchAng = (int)pitchAng ;
	if( iPitchAng >= 0 )
	{
		buf[0] = '0' + ( iPitchAng / 10 ) % 10 ;
		buf[1] = '0' + ( iPitchAng ) % 10 ;
		buf[2] = '\0' ;
	}
	else
	{
		iPitchAng = -iPitchAng ;
		buf[0] = '-' ;
		buf[1] = '0' + ( iPitchAng / 10 ) % 10 ;
		buf[2] = '0' + ( iPitchAng ) % 10 ;
		buf[3] = '\0' ;
	}
	pitchText->string.setValue( buf ) ;

	pitchIndicator->translation.setValue( 0.35f, 0.0f,
			pitchAng * 0.00277778f ) ; // 0.25 / 90.


	return( 0.0f ) ;
}



//------------------------------------------------------------------------------
// Move the con view.
//------------------------------------------------------------------------------
float
Flyer::moveConView(
	float	mx,
	float	my,
	float	dt,
	SoNode	*scene
	)
{
	return( Flyer::move( mx, my, dt, scene ) ) ;
}



//------------------------------------------------------------------------------
// Handle a key press.
//------------------------------------------------------------------------------
SbBool
Flyer::handleKeyPress(
	KeySym	ks
	)
{
	SbBool	handled = TRUE ;

	switch( ks )
	{
		case XK_a :
			thrust -= 0.05f ;
			if( thrust < minThrust )
			{
				thrust = minThrust ;
			}
			leftEngine = thrust ;
			rightEngine = thrust ;
			break ;

		case XK_s :
			thrust += 0.05f ;
			if( thrust > 1.0f )
			{
				thrust = 1.0f ;
			}
			leftEngine = thrust ;
			rightEngine = thrust ;
			break ;

		default :
			handled = FALSE ;
			break ;
	}

	return( handled ) ;
}



//------------------------------------------------------------------------------
// Position a newly launched weapon.
//------------------------------------------------------------------------------
void
Flyer::launchWeapon(
	Missile				*weapon,
	WeaponControl::WeaponPod	pod
	)
{
	SbVec3f	up ;
	SbVec3f	ahead ;
	SbVec3f	right ;
	SbRotation rot = rotation.getValue() ;
	Utilities::setUpVector( rot, up ) ;
	Utilities::setForwardVector( rot, ahead ) ;
	Utilities::setRightVector( rot, right ) ;

	SbVec3f lp ;
	if( pod == WeaponControl::MAIN_POD )
	{
		lp = launchPoint + getCenterOffset() ;
	}
	else
	{
		lp = auxViewpoint + getCenterOffset() -
			SbVec3f( 0.0f, 0.0f, 0.02f ) ;
	}
	SbVec3f pos = position.getValue() + lp[0] * right + lp[1] * ahead +
			lp[2] * up ;

	weapon->position.setValue( pos ) ;
	weapon->rotation.setValue( rot ) ;
	weapon->setVelocity( getVelocity() ) ;
}



//------------------------------------------------------------------------------
// Build the first person graphics.
//------------------------------------------------------------------------------
void
Flyer::addFirstPersonGraphics()
{
	// Main view graphics.
	SoSeparator *mainBase = new SoSeparator ;
	fpGraphics->addChild( mainBase ) ;
	mainBase->renderCulling = SoSeparator::OFF ;
	mainBase->pickCulling = SoSeparator::OFF ;
	mainBase->boundingBoxCaching = SoSeparator::OFF ;
	mainBase->renderCaching = SoSeparator::OFF ;

	SoTranslation *t = new SoTranslation ;
	mainBase->addChild( t ) ;
	t->translation.setValue( mainViewpoint + SbVec3f( 0.0f, 1.1f, 0.0f ) ) ;

	SoGroup *common = new SoGroup ;
	mainBase->addChild( common ) ;

	ZbufferMode *zb = new ZbufferMode ;
	common->addChild( zb ) ;
	zb->value = ZbufferMode::ALWAYS ;

	SoLightModel *lm = new SoLightModel ;
	common->addChild( lm ) ;
	lm->model = SoLightModel::BASE_COLOR ;

	SoBaseColor *bc = new SoBaseColor ;
	common->addChild( bc ) ;
	bc->rgb.setValue( 1.0f, 0.0f, 0.0f ) ;

	SoFont *font = new SoFont ;
	common->addChild( font ) ;
	font->name.setValue( "HelveticaBold" ) ;
	font->size = 15.0f ;

	SoDrawStyle *ds = new SoDrawStyle ;
	common->addChild( ds ) ;
	ds->lineWidth = 3.0f ;

	SoSeparator *altSep = new SoSeparator ;
	common->addChild( altSep ) ;
	altSep->renderCulling = SoSeparator::OFF ;
	altSep->pickCulling = SoSeparator::OFF ;
	altSep->boundingBoxCaching = SoSeparator::OFF ;
	altSep->renderCaching = SoSeparator::OFF ;

	t = new SoTranslation ;
	altSep->addChild( t ) ;
	t->translation.setValue( -0.30f, 0.0f, 0.32f ) ;

	altText = new SoText2 ;
	altSep->addChild( altText ) ;
	altText->justification = SoText2::LEFT ;
	altText->string.setValue( "000" ) ;

	SoSeparator *prSep = new SoSeparator ;
	common->addChild( prSep ) ;
	prSep->renderCulling = SoSeparator::OFF ;
	prSep->pickCulling = SoSeparator::OFF ;
	prSep->boundingBoxCaching = SoSeparator::OFF ;
	prSep->renderCaching = SoSeparator::OFF ;

	rollIndicator = new SoRotationXYZ ;
	prSep->addChild( rollIndicator ) ;
	rollIndicator->axis = SoRotationXYZ::Y ;
	rollIndicator->angle = 0.0f ;

	SoCoordinate3 *c = new SoCoordinate3 ;
	prSep->addChild( c ) ;
	c->point.set1Value( 0, -0.2f, 0.0f, -0.05f ) ;
	c->point.set1Value( 1, -0.2f, 0.0f, 0.0f ) ;
	c->point.set1Value( 2,  0.2f, 0.0f, 0.0f ) ;
	c->point.set1Value( 3,  0.2f, 0.0f, -0.05f ) ;

	SoLineSet *ls = new SoLineSet ;
	prSep->addChild( ls ) ;
	ls->numVertices = 4 ;
	ls->startIndex = 0 ;

	SoSeparator *piSep = new SoSeparator ;
	common->addChild( piSep ) ;
	piSep->renderCulling = SoSeparator::OFF ;
	piSep->pickCulling = SoSeparator::OFF ;
	piSep->boundingBoxCaching = SoSeparator::OFF ;
	piSep->renderCaching = SoSeparator::OFF ;

	c = new SoCoordinate3 ;
	piSep->addChild( c ) ;
	c->point.set1Value( 0, 0.27, 0.0f, 0.25f ) ;
	c->point.set1Value( 1, 0.25, 0.0f, 0.25f ) ;
	c->point.set1Value( 2, 0.25, 0.0f, -0.25f ) ;
	c->point.set1Value( 3, 0.27, 0.0f, -0.25f ) ;
	c->point.set1Value( 4, 0.25, 0.0f, 0.0f ) ;
	c->point.set1Value( 5, 0.27, 0.0f, 0.0f ) ;

	ls = new SoLineSet ;
	piSep->addChild( ls ) ;
	ls->numVertices.set1Value( 0, 4 ) ;
	ls->numVertices.set1Value( 1, 2 ) ;
	ls->startIndex = 0 ;

	pitchIndicator = new SoTranslation ;
	piSep->addChild( pitchIndicator ) ;
	pitchIndicator->translation.setValue( 0.35f, 0.0f, 0.0f ) ;

	pitchText = new SoText2 ;
	piSep->addChild( pitchText ) ;
	pitchText->justification = SoText2::LEFT ;
	pitchText->string.setValue( "00" ) ;

	c = new SoCoordinate3 ;
	piSep->addChild( c ) ;
	c->point.set1Value( 0, -0.06, 0.0f, 0.0f ) ;
	c->point.set1Value( 1, -0.02, 0.0f, 0.025f ) ;
	c->point.set1Value( 2, -0.02, 0.0f, -0.025f ) ;
	c->point.set1Value( 3, -0.06, 0.0f, 0.0f ) ;

	ls = new SoLineSet ;
	piSep->addChild( ls ) ;
	ls->numVertices = 4 ;
	ls->startIndex = 0 ;

	SoSeparator *auxBase = new SoSeparator ;
	fpGraphics->addChild( auxBase ) ;
	auxBase->renderCulling = SoSeparator::OFF ;
	auxBase->pickCulling = SoSeparator::OFF ;
	auxBase->boundingBoxCaching = SoSeparator::OFF ;
	auxBase->renderCaching = SoSeparator::OFF ;

	t = new SoTranslation ;
	auxBase->addChild( t ) ;
	t->translation.setValue( auxViewpoint + SbVec3f( 0.0f, 1.1f, 0.0f ) ) ;

	auxBase->addChild( common ) ;

	ds = new SoDrawStyle ;
	common->addChild( ds ) ;
	ds->lineWidth = 1.0f ;

	bombSite = new SoTranslation ;
	auxBase->addChild( bombSite ) ;
	bombSite->translation.setValue( 0.0f, 0.0f, 0.0f ) ;

	SoSeparator *sep = new SoSeparator ;
	auxBase->addChild( sep ) ;
	sep->renderCulling = SoSeparator::OFF ;
	sep->pickCulling = SoSeparator::OFF ;
	sep->boundingBoxCaching = SoSeparator::OFF ;
	sep->renderCaching = SoSeparator::ON ;

	c = new SoCoordinate3 ;
	sep->addChild( c ) ;
	c->point.set1Value( 0, 0.0f, 0.0f, 0.1f ) ;
	c->point.set1Value( 1, 0.075f, 0.0f, 0.0f ) ;
	c->point.set1Value( 2, 0.0f, 0.0f, -0.1f ) ;
	c->point.set1Value( 3, -0.075f, 0.0f, 0.0f ) ;
	c->point.set1Value( 4, 0.0f, 0.0f, 0.1f ) ;

	ls = new SoLineSet ;
	sep->addChild( ls ) ;
	ls->numVertices = 5 ;
	ls->startIndex = 0 ;
}



//------------------------------------------------------------------------------
// Indicate if it is okay for the vehicle to fire under current conditions
// (neglecting if weapon is ready or not, that's checked elsewhere).
//------------------------------------------------------------------------------
SbBool
Flyer::okayToFire(
	WeaponControl::WeaponPod pod
	)
{
	if( pod == WeaponControl::MAIN_POD )
	{
		return( TRUE ) ;
	}
	else
	{
		float q0, q1, q2, q3 ;
		rotation.getValue().getValue( q0, q1, q2, q3 ) ;
		float upZ = 1.0f - 2.0f * ( q1 * q1 + q0 * q0 ) ;
		return( upZ > 0.1f ) ;
	}
}



//------------------------------------------------------------------------------
// Show vehicle-specific help text.
//------------------------------------------------------------------------------
void
Flyer::showHelp(
	SoMFString	*str,
	int		&n
	)
{
	str->set1Value( n++, "s - Increase speed" ) ;
	str->set1Value( n++, "a - Decrease speed" ) ;
	str->set1Value( n++, "Mouse - control pitch and roll" ) ;
	str->set1Value( n++, "Right mouse - Toggle between Main and "
				"Bombing views" ) ;
	str->set1Value( n++, "" ) ;
	str->set1Value( n++, "MAIN VIEW" ) ;
	str->set1Value( n++, "Left mouse - fire cannon" ) ;
	str->set1Value( n++, "" ) ;
	str->set1Value( n++, "BOMBING VIEW" ) ;
	str->set1Value( n++, "Left mouse - drop bomb" ) ;
	str->set1Value( n++, "Bombs can only be dropped when upright." ) ;
	str->set1Value( n++, "A red diamond indicates the estimated" ) ;
	str->set1Value( n++, "impact site of the bomb." ) ;
}



