//------------------------------------------------------------------------------
// Action class to follow terrain using GroundShape nodes.
//------------------------------------------------------------------------------

#include <assert.h>
#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoComplexityElement.h>
#include <Inventor/elements/SoComplexityTypeElement.h>
#include <Inventor/elements/SoCoordinateElement.h>
#include <Inventor/elements/SoFontNameElement.h>
#include <Inventor/elements/SoFontSizeElement.h>
#include <Inventor/elements/SoProfileCoordinateElement.h>
#include <Inventor/elements/SoProfileElement.h>
#include <Inventor/elements/SoSwitchElement.h>
#include <Inventor/elements/SoUnitsElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/elements/SoViewingMatrixElement.h>
#include <Inventor/elements/SoViewportRegionElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include "GroundShape.h"
#include "FollowTerrainAction.h"


SO_ACTION_SOURCE( FollowTerrainAction ) ;


//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
FollowTerrainAction::initClass()
{
	SO_ACTION_INIT_CLASS( FollowTerrainAction, SoAction ) ;

	// Enable the elements that are involved.
	SO_ENABLE( FollowTerrainAction, SoModelMatrixElement ) ;
	SO_ENABLE( FollowTerrainAction, SoComplexityElement ) ;
	SO_ENABLE( FollowTerrainAction, SoComplexityTypeElement ) ;
	SO_ENABLE( FollowTerrainAction, SoCoordinateElement ) ;
	SO_ENABLE( FollowTerrainAction, SoFontNameElement ) ;
	SO_ENABLE( FollowTerrainAction, SoFontSizeElement ) ;
	SO_ENABLE( FollowTerrainAction, SoProfileCoordinateElement ) ;
	SO_ENABLE( FollowTerrainAction, SoProfileElement ) ;
	SO_ENABLE( FollowTerrainAction, SoSwitchElement ) ;
	SO_ENABLE( FollowTerrainAction, SoUnitsElement ) ;
	SO_ENABLE( FollowTerrainAction, SoViewVolumeElement ) ;
	SO_ENABLE( FollowTerrainAction, SoViewingMatrixElement ) ;
	SO_ENABLE( FollowTerrainAction, SoViewportRegionElement ) ;
	SO_ENABLE( FollowTerrainAction, SoNormalBindingElement ) ;

	// Register static methods fo the action.
	SO_ACTION_ADD_METHOD( SoNode, doNada ) ;

	SO_ACTION_ADD_METHOD( SoGroup, callDoAction ) ;
	SO_ACTION_ADD_METHOD( SoNormalBinding, callDoAction ) ;
	if( GroundShape::getClassTypeId() == SoType::badType() )
	{
		GroundShape::initClass() ;
	}
	SO_ACTION_ADD_METHOD( GroundShape, GroundShape::followTerrainS ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
FollowTerrainAction::FollowTerrainAction() :
	searchPoint( 0.0f, 0.0f, 0.0f )
{
	SO_ACTION_CONSTRUCTOR( FollowTerrainAction ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
FollowTerrainAction::~FollowTerrainAction()
{
	// Empty.
}



//------------------------------------------------------------------------------
// Initiate the action.
//------------------------------------------------------------------------------
void
FollowTerrainAction::beginTraversal(
	SoNode	*node
	)
{
	iList.truncate( 0 ) ;
	traverse( node ) ;
}



//------------------------------------------------------------------------------
// Generic action.
//------------------------------------------------------------------------------
void
FollowTerrainAction::callDoAction(
	SoAction	*action,
	SoNode		*node
	)
{
	node->doAction( action ) ;
}



//------------------------------------------------------------------------------
// Static method for ignoring the action.
//------------------------------------------------------------------------------
void
FollowTerrainAction::doNada(
	SoAction *,
	SoNode	 *
	)
{
}



//------------------------------------------------------------------------------
// Set the search point for the intersection.
//------------------------------------------------------------------------------
void
FollowTerrainAction::setSearchPoint(
	const SbVec3f &worldSpacePoint
	)
{
	searchPoint = worldSpacePoint ;
}



//------------------------------------------------------------------------------
// Add an intersection point.
//------------------------------------------------------------------------------
GroundPoint &
FollowTerrainAction::addIntersection(
	const SbVec3f &worldSpacePoint,
	const SbVec3f &normal
	)
{
	// Assume world space == local space (we want the terrain geometry
	// flattened for speed purposes).
	GroundPoint *gp = new GroundPoint( worldSpacePoint, normal,
						searchPoint ) ;

	int i ;
	for( i = 0 ; i < iList.getLength() ; i++ )
	{
		if( gp->getHeight() > iList[i]->getHeight() )
		{
			break ;
		}
	}

	iList.insert( gp, i ) ;

	return( *gp ) ;
}



//------------------------------------------------------------------------------
// GroundPoint class.
//------------------------------------------------------------------------------
GroundPoint::GroundPoint(
	const SbVec3f	&pt,
	const SbVec3f	&norm,
	const SbVec3f	&searchPoint
	) :
	point( pt ),
	normal( norm )
{
	height = point[2] - searchPoint[2] ;
	distance = height < 0.0f ? -height : height ;
}



//------------------------------------------------------------------------------
// GroundPoint destructor.
//------------------------------------------------------------------------------
GroundPoint::~GroundPoint()
{
}



//------------------------------------------------------------------------------
// Set the polygon information.
//------------------------------------------------------------------------------
void
GroundPoint::setPoly(
	int		nPts,
	const SbVec3f	*pts
	)
{
	assert( nPts <= 4 ) ;

	for( int i = 0 ; i < nPts ; i++ )
	{
		polyPt[i] = pts[i] ;
	}

	numPoints = nPts ;
}



//------------------------------------------------------------------------------
// Remove (and delete) an item from the list.
//------------------------------------------------------------------------------
void
GroundPointList::remove(
	int	which
	)
{
	if( (*this)[which] != NULL )
	{
		delete (*this)[which] ;
	}
	SbPList::remove( which ) ;
}



//------------------------------------------------------------------------------
// Delete all GroundPoints on the list from the given index on.
//------------------------------------------------------------------------------
void
GroundPointList::truncate(
	int	start
	)
{
	for( int i = start ; i < getLength() ; i++ )
	{
		if( (*this)[i] != NULL )
		{
			delete (*this)[i] ;
		}
	}

	SbPList::truncate( start ) ;
}
