//------------------------------------------------------------------------------
// Main game controller.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <ctype.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/SbLinear.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoTransformVec3f.h>
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoInfo.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoRotor.h>
#include <Inventor/nodes/SoText3.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoTranslation.h>
#include "Automaton.h"
#include "Cannon.h"
#include "CockPit.h"
#include "Dashboard.h"
#include "DebrisControl.h"
#include "DestructibleObstacle.h"
#include "ExplosionControl.h"
#include "CruiseMissile.h"
#include "GameScenario.h"
#include "Goal.h"
#include "GuidedMissile.h"
#include "Missile.h"
#include "Bomb.h"
#include "NameList.h"
#include "DamageControl.h"
#include "RadarControl.h"
#include "RotateRotation.h"
#include "Utilities.h"
#include "Vehicle.h"
#include "Flyer.h"
#include "WeaponControl.h"
#include "LosAction.h"
#include "GameIO.h"
#include "GamePackets.h"
#include "GameControl.h"
#include "InterComm.h"
#include "Speedometer.h"
#include "FollowTerrainAction.h"
#include "Sfx.h"
#include "SfxLoop.h"
#include "SfxStereoPitch.h"


static const float	MainViewLeft = 0.04f ;
static const float	MainViewRight = 0.96f ;
static const float	MainViewBottom = 0.4025f ;
static const float	MainViewTop = 0.948f ;
#ifdef VERBOSE
static SbBool		godMode = FALSE ;
#endif // VERBOSE

//------------------------------------------------------------------------------
// Player declaration.
//------------------------------------------------------------------------------
class Player {

    public :
	Player( HostAddr remoteId, Vehicle *v ) ;
	~Player() ;
	void		launchPlayerMissile( int i, Missile::MissileType type,
				WeaponControl *wc, const float *location ) ;
	unsigned short	status ;
	unsigned short	vehicleType ;
	HostAddr	hostAddr ;
	int		score ;
	int		team ;
	int		origTeam ;
	Vehicle		*vehicle ;
	HostAddr	missileStatus[WeaponControl::MAX_WEAPONS] ;
	Missile		*missile[WeaponControl::MAX_WEAPONS] ;

    private :
	Player() ;
} ;



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
GameControl::GameControl(
	CockPit		*cp,
	GameIO		*spigot,
	GameScenario	*scenario
	) :
	name( "GameControl" ),
	mainViewCenter( 0, 0 )
{
	numTeams = scenario->getNumTeams() ;
	mode = scenario->getGameMode() ;

	wind = new SfxLoop( "wind.aifc" ) ;
	wind->play() ;

	//
	// Seed the random number generator.
	//
	srand( (unsigned int)SbTime::getTimeOfDay().getMsecValue() ) ;
	gameTime = 0.0f ;

	gameRoot = NULL ;
	vehicle = NULL ;
	realStuff = NULL ;
	setUpCamera() ;
	cockPit = cp ;
        cockPit->drainEvents() ;
	dBoard = new Dashboard ;
	damageControl = dBoard->getDamageControl() ;
	radarControl = dBoard->getRadarControl() ;
	weaponControl = dBoard->getWeaponControl() ;
	weaponControl->setMissileStrikeHandler(
			(WeaponControlStrikeHandler *)missileStrike, this ) ;
	speedometer = dBoard->getSpeedometer() ;
        cockPit->drainEvents() ;
	la = new LosAction ;
	gio = spigot ;

	gio->addInputPacketType( PLAYER_PKT, sizeof( PlayerPacket ),
			(GamePacketHandler *)processPlayerPacket, this ) ;
	gio->addInputPacketType( INFO_PKT, sizeof( InfoPacket ),
			(GamePacketHandler *)processInfoPacket, this ) ;
	gio->addInputPacketType( GOAL_PKT, sizeof( GoalPacket ),
			(GamePacketHandler *)processGoalPacket, this ) ;

	mouseX = 0.0f ;
	mouseY = 0.0f ;
	mouseWinX = 0 ;
	mouseWinY = 0 ;

        cockPit->drainEvents() ;
	struckPlayer = new Sfx( "struckPlayer.aifc" ) ;
        cockPit->drainEvents() ;
	dieNoise = new Sfx( "dead.aifc" ) ;
        cockPit->drainEvents() ;
	hitNoise = new Sfx( "hit.aifc" ) ;
        cockPit->drainEvents() ;
        cockPit->drainEvents() ;
	engine = new SfxStereoPitch( "engine.aifc" ) ;
        cockPit->drainEvents() ;
	buzzer = new Sfx( "buzzer.aifc" ) ;
        cockPit->drainEvents() ;
	gong = new Sfx( "gong.aifc" ) ;
        cockPit->drainEvents() ;
	hello = new Sfx( "new_player.aifc" ) ;
        cockPit->drainEvents() ;

	topRoot = new SoSeparator ;
	topRoot->ref() ;
	topRoot->renderCaching = SoSeparator::OFF ;

	SoComplexity *comp = new SoComplexity ;
	topRoot->addChild( comp ) ;
	comp->textureQuality = 0.1f ;

	SoEnvironment *env = new SoEnvironment ;
	topRoot->addChild( env ) ;
	env->fogType = SoEnvironment::NONE ;

	modeSwitch = new SoSwitch ;
	topRoot->addChild( modeSwitch ) ;

        cockPit->drainEvents() ;
	setUpVehicleChooser() ;

	setUpTeamChooser() ;
        cockPit->drainEvents() ;

	vehicleChoice = 0 ;

	gio->setDropHostHandler( (DropHostHandler *)droppedHost, this ) ;
	gio->setBanHostHandler( (DropHostHandler *)bannedHost, this ) ;
 
	myScore = 0 ;
	myTeam = 0 ;
	myOrigTeam = 0 ;
	dBoard->updateScoreDisplay( myScore ) ;

	comm = new InterComm( gio ) ;
	comm->setTeam( myTeam ) ;

	scavengeTime = 0.0f ;

	ray = new SoRayPickAction( SbViewportRegion( 1, 1 ) ) ;

	scoresOn = FALSE ;
	helpOn = FALSE ;

	//
	// Create the debris and explosion controllers.
	//
	debrisControl = new DebrisControl ;
	explosionControl = new ExplosionControl ;

	// Initialize the level for the solo game.
	level = 1 ;

	if( !loadGame( scenario->getFilename().getString() ) )
	{
		exit( -1 ) ;
	}

	setUpOverlay() ;

	if( mode == GOAL )
	{
		setUpBreakMessage() ;
		goalRound = -1 ;
		initGoals() ;
		status = INITIALIZING ;
		modeSwitch->whichChild = 4 ;
		vehicleRotor->on = TRUE ;

		for( int i = 0 ; i < numTeams ; i++ )
		{
			teamScores[i] = 0 ;
		}
	}
	else if( mode == SOLO )
	{
		Automaton::setStrikeHandler( automatonStrikeHandler,
			(void *)this ) ;
		Automaton::setLaunchHandler( automatonLaunchHandler,
			(void *)weaponControl ) ;
		setUpBreakMessage() ;
		status = INITIALIZING ;
		modeSwitch->whichChild = 4 ;
		numTanks = 2 ;
		dBoard->updateNumPlayers( numTanks ) ;
	}
	else
	{
		if( numTeams < 2 || mode == SOLO )
		{
			myTeam = numTeams ;
			myOrigTeam = myTeam ;
			comm->setTeam( myTeam ) ;

			startChoosingTank() ;
		}
		else
		{
			status = CHOOSING_TEAM ;
			modeSwitch->whichChild = 1 ;
		}
	}

	buildCaches() ;

	wind->stop() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GameControl::~GameControl()
{
	engine->stop() ;
	wind->stop() ;
	dieNoise->play() ;
	InfoPacket packet( InfoPacket::QUIT_GAME, 0 ) ;
	gio->send( &packet, sizeof( packet ) ) ;
	gio->send( &packet, sizeof( packet ) ) ;
	sleep( 1 ) ;
	gio->send( &packet, sizeof( packet ) ) ;

	if( gameRoot )
	{
		gameRoot->unref() ;
	}
	if( overlayRoot )
	{
		overlayRoot->unref() ;
	}
	if( overlaySwitch )
	{
		overlaySwitch->unref() ;
	}
	if( radarColors )
	{
		radarColors->unref() ;
	}
	if( cam )
	{
		cam->unref() ;
	}
	if( topRoot )
	{
		topRoot->unref() ;
	}

	if( mode == SOLO && vehicleChoices )
	{
		vehicleChoices->unref() ;
	}

	delete comm ;
	delete dBoard ;
	delete la ;
	delete struckPlayer ;
	delete dieNoise ;
	delete hitNoise ;
	delete wind ;
	delete engine ;
	delete ray ;
	delete buzzer ;
	delete gong ;
	delete hello ;
	delete speedometer ;
}



//------------------------------------------------------------------------------
// Load in a game file.
//------------------------------------------------------------------------------
SbBool
GameControl::loadGame(
	const char	*filename
	)
{
	SoInput	*in = Utilities::createInput() ;

	if( !in->openFile( filename, TRUE ) )
	{
		SoReadError::post( in, "%s: file not opened\n", filename ) ;
		return( FALSE ) ;
	}

	gameRoot = SoDB::readAll( in ) ;
	in->closeFile() ;

	if( gameRoot == NULL )
	{
		SoReadError::post( in, "%s: no data read\n", filename ) ;
		delete in ;
		return( FALSE ) ;
	}

	gameRoot->ref() ;

	SoSeparator *oldRealStuff = realStuff ;

        cockPit->drainEvents() ;

	setTeamColors() ;

	if( !checkForCameras() || !setUpDropZones() || !getVehicleTypes() ||
		!findTerrain() || ( mode == GOAL && !setUpGoals() ) )
	{
		gameRoot->unref() ;
		gameRoot = NULL ;
		return( FALSE ) ;
	}

	realStuff->insertChild( cam, 0 ) ;

	if( oldRealStuff )
	{
		modeSwitch->replaceChild( oldRealStuff, realStuff ) ;
	}
	else
	{
		modeSwitch->addChild( realStuff ) ;
	}

	litObjects = new SoSeparator ;
	realStuff->addChild( litObjects ) ;
	litObjects->renderCaching = SoSeparator::OFF ;
	litObjects->renderCulling = SoSeparator::OFF ;
	litObjects->boundingBoxCaching = SoSeparator::OFF ;
	litObjects->pickCulling = SoSeparator::OFF ;

	SoLightModel *lm = new SoLightModel ;
	litObjects->addChild( lm ) ;
	lm->model = SoLightModel::PHONG ;

	vehicleGroup = new SoSeparator ;
	litObjects->addChild( vehicleGroup ) ;
	vehicleGroup->renderCaching = SoSeparator::OFF ;
	vehicleGroup->renderCulling = SoSeparator::OFF ;
	vehicleGroup->boundingBoxCaching = SoSeparator::OFF ;
	vehicleGroup->pickCulling = SoSeparator::OFF ;
	vehicleGroup->setName( "VehicleGroup" ) ;

	SoSeparator *teamSeparator ;
	SoBaseColor *teamHue ;
	//
	// First team is for the neutral players who haven't picked a name
	// yet.
	//
	for( int i = 0 ; i <= numTeams ; i++ )
	{
		teamSeparator = new SoSeparator ;
		vehicleGroup->addChild( teamSeparator ) ;
		teamSeparator->renderCaching = SoSeparator::OFF ;
		teamSeparator->renderCulling = SoSeparator::OFF ;
		teamSeparator->boundingBoxCaching = SoSeparator::OFF ;
		teamSeparator->pickCulling = SoSeparator::OFF ;

		teamHue = new SoBaseColor ;
		teamHue->rgb.setValue( teamColors->rgb[i] ) ;
		teamSeparator->addChild( teamHue ) ;
	}

	missileGroup = weaponControl->getWeaponGroup() ;
	missileGroup->setName( "MissileGroup" ) ;
	litObjects->addChild( missileGroup ) ;

	realStuff->addChild( debrisControl->getSceneGraph() ) ;

	//
	// Make sure the explosion group is last in the scene since it uses
	// transparency.
	//
	explosionControl->setCamera( cam ) ;
	realStuff->addChild( explosionControl->getSceneGraph() ) ;
	weaponControl->setExplosionControl( explosionControl ) ;

        cockPit->drainEvents() ;
	cockPit->clearScenes() ;
	cockPit->addScene( dBoard->getRoot(), 0.0f, 1.0f, 0.0f, 1.0f, FALSE ) ;
	cockPit->addScene( topRoot, MainViewLeft, MainViewRight,
				MainViewBottom, MainViewTop, TRUE ) ;

	delete in ;

	//
	// If in solo mode, find the automatons.
	//
	if( mode == SOLO )
	{
		setUpSoloGame() ;
	}
	//
	// Otherwise, destroy all DestructibleObstacles.
	//
	else
	{
		SoSearchAction sa ;
		sa.setInterest( SoSearchAction::ALL ) ;
		sa.setFind( SoSearchAction::TYPE ) ;
		sa.setType( DestructibleObstacle::getClassTypeId() ) ;
		sa.setSearchingAll( TRUE ) ;
		sa.apply( gameRoot ) ;

		SoPathList &pl = sa.getPaths() ;
		for( int i = 0 ; i < pl.getLength() ; i++ )
		{
			DestructibleObstacle *dob =
				(DestructibleObstacle *)pl[i]->getTail() ;
			dob->destroy( dob->armor.getValue() + 1.0f ) ;
		}

		if( i > 0 )
		{
			SoDebugError::post( name.getString(),
				"DestructibleObstacles are only allowed in the"
				"solo version of the game" ) ;
		}
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Initialize the overlay displays.
//------------------------------------------------------------------------------
void
GameControl::setUpOverlay()
{
	SoInput	*in = Utilities::createInput() ;

	if( !in->openFile( "graphs/overlay.iv", TRUE ) )
	{
		SoReadError::post( in, "Can't open overlay.iv" ) ;
		overlayRoot = NULL ;
	}
	else
	{
		overlayRoot = SoDB::readAll( in ) ;
		if( overlayRoot != NULL )
		{
			overlayRoot->ref() ;
			cockPit->setOverlaySceneGraph( overlayRoot ) ;
		}
		else
		{
			SoReadError::post( in, "Nothing read from "
				"'overlay.iv'" ) ;
			exit( -1 ) ;
		}
	}

	overlaySwitch = (SoSwitch *)Utilities::getNamedNode( "OverlaySwitch",
						SoSwitch::getClassTypeId() ) ;

	if( overlaySwitch == NULL )
	{
		SoReadError::post( in, "File `overlay.iv' does not contain "
			"a switch node named `OverlaySwitch'" ) ;
		exit( -1 ) ;
	}

	overlaySwitch->ref() ;
	overlaySwitch->whichChild = -1 ;

	overlayTextSwitch = (SoSwitch *)Utilities::getNamedNode(
			"OverlayTextSwitch", SoSwitch::getClassTypeId() ) ;

	if( overlayTextSwitch == NULL )
	{
		SoReadError::post( in, "File `overlay.iv' does not contain "
			"a switch node named `OverlayTextSwitch'" ) ;
		exit( -1 ) ;
	}

	overlayTextSwitch->whichChild = -1 ;

	overlayText = (SoText3 *)Utilities::getNamedNode(
			"OverlayText", SoText3::getClassTypeId() ) ;

	if( overlayText == NULL )
	{
		SoReadError::post( in, "File `overlay.iv' does not contain "
			"a text3 node named `OverlayText'" ) ;
		exit( -1 ) ;
	}

	in->closeFile() ;
	delete in ;
}



//------------------------------------------------------------------------------
// Check for cameras.
//------------------------------------------------------------------------------
SbBool
GameControl::checkForCameras()
{
	//
	// Search for a camera.
	//
	SoSearchAction sa ;
	sa.setInterest( SoSearchAction::FIRST ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setType( SoCamera::getClassTypeId() ) ;
	sa.setSearchingAll( FALSE ) ;
	sa.apply( gameRoot ) ;

	SoPath *path = sa.getPath() ;
	if( path != NULL )
	{
		SoDebugError::post( name.getString(),
			 "Scene shouldn't have a camera in it!\n" ) ;
		return( FALSE ) ;
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Initialize any team color nodes in the game file.
//------------------------------------------------------------------------------
void
GameControl::setTeamColors()
{
	char *colorNames[] = { "TeamOneColor", "TeamTwoColor", "TeamThreeColor",
				"TeamFourColor" } ;

	SoSearchAction sa ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::NAME ) ;
	sa.setSearchingAll( TRUE ) ;

	for( int i = 0 ; i < sizeof( colorNames ) / sizeof( colorNames[0] ) ;
		i++ )
	{
		sa.setName( colorNames[i] ) ;
		sa.apply( gameRoot ) ;

		SoPathList &pl = sa.getPaths() ;
		for( int j = 0 ; j < pl.getLength() ; j++ )
		{
			SoNode *node = pl[j]->getTail() ;
			setTeamColorNode( node, i ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Set an SoBaseColor or SoMaterial node to the proper team color.
//------------------------------------------------------------------------------
void
GameControl::setTeamColorNode(
	SoNode	*node,
	int	team
	)
{
	if( node->isOfType( SoBaseColor::getClassTypeId() ) )
	{
		SoBaseColor *bc = (SoBaseColor *)node ;
		bc->rgb.set1Value( 0, teamColors->rgb[team+1] ) ;
	}
	else if( node->isOfType( SoMaterial::getClassTypeId() ))
	{
		SoMaterial *mat = (SoMaterial *)node ;
		SbColor c = 0.8f * teamColors->rgb[team+1] ;
		mat->diffuseColor.set1Value( 0, c ) ;
		c *= 0.25f ;
		mat->ambientColor.set1Value( 0, c ) ;
	}
}



//------------------------------------------------------------------------------
// Initialize the background color based on the fog color of the game scene.
//------------------------------------------------------------------------------
void
GameControl::setBackgroundColor()
{
	//
	// Search for fog color and set background to that color
	//
	SoSearchAction sa ;
	sa.setInterest( SoSearchAction::FIRST ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setType( SoEnvironment::getClassTypeId() ) ;
	sa.setSearchingAll( FALSE ) ;
	sa.apply( gameRoot ) ;
	SoPath *path = sa.getPath() ;
	if( path == NULL )
	{
		cockPit->setSceneBackgroundColor( SbColor( 0.0f, 0.0f, 0.0f ),
							1 ) ;
	}
	else
	{
		SoEnvironment *env = (SoEnvironment *)path->getTail() ;
		cockPit->setSceneBackgroundColor( env->fogColor.getValue(), 1 );
	}
}



//------------------------------------------------------------------------------
// Set up the drop zones.
//------------------------------------------------------------------------------
SbBool
GameControl::setUpDropZones()
{
	dropZones = (SoCoordinate3 *)Utilities::getNamedNode( "DropZones",
					SoCoordinate3::getClassTypeId() ) ;
	if( dropZones == NULL )
	{
		SoDebugError::post( name.getString(),
			"No drop zones in the game file\n" ) ;
		return( FALSE ) ;
	}

	if( mode == SOLO )
	{
		lastPosition = dropZones->point[0] ;
	}
	else if( dropZones->point.getNum() < 2 )
	{
		SoDebugError::post( name.getString(),
			"Not enough drop zones in the game file\n" ) ;
		return( FALSE ) ;
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Set up the goals.
//------------------------------------------------------------------------------
SbBool
GameControl::setUpGoals()
{
	goalGroup = (SoGroup *)Utilities::getNamedNode( "GoalHolder",
					SoGroup::getClassTypeId() ) ;
	if( goalGroup == NULL )
	{
		SoDebugError::post( name.getString(),
			"No goal holder in the game file\n" ) ;
		return( FALSE ) ;
	}

	if( goalGroup->getNumChildren() < numTeams )
	{
		SoDebugError::post( name.getString(),
			"Incorrect number of goals in the game file\n" ) ;
		return( FALSE ) ;
	}

	goals.truncate( 0 ) ;

	// Check that goals are set up correctly.
	//
	// Each child should be a SoGroup that contains a SoMaterial or
	// SoBaseColor node and a Goal node.
	for( int i = 0 ; i < goalGroup->getNumChildren() ; i++ )
	{
		SbBool st = TRUE ;
		SoNode *node = goalGroup->getChild( i ) ;
		if( !node->isOfType( SoGroup::getClassTypeId() ) )
		{
			st = FALSE ;
		}
		else
		{
			SoGroup *g = (SoGroup *)node ;
			if( g->getNumChildren() != 2 ||
				!( g->getChild( 0 )->isOfType(
				    SoBaseColor::getClassTypeId() ) ||
				g->getChild( 0 )->isOfType(
				    SoMaterial::getClassTypeId() ) ) ||
				!g->getChild( 1 )->isOfType(
				    Goal::getClassTypeId() ) )
			{
				st = FALSE ;
			}
			else
			{
				goals.append( g->getChild( 1 ) ) ;
				setTeamColorNode( g->getChild( 0 ), i ) ;
			}
		}
		if( !st )
		{
			SoDebugError::post( name.getString(),
				"GoalHolder is not set up properly in the game"
				" file\n" ) ;
			return( FALSE ) ;
		}
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Get the Vehicle selection node.
//------------------------------------------------------------------------------
SbBool
GameControl::getVehicleTypes()
{
	//
	// Don't change the vehicle choices after the first level.
	//
	if( mode == SOLO && level > 1 )
	{
		return( TRUE ) ;
	}

	vehicleChoices = (NameList *)Utilities::getNamedNode(
			"AvailableVehicles", NameList::getClassTypeId() ) ;

	if( vehicleChoices == NULL )
	{
		SoDebugError::post( name.getString(),
			"No list of available vehicles in game file\n" ) ;
		return( FALSE ) ;
	}
	else if( vehicleChoices->name.getNum() < 1 )
	{
		SoDebugError::post( name.getString(),
			"Not enough choices in the available vehicle list\n" ) ;
		return( FALSE ) ;
	}

	if( mode == SOLO )
	{
		vehicleChoices->ref() ;
	}

	vehicleHolder->removeAllChildren() ;
	for( int i = 0 ; i < vehicleChoices->name.getNum() ; i++ )
	{
		SoType type = SoType::fromName( vehicleChoices->name[i] ) ;
		if( !type.isDerivedFrom( Vehicle::getClassTypeId() ) )
		{
			SoDebugError::post( name.getString(),
				"A bad Vehicle type (%s) was encountered in"
				"the available vehicle list",
				vehicleChoices->name[i].getString() ) ;
			return( FALSE ) ;
		}

		Vehicle *v = (Vehicle *)type.createInstance() ;
		if( v == NULL )
		{
			SoDebugError::post( name.getString(),
				"Can not instantiate Vehicle type `%s'",
				vehicleChoices->name[i].getString() ) ;
			exit( 1 ) ;
		}

		vehicleHolder->addChild( v ) ;
	}
	vehicleHolder->whichChild = 0 ;

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Find the terrain and scenery node.
//------------------------------------------------------------------------------
SbBool
GameControl::findTerrain()
{
	realStuff = (SoSeparator *)Utilities::getNamedNode( "World",
			SoSeparator::getClassTypeId() ) ;

	if( realStuff == NULL )
	{
		SoDebugError::post( name.getString(),
			"No world (scenery) data found\n" ) ;
		return( FALSE ) ;
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Set up the camera.
//------------------------------------------------------------------------------
void
GameControl::setUpCamera()
{
	cam = new SoPerspectiveCamera ;
	cam->ref() ;
	cam->nearDistance = 0.1f ;
	cam->farDistance = 40.0f ;
	cam->focalDistance = 1 ;
	cam->heightAngle = 0.7f ;

	SoComposeMatrix *cm = new SoComposeMatrix ;
	cm->ref() ;

	xfv = new SoTransformVec3f ;
	xfv->ref() ;
	xfv->matrix.connectFrom( &cm->matrix ) ;

	camCalc = new SoCalculator ;
	camCalc->ref() ;
	camCalc->B.connectFrom( &xfv->direction ) ;
	camCalc->expression.setValue( "oA = A + B" ) ;
	cam->position.connectFrom( &camCalc->oA ) ;

	rr = new RotateRotation ;
	rr->ref() ;
	cam->orientation.connectFrom( &rr->rotation ) ;

	cm->rotation.connectFrom( &rr->rotationB ) ;
}



//------------------------------------------------------------------------------
// Choose a vehicle.
//------------------------------------------------------------------------------
void
GameControl::chooseVehicle()
{
	vehicleChoice = vehicleHolder->whichChild.getValue() ;

	if( vehicle != NULL )
	{
		removeVehicleFromDisplay( vehicle ) ;
		vehicle = NULL ;
	}

	SoType type = SoType::fromName( vehicleChoices->name[vehicleChoice] ) ;
	vehicle = (Vehicle *)type.createInstance() ;
	weaponPod = WeaponControl::MAIN_POD ;

	addVehicleToDisplay( vehicle, myTeam ) ;
	radarControl->attach( &vehicle->position, &vehicle->radarRotation ) ;
	explosionControl->setVehicle( vehicle ) ;
	resetVehicle() ;

	la->setSelf( NULL ) ;
	la->setMode( LosAction::GROSS_CHECK ) ;

	weaponControl->setNumberOfBays( vehicle->getNumBays() ) ;
	weaponControl->clearWeaponTypes() ;
	weaponControl->setBayPod( vehicle->getBayPod() ) ;

	for( int i = 0 ; i < vehicle->getNumWeapons( WeaponControl::MAIN_POD ) ;
		i++ )
	{
		weaponControl->addWeaponType( vehicle->getWeaponType( i,
			WeaponControl::MAIN_POD ), WeaponControl::MAIN_POD ) ;
	}

	for( i = 0 ; i < vehicle->getNumWeapons( WeaponControl::AA_POD ) ; i++ )
	{
		weaponControl->addWeaponType( vehicle->getWeaponType( i,
			WeaponControl::AA_POD ), WeaponControl::AA_POD ) ;
	}

	damageControl->restoreArmor( vehicle->getMaxDamage() ) ;

	status = ALIVE ;
	engine->setPitch( 1.0f, 1.0f ) ;
	engine->play() ;
	statusTime = gameTime ;
	overlaySwitch->whichChild = 0 ;

	if( mode == SOLO )
	{
		for( i = 0 ; i < automatons.getLength() ; i++ )
		{
			( (Automaton *)automatons[i] )->reset() ;
		}
	}

	resetMouse() ;
}



//------------------------------------------------------------------------------
// Control how the camera views the tank.
//------------------------------------------------------------------------------
void
GameControl::setCameraMode(
	Views		mode
	)
{
	SbRotation	rot ;
	static Views	lastMode = COCKPIT ;
	static float	distance = 5.0f ;

	if( mode == INC_DISTANCE )
	{
		distance *= 1.1f ;
		mode = lastMode ;
	}
	else if( mode == DEC_DISTANCE )
	{
		distance /= 1.1f ;
		mode = lastMode ;
	}
	else if( mode == LAST_MODE )
	{
		mode = lastMode ;
	}
	else
	{
		lastMode = mode ;
	}

	SbVec3f viewVector ;
	float	d = distance ;
	SbBool	connectRotation = TRUE ;

	if( weaponControl->isGuidingWeapon() )
	{
		mode = MISSILE ;
	}

	switch( mode )
	{
		default :
		case COCKPIT :
			d = 1.0f ;
			if( weaponPod == WeaponControl::MAIN_POD )
			{
				viewVector = vehicle->getMainViewpoint() +
					vehicle->getCenterOffset() ;
				rr->rotationA.disconnect() ;
				rr->rotationA.setValue(
					SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_2 ) ;
			}
			else
			{
				viewVector = vehicle->getAuxViewpoint() +
					vehicle->getCenterOffset() ;
				rr->rotationA.connectFrom( &vehicle->conView ) ;
			}
			vehicle->showFirstPersonGraphics( TRUE, weaponPod ) ;
			break ;

		case OVERHEAD :
			viewVector.setValue( 0.0f, 0.0f, 1.0f ) ;
			rr->rotationA.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
						0.0f ) ;
			break ;

		case UNDER :
			viewVector.setValue( 0.0f, 0.0f, -1.0f ) ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
						(float)M_PI ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, 1.0f ),
						(float)M_PI ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case AHEAD :
			viewVector.setValue( 0.0f, 1.0f, 0.0f ) ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
						(float)M_PI_2 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, 1.0f ),
						(float)M_PI ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case BEHIND :
			viewVector.setValue( 0.0f, -1.0f, 0.0f ) ;
			rr->rotationA.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
						(float)M_PI_2 ) ;
			break ;

		case LEFT :
			viewVector.setValue( -1.0f, 0.0f, 0.0f ) ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_2 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, -1.0f ),
					(float)M_PI_2 ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case RIGHT :
			viewVector.setValue( 1.0f, 0.0f, 0.0f ) ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_2 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, 1.0f ),
					(float)M_PI_2 ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case LEFT_REAR :
			viewVector.setValue( -1.5f, -1.5f, 2.0f ) ;
			viewVector.normalize() ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_4 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, -1.0f ),
					(float)M_PI_4 ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case LEFT_FRONT :
			viewVector.setValue( -1.5f, 1.5f, 2.0f ) ;
			viewVector.normalize() ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_4 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, -1.0f ),
					3.0f * (float)M_PI_4 ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case RIGHT_REAR :
			viewVector.setValue( 1.5f, -1.5f, 2.0f ) ;
			viewVector.normalize() ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_4 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, 1.0f ),
					(float)M_PI_4 ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case RIGHT_FRONT :
			viewVector.setValue( 1.5f, 1.5f, 2.0f ) ;
			viewVector.normalize() ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_4 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, 1.0f ),
					3.0f * (float)M_PI_4 ) ;
			rr->rotationA.setValue( rot ) ;
			break ;

		case NEGATIVE_X :
			viewVector.setValue( -1.0f, 0.0f, 0.0f ) ;
			rot.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
					(float)M_PI_2 ) ;
			rot *= SbRotation( SbVec3f( 0.0f, 0.0f, -1.0f ),
					(float)M_PI_2 ) ;
			rr->rotationA.setValue( rot ) ;
			rr->rotationB.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
							0.0f ) ;
			connectRotation = FALSE ;
			break ;

		case NEGATIVE_Y :
			viewVector.setValue( 0.0f, -1.0f, 0.0f ) ;
			rr->rotationA.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
							(float)M_PI_2 ) ;
			rr->rotationB.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
							0.0f ) ;
			connectRotation = FALSE ;
			break ;

		case POSITIVE_Z :
			viewVector.setValue( 0.0f, 0.0f, 1.0f ) ;
			rr->rotationA.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
							0.0f ) ;
			rr->rotationB.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
							0.0f ) ;
			connectRotation = FALSE ;
			break ;

		case MISSILE :
			viewVector.setValue( 0.0f, 0.15f, 0.0f ) ;
			d = 1.0f ;
			rr->rotationA.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ),
						(float)M_PI_2 ) ;
			vehicle->showFirstPersonGraphics( FALSE, weaponPod ) ;
			break ;

	}
	xfv->vector.setValue( d * viewVector ) ;

	if( !weaponControl->isGuidingWeapon() )
	{
		cam->heightAngle = 0.7f ;
		if( connectRotation )
		{
			rr->rotationB.connectFrom( &vehicle->rotation ) ;
		}
		else
		{
			rr->rotationB.disconnect() ;
		}
		camCalc->A.connectFrom( &vehicle->position ) ;
	}
	else
	{
		cam->heightAngle = 0.4f ;
		Missile *mis = weaponControl->getGuidedWeapon() ;
		if( connectRotation )
		{
			rr->rotationB.connectFrom( &mis->rotation ) ;
		}
		else
		{
			rr->rotationB.disconnect() ;
		}
		camCalc->A.connectFrom( &mis->position ) ;
	}
}



//------------------------------------------------------------------------------
// Reset the vehicle.
//------------------------------------------------------------------------------
void
GameControl::resetVehicle()
{
	float	DropRange = 25.0f ;
	SbBool	okay = FALSE ;
	SbVec3f location ;
	int	count = 0 ;
	int	i ;
	int	dp ;

	while( !okay )
	{
		if( mode == SOLO )
		{
			int closestIndex = 0.0f ;
			float closestDist =
				(lastPosition - dropZones->point[0]).length() ;
			for( i = 1 ; i < dropZones->point.getNum() ; i++ )
			{
				float d = ( lastPosition -
						dropZones->point[i] ).length() ;
				if( d < closestDist )
				{
					closestDist = d ;
					closestIndex = i ;
				}
			}

			location = dropZones->point[closestIndex] ;
			okay = TRUE ;
		}
		else
		{
			int offset = 0 ;
			int modulus = dropZones->point.getNum() ;

			if( mode == GOAL )
			{
				int n = dropZones->point.getNum() / numTeams ;
				if( n > 0 && myTeam > 0 )
				{
					modulus = n ;
					offset = ( myTeam - 1 ) * n ;
				}
			}
			dp = offset + ( rand() % modulus ) ;
			location = dropZones->point[dp] ;
			okay = TRUE ;
			if( count++ > 20 )
			{
				DropRange *= 0.5f ;
				count = 0 ;
			}

			for( i = 0 ; okay &&
				i < vehicleGroup->getNumChildren() ; i++ )
			{
				SoSeparator *sep = (SoSeparator *)
						vehicleGroup->getChild( i ) ;

				//
				// First child is the team color, so skip it.
				//
				for( int n = 1 ; okay &&
					n < sep->getNumChildren() ; n++ )
				{
					Vehicle *other = (Vehicle *)
							sep->getChild( n ) ;
					SbVec3f &dist = location -
						other->position.getValue() ;
					// Just check X and Y.
					dist[2] = 0.0f ;
					if( other != vehicle &&
						dist.dot( dist ) < DropRange )
					{
						okay = FALSE ;
					}
				}
			}
		}
	}

	//
	// Now adjust Z value for terrain.
	FollowTerrainAction fta ;
	fta.setSearchPoint( location ) ;
	fta.apply( realStuff ) ;
	const GroundPointList &gpl = fta.getGroundPointList() ;
	if( gpl.getLength() > 0 )
	{
		location = gpl[0]->getPoint() ;
		location[2] += 0.25f ;
	}
	else
	{
		printf( "Starting point %d is off the map!!!\n", dp ) ;
	}

	vehicle->position.setValue( location ) ;
	vehicle->rotation.setValue( SbVec3f( 0.0f, 0.0f, 1.0f ),
		(float)( rand() % 360 ) * (float)M_PI / 180.0f ) ;
	vehicle->reset() ;

	weaponControl->reset() ;
	setCameraMode( COCKPIT ) ;
	noVehicleInput = FALSE ;
	otherHost = 0 ;
}



//------------------------------------------------------------------------------
// Handle mouse motion events.
//------------------------------------------------------------------------------
void
GameControl::motionEvent(
	XMotionEvent	*motion
	)
{
	SbVec2s winSize = cockPit->getSize() ;
	const SbVec2s &center = cockPit->getSceneCenter( 1 ) ;
	const SbVec2s &mvSize = cockPit->getSceneSize( 1 ) ;

	mouseWinX = motion->x ;
	mouseWinY = winSize[1] - motion->y ;
	mouseX = ( (float)mouseWinX - (float)center[0] ) ;
	mouseY = ( (float)mouseWinY - (float)center[1] ) ;

	float div ;
	if( mvSize[1] < mvSize[0] )
	{
		div = 2.0f / ( 0.75f * (float)mvSize[1] ) ;
	}
	else
	{
		div = 2.0f / ( 0.75f * (float)mvSize[0] ) ;
	}

	mouseX *= div ;
	mouseY *= div ;

	if( mouseX < -1.0f )
	{
		mouseX = -1.0f ;
	}
	else if( mouseX > 1.0f )
	{
		mouseX = 1.0f ;
	}

	if( mouseY < -1.0f )
	{
		mouseY = -1.0f ;
	}
	else if( mouseY > 1.0f )
	{
		mouseY = 1.0f ;
	}

	if( weaponPod == WeaponControl::MAIN_POD )
	{
		float tol = 0.1f ;

		if( -tol < mouseX && mouseX < tol &&
			-tol < mouseY && mouseY < tol )
		{
			mouseX = 0.0f ;
			mouseY = 0.0f ;
		}
	}
}



//------------------------------------------------------------------------------
// Handle a button event.
//------------------------------------------------------------------------------
void
GameControl::buttonEvent(
	XButtonEvent	*be
	)
{
	if( status == ALIVE )
	{
		if( be->type == ButtonRelease )
		{
			if( be->button == Button1 )
			{
				if( weaponControl->isGuidingWeapon() )
				{
					weaponControl->stopGuidingWeapon() ;
					wind->stop() ;
					resetMouse() ;
					engine->setPitch( 1.0f, 1.0f ) ;
					engine->play() ;
					overlaySwitch->whichChild = 0 ;
					setCameraMode( LAST_MODE ) ;
				}
				else if( !weaponControl->isGuidedWeaponReady(
					weaponPod ) || noVehicleInput )
				{
					if( vehicle->okayToFire( weaponPod ) )
					{
						launchMissile() ;
					}
					else
					{
						buzzer->play() ;
					}
					if( weaponControl->isGuidingWeapon() )
					{
						engine->stop() ;
						wind->play() ;
					}
				}
				noVehicleInput = FALSE ;
			}
		}
		else if( be->type == ButtonPress )
		{
			if( be->button == Button1 )
			{
				if( weaponControl->isGuidedWeaponReady(
					weaponPod ) )
				{
					noVehicleInput = TRUE ;
				}
			}
			else if( be->button == Button2 )
			{
				weaponControl->changeWeapon( weaponPod ) ;
			}
			else if( be->button == Button3 )
			{
				if( ( vehicle->hasConView() ||
					weaponPod == WeaponControl::AA_POD ) &&
					!weaponControl->isGuidingWeapon() )
				{
					if( weaponPod ==
						WeaponControl::MAIN_POD )
					{
						weaponPod =
							WeaponControl::AA_POD ;
						overlaySwitch->whichChild = 3 ;
						vehicle->resetConView() ;
					}
					else
					{
						weaponPod =
						    WeaponControl::MAIN_POD ;
						overlaySwitch->whichChild = 0 ;
					}
					if( !vehicle->isOfType(
						Flyer::getClassTypeId() ) )
					{
						resetMouse() ;
					}
					setCameraMode( LAST_MODE ) ;
				}
			}
		}
	}
	else if( status == CHOOSING_TANK && be->type == ButtonPress &&
		be->button == Button1 )
	{
		pickTank( be ) ;
	}
	else if( status == CHOOSING_TEAM && be->type == ButtonPress &&
		be->button == Button1 )
	{
		pickTeam( be ) ;
	}
	else if( status == INITIALIZING && be->type == ButtonPress )
	{
		struckPlayer->play() ;

		if( mode == GOAL )
		{
			vehicleRotor->on = FALSE ;
		}
		if( numTeams < 2 || mode == SOLO )
		{
			myTeam = numTeams ;
			myOrigTeam = myTeam ;
			comm->setTeam( myTeam ) ;

			startChoosingTank() ;
		}
		else
		{
			status = CHOOSING_TEAM ;
			modeSwitch->whichChild = 1 ;
		}
	}
}



//------------------------------------------------------------------------------
// Pick a tank from the display.
//------------------------------------------------------------------------------
void
GameControl::pickTank(
	XButtonEvent	*be
	)
{
	const SbViewportRegion &vp = cockPit->getSceneViewportRegion( 1 ) ;
	SbViewportRegion reg( vp.getViewportSizePixels() ) ;
	ray->setViewportRegion( reg ) ;
	SbVec2s winSize = cockPit->getSize() ;
	SbVec2s pt( be->x, winSize[1] - be->y ) ;
	pt -= vp.getViewportOriginPixels() ;
	ray->setPoint( pt ) ;
	ray->setPickAll( TRUE ) ;
	ray->setRadius( 5.0f ) ;
	ray->apply( topRoot ) ;
	const SbName next( "NextVehicle" ) ;
	const SbName prev( "PrevVehicle" ) ;
	const SbName enter( "ThisVehicle" ) ;

	const SoPickedPointList &pp = ray->getPickedPointList() ;
	int maxVehicle = vehicleHolder->getNumChildren() ;
	int curVehicle = vehicleHolder->whichChild.getValue() ;
	for( int i = 0 ; i < pp.getLength() ; i++ )
	{
		const SbName &pickName = pp[i]->getPath()->getTail()->getName();

		if( pickName == next )
		{
			vehicleHolder->whichChild = ( curVehicle + 1 ) %
							maxVehicle ;
			describeVehicle() ;
		}
		else if( pickName == prev )
		{
			vehicleHolder->whichChild = ( curVehicle + maxVehicle -
							1 ) % maxVehicle ;
			describeVehicle() ;
		}
		else if( pickName == enter )
		{
			chooseVehicle() ;
			modeSwitch->whichChild = 2 ;
			overlaySwitch->whichChild = 0 ;
			vehicleRotor->on = FALSE ;
			setBackgroundColor() ;
		}
	}
}



//------------------------------------------------------------------------------
// Pick a team from the display.
//------------------------------------------------------------------------------
void
GameControl::pickTeam(
	XButtonEvent	*be
	)
{
	const SbViewportRegion &vp = cockPit->getSceneViewportRegion( 1 ) ;
	SbViewportRegion reg( vp.getViewportSizePixels() ) ;
	ray->setViewportRegion( reg ) ;
	SbVec2s winSize = cockPit->getSize() ;
	SbVec2s pt( be->x, winSize[1] - be->y ) ;
	pt -= cockPit->getSceneViewportRegion( 1 ).getViewportOriginPixels() ;
	ray->setPoint( pt ) ;
	ray->setPickAll( TRUE ) ;
	ray->setRadius( 5.0f ) ;
	ray->apply( topRoot ) ;

	const SoPickedPointList &pp = ray->getPickedPointList() ;
	if( pp.getLength() > 1 )
	{
		printf( "Team picking scene graph not correct since more than"
			" one object was picked\n" ) ;
		exit( 1 ) ;
	}
	else if( pp.getLength() == 1 )
	{
		int pickedTeam = pp[0]->getMaterialIndex() ;
		if( pickedTeam < 1 || pickedTeam > numTeams )
		{
			printf( "Bad team number (%d) picked.  Check team "
				"picking scene graph\n" ) ;
			exit( 1 ) ;
		}
		myTeam = pickedTeam ;
		myOrigTeam = myTeam ;

		//
		// Check to see if my team has been "absorbed" by another
		// team temporarily.
		//
		if( mode == GOAL )
		{
			if( teamAssignment[myTeam-1] != myTeam )
			{
				myTeam = teamAssignment[myTeam-1] ;
				dBoard->addMessage( "You are now on "
				    "the %s team",
				    teamNames-> name[myTeam].getString() ) ;
			}
		}
		comm->setTeam( myTeam ) ;

		startChoosingTank() ;
	}
}



//------------------------------------------------------------------------------
// Launch a missile.
//------------------------------------------------------------------------------
void
GameControl::launchMissile()
{
	Missile *rocket = weaponControl->launchWeapon( weaponPod ) ;

	if( rocket == NULL )
	{
		return ;
	}

	float h = 0.2f ;

	if( weaponControl->isGuidingWeapon() )
	{
		overlaySwitch->whichChild = 1 ;
	}
	setCameraMode( LAST_MODE ) ;

	rocket->setHeight( h ) ;
	rocket->setOwner( vehicle ) ;

	vehicle->launchWeapon( rocket, weaponPod ) ;
}



//------------------------------------------------------------------------------
// Handle a key press.
//------------------------------------------------------------------------------
SbBool
GameControl::keyPressEvent(
	KeySym	ks
	)
{
	SbBool handled = TRUE ;

	switch( ks )
	{
		case XK_r :
			if( status == ALIVE )
			{
				die( 0 ) ;
			}
			break ;

#ifdef VERBOSE
		case XK_g :
			if( mode == SOLO )
			{
				godMode = !godMode ;
				printf( "godMode o%s\n", godMode ? "n" : "ff" );
			}
			break ;

		case XK_d :
			float armor = damageControl->sufferDamage( 2.0f ) ;
			if( armor <= 0.0f )
			{
				die( 0 ) ;
			}
			break ;

		case XK_KP_Up :
			setCameraMode( POSITIVE_Z ) ;
			break ;

		case XK_KP_Left :
			setCameraMode( LEFT ) ;
			break ;

		case XK_KP_Right :
			setCameraMode( RIGHT ) ;
			break ;

		case XK_KP_Begin :
			setCameraMode( COCKPIT ) ;
			break ;

		case XK_Left :
			setCameraMode( DEC_DISTANCE ) ;
			break ;

		case XK_Right :
			setCameraMode( INC_DISTANCE ) ;
			break ;

		case XK_F1 :
			Utilities::dumpIt( "dashboard.iv",
						dBoard->getRoot() ) ;
			printf( "Dashboard written to dashboard.iv\n" );
			Utilities::dumpIt( "tracker.iv", topRoot ) ;
			printf( "Scene written to tracker.iv\n" ) ;
			break ;
#endif /* VERBOSE */

		case XK_Up:
			Sfx::changeVolume( Sfx::LOUDER ) ;
			struckPlayer->play() ;
			break ;

		case XK_Down:
			Sfx::changeVolume( Sfx::SOFTER ) ;
			struckPlayer->play() ;
			break ;

		case XK_x:
			if( status == ALIVE )
			{
				radarControl->increaseRange() ;
			}
			break ;

		case XK_l:
			if( status == ALIVE )
			{
				const SbVec3f &loc =
					vehicle->position.getValue() ;
				dBoard->addMessage( "Location: %.1f, %.1f",
					loc[0], loc[1] ) ;
			}
			break ;

		case XK_z:
			if( status == ALIVE )
			{
				radarControl->decreaseRange() ;
			}
			break ;

		case XK_Alt_L :
			comm->changeMode( InterComm::PUBLIC ) ;
			break ;

		case XK_Control_L :
			comm->changeMode( InterComm::PRIVATE ) ;
			break ;

		case XK_space :
			if( status == ALIVE && vehicle != NULL &&
				vehicle->hasRangeFinder() )
			{
				findRange() ;
			}
			break ;

		case XK_t :
			if( status == ALIVE && mode != SOLO )
			{
				if( scoresOn )
				{
					overlayTextSwitch->whichChild = -1 ;
					scoresOn = FALSE ;
				}
				else
				{
					showScores() ;
					scoresOn = TRUE ;
					helpOn = FALSE ;
				}
			}
			break ;

		case XK_h :
			if( status == ALIVE )
			{
				if( helpOn )
				{
					overlayTextSwitch->whichChild = -1 ;
					helpOn = FALSE ;
				}
				else
				{
					showHelp() ;
					helpOn = TRUE ;
					scoresOn = FALSE ;
				}
			}
			else
			{
				SbString cmd( "showcase -v " ) ;
				cmd += GameScenario::getDataDir() ;
				cmd += "PointBlankHelp.sc" ;
				sproc( GameControl::runCommand, PR_SDIR,
					cmd.getString() ) ;
			}
			break ;

		default :
			if( status == ALIVE && vehicle != NULL )
			{
				handled = vehicle->handleKeyPress( ks ) ;
			}
			else
			{
				handled = FALSE ;
			}
			break ;
	}

	return( handled ) ;
}



//------------------------------------------------------------------------------
// Handle a key release.
//------------------------------------------------------------------------------
SbBool
GameControl::keyReleaseEvent(
	KeySym	ks
	)
{
	SbBool handled = TRUE ;

	switch( ks )
	{
		case XK_Alt_L :
		case XK_Control_L :
			comm->changeMode( InterComm::QUIET ) ;
			break ;

		default :
			handled = FALSE ;
			break ;
	}

	return( handled ) ;
}



//------------------------------------------------------------------------------
// Do a frame.
//------------------------------------------------------------------------------
SbBool
GameControl::doOneFrame(
	float	dt
	)
{
	float	mx = mouseX ;
	float	my = mouseY ;

	SbBool guided = weaponControl->isGuidingWeapon() ;

	gameTime += dt ;

	if( status == ALIVE )
	{
		weaponControl->updateWeapons( mx, my, dt, TRUE, realStuff ) ;

		if( guided && guided != weaponControl->isGuidingWeapon() &&
			status == ALIVE )
		{
			overlaySwitch->whichChild = 0 ;
			setCameraMode( LAST_MODE ) ;
			wind->stop() ;
			resetMouse() ;
			engine->setPitch( 1.0f, 1.0f ) ;
			engine->play() ;
		}

		if( guided || noVehicleInput )
		{
			mx = 0.0f ;
			my = 0.0f ;
		}

		float	energy ;
		if( weaponPod == WeaponControl::AA_POD )
		{
			energy = vehicle->moveConView( mx, my, dt, realStuff ) ;
		}
		else
		{
			energy = vehicle->move( mx, my, dt, realStuff ) ;
		}
		float q0, q1, q2, q3 ;
		vehicle->rotation.getValue().getValue( q0, q1, q2, q3 );
		float hdgX = 2.0f * ( q0 * q1 - q2 * q3 ) ;
		float hdgY = 1.0f - 2.0f * ( q2 * q2 + q0 * q0 ) ;
		int hdg = (int)( 180.0f + atan2f( -hdgX, -hdgY ) *
				( 180.0f / (float)M_PI ) ) % 360 ;
		dBoard->updateHeading( hdg ) ;

		if( energy > 50.0f )
		{
			float armor = damageControl->sufferDamage(
						0.3f * ( energy - 50.0f ) ) ;
			if( armor <= 0.0f )
			{
				dBoard->addMessage( "Too rough a ride" ) ;
				die( 0 ) ;
			}
		}

		// Check to see if we can scavenge some armor.
		float leftEngine, rightEngine ;
		vehicle->getEngineSpeeds( leftEngine, rightEngine ) ;
		speedometer->setSpeed( 0.5f * ( leftEngine + rightEngine ) ) ;
		float amount ;
		if( leftEngine == 0 && rightEngine == 0 &&
			( amount = dt * debrisControl->scavenge(
				vehicle->position.getValue() ) ) > 0.0f )
		{
			if( scavengeTime < 2.0f )
			{
				amount *= scavengeTime * 0.5f ;
				if( scavengeTime == 0.0f )
				{
					dBoard->addMessage(
						"Starting armor repairs" ) ;
				}
			}
			damageControl->repair( 2.0f * amount ) ;
			scavengeTime += dt ;
		}
		else if( scavengeTime != 0.0f )
		{
			dBoard->addMessage( "Ending armor repairs" ) ;
			scavengeTime = 0.0f ;
		}
		float l = ( leftEngine > 0.0f ) ? leftEngine : -leftEngine ;
		float r = ( rightEngine > 0.0f ) ? rightEngine : -rightEngine ;
		engine->setPitch( 1.0f + l, 1.0f + r ) ;
	}
	else
	{
		if( status == DEAD && statusTime < gameTime )
		{
			updateStatus() ;
		}
		weaponControl->updateWeapons( 0.0f, 0.0f, dt, FALSE,
						realStuff ) ;
	}

	if( mode == GOAL )
	{
		goalAdmin() ;
	}
	else if( mode == SOLO )
	{
		if( status == ROUND_OVER )
		{
			// Check to see if people are still playing instead of
			// taking the round break.
			if( GameIO::getGameTime() > roundOverTime )
			{
				if( nextSoloLevel.getLength() <= 0 ||
					!loadGame( nextSoloLevel.getString() ) )
				{
					return( TRUE ) ;
				}
				buildCaches() ;
				if( lastStatus == ALIVE )
				{
					chooseVehicle() ;
					modeSwitch->whichChild = 2 ;
					overlaySwitch->whichChild = 0 ;
					vehicleRotor->on = FALSE ;
					setBackgroundColor() ;
				}
				else
				{
					startChoosingTank() ;
				}
			}
		}

		for( int i = 0 ; i < automatons.getLength() ; i++ )
		{
			( (Automaton *)automatons[i] )->update( dt, realStuff,
					( status == ALIVE ? vehicle : NULL ) ) ;
		}
	}

	debrisControl->update( dt, realStuff ) ;

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Update the radar display by checking the line of sight of every object.
//------------------------------------------------------------------------------
void
GameControl::updateRadar()
{
	int i ;
	SbVec3f &origination = vehicle->position.getValue() +
					SbVec3f( 0.0f, 0.0f, 9000.0f ) ;

	//
	// Prepare the radar.
	//
	radarControl->clear() ;

	//
	// See if we can "see" our satellite.
	//
	la->setRay( origination, vehicle->position.getValue() ) ;
	la->setTarget( vehicle ) ;
	la->apply( realStuff ) ;
	// Don't show anything if our view to the satellite is blocked.
	if( !la->isObstructed() )
	{
		for( i = 0 ; i < vehicleGroup->getNumChildren() ; i++ )
		{
			SoSeparator *sep =
				(SoSeparator *)vehicleGroup->getChild( i ) ;
			//
			// First child is the team color, so skip it.
			//
			for( int n = 1 ; n < sep->getNumChildren() ; n++ )
			{
				Vehicle *target = (Vehicle *)sep->getChild(n) ;

				if( target != vehicle &&
					radarControl->isInRange(
						target->position.getValue() ) )
				{
					la->setRay( origination,
						target->position.getValue() ) ;
					la->setTarget( target ) ;
					la->apply( realStuff ) ;

					if( !la->isObstructed() )
					{
						radarControl->addTarget(
						    target->position.getValue(),
						    radarColors->rgb[i] ) ;
					}
				}
			}
		}

		for( i = 0 ; i < missileGroup->getNumChildren() ; i++ )
		{
			Missile *target = (Missile *)missileGroup->getChild(i) ;

			if( radarControl->isInRange(
				target->position.getValue() ) )
			{
				la->setRay( origination,
					target->position.getValue() ) ;
				la->setTarget( target ) ;
				la->apply( realStuff ) ;

				if( !la->isObstructed() )
				{
					radarControl->addTarget(
						target->position.getValue(),
						SbColor( 1.0f, 1.0f, 1.0f ) ) ;
				}
			}
		}
	}

	// Check to show goals.
	if( mode == GOAL )
	{
		Goal *goal ;
		for( i = 0 ; i < numTeams ; i++ )
		{
			int mask = 1 << i ;
			if( ( goalsStruck & mask ) == 0 )
			{
				goal = (Goal *)goals[i] ;
				if( goal->onRadar.getValue() &&
					radarControl->isInRange(
					goal->position.getValue() ) )
				{
					radarControl->addTarget(
						    goal->position.getValue(),
						    radarColors->rgb[i+1] ) ;
				}
			}
		}
	}
	else if( mode == SOLO )
	{
		for( i = 0 ; i < automatons.getLength() ; i++ )
		{
			Automaton *a = (Automaton *)automatons[i] ;
			if( a->isAlive() && radarControl->isInRange(
				a->position.getValue() ) )
			{
				la->setRay( origination,
						a->position.getValue() ) ;
				la->setTarget( a ) ;
				la->apply( realStuff ) ;

				if( !la->isObstructed() )
				{
				    radarControl->addTarget(
						a->position.getValue(),
						radarColors->rgb[1] ) ;
				}
			}
		}
		Goal *goal ;
		for( i = 0 ; i < goals.getLength() ; i++ )
		{
			goal = (Goal *)goals[i] ;
			if( goal->onRadar.getValue() &&
				radarControl->isInRange(
					goal->position.getValue() ) )
			{
				radarControl->addTarget(
					goal->position.getValue(),
					radarColors->rgb[2] ) ;
			}
		}
	}
}



//------------------------------------------------------------------------------
// Broadcast data through GameIO spigot.
//------------------------------------------------------------------------------
void
GameControl::broadcast(
	float	dt
	)
{
	static const float zeros[] = { 0.0f, 0.0f, 0.0f, 0.0f } ;

	//
	// The following 2 variables insure that we don't flood the net at
	// rates higher than 36 fps.
	//
	static float gateTime = 0.0f ;
	static const float minTime = 1.0f / 36.0f ;

	GameIO::updateGameTime( dt ) ;

	gateTime += dt ;
	if( gateTime < minTime )
	{
		return ;
	}
	else
	{
		gateTime -= minTime ;
	}
		
	if( ( status & HAS_VEHICLE ) != 0 )
	{
		PlayerPacket packet( (unsigned short)status, myTeam, myOrigTeam,
				vehicleChoice, otherHost, myScore,
				vehicle->position.getValue().getValue(),
				vehicle->rotation.getValue().getValue() ) ;
		weaponControl->loadMissileInfo( packet.missileInfo, dt ) ;
		gio->send( &packet, sizeof( packet ) ) ;
	}
	else
	{
		PlayerPacket packet( status, myTeam, myOrigTeam, vehicleChoice,
					0, myScore, zeros, zeros ) ;
		gio->send( &packet, sizeof( packet ) ) ;
	}
}



//------------------------------------------------------------------------------
// Process a player packet (static method).
//------------------------------------------------------------------------------
void
GameControl::processPlayerPacket(
	PlayerPacket	*pkt,
	GameControl	*game
	)
{
	game->processPlayer( pkt ) ;
}



//------------------------------------------------------------------------------
// Really process the player packet.
//------------------------------------------------------------------------------
void
GameControl::processPlayer(
	PlayerPacket	*pkt
	)
{
	Player *player ;

	int n = playerList.find( pkt->fromAddress ) ;

	if( n < 0 )
	{
		player = addPlayer( pkt ) ;
		if( scoresOn )
		{
			showScores() ;
		}
	}
	else
	{
		player = playerList[n] ;
	}

	player->vehicle->position.setValue( pkt->location.pos ) ;
	player->vehicle->rotation.setValue( pkt->orientation.o ) ;

	//
	// Check for a change in the player's status.
	//
	if( player->status != pkt->playerStatus() )
	{
		switch( pkt->playerStatus() )
		{
			case DEAD :
				if( pkt->getCulprit() == gio->getMyId() )
				{
					myScore += 100 +
						player->vehicle->getPoints() -
						vehicle->getPoints() ;
					if( pkt->getTeam() == myTeam )
					{
						dBoard->addMessage(
							"Oops, killed %s",
							gio->getPlayerName(
							pkt->fromAddress ) ) ;
						myScore -= 250 ;
						gong->play() ;
					}
					else
					{
						dBoard->addMessage(
							"You killed %s",
							gio->getPlayerName(
							pkt->fromAddress ) ) ;
					}
					dBoard->updateScoreDisplay( myScore ) ;
					if( scoresOn )
					{
						showScores() ;
					}
				}
				else
				{
					dBoard->addMessage( "%s is dead",
						gio->getPlayerName(
						pkt->fromAddress ) ) ;
				}
				explosionControl->addExplosion(
					    pkt->location.pos ) ;
				debrisControl->addDebris( pkt->location.pos,
					player->vehicle->getMaxDamage() ) ;
				removeVehicleFromDisplay( player->vehicle ) ;
				break ;

			case ALIVE :
				addVehicleToDisplay( player->vehicle,
							pkt->getTeam() ) ;
				break ;

			case ROUND_OVER :
				removeVehicleFromDisplay( player->vehicle ) ;
				break ;

			case CHOOSING_TANK :
			case CHOOSING_TEAM :
				break ;
		}
		player->status = pkt->playerStatus() ;
	}

	//
	// Check for a change in the player's vehicle type.
	//
	if( player->vehicleType != pkt->getVehicleType() )
	{
		if( player->vehicle != NULL )
		{
			removeVehicleFromDisplay( player->vehicle ) ;
			player->vehicle->unref() ;
		}
		player->vehicle = createVehicle( pkt->getVehicleType() ) ;
		player->vehicle->ref() ;
		if( pkt->playerStatus() == ALIVE )
		{
			addVehicleToDisplay( player->vehicle, player->team ) ;
		}
		player->vehicleType = pkt->getVehicleType() ;
	}

	//
	// Check for a change in the player's team
	//
	if( player->team != pkt->getTeam() )
	{
		if( player->origTeam == 0 )
		{
			dBoard->addMessage( "%s joined the %s team",
				gio->getPlayerName( player->hostAddr ),
				teamNames->name[pkt->getOrigTeam()].getString()
				) ;
			player->origTeam = pkt->getOrigTeam() ;
			if( scoresOn )
			{
				showScores() ;
			}
		}
		if( pkt->playerStatus() == ALIVE )
		{
			removeVehicleFromDisplay( player->vehicle ) ;
			addVehicleToDisplay( player->vehicle, pkt->getTeam() ) ;
		}
		player->team = pkt->getTeam() ;
		updateTeamCounts() ;
	}

	//
	// Check for a change in the player's score.
	//
	if( player->score != pkt->score )
	{
		player->score = pkt->score ;
		if( scoresOn )
		{
			showScores() ;
		}
	}

	if( ( player->status & HAS_VEHICLE ) == 0 )
	{
		return ;
	}

	// Update the player's missiles.
	for( int i = 0 ; i < WeaponControl::MAX_WEAPONS ; i++ )
	{
		// Check for changes in the missile status.
		if( pkt->missileStatus( i ) != player->missileStatus[i] )
		{
			if( pkt->missileStatus( i ) ==
				WeaponControl::ActiveMissile )
			{
				player->launchPlayerMissile( i,
					pkt->missileType( i ), weaponControl,
					pkt->missileLocation( i ) ) ;
			}
			else
			{
				//
				// If the missile was launched was really
				// close, we may not have a received a
				// packet with the ActiveMissile status set
				// for that missile.
				//
				if( player->missileStatus[i] !=
					WeaponControl::ActiveMissile )
				{
					player->launchPlayerMissile( i,
						pkt->missileType( i ),
						weaponControl,
						pkt->missileLocation( i ) ) ;
				}
				else
				{
					player->missile[i]->position.setValue(
						pkt->missileLocation( i ) ) ;
				}
				if( player->missile[i] == NULL )
				{
					continue ;
				}
				if( pkt->missileStatus( i ) == gio->getMyId() &&
					status == ALIVE )
				{
					hitNoise->play() ;
					float armor = damageControl->
					    sufferDamage( pkt->getDamage( i ) );
					if( armor <= 0.0f )
					{
						die( pkt->fromAddress ) ;
						dBoard->addMessage(
							"Killed by %s",
							gio->getPlayerName(
							pkt->fromAddress ) ) ;
					}
					if( weaponControl->isGuidingWeapon() )
					{
						weaponControl->
							stopGuidingWeapon() ;
						overlaySwitch->whichChild = 0 ;
						setCameraMode( LAST_MODE ) ;
						wind->stop() ;
						resetMouse() ;
						engine->setPitch( 1.0f, 1.0f ) ;
						engine->play() ;
					}
				}
				if( weaponControl->isBomb(pkt->missileType(i))&&
					vehicle )
				{
					if( checkBombHit( player->missile[i] ) )
					{
						die( pkt->fromAddress ) ;
						dBoard->addMessage(
							"Killed by %s",
							gio->getPlayerName(
							pkt->fromAddress ) ) ;
					}
				}
				weaponControl->terminatePlayerMissile(
							player->missile[i] ) ;
				player->missile[i]->unref() ;
				player->missile[i] = NULL ;
			}
			player->missileStatus[i] = pkt->missileStatus( i ) ;
		}

		// Move the missile if it is running.
		if( pkt->missileStatus( i ) == WeaponControl::ActiveMissile )
		{
			player->missile[i]->position.setValue(
				pkt->missileLocation( i ) ) ;
			player->missile[i]->rotation.setValue(
				pkt->missileOrientation( i ) ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Process an info packet (static method).
//------------------------------------------------------------------------------
void
GameControl::processInfoPacket(
	InfoPacket	*pkt,
	GameControl	*game
	)
{
	game->processInfo( pkt ) ;
}



//------------------------------------------------------------------------------
// Really process the info packet.
//------------------------------------------------------------------------------
void
GameControl::processInfo(
	InfoPacket	*pkt
	)
{
	switch( pkt->data )
	{
		case InfoPacket::QUIT_GAME :
			dBoard->addMessage( "%s quit",
				gio->getPlayerName( pkt->fromAddress ) ) ;
			removePlayer( pkt->fromAddress, TRUE ) ;
			break ;

		case InfoPacket::BANNING :
			if( pkt->remoteHost == gio->getMyId() )
			{
				dBoard->addMessage( "Banning %s",
					gio->getPlayerName(
						pkt->fromAddress ) ) ;
				removePlayer( pkt->fromAddress, FALSE ) ;
				gio->banHost( pkt->fromAddress ) ;
			}
			break ;

		case InfoPacket::DROPPING :
			if( pkt->remoteHost == gio->getMyId() )
			{
				dBoard->addMessage( "Dropped by %s",
					gio->getPlayerName(
						pkt->fromAddress ) ) ;
				removePlayer( pkt->fromAddress, FALSE ) ;
				gio->dropHost( pkt->fromAddress ) ;
			}
			break ;

		default :
			printf( "unknown data type in info packet (%d)\n",
				(int)pkt->data ) ;
			break ;
	}
			
}



//------------------------------------------------------------------------------
// Add a player to the game.
//------------------------------------------------------------------------------
Player *
GameControl::addPlayer(
	PlayerPacket	*pkt
	)
{
	dBoard->addMessage( "Added %s", gio->getPlayerName( pkt->fromAddress ));
	if( status != CHOOSING_TEAM )
	{
		hello->play() ;
	}
	Vehicle *hisVehicle = createVehicle( pkt->getVehicleType() ) ;

	Player *player = new Player( pkt->fromAddress, hisVehicle ) ;
	playerList.append( player ) ;
	dBoard->updateNumPlayers( playerList.getLength() ) ;

	return( player ) ;
}



//------------------------------------------------------------------------------
// Create a vehicle given the type number.
//------------------------------------------------------------------------------
Vehicle *
GameControl::createVehicle(
	int vType
	)
{
	if( vType < 0 || vType >= vehicleChoices->name.getNum() )
	{
		vType = 0 ;
		printf( "Trying to create a bad vehicle type (%d)!\n", vType ) ;
	}

	SoType type = SoType::fromName( vehicleChoices->name[vType] ) ;
	return( (Vehicle *)type.createInstance() ) ;
}



//------------------------------------------------------------------------------
// Remove a player from the game.
//------------------------------------------------------------------------------
void
GameControl::removePlayer(
	HostAddr	remoteId,
	SbBool		callGameIO
	)
{
	int n = playerList.find( remoteId ) ;

	if( n != -1 )
	{
		printf( "player 0x%p left the game\n", remoteId ) ;
		dBoard->addMessage( "%s left the game",
			gio->getPlayerName( remoteId ) ) ;
		if( playerList[n]->status == ALIVE )
		{
			explosionControl->addExplosion( playerList[n]->
				vehicle->position.getValue().getValue() ) ;
		}
		for( int i = 0 ; i < WeaponControl::MAX_WEAPONS ; i++ )
		{
			if( playerList[n]->missileStatus[i] ==
				WeaponControl::ActiveMissile )
			{
				weaponControl->terminatePlayerMissile(
						playerList[n]->missile[i] ) ;
			}
		}
		removeVehicleFromDisplay( playerList[n]->vehicle ) ;
		playerList.remove( n ) ;
	}

	if( callGameIO )
	{
		gio->removeHost( remoteId ) ;
	}

	updateTeamCounts() ;
	dBoard->updateNumPlayers( playerList.getLength() ) ;
}



//------------------------------------------------------------------------------
// Static method to handle missile strikes.
//------------------------------------------------------------------------------
HostAddr
GameControl::missileStrike(
	SoNode		*node,
	Missile		*missile,
	GameControl	*gc
	)
{
	return( gc->checkMissileStrike( node, missile ) ) ;
}



//------------------------------------------------------------------------------
// Check a missile strike to see if it struck another player.
//------------------------------------------------------------------------------
HostAddr
GameControl::checkMissileStrike(
	SoNode	*node,
	Missile	*missile
	)
{
	if( mode == GOAL && node->isOfType( Goal::getClassTypeId() ) )
	{
		int nt = getGoalsTeam( (Goal *)node ) ;
		struckPlayer->play() ;
		if( nt == myTeam )
		{
			dBoard->addMessage( "Idiot." ) ;
		}
		else if( nt > 0 )
		{
			struckGoal( nt ) ;
		}
	}
	else if( mode == SOLO )
	{
		if( node->isOfType( Automaton::getClassTypeId() ) )
		{
			Automaton *a = (Automaton *)node ;
			struckPlayer->play() ;
			hitAutomaton( a, missile->getDamage() ) ;
		}
		else if( node->isOfType( Goal::getClassTypeId() ) )
		{
			struckPlayer->play() ;
			Goal *goal =
				(Goal *)missile->getObstructor()->getTail() ;
			if( goal->destroy( missile->getDamage() ) )
			{
				int n = goals.find( node ) ;
				if( n >= 0 )
				{
					goals.remove( n ) ;
					myScore += 50 ;
					dBoard->updateScoreDisplay( myScore ) ;
					if( goals.getLength() == 0 )
					{
						endSoloLevel() ;
					}
					else
					{
						dBoard->addMessage( "One down, "
							"%d %s to go",
							goals.getLength(),
							goals.getLength() > 1 ?
							"goals" : "goal" ) ;
					}
				}
			}
		}
		else if( node->isOfType(
				DestructibleObstacle::getClassTypeId() ) )
		{
			struckPlayer->play() ;
			( (DestructibleObstacle *)node )->destroy(
						missile->getDamage() ) ;
		}
		//
		// Check for bomb damage to automatons.
		//
		if( missile->isOfType( Bomb::getClassTypeId() ) )
		{
			float d ;
			for( int i = 0 ; i < automatons.getLength() ; i++ )
			{
				Automaton *a = (Automaton *)automatons[i] ;
				if( ( d = checkBombDamage( missile, a )) > 0.0f)
				{
					hitAutomaton( a, d ) ;
				}
			}
		}
	}
	else
	{
		for( int i = 0 ; i < playerList.getLength() ; i++ )
		{
			if( node == playerList[i]->vehicle )
			{
				struckPlayer->play() ;
				return( playerList[i]->hostAddr ) ;
			}
		}
	}

	return( WeaponControl::DeadMissile ) ;
}



//------------------------------------------------------------------------------
// Inflict damage on an automaton.
//------------------------------------------------------------------------------
void
GameControl::hitAutomaton(
	Automaton	*a,
	float		damage
	)
{
	int pts = a->sufferDamage( damage ) ;
	if( pts > 0 )
	{
		myScore += pts ;
		dBoard->updateScoreDisplay( myScore ) ;
		debrisControl->addDebris( a->position.getValue(),
					a->armor.getValue() ) ;
		dBoard->addMessage( "Destroyed automaton." ) ;
	}
}



//------------------------------------------------------------------------------
// Determine which team's goal you hit.
//------------------------------------------------------------------------------
int
GameControl::getGoalsTeam(
	Goal *goal
	)
{
	for( int i = 0 ; i < goals.getLength() ; i++ )
	{
		if( goal == (Goal *)goals[i] )
		{
			return( i + 1 ) ;
		}
	}

	return( 0 ) ;
}



//------------------------------------------------------------------------------
// Destroy your tank.
//------------------------------------------------------------------------------
void
GameControl::die(
	HostAddr culprit
	)
{
	status = DEAD ;
	engine->stop() ;
	statusTime = gameTime + 5.0f ;
	weaponControl->stopGuidingWeapon() ;
	weaponControl->reset() ;
	setCameraMode( LAST_MODE ) ;
	overlaySwitch->whichChild = 2 ;

	if( scoresOn || helpOn )
	{
		scoresOn = FALSE ;
		helpOn = FALSE ;
		overlayTextSwitch->whichChild = -1 ;
	}

	otherHost = culprit ;
	wind->stop() ;
	speedometer->setSpeed( 0.0f ) ;

	if( mode == SOLO )
	{
		if( culprit == 0 )
		{
			dieNoise->play() ;
			myScore -= 100 ;
			dBoard->updateScoreDisplay( myScore ) ;
			numTanks-- ;
			if( numTanks >= 0 )
			{
				dBoard->updateNumPlayers( numTanks ) ;
			}
		}
		else
		{
			gong->play() ;
		}
		lastPosition = vehicle->position.getValue() ;
	}
	else
	{
		dieNoise->play() ;
		int n = playerList.find( culprit ) ;

		int points ;
		if( n != -1 )
		{
			points = playerList[n]->vehicle->getPoints() ;
		}
		else
		{
			points = vehicle->getPoints() ;
		}

		myScore -= 100 - points + vehicle->getPoints() ;
		dBoard->updateScoreDisplay( myScore ) ;
		scavengeTime = 0.0f ;
		if( scoresOn )
		{
			showScores() ;
		}
	}
}



//------------------------------------------------------------------------------
// Update the status if the time is right.
//------------------------------------------------------------------------------
void
GameControl::updateStatus()
{
	switch( status )
	{
		case DEAD :
			if( numTanks < 0 )
			{
				endSoloGame() ;
			}
			else
			{
				startChoosingTank() ;
			}
			break ;
	}
}



//------------------------------------------------------------------------------
// Set up the graph for choosing a tank.
//------------------------------------------------------------------------------
void
GameControl::setUpVehicleChooser()
{
	SoInput	*in = Utilities::createInput() ;
	SbString file( "graphs/vehicle_chooser.iv" ) ;

	if( !in->openFile( file.getString(), TRUE ) )
	{
		SoReadError::post( in, "Can't open %s", file.getString() ) ;
		exit( -1 ) ;
	}

	SoSeparator *root = SoDB::readAll( in ) ;

	if( root == NULL )
	{
		SoReadError::post( in, "Can't read %s", file.getString() ) ;
		exit( -1 ) ;
	}

	in->closeFile() ;

	modeSwitch->addChild( root ) ;

	const char *id = "VehicleChoiceHolder" ;

	vehicleHolder = (SoSwitch *)Utilities::getNamedNode( id,
						SoSwitch::getClassTypeId() ) ;
	if( vehicleHolder == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s'"
			" does not contain a switch node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}

	id = "VehicleChoiceColor" ;

	vehicleChoiceColor = (SoBaseColor *)Utilities::getNamedNode( id,
					SoBaseColor::getClassTypeId() ) ;
	if( vehicleChoiceColor == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s'"
			" does not contain a basecolor node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}

	id = "VehicleChoiceRotor" ;

	vehicleRotor = (SoRotor *)Utilities::getNamedNode( id,
						SoRotor::getClassTypeId() ) ;
	if( vehicleRotor == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s' "
			"does not contain a rotor node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}

	id = "VehicleDescription" ;

	vehicleText = (SoText3 *)Utilities::getNamedNode( id,
						SoText3::getClassTypeId() ) ;
	if( vehicleText == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s' "
			"does not contain a text3 node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}

	id = "WeaponDescription" ;

	weaponText = (SoText3 *)Utilities::getNamedNode( id,
						SoText3::getClassTypeId() ) ;
	if( weaponText == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s' "
			"does not contain a text3 node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}

	delete in ;
}



//------------------------------------------------------------------------------
// Set up the graph for choosing a team.
//------------------------------------------------------------------------------
void
GameControl::setUpTeamChooser()
{
	SoInput	*in = Utilities::createInput() ;
	SbString file( "graphs/team_chooser.iv" ) ;

	if( !in->openFile( file.getString(), TRUE ) )
	{
		SoReadError::post( in, "Can't open %s", file.getString() ) ;
		exit( -1 ) ;
	}

	SoSeparator *root = SoDB::readAll( in ) ;

	if( root == NULL )
	{
		SoReadError::post( in, "Can't read %s", file.getString() ) ;
		exit( -1 ) ;
	}

	in->closeFile() ;

	modeSwitch->addChild( root ) ;

	const char *id = "TeamColors" ;

	teamColors = (SoBaseColor *)Utilities::getNamedNode( id,
					SoBaseColor::getClassTypeId() ) ;
	if( teamColors == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s'"
			" does not contain a basecolor node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}
	teamColors->rgb.setNum( 5 ) ;
	teamColors->rgb.set1Value( 0, 0.8f, 0.8f, 0.8f ) ;
	teamColors->rgb.set1Value( 1, 0.8f, 0.4f, 0.4f ) ;
	teamColors->rgb.set1Value( 2, 0.4f, 0.7f, 0.8f ) ;
	teamColors->rgb.set1Value( 3, 0.46f, 0.8f, 0.4f ) ;
	teamColors->rgb.set1Value( 4, 0.55f, 0.4f, 0.8f ) ;

	id = "TeamButtonHolder" ;

	SoGroup *tbh = (SoGroup *)Utilities::getNamedNode( id,
					SoGroup::getClassTypeId() ) ;
	if( tbh == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s'"
			" does not contain a group node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}
	while( tbh->getNumChildren() > numTeams )
	{
		tbh->removeChild( tbh->getNumChildren() - 1 ) ;
	}

	id = "TeamNames" ;
	teamNames = (NameList *)Utilities::getNamedNode( id,
					NameList::getClassTypeId() ) ;
	if( teamNames == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s'"
			" does not contain a namelist node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}
	teamNames->name.setNum( 5 ) ;
	teamNames->name.set1Value( 0, "White" ) ;
	teamNames->name.set1Value( 1, "Red" ) ;
	teamNames->name.set1Value( 2, "Blue" ) ;
	teamNames->name.set1Value( 3, "Green" ) ;
	teamNames->name.set1Value( 4, "Purple" ) ;

	radarColors = new SoBaseColor ;
	radarColors->ref() ;

	for( int i = 0 ; i <= MAX_TEAMS ; i++ )
	{
		float h, s, v ;
		teamColors->rgb[i].getHSVValue( h, s, v ) ;
		radarColors->rgb.set1HSVValue( i, h, 1.0f, 1.0f ) ;
	}

	id = "TeamCounts" ;

	teamCounts = (SoText3 *)Utilities::getNamedNode( id,
						SoText3::getClassTypeId() ) ;
	if( teamCounts == NULL )
	{
		SoDebugError::post( name.getString(), "File `%s' "
			"does not contain a text3 node named `%s'",
			file.getString(), id ) ;
		exit( -1 ) ;
	}

	teamCounts->string.setNum( numTeams ) ;

	delete in ;
}



//------------------------------------------------------------------------------
// Do a time step.
//------------------------------------------------------------------------------
void
GameControl::step(
	float	dT
	)
{
	broadcast( dT ) ;
	comm->startCommProcessing() ;
	gio->processInput() ;
	comm->finishCommProcessing() ;
	if( status & HAS_VEHICLE )
	{
		updateRadar() ;
	}
	else
	{
		radarControl->clear() ;
	}

	cockPit->render() ;
}



//------------------------------------------------------------------------------
// Initialize for choosing a tank.
//------------------------------------------------------------------------------
void
GameControl::startChoosingTank()
{
	vehicleChoiceColor->rgb.setValue( teamColors->rgb[myTeam] ) ;
	status = CHOOSING_TANK ;
	modeSwitch->whichChild = 0 ;
	overlaySwitch->whichChild = -1 ;
	vehicleRotor->on = TRUE ;
	cockPit->setSceneBackgroundColor( SbColor( 0.0f, 0.0f, 0.0f ), 1 ) ;
	describeVehicle() ;
}



//------------------------------------------------------------------------------
// Add vehicle description.
//------------------------------------------------------------------------------
void
GameControl::describeVehicle()
{
	int n = 0 ;
	vehicleText->string.setNum( n ) ;

	Vehicle *v = (Vehicle *)vehicleHolder->getChild(
				vehicleHolder->whichChild.getValue() ) ;

	SbString s( "Name: " ) ;
	s += v->getLabel().getString() ;
	vehicleText->string.set1Value( n++, s ) ;

	if( v->isOfType( Flyer::getClassTypeId() ) )
	{
		s = "Type: Air" ;
	}
	else
	{
		s = "Type: Ground" ;
	}
	vehicleText->string.set1Value( n++, s ) ;

	s = "Speed: " ;
	s += SbString( (int)( v->getMaxSpeed() * 10.0f ) ) ;
	vehicleText->string.set1Value( n++, s ) ;

	s = "Armor: " ;
	s += SbString( (int)( v->getMaxDamage() ) ) ;
	vehicleText->string.set1Value( n++, s ) ;

	s = "Main bays: " ;
	s += SbString( v->getNumBays() ) ;
	vehicleText->string.set1Value( n++, s ) ;

	s = "AA view: " ;
	s += v->hasConView() ? "Y" : "N" ;
	vehicleText->string.set1Value( n++, s ) ;

	s = "Points: " ;
	s += v->getPoints() ;
	vehicleText->string.set1Value( n++, s ) ;

	n = 0 ;
	weaponText->string.setNum( n ) ;

	s = "Weapon List" ;
	weaponText->string.set1Value( n++, s ) ;

	for( int i = 0 ; i < v->getNumWeapons( WeaponControl::MAIN_POD ) ; i++ )
	{
		s = weaponControl->getWeaponName(
			v->getWeaponType( i, WeaponControl::MAIN_POD ) ) ;
		weaponText->string.set1Value( n++, s ) ;
	}

	for( i = 0 ; i < v->getNumWeapons( WeaponControl::AA_POD ) ; i++ )
	{
		s = weaponControl->getWeaponName(
			v->getWeaponType( i, WeaponControl::AA_POD ) ) ;
		weaponText->string.set1Value( n++, s ) ;
	}
}



//------------------------------------------------------------------------------
// Check for a bomb near-hit.
//------------------------------------------------------------------------------
SbBool
GameControl::checkBombHit(
	Missile	*bomb
	)
{
	float damage = checkBombDamage( bomb, vehicle ) ;

	if( damage > 0.0f )
	{
		float armor = damageControl->sufferDamage( damage ) ;
		if( armor <= 0.0f )
		{
			return( TRUE ) ;
		}
	}
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Check to see if a bomb inflicts damage.
//------------------------------------------------------------------------------
float
GameControl::checkBombDamage(
	Missile		*bomb,
	Obstacle	*obst
	)
{
	const float maxDistance = 9.0f ;
	float dam = 0.0f ;

	// Check distance between bomb hit and current location.
	SbVec3f dist = bomb->position.getValue() - obst->position.getValue();

	float d = dist.dot( dist ) ;
	if( d > maxDistance )
	{
		return( dam ) ;
	}

	// Close enough, so do a line-of-sight check to see if anything between
	// the bomb and the vehicle.
	la->setRay( bomb->position.getValue() + SbVec3f( 0.0f, 0.0f, 0.1f ),
			obst->position.getValue() ) ;
	la->setSelf( bomb ) ;
	la->setTarget( obst ) ;
	la->apply( realStuff ) ;
	la->setSelf( NULL ) ;
	// Clear line of sight, so inflict damage.
	if( !la->isObstructed() )
	{
		dam = ( maxDistance - d ) * bomb->getDamage() / maxDistance ;
	}

	return( dam ) ;
}



//------------------------------------------------------------------------------
// Callback for when a host is dropped.
//------------------------------------------------------------------------------
void
GameControl::droppedHost(
	HostAddr 	remoteId,
	GameControl	*game
	)
{
	game->dBoard->addMessage( "Dropped %s",
		GameIO::getHostName( remoteId ) ) ;
	game->removePlayer( remoteId, TRUE ) ;
	InfoPacket packet( InfoPacket::DROPPING, remoteId ) ;
	game->gio->send( &packet, sizeof( packet ) ) ;
	game->gio->send( &packet, sizeof( packet ) ) ;
}



//------------------------------------------------------------------------------
// Callback for when a host is banned.
//------------------------------------------------------------------------------
void
GameControl::bannedHost(
	HostAddr 	remoteId,
	GameControl	*game
	)
{
	game->dBoard->addMessage( "Banned %s",
		GameIO::getHostName( remoteId ) ) ;
	game->removePlayer( remoteId, FALSE ) ;
	InfoPacket packet( InfoPacket::BANNING, remoteId ) ;
	game->gio->send( &packet, sizeof( packet ) ) ;
	game->gio->send( &packet, sizeof( packet ) ) ;
}



//------------------------------------------------------------------------------
// Remove a vehicle from the display.
//------------------------------------------------------------------------------
void
GameControl::removeVehicleFromDisplay(
	Vehicle	*v
	)
{
	int n ;

	for( int i = 0 ; i < vehicleGroup->getNumChildren() ; i++ )
	{
		SoSeparator *sep = (SoSeparator *)vehicleGroup->getChild( i ) ;

		if( ( n = sep->findChild( v ) ) != -1 )
		{
			sep->removeChild( n ) ;
			return ;
		}
	}
}



//------------------------------------------------------------------------------
// Add a vehicle to the proper group for display.
//------------------------------------------------------------------------------
void
GameControl::addVehicleToDisplay(
	Vehicle	*v,
	int	team
	)
{
	if( team >= 0 && team <= numTeams )
	{
		SoSeparator *sep = (SoSeparator *)vehicleGroup->getChild(team);

		sep->addChild( v ) ;
	}
	else
	{
		printf( "tried to add vehicle to bad team number %d\n", team ) ;
	}
}



//------------------------------------------------------------------------------
// Position the pointer to the center of the main view.
//------------------------------------------------------------------------------
void
GameControl::resetMouse()
{
	SbVec2s winSize = cockPit->getSize() ;
	SbVec2s center = cockPit->getSceneCenter( 1 ) ;

	XWarpPointer( cockPit->getDisplay(), None,
		XtWindow( cockPit->getWidget() ), 0, 0, 0, 0,
		center[0], winSize[1] - center[1] ) ;
	XFlush( cockPit->getDisplay() ) ;
}



//------------------------------------------------------------------------------
// Update the counts of the number of team members.
//------------------------------------------------------------------------------
void
GameControl::updateTeamCounts()
{
	int team[MAX_TEAMS] ;

	for( int i = 0 ; i < numTeams ; i++ )
	{
		team[i] = 0 ;
	}

	if( myOrigTeam > 0 && myOrigTeam <= numTeams )
	{
		team[myTeam-1] = 1 ;
	}

	for( i = 0 ; i < playerList.getLength() ; i++ )
	{
		if( playerList[i]->origTeam > 0 &&
			playerList[i]->origTeam <= numTeams )
		{
			team[playerList[i]->origTeam-1] += 1 ;
		}
	}
			
	for( i = 0 ; i < numTeams ; i++ )
	{
		teamCounts->string.set1Value( i, SbString( team[i] ) ) ;
	}
}



//------------------------------------------------------------------------------
// Find the range to what is pointed to by the mouse.
//------------------------------------------------------------------------------
void
GameControl::findRange()
{
	const SbViewportRegion &vp = cockPit->getSceneViewportRegion( 1 ) ;
	SbViewportRegion reg( vp.getViewportSizePixels() ) ;
	ray->setViewportRegion( reg ) ;
	SbVec2s pt( mouseWinX, mouseWinY ) ;
	pt -= vp.getViewportOriginPixels() ;
	ray->setPoint( pt ) ;
	ray->setPickAll( FALSE ) ;
	ray->setRadius( 1.0f ) ;
	ray->apply( topRoot ) ;

	const SoPickedPointList &pp = ray->getPickedPointList() ;
	if( pp.getLength() > 0 )
	{
		SbVec3f &d = pp[0]->getPoint() - vehicle->position.getValue() ;
		vehicle->setRange( d.length() ) ;
	}
	else
	{
		vehicle->setRange( 0.0f ) ;
	}
}



//------------------------------------------------------------------------------
// Draw the scene so that the caches build.
//------------------------------------------------------------------------------
void
GameControl::buildCaches()
{
	int nc = modeSwitch->whichChild.getValue() ;

	modeSwitch->whichChild = 2 ;
	SoPerspectiveCamera *tcam = new SoPerspectiveCamera ;
	tcam->ref() ;
	tcam->nearDistance = 0.1f ;
	tcam->farDistance = 10000.0f ;
	tcam->focalDistance = 1 ;
	tcam->heightAngle = 0.7f ;
	
	realStuff->replaceChild( cam, tcam ) ;

	tcam->viewAll( realStuff, cockPit->getSceneViewportRegion( 1 ) ) ;
	cockPit->renderNoSwap() ;
	cockPit->renderNoSwap() ;

	realStuff->replaceChild( tcam, cam ) ;

	modeSwitch->whichChild = nc ;
}



//------------------------------------------------------------------------------
// Show a snapshot of the current players and scores.
//------------------------------------------------------------------------------
void
GameControl::showScores()
{
	// Determine team scores.
	int teamCnt[MAX_TEAMS] ;
	int topScore[5] ;

	for( int i = 0 ; i < numTeams ; i++ )
	{
		if( mode == FREE )
		{
			teamScores[i] = 0 ;
		}
		teamCnt[i] = 0 ;
	}

	topScore[0] = -1 ;
	topScore[1] = -2 ;
	topScore[2] = -2 ;
	topScore[3] = -2 ;
	topScore[4] = -2 ;

	if( myTeam > 0 && myTeam <= numTeams )
	{
		if( mode == FREE )
		{
			teamScores[myTeam-1] = myScore ;
		}
		teamCnt[myTeam-1] = 1 ;
	}

	for( i = 0 ; i < playerList.getLength() ; i++ )
	{
		if( playerList[i]->origTeam > 0 &&
			playerList[i]->origTeam <= numTeams )
		{
			teamCnt[playerList[i]->origTeam-1] += 1 ;
			if( mode == FREE )
			{
				teamScores[playerList[i]->origTeam-1] +=
							playerList[i]->score ;
			}
		}

		// Sort top 5 scores.
		int k = 4 ;
		while( k > 0 )
		{
			int m = topScore[k] ;
			int score ;
			if( m == -1 )
			{
				score = myScore ;
			}
			else if( m == -2 )
			{
				score = -30000 ;
			}
			else
			{
				score = playerList[m]->score ;
			}
			if( playerList[i]->score > score )
			{
				if( k < 4 )
				{
					topScore[k+1] = topScore[k] ;
				}
				topScore[k] = i ;
				k-- ;
			}
			else
			{
				k = 0 ;
			}
		}
	}

	int n = 0 ;
	overlayText->string.set1Value( n++, "TEAM SCORES" ) ;
	for( i = 0 ; i < numTeams ; i++ )
	{
		if( teamCnt[i] > 0 )
		{
			SbString s( teamNames->name[i+1].getString() ) ;
			s += ": " ;
			s += SbString( teamScores[i] ) ;
			overlayText->string.set1Value( n++, s.getString() ) ;
		}
	}

	overlayText->string.set1Value( n++, "" ) ;
	int max = playerList.getLength() ;
	if( max > 4 )
	{
		overlayText->string.set1Value( n++, "Top 5 Scores" ) ;
		max = 5 ;
	}
	else
	{
		overlayText->string.set1Value( n++, "Scores" ) ;
		max++ ;
	}

	for( i = 0 ; i < max ; i++ )
	{
		int score ;
		SbString s ;
		if( topScore[i] == -1 )
		{
			score = myScore ;
			s = GameIO::getMyName() ;
		}
		else
		{
			int m = topScore[i] ;
			score = playerList[m]->score ;
			s = gio->getPlayerName( playerList[m]->hostAddr ) ;
		}
		s += ": " ;
		s += score ;
		overlayText->string.set1Value( n++, s ) ;
	}

	overlayText->string.setNum( n ) ;
	overlayTextSwitch->whichChild = 0 ;
}



//------------------------------------------------------------------------------
// Show help text.
//------------------------------------------------------------------------------
void
GameControl::showHelp()
{
	int n = 0 ;
	SoMFString *str = &overlayText->string ;

	str->set1Value( n++, "HELP" ) ;
	str->set1Value( n++, "h - Toggle help message" ) ;
	str->set1Value( n++, "r - Reset (destroy current vehicle)" ) ;
	if( mode != SOLO )
	{
		str->set1Value( n++, "t - Toggle top scores" ) ;
	}
	str->set1Value( n++, "x - Unzoom radar" ) ;
	str->set1Value( n++, "z - Zoom radar" ) ;
	str->set1Value( n++, "l - Show location" ) ;
	str->set1Value( n++, "Left Alt - Broadcast voice to all" ) ;
	str->set1Value( n++, "Left Ctrl - Broadcast voice to team" ) ;
	str->set1Value( n++, "Up - Increase volume" ) ;
	str->set1Value( n++, "Down - Decrease volume" ) ;
	if( vehicle )
	{
		vehicle->showHelp( str, n ) ;
	}
	str->setNum( n ) ;

	overlayTextSwitch->whichChild = 0 ;
}



//------------------------------------------------------------------------------
// Process an goal packet (static method).
//------------------------------------------------------------------------------
void
GameControl::processGoalPacket(
	GoalPacket	*pkt,
	GameControl	*game
	)
{
	game->processGoal( pkt ) ;
}



//------------------------------------------------------------------------------
// Really process the goal packet.
//------------------------------------------------------------------------------
void
GameControl::processGoal(
	GoalPacket	*pkt
	)
{
	if( pkt->data != GoalPacket::UPDATE &&
		( pkt->owner < 1 || pkt->owner > numTeams ) )
	{
		return ;
	}

	switch( pkt->data )
	{
		case GoalPacket::STRIKE :
			//
			// If I am also claiming to have destroyed the same
			// goal, give up my claim if the other host has a
			// lower address (the odds of this happening are
			// pretty low).  Otherwise, wait a little longer to
			// make sure no one else is claiming the goal.
			//
			if( goalMode == STRUCK_GOAL &&
				pkt->owner == myStruckGoal )
			{
				if( pkt->fromAddress < gio->getMyId() )
				{
					goalMode = OPEN ;
				}
				else
				{
					goalStrikeTime = 1.0f +
						GameIO::getGameTime() ;
					strikeAnnouncements = 0 ;
				}
			}
			//
			// Destroy the goal.
			//
			else
			{
				blowUpGoal( pkt->owner ) ;
			}
			break ;

		case GoalPacket::CLAIM :
			// Propagate the goal claimed message once
			// (just for safety's sake).
			if( goalClaimed( pkt->owner, pkt->killer ) )
			{
				gio->send( pkt, sizeof( *pkt ) ) ;
			}
			break ;
	}

	//
	// Check round number to see if it matches up.
	//
	if( pkt->round > goalRound )
	{
printf( "got a packet with a higher round number\n" ) ;
		initGoals() ;
		goalRound = pkt->round ;
		if( pkt->flagBits != goalsStruck )
		{
			for( int i = 0 ; i < numTeams ; i++ )
			{
				int mask = 1 << i ;
				if( pkt->flagBits & mask )
				{
					blowUpGoal( i + 1 ) ;
				}
			}
		}

		if( status == ROUND_OVER )
		{
			int n = playerList.find( pkt->fromAddress ) ;
			if( n != -1 && playerList[n]->status != ROUND_OVER )
			{
				startChoosingTank() ;
			}
		}
	}
	else if( pkt->flagBits != goalsStruck &&
		pkt->data == GoalPacket::UPDATE )
	{
printf( "flagBits differ 0x%d <-> 0x%d\n", pkt->flagBits, goalsStruck ) ;
		for( int i = 0 ; i < numTeams ; i++ )
		{
			int mask = 1 << i ;
			//
			// If other player doesn't know about the goal
			// blowing up, inform him.
			//
			if( ( pkt->flagBits & mask ) == 0 &&
				( goalsStruck & mask ) != 0 )
			{
printf( "Indicating that team %d goal is gone\n", i+1 ) ;
				GoalPacket gp( GoalPacket::CLAIM, 
					i+1, goalTaker[i], goalRound,
					goalsStruck, GameIO::getGameTime() ) ;
				gio->send( &gp, sizeof( gp ) ) ;
			}
		}
	}
}



//------------------------------------------------------------------------------
// Struck a goal, so stake a claim for it.
//------------------------------------------------------------------------------
void
GameControl::struckGoal(
	int	goalTeam
	)
{
	blowUpGoal( goalTeam ) ;
	goalStrikeTime = 1.0f + GameIO::getGameTime() ;
	strikeAnnouncements = 0 ;
	goalMode = STRUCK_GOAL ;
	myStruckGoal = goalTeam ;
}



//------------------------------------------------------------------------------
// Do logic for when a goal is claimed.
//------------------------------------------------------------------------------
SbBool
GameControl::goalClaimed(
	int	owner,
	int	killer
	)
{
	SbBool st = FALSE ;

	// See if this is a new goal claim or just an update.
	int mask = ( 1 << ( owner - 1 ) ) ;
	if( ( goalsStruck & mask ) == 0 )
	{
		goalsStruck += mask ;
		numGoalsLeft-- ;
		goalTaker[owner-1] = killer ;
printf( "a goal was destroyed, owner = %d, killer = %d\n", owner, killer ) ;
		dBoard->addMessage( "The %s goal was destroyed",
				teamNames->name[owner].getString()) ;

		for( int i = 0 ; i < numTeams ; i++ )
		{
			if( teamAssignment[i] == owner )
			{
				teamScores[i] -= 100 ;
				teamAssignment[i] = killer ;
			}
			else if( teamAssignment[i] == killer )
			{
				teamScores[i] += 100 ;
			}
			if( scoresOn )
			{
				showScores() ;
			}
		}

		// See if my goal was destroyed and I should change
		// teams.
		if( owner == myTeam )
		{
			if( status == ALIVE )
			{
				die( 0 ) ;
			}
			myTeam = killer ;
			comm->setTeam( myTeam ) ;
			vehicleChoiceColor->rgb.setValue(
						teamColors->rgb[myTeam] ) ;
			if( numGoalsLeft > 1 )
			{
				dBoard->addMessage( "You are now on "
				    "the %s team",
				    teamNames-> name[myTeam].getString() ) ;
			}
		}
		// See if the round is over.
		if( numGoalsLeft < 2 )
		{
			endGoalGame() ;
		}

		st = TRUE ;
	}

	return( st ) ;
}



//------------------------------------------------------------------------------
// Blow up a goal.
//------------------------------------------------------------------------------
void
GameControl::blowUpGoal(
	int n
	)
{
	if( n < 1 || n > goalGroup->getNumChildren() )
	{
		return ;
	}

	Goal *g = (Goal *)goals[n-1] ;

	g->destroy( g->armor.getValue() + 1.0f ) ;
}



//------------------------------------------------------------------------------
// The goal game has ended.  Prepare to reset.
//------------------------------------------------------------------------------
void
GameControl::endGoalGame()
{
	myScore += 100 ;
	die( 0 ) ;

	status = ROUND_OVER ;
	roundOverTime = GameIO::getGameTime() + 10.0f ;

	modeSwitch->whichChild = 3 ;

	int winner = 0 ;
	// Determine which team won.
	for( int i = 0 ; i < numTeams ; i++ )
	{
		int mask = 1 << i ;
		if( ( goalsStruck & mask ) == 0 )
		{
			winner = i + 1 ;
		}
	}

	if( winner == 0 )
	{
		breakMessage->string.set1Value( 0, "It's a draw." ) ;
	}
	else
	{
		SbString msg = teamNames->name[winner].getString() ;
		msg += " team won." ;
		breakMessage->string.set1Value( 0, msg ) ;
	}

	// Determine team scores.
	int teamCnt[MAX_TEAMS] ;

	for( i = 0 ; i < numTeams ; i++ )
	{
		teamCnt[i] = ( i + 1 == myOrigTeam ) ? 1 : 0 ;
	}

	for( i = 0 ; i < playerList.getLength() ; i++ )
	{
		if( playerList[i]->origTeam > 0 &&
			playerList[i]->origTeam <= numTeams )
		{
			teamCnt[playerList[i]->origTeam-1] += 1 ;
		}
	}

	int n = 1 ;
	breakMessage->string.set1Value( n++, "" ) ;
	for( i = 0 ; i < numTeams ; i++ )
	{
		if( teamCnt[i] > 0 )
		{
			SbString s( teamNames->name[i+1].getString() ) ;
			s += ": " ;
			s += SbString( teamScores[i] ) ;
			breakMessage->string.set1Value( n++, s.getString() ) ;
		}
	}

	if( myTeam != myOrigTeam )
	{
		breakMessage->string.set1Value( n++, "" ) ;
		SbString tmp( "You are now back on the " ) ;
		tmp += teamNames->name[myOrigTeam].getString() ;
		tmp += " team." ;
		breakMessage->string.set1Value( n++, tmp ) ;
	}
	breakMessage->string.set1Value( n++, "" ) ;
	breakMessage->string.set1Value( n++, "Prepare for the next round..." ) ;
	breakMessage->string.setNum( n ) ;

	overlaySwitch->whichChild = -1 ;
	cockPit->setSceneBackgroundColor( SbColor( 0.0f, 0.0f, 0.0f ), 1 ) ;
}



//------------------------------------------------------------------------------
// The solo level has ended.  Either go to the next level or exit.
//------------------------------------------------------------------------------
void
GameControl::endSoloLevel()
{
	SbString str ;

	myScore += 100 ;
	dBoard->updateScoreDisplay( myScore ) ;
	numTanks++ ;
	dBoard->updateNumPlayers( numTanks ) ;
	lastStatus = status ;
	die( 1 ) ;

	status = ROUND_OVER ;
	roundOverTime = GameIO::getGameTime() + 5.0f ;

	modeSwitch->whichChild = 3 ;

	int n = 0 ;
	breakMessage->string.set1Value( n++, "" ) ;
	breakMessage->string.set1Value( n++, "" ) ;
	str = "Level " ;
	str += level ;
	str += " completed." ;
	level++ ;
	breakMessage->string.set1Value( n++, str ) ;
	breakMessage->string.set1Value( n++, "" ) ;

	SoInfo *nextLevel = (SoInfo *)Utilities::getNamedNode( "NextLevel",
					SoInfo::getClassTypeId() ) ;
	if( nextLevel == NULL )
	{
		breakMessage->string.set1Value( n++, "Game over." ) ;
		breakMessage->string.set1Value( n++, "" ) ;
		str = "Your final score: " ;
		str += myScore ;
		breakMessage->string.set1Value( n++, str ) ;
		nextSoloLevel.makeEmpty() ;
		roundOverTime += 2.5f ;
	}
	else
	{
		breakMessage->string.set1Value( n++, "Loading new level..." ) ;
		nextSoloLevel = nextLevel->string.getValue() ;
	}
	breakMessage->string.setNum( n ) ;

	gameRoot->unref() ;
	gameRoot = NULL ;

	overlaySwitch->whichChild = -1 ;
	cockPit->setSceneBackgroundColor( SbColor( 0.0f, 0.0f, 0.0f ), 1 ) ;
}



//------------------------------------------------------------------------------
// The solo game has ended.  Show game summary.
//------------------------------------------------------------------------------
void
GameControl::endSoloGame()
{
	SbString str ;

	status = ROUND_OVER ;
	roundOverTime = GameIO::getGameTime() + 7.5f ;

	modeSwitch->whichChild = 3 ;

	int n = 0 ;
	breakMessage->string.set1Value( n++, "" ) ;
	breakMessage->string.set1Value( n++, "" ) ;
	breakMessage->string.set1Value( n++, "Game over." ) ;
	breakMessage->string.set1Value( n++, "" ) ;
	str = "Your final score: " ;
	str += myScore ;
	breakMessage->string.set1Value( n++, str ) ;
	breakMessage->string.set1Value( n++, "" ) ;
	str = level - 1 ;
	str += ( level == 2 ) ? " level" : " levels" ;
	str += " completed." ;
	level++ ;
	breakMessage->string.set1Value( n++, str ) ;
	breakMessage->string.setNum( n ) ;

	nextSoloLevel.makeEmpty() ;

	gameRoot->unref() ;
	gameRoot = NULL ;

	overlaySwitch->whichChild = -1 ;
	cockPit->setSceneBackgroundColor( SbColor( 0.0f, 0.0f, 0.0f ), 1 ) ;
}



//------------------------------------------------------------------------------
// Initialize the game for a round of playing with goals.
//------------------------------------------------------------------------------
void
GameControl::initGoals()
{
	for( int i = 0 ; i < goalGroup->getNumChildren() ; i++ )
	{
		Goal *goal = (Goal *)goals[i] ;
		goal->restore() ;
	}

	for( i = 0 ; i < numTeams ; i++ )
	{
		teamAssignment[i] = i + 1 ;
	}

	goalsStruck = 0 ;
	numGoalsLeft = numTeams ;
	goalMode = OPEN ;
	myTeam = myOrigTeam ;
	comm->setTeam( myTeam ) ;
	goalRound++ ;
}



//------------------------------------------------------------------------------
// Administer the goalMode status.
//------------------------------------------------------------------------------
void
GameControl::goalAdmin()
{
	static float nextUpdate = 0.0f ;
	float theTime = GameIO::getGameTime() ;

	switch( goalMode )
	{
		case STRUCK_GOAL :
			if( goalStrikeTime < GameIO::getGameTime() &&
				strikeAnnouncements > 5 )
			{
				printf( "looks like you claim the goal\n" ) ;
				goalMode = CLAIMING ;
				myScore += 100 ;
				dBoard->updateScoreDisplay( myScore ) ;
				goalClaimed( myStruckGoal, myTeam ) ;
				strikeAnnouncements = 0 ;
			}
			else
			{
				GoalPacket packet( GoalPacket::STRIKE, 
					myStruckGoal, myTeam, goalRound,
					goalsStruck, theTime ) ;
				gio->send( &packet, sizeof( packet ) ) ;
				strikeAnnouncements++ ;
			}
			break ;

		case CLAIMING :
			int sendClaim = ( strikeAnnouncements < 5 ) ? 1 : 0 ;
			for( int i = 0 ; sendClaim == 0 &&
				i < playerList.getLength() ; i++ )
			{
				if( playerList[i]->team == myStruckGoal )
				{
					sendClaim = 1 ;
				}
			}
			if( sendClaim )
			{
printf( "sending claim\n" ) ;
				strikeAnnouncements++ ;
				GoalPacket packet( GoalPacket::CLAIM,
					myStruckGoal, myTeam, goalRound,
					goalsStruck, theTime ) ;
				gio->send( &packet, sizeof( packet ) ) ;
			}
			else
			{
printf( "through with claim\n" ) ;
				goalMode = OPEN ;
			}
			break ;

		default :
			break ;
	}

	if( status == ROUND_OVER )
	{
		// Check to see if people are still playing instead of
		// taking the round break.
		if( theTime > roundOverTime )
		{
			initGoals() ;
			startChoosingTank() ;
		}
	}

	if( theTime > nextUpdate )
	{
		nextUpdate = theTime + 2.0f ;
		GoalPacket packet( GoalPacket::UPDATE, 0, myTeam,
					goalRound, goalsStruck, theTime ) ;
		gio->send( &packet, sizeof( packet ) ) ;
	}
}



//------------------------------------------------------------------------------
// Set up the database to display a message between goal and solo rounds.
//------------------------------------------------------------------------------
void
GameControl::setUpBreakMessage()
{
	SoSeparator *root = new SoSeparator ;
	modeSwitch->addChild( root ) ;

	SoOrthographicCamera *oc = new SoOrthographicCamera ;
	root->addChild( oc ) ;
	oc->height = 2.0f ;
	oc->position.setValue( 0.0f, 0.0f, 5.0f ) ;
	oc->orientation.setValue( SbVec3f( 0.0f, 0.0f, 1.0f ), 0.0f ) ;
	oc->aspectRatio = 1.0f ;
	oc->nearDistance = 1.0f ;
	oc->farDistance = 11.0f ;

	SoLightModel *lm = new SoLightModel ;
	root->addChild( lm ) ;
	lm->model = SoLightModel::BASE_COLOR ;

	SoComplexity *comp = new SoComplexity ;
	root->addChild( comp ) ;
	comp->value = 1.0f ;

	SoFont *font = new SoFont ;
	root->addChild( font ) ;
	font->name = "HelveticaBold" ;
	font->size = 0.15f ;

	SoBaseColor *bc = new SoBaseColor ;
	root->addChild( bc ) ;
	bc->rgb.setValue( 1.0f, 1.0f, 0.0f ) ;

	SoTranslation *t = new SoTranslation ;
	root->addChild( t ) ;
	t->translation.setValue( 0.0f, 0.6f, 0.0f ) ;

	breakMessage = new SoText3 ;
	root->addChild( breakMessage ) ;
	breakMessage->justification = SoText3::CENTER ;
	breakMessage->parts = SoText3::FRONT ;
	breakMessage->spacing = 1.1f ;

	// Add intro message
	root = new SoSeparator ;
	modeSwitch->addChild( root ) ;
	root->renderCaching = SoSeparator::OFF ;
	root->renderCulling = SoSeparator::OFF ;

	SoDirectionalLight *dl = new SoDirectionalLight ;
	root->addChild( dl ) ;
	dl->direction.setValue( -1.0f, -1.0f, -5.0f ) ;
	dl->color.setValue( 1.0f, 1.0f, 1.0f ) ;

	SoPerspectiveCamera *pc = new SoPerspectiveCamera ;
	root->addChild( pc ) ;
	pc->position.setValue( 0.0f, 0.0f, 2.0f ) ;
	pc->orientation.setValue( SbVec3f( 0.0f, 0.0f, 1.0f ), 0.0f ) ;
	pc->aspectRatio = 1.0f ;
	pc->nearDistance = 0.5f ;
	pc->farDistance = 100.0f ;
	pc->focalDistance = 2.0f ;
	pc->heightAngle = 0.785398f ;

	font = new SoFont ;
	root->addChild( font ) ;
	font->size = .1f ;
	font->name = "HelveticaBold" ;

	root->addChild( comp ) ;

	if( mode == GOAL )
	{
		SoSeparator *sep = new SoSeparator ;
		root->addChild( sep ) ;
		sep->renderCulling = SoSeparator::OFF ;
		sep->renderCaching = SoSeparator::OFF ;

		SoTransform *xf = new SoTransform ;
		sep->addChild( xf ) ;
		xf->translation.setValue( 0.0f, 0.0f, -2.0f ) ;
		xf->rotation.setValue( SbVec3f( 1.0f, 0.0f, 0.0f ), -1.2f ) ;

		sep->addChild( vehicleRotor ) ;

		SoMaterial *mat = new SoMaterial ;
		sep->addChild( mat ) ;
		mat->ambientColor.setValue( .1f, .1f, .1f ) ;
		mat->diffuseColor.setValue( .4f, .4f, .4f ) ;

		t = new SoTranslation ;
		sep->addChild( t ) ;

		// Get the goal model.
		Goal *goal = (Goal *)goals[0] ;
		sep->addChild( goal ) ;

		const SbVec3f &gp = goal->position.getValue() ;
		t->translation.setValue( -gp ) ;
	}

	t = new SoTranslation ;
	root->addChild( t ) ;
	t->translation.setValue( 0.0f, .65f, 0.0f ) ;

	root->addChild( lm ) ;
	root->addChild( bc ) ;

	static char *goalIntroMsg[] = {
		"GOAL MODE",
		"",
		"Destroy the other teams' goals",
		"while protecting your own.",
		"When your goal is destroyed, you will be",
		"switched temporarily to the team that",
		"claimed your goal.",
		"When all but one team's goal is destroyed,",
		"the round will end and the game will be",
		"reset for the next round.",
		"",
		"Click any button to continue...",
		} ;
	static char *soloIntroMsg[] = {
		"SOLO MODE",
		"",
		"Destroy all of the enemy goals",
		"(displayed in blue on the radar)",
		"while avoiding enemy fire (enemy is",
		"displayed in red on the radar).",
		"The game is over after all levels",
		"have been cleared or when all of",
		"your tanks have been destroyed.",
		"",
		"Click any button to continue...",
		} ;
	SoText3 *t3 = new SoText3 ;
	root->addChild( t3 ) ;
	t3->justification = SoText3::CENTER ;
	t3->spacing = 1.1f ;
	if( mode == GOAL )
	{
		t3->string.setValues( 0,
			sizeof( goalIntroMsg ) / sizeof( goalIntroMsg[0] ),
			goalIntroMsg ) ;
	}
	else
	{
		t3->string.setValues( 0,
			sizeof( soloIntroMsg ) / sizeof( soloIntroMsg[0] ),
			soloIntroMsg ) ;
	}
}



//------------------------------------------------------------------------------
// Initialize a round for the solo game.
//------------------------------------------------------------------------------
void
GameControl::setUpSoloGame()
{
	SoSearchAction sa ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setType( Automaton::getClassTypeId() ) ;
	sa.setSearchingAll( TRUE ) ;
	sa.apply( realStuff ) ;

	SoPathList &pl = sa.getPaths() ;
	automatons.truncate( 0 ) ;
	for( int i = 0 ; i < pl.getLength() ; i++ )
	{
		automatons.append( pl[i]->getTail() ) ;
	}

	sa.setType( Goal::getClassTypeId() ) ;
	sa.apply( realStuff ) ;
	goals.truncate( 0 ) ;
	for( i = 0 ; i < sa.getPaths().getLength() ; i++ )
	{
		goals.append( sa.getPaths()[i]->getTail() ) ;
	}
}



//------------------------------------------------------------------------------
// Handle automaton missile strikes.
//------------------------------------------------------------------------------
void
GameControl::automatonStrikeHandler(
	Missile	*missile,
	SoNode	*node,
	void	*userData
	)
{
	GameControl *gc = (GameControl *)userData ;

	gc->automatonStrike( missile, node ) ;
}



//------------------------------------------------------------------------------
// Automaton missile strike.
//------------------------------------------------------------------------------
void
GameControl::automatonStrike(
	Missile	*missile,
	SoNode	*node
	)
{
	if( node )
	{
		if( node->isOfType( Automaton::getClassTypeId() ) )
		{
			Automaton *a = (Automaton *)node ;
			if( a->sufferDamage( missile->getDamage() ) > 0 )
			{
				debrisControl->addDebris(
					a->position.getValue(),
					a->armor.getValue() ) ;
			}
		}
		else if( node->isOfType(
			DestructibleObstacle::getClassTypeId() ) )
		{
			( (DestructibleObstacle *)node )->destroy(
					missile->getDamage() ) ;
		}
	}

	if( status == ALIVE )
	{
#ifdef VERBOSE
		if( godMode )
		{
			weaponControl->terminatePlayerMissile( missile ) ;
			return ;
		}
#endif // VERBOSE
		float armor ;
		if( node == vehicle )
		{
			hitNoise->play() ;
			armor = damageControl->sufferDamage(
							missile->getDamage() ) ;
		}
		else if( missile->isOfType( Bomb::getClassTypeId() ) &&
			checkBombHit( missile ) )
		{
			armor = -1.0f ;
		}
		else
		{
			weaponControl->terminatePlayerMissile( missile ) ;
			return ;
		}

		if( armor <= 0.0f )
		{
			die( 0 ) ;
			dBoard->addMessage( "Killed by automatons" ) ;
		}
		else if( weaponControl->isGuidingWeapon() )
		{
			weaponControl->stopGuidingWeapon() ;
			overlaySwitch->whichChild = 0 ;
			setCameraMode( LAST_MODE ) ;
			wind->stop() ;
			resetMouse() ;
			engine->setPitch( 1.0f, 1.0f ) ;
			engine->play() ;
		}
	}

	weaponControl->terminatePlayerMissile( missile ) ;
}



//------------------------------------------------------------------------------
// Handle automaton missile launches.
//------------------------------------------------------------------------------
Missile *
GameControl::automatonLaunchHandler(
	SoType	missileType,
	void	*userData
	)
{
	WeaponControl *wc = (WeaponControl *)userData ;

	return( wc->launchAutomatonMissile( missileType ) ) ;
}



//------------------------------------------------------------------------------
// Player definitions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Player::Player(
	HostAddr	remoteId,
	Vehicle		*v
	) :
	hostAddr( remoteId ),
	vehicle( v )
{
	vehicle->ref() ;
	for( int i = 0 ; i < WeaponControl::MAX_WEAPONS ; i++ )
	{
		missile[i] = NULL ;
		missileStatus[i] = WeaponControl::DeadMissile ;
	}
	status = GameControl::CHOOSING_TEAM ;
	score = 0 ;
	team = 0 ;
	origTeam = 0 ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Player::~Player()
{
	vehicle->unref() ;

	for( int i = 0 ; i < WeaponControl::MAX_WEAPONS ; i++ )
	{
		if( missile[i] != NULL )
		{
			missile[i]->unref() ;
		}
	}
}



//------------------------------------------------------------------------------
// Launch a player's missile.
//------------------------------------------------------------------------------
void
Player::launchPlayerMissile(
	int			i,
	Missile::MissileType	type,
	WeaponControl		*wc,
	const float		*location
	)
{
	// Safety check.
	if( missile[i] != NULL )
	{
		missile[i]->position.setValue( location ) ;
		wc->terminatePlayerMissile( missile[i] ) ;
		missile[i]->unref() ;
		missile[i] = NULL ;
	}

	if( 0 <= type && type < Missile::BAD_TYPE )
	{
		missile[i] = wc->launchPlayerMissile( type ) ;
		missile[i]->ref() ;
	}
}



//------------------------------------------------------------------------------
// PlayerList definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Truncate the PlayerList and delete the Player's.
//------------------------------------------------------------------------------
void
PlayerList::truncate(
	int	start
	)
{
	for( int i = start ; i < getLength() ; i++ )
	{
		delete (*this)[i] ;
	}

	SbPList::truncate( start ) ;
}



//------------------------------------------------------------------------------
// Find a Player within the PlayerList based on the host id.
//------------------------------------------------------------------------------
int
PlayerList::find(
	HostAddr remoteId
	)
{
	for( int i = 0 ; i < getLength() ; i++ )
	{
		if( (*this)[i]->hostAddr == remoteId )
		{
			return( i ) ;
		}
	}

	return( -1 ) ;
}



//------------------------------------------------------------------------------
// Remove an entry from the PlayerList and delete it.
//------------------------------------------------------------------------------
void
PlayerList::remove(
	int	which
	)
{
	delete (*this)[which] ;
	SbPList::remove( which ) ;
}



//------------------------------------------------------------------------------
// Sproc off a command.
//------------------------------------------------------------------------------
void
GameControl::runCommand(
	void	*s
	)
{
	int	argc = 0 ;
	int	inword = 0 ;
	char	*p = (char *)s ;

	while( *p )
	{
		if( isspace( *p ) )
		{
			if( inword )
			{
				inword = 0 ;
			}
		}
		else if( !inword )
		{
			inword = 1 ;
			argc++ ;
		}
		p++ ;
	}

	char **argv = new char *[argc+1] ;

	p = (char *)s ;
	argc = 0 ;
	inword = 0 ;
	while( *p )
	{
		if( isspace( *p ) )
		{
			if( inword )
			{
				inword = 0 ;
				*p = '\0' ;
			}
		}
		else if( !inword )
		{
			inword = 1 ;
			argv[argc++] = p ;
		}
		p++ ;
	}
	argv[argc] = NULL ;

#ifndef NDEBUG
	printf( "runCommand %s, argc = %d\n", (char *)s, argc ) ;
	for( int i = 0 ; i < argc ; i++ )
	{
		printf( "\targv[%d] = %s\n", i, argv[i] ) ;
	}
#endif /* NDEBUG */

	execvp( argv[0], argv ) ;
	perror( "execl" ) ;
}
