//------------------------------------------------------------------------------
// GameOutput header.
//------------------------------------------------------------------------------

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "GameInput.h"
#include "GameOutput.h"
#include "GameIO.h"

#include <Inventor/errors/SoDebugError.h>


static const unsigned short	MAX_PKT = 0x80 ;
static const unsigned short	ACK_PKT = 0x80 ;
DropHostHandler	*GameIO::dropHandler = NULL ;
DropHostHandler	*GameIO::banHandler = NULL ;

void	*GameIO::dropData = NULL ;
void	*GameIO::banData = NULL ;
float	GameIO::gameTime = 0.0f ;
char	GameIO::myName[GameIO::MAX_NAME_LEN] = "Anonymous" ;

static const char	*defaultName = "Unknown" ;


//------------------------------------------------------------------------------
// Acknowledgement packet.
//------------------------------------------------------------------------------
class AckPacket : public GamePacket {
    public :
	AckPacket( HostAddr host, const char *theName ) ;
	~AckPacket() {}

	HostAddr	remoteHost ;
	char		name[GameIO::MAX_NAME_LEN] ;
    private :
	AckPacket() ;
	AckPacket( const AckPacket &a ) ;
} ;



//------------------------------------------------------------------------------
// Handler declaration.
//------------------------------------------------------------------------------
class Handler {

    public :
	Handler( unsigned char token, int size, GamePacketHandler *callback,
			void *data ) ;
	~Handler() {}
	int	handlePacket( void *buffer, int size ) ;
	int	getSize() { return( length ) ; }

    private :
	unsigned char		token ;
	int			length ;
	GamePacketHandler	*func ;
	void			*userData ;
} ;



//------------------------------------------------------------------------------
// HostInfo declaration.
//------------------------------------------------------------------------------
class HostInfo {
    public :
	enum HostStatus {
		ACCEPTED,
		BANNED,
		DROPPED,
		UNKNOWN,
		} ;
	HostInfo( HostAddr remoteId, GameOutput *out ) ;
	~HostInfo() {}
	SbBool	isOkay() ;
	void	accept( const char *name ) ;
	void	ban() { status = BANNED ; }
	void	drop() { status = DROPPED ; }
	HostAddr	hostAddr ;
	SbBool	checkAge() ;
	const char *getName() { return( playerName ) ; }

    private :
	HostInfo() ;
	HostStatus	status ;
	int		ackCounts ;
	float		timeToSendAck ;
	float		age ;
	GameOutput	*out ;
	char		playerName[GameIO::MAX_NAME_LEN] ;
} ;



//------------------------------------------------------------------------------
// Constructor for doing file I/O.
//------------------------------------------------------------------------------
GameIO::GameIO(
	const char	*inFileName,
	const char	*outFileName
	)
{
	in = new GameInput( inFileName ) ;
	out = new GameOutput( outFileName ) ;
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Constructor for doing network I/O.
//------------------------------------------------------------------------------
GameIO::GameIO(
	const char	*serviceName,
	const char	*groupName,
	const char	*interface,
	int		defaultPort,
	int		loopback,
	int		timeToLive
	)
{
	in = new GameInput( serviceName, groupName, interface, defaultPort ) ;
	out = new GameOutput( serviceName, groupName, interface, defaultPort,
				loopback, timeToLive ) ;
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common constructor steps.
//------------------------------------------------------------------------------
void
GameIO::buildCommon()
{
	gameIdentifier = 0 ;
	gameVersion = 0 ;
	buffer = NULL ;
	bufferSize = 0 ;
	gameTime = 0.0f ;
	hostList.setGameOutput( out ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GameIO::~GameIO()
{
	delete in ;
	delete out ;
}



//------------------------------------------------------------------------------
// Set the identifier.
//------------------------------------------------------------------------------
void
GameIO::setIdentifier(
	unsigned short	id,
	unsigned char	version
	)
{
	gameIdentifier = id ;
	gameVersion = version ;
	GamePacket::setGameId( id, version ) ;
}



//------------------------------------------------------------------------------
// Add a packet handler for a specific type of packet.
//------------------------------------------------------------------------------
void
GameIO::addInputPacketType(
	unsigned char		token,
	int			size,
	GamePacketHandler	*callback,
	void			*data
	)
{
	if( ( token & 0x80 ) != 0 )
	{
		SoDebugError::post( "GameIO::addInputPacketType",
			"Packet type must be less than %d\n", MAX_PKT ) ;
		return ;
	}

	handlerList.append( new Handler( token, size, callback, data ) ) ;

	if( size > bufferSize )
	{
		delete buffer ;
		buffer = (void *)new char[size] ;
		bufferSize = size ;
	}
}



//------------------------------------------------------------------------------
// Process input.
//------------------------------------------------------------------------------
void
GameIO::processInput()
{
	SbBool	dataPresent = TRUE ;
	int	nRead ;
	int	i ;
	int	st ;
	GamePacket	*pkt ;

	while( dataPresent )
	{
		nRead = in->get( buffer, bufferSize ) ;
		pkt = (GamePacket *)buffer ;

		if( nRead == 0 )
		{
			dataPresent = FALSE ;
		}
		//
		// Make sure we at least have identifier info
		//
		else if( nRead < sizeof( int ) )
		{
			fprintf( stderr, "processInput:  bad data received\n" );
		}
		//
		// Make sure it is the right game.
		//
		else if( pkt->identifier != gameIdentifier )
		{
			// Ignore.
			continue ;
		}
		//
		// Make sure it is the right version.
		//
		else if( pkt->version != gameVersion )
		{
			if( pkt->version > gameVersion )
			{
				static int notice = 0 ;
				if( !notice )
				{
					printf( "***********************\n" ) ;
					printf( "A newer version of the\n"
						"game is being played by\n"
						"%s!\n",
						getHostName( pkt->fromAddress )
						) ;
					printf( "***********************\n" ) ;
					notice = 1 ;
				}
			}
		}
		//
		// Handle ACK_PKT's directly.
		//
		else if( pkt->type == ACK_PKT )
		{
			processAckPacket( (AckPacket *)pkt ) ;
		}
		//
		// Other packet types.
		//
		else if( hostList.isOkay( pkt->fromAddress ) )
		{
			//
			// Check packet identifier.
			//
			for( i = 0 ; i < handlerList.getLength() ; i++ )
			{
				st = handlerList[i]->handlePacket( buffer,
								nRead ) ;
				if( st == 1 )
				{
					break ;
				}
				else if( st == -1 )
				{
					fprintf( stderr, "processInput: "
						"received odd packet size of %d"
						" instead of %d bytes (id = "
						"0x%08x)\n", nRead,
						handlerList[i]->getSize(),
						(int *)buffer ) ;
					break ;
				}
			}
		}
	}

	// Check for hosts that we haven't heard of for a while.
	for( i = 0 ; i < hostList.getLength() ; i++ )
	{
		if( !hostList[i]->checkAge() )
		{
			printf( "hadn't heard from 0x%p in a while\n",
				hostList[i]->hostAddr ) ;
			if( dropHandler )
			{
				dropHandler( hostList[i]->hostAddr, dropData ) ;
			}
		}
	}
}



//------------------------------------------------------------------------------
// Process an Acknowledgement packet and add the player to the list of valid
// hosts.
//------------------------------------------------------------------------------
void
GameIO::processAckPacket(
	AckPacket *pkt
	)
{
	if( pkt->remoteHost == out->getMyAddress() )
	{
		hostList.accept( pkt->fromAddress, pkt->name ) ;
	}
}



//------------------------------------------------------------------------------
// Remove a host from the game.
//------------------------------------------------------------------------------
void
GameIO::removeHost(
	HostAddr remoteId
	)
{
	int n = hostList.find( remoteId ) ;

	if( n != -1 )
	{
		hostList.remove( n ) ;
	}
}



//------------------------------------------------------------------------------
// Ban a host.
//------------------------------------------------------------------------------
void
GameIO::banHost(
	HostAddr remoteId
	)
{
	int n = hostList.find( remoteId ) ;

	if( n != -1 )
	{
		hostList[n]->ban() ;
		printf( "banned 0x%p\n", remoteId ) ;
	}
}



//------------------------------------------------------------------------------
// Drop a host.
//------------------------------------------------------------------------------
void
GameIO::dropHost(
	HostAddr remoteId
	)
{
	int n = hostList.find( remoteId ) ;

	if( n != -1 )
	{
		hostList[n]->drop() ;
		printf( "dropped 0x%p\n", remoteId ) ;
	}
}



//------------------------------------------------------------------------------
// Set the callback routine to handled a dropped host.
//------------------------------------------------------------------------------
void
GameIO::setDropHostHandler(
	DropHostHandler	*h,
	void		*userData
	)
{
	dropHandler = h ;
	dropData = userData ;
}



//------------------------------------------------------------------------------
// Set the callback routine to handled a banned host.
//------------------------------------------------------------------------------
void
GameIO::setBanHostHandler(
	DropHostHandler	*h,
	void		*userData
	)
{
	banHandler = h ;
	banData = userData ;
}



//------------------------------------------------------------------------------
// Set the name to be used by other hosts.
//------------------------------------------------------------------------------
void
GameIO::setMyName(
	const char *newName
	)
{
	if( newName != NULL && newName[0] != '\0' )
	{
		strncpy( myName, newName, sizeof( myName ) ) ;
	}
	myName[GameIO::MAX_NAME_LEN-1] = '\0' ;
}



//------------------------------------------------------------------------------
// Return the name of a player corresponding to a host.
//------------------------------------------------------------------------------
const char *
GameIO::getPlayerName(
	HostAddr remoteId
	)
{
	int n = hostList.find( remoteId ) ;

	if( n != -1 )
	{
		return( hostList[n]->getName() ) ;
	}
	else
	{
		return( defaultName ) ;
	}
}



//------------------------------------------------------------------------------
// Return a string containing the readable name corresponding to an IP address.
//------------------------------------------------------------------------------
const char *
GameIO::getHostName(
	HostAddr remoteId
	)
{
	struct in_addr	addr ;
	struct hostent	*hp ;
	static char	buffer[512] ;

	addr.s_addr = remoteId ;
	hp = gethostbyaddr( &addr, sizeof(addr), AF_INET ) ;
	if( hp )
	{
		sprintf( buffer, "%s", hp->h_name ) ;
	}
	else
	{
		sprintf( buffer, "%s", inet_ntoa(addr) ) ;
	}

	return( buffer ) ;
}



//------------------------------------------------------------------------------
// Handler definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Handler::Handler(
	unsigned char		id,
	int			size,
	GamePacketHandler	*callback,
	void			*data
	)
{
	token = id ;
	length = size ;
	func = callback ;
	userData = data ;
}



//------------------------------------------------------------------------------
// Check a packet and handle it if the type and size matches.  Return 0 if not
// handled, 1 if handled, and -1 if type was right but size was wrong.
//------------------------------------------------------------------------------
int
Handler::handlePacket(
	void	*buffer,
	int	size
	)
{
	GamePacket *packet = (GamePacket *)buffer ;

	if( packet->type != token )
	{
		return( 0 ) ;
	}
	else if( size != length )
	{
		return( -1 ) ;
	}

	func( packet, userData ) ;

	return( 1 ) ;
}



//------------------------------------------------------------------------------
// HandlerList definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Truncate the HandlerList and delete the Handler's.
//------------------------------------------------------------------------------
void
HandlerList::truncate(
	int	start
	)
{
	for( int i = start ; i < getLength() ; i++ )
	{
		delete (*this)[i] ;
	}

	SbPList::truncate( start ) ;
}



//------------------------------------------------------------------------------
// HostInfo definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
HostInfo::HostInfo(
	HostAddr	remoteId,
	GameOutput	*gameOut
	)
{
	hostAddr = remoteId ;
	status = UNKNOWN ;
	ackCounts = 5 ;
	timeToSendAck = GameIO::getGameTime() + 1.0f ;
	out = gameOut ;
	age = 0.0f ;

	strncpy( playerName, defaultName, sizeof( playerName ) ) ;
	playerName[GameIO::MAX_NAME_LEN-1] = '\0' ;

	AckPacket ack( remoteId, GameIO::getMyName() ) ;
	out->send( &ack, sizeof( ack ) ) ;
}



//------------------------------------------------------------------------------
// Accept a host for further info.
//------------------------------------------------------------------------------
void
HostInfo::accept(
	const char *name
	)
{
	status = ACCEPTED ;
	age = GameIO::getGameTime() ;
	strncpy( playerName, name, sizeof( playerName ) ) ;
	playerName[GameIO::MAX_NAME_LEN-1] = '\0' ;
}



//------------------------------------------------------------------------------
// Check to see if the host is accepted.  If not banned, see if an ack packet
// needs to be sent.
//------------------------------------------------------------------------------
SbBool
HostInfo::isOkay()
{
	age = GameIO::getGameTime() ;
	if( status == ACCEPTED )
	{
		return( TRUE ) ;
	}
	else if( status == BANNED )
	{
		return( FALSE ) ;
	}
	else if( status == DROPPED )
	{
		status = UNKNOWN ;
		ackCounts = 5 ;
		timeToSendAck = GameIO::getGameTime() + 1.0f ;
		AckPacket ack( hostAddr, GameIO::getMyName() ) ;
		out->send( &ack, sizeof( ack ) ) ;
	}

	if( GameIO::getGameTime() > timeToSendAck )
	{
		if( ackCounts <= 0 )
		{
			status = BANNED ;
			printf( "banning %08x\n", hostAddr ) ;
			if( GameIO::banHandler )
			{
				GameIO::banHandler( hostAddr, GameIO::banData );
			}
			return( FALSE ) ;
		}
		timeToSendAck = GameIO::getGameTime() + 1.0f ;
		ackCounts-- ;
		AckPacket ack( hostAddr, GameIO::getMyName() ) ;
		out->send( &ack, sizeof( ack ) ) ;
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Check the time since we last heard from a host and drop if not recent.
//------------------------------------------------------------------------------
SbBool
HostInfo::checkAge()
{
	if( status == ACCEPTED && GameIO::getGameTime() - age > 3.0f )
	{
		printf( "dropping %08x\n", hostAddr ) ;
		status = DROPPED ;
		return( FALSE ) ;
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// HostInfoList definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Truncate the HostInfoList and delete the HostInfo's.
//------------------------------------------------------------------------------
void
HostInfoList::truncate(
	int	start
	)
{
	for( int i = start ; i < getLength() ; i++ )
	{
		delete (*this)[i] ;
	}

	SbPList::truncate( start ) ;
}



//------------------------------------------------------------------------------
// Remove an entry from the HostInfoList and delete the entry.
//------------------------------------------------------------------------------
void
HostInfoList::remove(
	int	which
	)
{
	delete (*this)[which] ;

	SbPList::remove( which ) ;
}



//------------------------------------------------------------------------------
// Search for a host address in the current list and return the status if
// found, add and initialize if not found.
//------------------------------------------------------------------------------
SbBool
HostInfoList::isOkay(
	HostAddr remoteId
	)
{
	if( remoteId == 0 )
	{
		return( FALSE ) ;
	}

	for( int i = 0 ; i < getLength() ; i++ )
	{
		HostInfo *host = (*this)[i] ;
		if( host->hostAddr == remoteId )
		{
			return( host->isOkay() ) ;
		}
	}

	append( new HostInfo( remoteId, out ) ) ;

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Indicate that a host has been accepted.
//------------------------------------------------------------------------------
void
HostInfoList::accept(
	HostAddr	remoteId,
	const char	*name
	)
{
	for( int i = 0 ; i < getLength() ; i++ )
	{
		HostInfo *host = (*this)[i] ;
		if( host->hostAddr == remoteId )
		{
			host->accept( name ) ;
			return ;
		}
	}

	//
	// If we get here, it means that the first packet we received was
	// an ACK packet from the hosts.  Lets add him and accept him at
	// once.
	//
	HostInfo *host = new HostInfo( remoteId, out ) ;
	append( host ) ;
	host->accept( name ) ;
}



//------------------------------------------------------------------------------
// Search for a host entry based on the IP address.
//------------------------------------------------------------------------------
int
HostInfoList::find(
	HostAddr remoteId
	)
{
	for( int i = 0 ; i < getLength() ; i++ )
	{
		if( (*this)[i]->hostAddr == remoteId )
		{
			return( i ) ;
		}
	}

	return( -1 ) ;
}



//------------------------------------------------------------------------------
// AckPacket constructor.
//------------------------------------------------------------------------------
AckPacket::AckPacket(
	HostAddr	host,
	const char	*theName
	) :
	GamePacket( ACK_PKT ),
	remoteHost( host )
{
	strncpy( name, theName, sizeof( name ) ) ;
	name[GameIO::MAX_NAME_LEN-1] = '\0' ;
}



