//------------------------------------------------------------------------------
// GameInput class.  Reads data from a file or from the network, or from nada.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <Inventor/SbString.h>
#include <Inventor/errors/SoDebugError.h>
#include "GameInput.h"
#include "multicast.h"


extern int	errno ;

static const char *errorId = "GameInput" ;

class GameFileInput : public GameInput {

    public :
	GameFileInput( const char *name ) ;
	virtual ~GameFileInput() ;
	virtual int	get( void *buffer, int size ) ;

    private :
	SbString	filename ;
	int		fd ;
} ;



class GameNetworkInput : public GameInput {

    public :
	GameNetworkInput( const char *serviceName, const char *groupName,
				const char *interface, int defaultPort ) ;
	virtual ~GameNetworkInput() ;
	virtual int	get( void *buffer, int size ) ;

    private :
	int			fd ;
	struct sockaddr_in	inAddr ;
} ;



class GameNadaInput : public GameInput {

    public :
	GameNadaInput() {}
	virtual ~GameNadaInput() {}
	virtual int	get( void *, int ) { return( 0 ) ; }
} ;



//------------------------------------------------------------------------------
// Constructor for output to a file.
//------------------------------------------------------------------------------
GameInput::GameInput(
	const char	*filename
	)
{
	statusOk = TRUE ;
	if( filename )
	{
		rep = new GameFileInput( filename ) ;
		if( !statusOk )
		{
			delete rep ;
			rep = new GameNadaInput ;
		}
	}
	else
	{
		rep = new GameNadaInput ;
	}
}



//------------------------------------------------------------------------------
// Constructor for output to a network socket.
//------------------------------------------------------------------------------
GameInput::GameInput(
	const char	*serviceName,
	const char	*groupName,
	const char	*interface,
	int		defaultPort
	)
{
	statusOk = TRUE ;
	rep = new GameNetworkInput( serviceName, groupName, interface,
				defaultPort ) ;
	if( !statusOk )
	{
		delete rep ;
		rep = new GameNadaInput ;
	}
}



//------------------------------------------------------------------------------
// If get method fails, reset the reader to the Nada input class.
//------------------------------------------------------------------------------
int
GameInput::get(
	void	*buffer,
	int	size
	)
{
	int n = rep->get( buffer, size ) ;

	if( !rep->isOk() )
	{
		delete rep ;
		rep = new GameNadaInput() ;
	}

	return( n ) ;
}



//------------------------------------------------------------------------------
// GameFileInput constructor.
//------------------------------------------------------------------------------
GameFileInput::GameFileInput(
	const char	*name
	) :
	filename( name )
{
	if( ( fd = open( name, O_RDONLY ) ) < 0 )
	{
		SoDebugError::post( errorId, strerror( errno ), name ) ;
		statusOk = FALSE ;
	}
}



//------------------------------------------------------------------------------
// GameFileInput destructor.
//------------------------------------------------------------------------------
GameFileInput::~GameFileInput()
{
	if( fd >= 0 )
	{
		close( fd ) ;
	}
}



//------------------------------------------------------------------------------
// GameFileInput get method.
//------------------------------------------------------------------------------
int
GameFileInput::get(
	void	*buffer,
	int	size
	)
{
	GamePacket *packet = (GamePacket *)buffer ;
	int n = read( fd, packet, size ) ;

	if( n == 0 )
	{
		if( lseek( fd, 0, SEEK_SET ) != 0 )
		{
			n = -1 ;
		}
		else
		{
			n = read( fd, packet, size ) ;
		}
	}

	if( n < size )
	{
		SoDebugError::postWarning( errorId, strerror( errno ),
					filename.getString() ) ;
		statusOk = FALSE ;
	}

	return( n ) ;
}



//------------------------------------------------------------------------------
// GameNetworkInput constructor.
//------------------------------------------------------------------------------
GameNetworkInput::GameNetworkInput(
	const char	*serviceName,
	const char	*groupName,
	const char	*interface,
	int		defaultPort
	)
{
	int port = getPortFromService( serviceName, defaultPort ) ;

	uint32_t myAddr ;
	if( ( fd = openMulticastSocket( &inAddr, port, 0, 0, groupName,
					interface, "r", &myAddr ) ) < 0 )
	{
		SoDebugError::post( errorId, "Error opening multicast socket" );
		statusOk = FALSE ;
	}

	if( fcntl( fd, F_SETFL, FNDELAY ) < 0 )
	{
		SoDebugError::postWarning( errorId, strerror( errno ),
			"fcntl( F_SETFL, FNDELAY )" ) ;
		statusOk = FALSE ;
	}
}



//------------------------------------------------------------------------------
// GameNetworkInput destructor.
//------------------------------------------------------------------------------
GameNetworkInput::~GameNetworkInput()
{
	if( fd >= 0 )
	{
		close( fd ) ;
	}
}



//------------------------------------------------------------------------------
// GameNetworkInput get method.
//------------------------------------------------------------------------------
int
GameNetworkInput::get(
	void	*buffer,
	int	size
	)
{
	int inAddrLen ;

	GamePacket *packet = (GamePacket *)buffer ;
	int n = recvfrom( fd, packet, size, 0, &inAddr, &inAddrLen ) ;

	if( n < 0 && errno == EWOULDBLOCK )
	{
		n = 0 ;
		statusOk = TRUE ;
	}
	else if( n <= 0 )
	{
		SoDebugError::postWarning( errorId, strerror( errno ),
			"recvfrom" ) ;
		statusOk = FALSE ;
	}
/*
	else
	{
		packet->fromAddress = (HostAddr)inAddr.sin_addr.s_addr ;
		n += sizeof( HostAddr ) ;
	}
*/

	return( n ) ;
}
