//------------------------------------------------------------------------------
// GameOutput class.  Sends data to a file, to the network, or to nada.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <Inventor/SbString.h>
#include <Inventor/errors/SoDebugError.h>
#include "GameOutput.h"
#include "multicast.h"


extern int	errno ;

static const char *errorId = "GameOutput" ;

class GameFileOutput : public GameOutput {

    public :
	GameFileOutput( const char *name ) ;
	virtual ~GameFileOutput() ;
	virtual int	send( GamePacket *packet, int size ) ;
	virtual HostAddr getMyAddress() { return( 1 ) ; }

    private :
	SbString	filename ;
	int		fd ;
} ;



class GameNetworkOutput : public GameOutput {

    public :
	GameNetworkOutput( const char *serviceName, const char *groupName,
				const char *interface, int defaultPort,
				int loopback, int timeToLive ) ;
	virtual ~GameNetworkOutput() ;
	virtual int	send( GamePacket *packet, int size ) ;
	virtual HostAddr getMyAddress() { return( myAddr ) ; }

    private :
	int			fd ;
	struct sockaddr_in	outAddr ;
	HostAddr		myAddr ;
} ;



class GameNadaOutput : public GameOutput {

    public :
	GameNadaOutput() {}
	virtual ~GameNadaOutput() {}
	virtual int	send( GamePacket *, int size ) { return( size ) ; }
	virtual HostAddr getMyAddress() { return( 1 ) ; }
} ;



//------------------------------------------------------------------------------
// Constructor for output to a file.
//------------------------------------------------------------------------------
GameOutput::GameOutput(
	const char	*filename
	)
{
	statusOk = TRUE ;

	if( filename )
	{
		rep = new GameFileOutput( filename ) ;
		if( !statusOk )
		{
			delete rep ;
			rep = new GameNadaOutput ;
		}
	}
	else
	{
		rep = new GameNadaOutput ;
	}
}



//------------------------------------------------------------------------------
// Constructor for output to a network socket.
//------------------------------------------------------------------------------
GameOutput::GameOutput(
	const char	*serviceName,
	const char	*groupName,
	const char	*interface,
	int		defaultPort,
	int		loopback,
	int		timeToLive
	)
{
	statusOk = TRUE ;
	rep = new GameNetworkOutput( serviceName, groupName, interface,
				defaultPort, loopback, timeToLive ) ;
	if( !statusOk )
	{
		delete rep ;
		rep = new GameNadaOutput ;
	}
}



//------------------------------------------------------------------------------
// Reset the output handler to nada.
//------------------------------------------------------------------------------
int
GameOutput::send(
	GamePacket	*packet,
	int		size
	)
{
	int n = rep->send( packet, size ) ;

	if( n < size )
	{
		delete rep ;
		rep = new GameNadaOutput() ;
	}

	return( n ) ;
}



//------------------------------------------------------------------------------
// GameFileOutput constructor.
//------------------------------------------------------------------------------
GameFileOutput::GameFileOutput(
	const char	*name
	) :
	filename( name )
{
	if( ( fd = open( name, O_WRONLY | O_CREAT | O_TRUNC, 0644 ) ) < 0 )
	{
		SoDebugError::post( errorId, strerror( errno ), name ) ;
		statusOk = FALSE ;
	}
}



//------------------------------------------------------------------------------
// GameFileOutput destructor.
//------------------------------------------------------------------------------
GameFileOutput::~GameFileOutput()
{
	if( fd >= 0 )
	{
		close( fd ) ;
	}
}



//------------------------------------------------------------------------------
// GameFileOutput send method.
//------------------------------------------------------------------------------
int
GameFileOutput::send(
	GamePacket	*packet,
	int		size
	)
{
printf( "GameFileOutput::send\n" ) ;
	int n = write( fd, packet, size ) ;
	if( n < size )
	{
		SoDebugError::postWarning( errorId, strerror( errno ),
					filename.getString() ) ;
	}

	return( n ) ;
}



//------------------------------------------------------------------------------
// GameNetworkOutput constructor.
//------------------------------------------------------------------------------
GameNetworkOutput::GameNetworkOutput(
	const char	*serviceName,
	const char	*groupName,
	const char	*interface,
	int		defaultPort,
	int		loopback,
	int		timeToLive
	)
{
	int port = getPortFromService( serviceName, defaultPort ) ;

	if( timeToLive < MIN_TTL || timeToLive > MAX_TTL )
	{
		SoDebugError::postInfo( errorId, "Specified time-to-live (%d) "
			"does not fall within acceptable range (%d - %d).\nThe "
			"default value of %d will be used.\n", timeToLive,
			MIN_TTL, MAX_TTL, DEFAULT_TTL ) ;
		timeToLive = DEFAULT_TTL ;
	}

	u_char loop = loopback ;
	u_char ttl = timeToLive ;
	uint32_t theAddress ;
	if( ( fd = openMulticastSocket( &outAddr, port, ttl, loop, groupName,
					interface, "w", &theAddress ) ) < 0 )
	{
		SoDebugError::post( errorId, "Error opening multicast socket" );
		statusOk = FALSE ;
	}

	myAddr = (HostAddr)theAddress ;
}



//------------------------------------------------------------------------------
// GameNetworkOutput destructor.
//------------------------------------------------------------------------------
GameNetworkOutput::~GameNetworkOutput()
{
	if( fd >= 0 )
	{
		close( fd ) ;
	}
}



//------------------------------------------------------------------------------
// GameNetworkOutput send method.
//------------------------------------------------------------------------------
int
GameNetworkOutput::send(
	GamePacket	*packet,
	int		size
	)
{
	packet->fromAddress = myAddr ;

	int n = sendto( fd, packet, size, 0, &outAddr,
			sizeof( outAddr ) ) ;
	if( n < size )
	{
		SoDebugError::postWarning( errorId, strerror( errno ),
			"sendto" ) ;
	}

	return( n ) ;
}
