//------------------------------------------------------------------------------
// Class for holding game scenario data.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "GameScenario.h"
#include "GameControl.h"


DIR *GameScenario::listingDir = NULL ;
SbString GameScenario::listingDirName ;
SbString GameScenario::dataDir( "data/" ) ;



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
GameScenario::GameScenario(
	const char	*name
	)
{
	SbString dataname( dataDir ) ;
	dataname += SbString( name ) ;
	dataname += ".pbs" ;

	loadFile( dataname, TRUE ) ;
}



//------------------------------------------------------------------------------
// Private constructor.
//------------------------------------------------------------------------------
GameScenario::GameScenario()
{
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GameScenario::~GameScenario()
{
	if( description )
	{
		delete description ;
	}
}



//------------------------------------------------------------------------------
// Read in the data file.
//------------------------------------------------------------------------------
SbBool
GameScenario::loadFile(
	const SbString	&dataname,
	SbBool		fatalOnError
	)
{
	description = NULL ;

	FILE *fd = fopen( dataname.getString(), "r" ) ;
	if( fd == NULL )
	{
		if( fatalOnError )
		{
			perror( dataname.getString() ) ;
			exit( 1 ) ;
		}
		else
		{
			return( FALSE ) ;
		}
	}

	char buf[512] ;
	//
	// Check for valid header.
	//
	if( fgets( buf, sizeof( buf ), fd ) == NULL )
	{
		if( fatalOnError )
		{
			perror( dataname.getString() ) ;
			exit( 1 ) ;
		}
		else
		{
			return( FALSE ) ;
		}
	}

	const char *versionStr = "#Pointblank data V" ;
	int vl = strlen( versionStr ) ;
	if( strncmp( buf, versionStr, vl ) != 0 )
	{
		if( fatalOnError )
		{
			fprintf( stderr, "%s: incorrect header\n",
				dataname.getString() ) ;
			exit( 1 ) ;
		}
		else
		{
			return( FALSE ) ;
		}
	}

	float vers = atof( buf+vl ) ;
	if( vers != 1.0f )
	{
		if( fatalOnError )
		{
			fprintf( stderr, "%s: bad version number\n",
				dataname.getString() ) ;
			exit( 1 ) ;
		}
		else
		{
			return( FALSE ) ;
		}
	}

	id = -1 ;
	numTeams = -1 ;
	mode = GameControl::FREE ;
	filename.makeEmpty() ;

	int lineNo = 1 ;
	while( fgets( buf, sizeof( buf ), fd ) )
	{
		lineNo++ ;
		if( feof( fd ) )
		{
			break ;
		}
		if( buf[0] == '#' || buf[0] == '\n' )
		{
			continue ;
		}

		char *token = strtok( buf, ":" ) ;
		if( token == NULL )
		{
			continue ;
		}
		else if( strcmp( token, "Id" ) == 0 )
		{
			token = strtok( NULL, "\n" ) ;
			id = atoi( token ) ;
		}
		else if( strcmp( token, "Teams" ) == 0 )
		{
			token = strtok( NULL, "\n" ) ;
			numTeams = atoi( token ) ;
		}
		else if( strcmp( token, "Mode" ) == 0 )
		{
			token = strtok( NULL, "\n" ) ;
			while( *token == ' ' || *token == '\t' ) token++ ;
			if( strcmp( token, "free" ) == 0 )
			{
				mode = GameControl::FREE ;
			}
			else if( strcmp( token, "solo" ) == 0 )
			{
				mode = GameControl::SOLO ;
			}
			else if( strcmp( token, "goal" ) == 0 )
			{
				mode = GameControl::GOAL ;
			}
			else
			{
				if( fatalOnError )
				{
					fprintf( stderr,
						"%s: bad mode data '%s'\n",
						dataname.getString(), token ) ;
					exit( 1 ) ;
				}
				else
				{
					return( FALSE ) ;
				}
			}
		}
		else if( strcmp( token, "File" ) == 0 )
		{
			token = strtok( NULL, "\n" ) ;
			while( *token == ' ' || *token == '\t' ) token++ ;
			filename = token ;
		}
		else if( strcmp( token, "Description" ) == 0 )
		{
			int curPos = ftell( fd ) ;
			struct stat statBuf ;
			fstat( fileno( fd ), &statBuf ) ;
			int descSize = statBuf.st_size - curPos ;
			description = new char[statBuf.st_size - curPos] ;
			int actSize = fread( description, 1, descSize, fd ) ;
		}
		else
		{
			if( fatalOnError )
			{
				fprintf( stderr,
					"%s: unknown data at line %d\n",
					dataname.getString(), lineNo ) ;
			}
		}
	}

	if( id < 0 || numTeams < 0 || numTeams > GameControl::MAX_TEAMS )
	{
		if( fatalOnError )
		{
			fprintf( stderr, "%s: bad scenario data\n",
				dataname.getString() ) ;
			exit( 1 ) ;
		}
		else
		{
			return( FALSE ) ;
		}
	}
	fclose( fd ) ;

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// List available scenarios.
//------------------------------------------------------------------------------
void
GameScenario::listGames()
{
	const char	*nextFile ;
	GameScenario	*gs ;
	initListing( dataDir.getString() ) ;

	while( ( nextFile = nextListing() ) != NULL )
	{
		SbString title( nextFile, 0, strlen( nextFile ) - 5 ) ;
		SbString filename( listingDirName ) ;
		filename += nextFile ;
		gs = new GameScenario ;
		if( !gs->loadFile( filename.getString(), FALSE ) )
		{
			continue ;
		}
		printf( "Title: %s\n", title.getString() ) ;
		printf( "Mode: " ) ;
		switch( gs->mode )
		{
			case GameControl::FREE :
				printf( "FREE PLAY\n" ) ;
				break ;
			case GameControl::GOAL :
				printf( "GOAL\n" ) ;
				break ;
			case GameControl::SOLO :
				printf( "SOLO\n" ) ;
				break ;
		}
		if( gs->mode != GameControl::SOLO )
		{
			printf( "Number of teams: %d\n", gs->numTeams ) ;
		}
		if( gs->description )
		{
			printf( "Description:\n%s\n", gs->description ) ;
		}
		else
		{
			printf( "(No description)\n\n" ) ;
		}

		delete gs ;
	}

	endListing() ;
}



//------------------------------------------------------------------------------
// Initialize the data directory listing.
//------------------------------------------------------------------------------
void
GameScenario::initListing(
	const char *dirName
	)
{
	listingDir = opendir( dirName ) ;
	listingDirName = dirName ;
	if( listingDir == NULL )
	{
		perror( dirName ) ;
		exit( 1 ) ;
	}

	if( dirName[listingDirName.getLength() - 1] != '/' )
	{
		listingDirName += "/" ;
	}
}



//------------------------------------------------------------------------------
// Finish up the data directory listing.
//------------------------------------------------------------------------------
void
GameScenario::endListing()
{
	closedir( listingDir ) ;
}



//------------------------------------------------------------------------------
// Read the next Pointblank Scenario file.
//------------------------------------------------------------------------------
const char *
GameScenario::nextListing()
{
	while( 1 )
	{
		struct dirent *listing = readdir( listingDir ) ;

		if( listing == NULL )
		{
			break ;
		}

		int len = strlen( listing->d_name ) ;

		if( len > 4 && strcmp( listing->d_name+len-4, ".pbs" ) == 0 )
		{
			return( listing->d_name ) ;
		}
	}

	return( NULL ) ;
}



//------------------------------------------------------------------------------
// Static method to set the data directory.
//------------------------------------------------------------------------------
void
GameScenario::setDataDir(
	const char	*dirName
	)
{
	int l ;
	if( dirName && ( l = strlen( dirName ) ) > 0 )
	{
		dataDir = dirName ;
		if( dirName[l-1] != '/' )
		{
			dataDir += "/" ;
		}
	}
}
