//------------------------------------------------------------------------------
// Class for terrain-following an ordered grid.
//------------------------------------------------------------------------------
#include "GridTracker.h"
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoNormal.h>



//------------------------------------------------------------------------------
// Constructor.
//
// Assumptions:
//	X and Y values are evenly spaced.
//	X and Y are constant along rows and columns, respectively.
//	There is a one-to-one match for normals and coordinates.
//	All coordinates are used (startIndex = 0).
//	Y varies the fastest per index.
//	Both X and Y increment in the positive direction.
//------------------------------------------------------------------------------
GridTracker::GridTracker(
	SoCoordinate3	*terrainCoords,
	SoNormal	*terrainNormals,
	long		verticesPerColumn,
	long		verticesPerRow
	)
{
	coords = terrainCoords ;
	coords->ref() ;

	normals = terrainNormals ;
	normals->ref() ;

	vpCol = verticesPerColumn ;
	vpRow = verticesPerRow ;

	float z ;
	coords->point[0].getValue( minX, minY, z ) ;
	coords->point[vpCol*vpRow-1].getValue( maxX, maxY, z ) ;

	float dx, dy, dz ;
	( coords->point[vpRow] - coords->point[0] ).getValue( deltaX, dy, dz ) ;
	( coords->point[1] - coords->point[0] ).getValue( dx, deltaY, dz ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GridTracker::~GridTracker()
{
	coords->unref() ;
	normals->unref() ;
}



//------------------------------------------------------------------------------
// Track.
//
// Given a desired position, determine if the position is within the coordinates
//   being tracked (return !GridTracker::OK if so).
// Locate the cell containing the point.
// Determine if the point is in the lower-left or upper-right triangle of
//   the cell.
// In non-dimensional, projected XY space, weight the 3 corners of the triangle
//   based on the areas of the 3 triangles created by the desired position
//   point and the corner points:
//
//	C +		actual = A * area(BPC) / area(ABC)
//	  |\		       + B * area(APC) / area(ABC)
//	  | \		       + C * area(APB) / area(ABC)
//	  |  \
//	  |   \
//	  | P. \
//	  |     \
//      A +------+ B
//------------------------------------------------------------------------------
int
GridTracker::track(
	SbVec3f	&desiredPosition,
	SbVec3f	&actualPosition,
	SbVec3f	&actualNormal
	)
{
	float	x, y, z ;

	desiredPosition.getValue( x, y, z ) ;

	int status = GridTracker::OK ;
	if( x < minX ) status |= GridTracker::OFF_X_MIN ;
	if( x > maxX ) status |= GridTracker::OFF_X_MAX ;
	if( y < minY ) status |= GridTracker::OFF_Y_MIN ;
	if( y > maxY ) status |= GridTracker::OFF_Y_MAX ;

	if( status != GridTracker::OK ) return( status ) ;

	float s = x / deltaX ;
	float t = y / deltaY ;

	long nx = (long)s ;
	long ny = (long)t ;

	s -= (float)nx ;
	t -= (float)ny ;

	long n = nx * vpRow + ny ;
	long nA, nB, nC ;

	SbVec3f weight ;
	//
	// Upper right triangle of quad.
	//
	if( t + s > 1.0f )
	{
		nA = n + vpRow + 1 ;
		weight.setValue( s + t - 1.0f, 1.0f - t, 1.0f - s ) ;
	}
	//
	// Lower left triangle of quad.
	//
	else
	{
		nA = n ;
		weight.setValue( 1.0f - s - t, s, t ) ;
	}
	nB = n + vpRow ;
	nC = n + 1 ;

	actualPosition.setValue( weight, coords->point[nA], coords->point[nB],
				coords->point[nC] ) ;
	actualNormal.setValue( weight, normals->vector[nA], normals->vector[nB],
				normals->vector[nC] ) ;

	actualNormal.normalize() ;

	return( status ) ;
}
