//------------------------------------------------------------------------------
// GroundMesh Inventor node.
//------------------------------------------------------------------------------

#include <math.h>
#include <limits.h>
#include <assert.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/elements/SoGLNormalElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include "GroundMesh.h"
#include "FollowTerrainAction.h"
#include "LosAction.h"
#include "Utilities.h"

SO_NODE_SOURCE( GroundMesh ) ;

#define PRINT_VEC(s,v) \
	printf("%-20s: %7.4f %7.4f %7.4f\n",(s),(v)[0],(v)[1],(v)[2])

#define	XMIN_OUTER	( 1 << 0 )
#define	XMAX_OUTER	( 1 << 1 )
#define	YMIN_OUTER	( 1 << 2 )
#define	YMAX_OUTER	( 1 << 3 )

//------------------------------------------------------------------------------
// Initialize the GroundMesh class.
//------------------------------------------------------------------------------
void
GroundMesh::initClass()
{
	if( GroundShape::getClassTypeId() == SoType::badType() )
	{
		GroundShape::initClass() ;
	}
	SO_NODE_INIT_CLASS( GroundMesh, GroundShape, "GroundShape" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
GroundMesh::GroundMesh()
{
	SO_NODE_CONSTRUCTOR( GroundMesh ) ;
	SO_NODE_ADD_FIELD( verticesPerRow, ( 1 ) ) ;
	SO_NODE_ADD_FIELD( verticesPerColumn, ( 1 ) ) ;
	SO_NODE_ADD_FIELD( maxSize, ( 4 ) ) ;

	vprChangedSensor = new SoFieldSensor( GroundShape::somethingChangedCB,
						this ) ;
	vprChangedSensor->setPriority( 0 ) ;
	vprChangedSensor->attach( &verticesPerRow ) ;

	vpcChangedSensor = new SoFieldSensor( GroundShape::somethingChangedCB,
						this ) ;
	vpcChangedSensor->setPriority( 0 ) ;
	vpcChangedSensor->attach( &verticesPerColumn ) ;

	maxSizeChangedSensor = new SoFieldSensor(
				GroundShape::somethingChangedCB, this ) ;
	maxSizeChangedSensor->setPriority( 0 ) ;
	maxSizeChangedSensor->attach( &maxSize ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GroundMesh::~GroundMesh()
{
	delete vprChangedSensor ;
	delete vpcChangedSensor ;
	delete maxSizeChangedSensor ;
}



//------------------------------------------------------------------------------
// Read in the GroundMesh node.
//------------------------------------------------------------------------------
SbBool
GroundMesh::readInstance(
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoInput		*in,
	unsigned short	flags
#else
	SoInput	*in
#endif
	)
{
	// Detach sensors temporarily
	vprChangedSensor->detach() ;
	vpcChangedSensor->detach() ;
	maxSizeChangedSensor->detach() ;

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SbBool status = GroundShape::readInstance( in, flags ) ;
#else
	SbBool status = GroundShape::readInstance( in ) ;
#endif

	// Reattach sensors
	vprChangedSensor->attach( &verticesPerRow ) ;
	vpcChangedSensor->attach( &verticesPerColumn ) ;
	maxSizeChangedSensor->attach( &maxSize ) ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Helper utility to split a quad-like mesh of points into smaller tristrips.
//------------------------------------------------------------------------------
static void
splitQuad(
	SoSeparator	*parent,
	int32_t		vpRow,
	int32_t		cMin,
	int32_t		cMax,
	int32_t		rMin,
	int32_t		rMax,
	int32_t		maxSize
	)
{
	int32_t	cSize = cMax - cMin ;
	int32_t	rSize = rMax - rMin ;

	if( cSize >= maxSize || rSize >= maxSize )
	{
		int32_t cMid = ( cMin + cMax ) / 2 ;
		int32_t rMid = ( rMin + rMax ) / 2 ;

		SoSeparator *newParent = new SoSeparator ;
		parent->addChild( newParent ) ;
		newParent->renderCaching = SoSeparator::OFF ;
		newParent->boundingBoxCaching = SoSeparator::ON ;
		newParent->renderCulling = SoSeparator::ON ;
		newParent->pickCulling = SoSeparator::ON ;

		if( cSize >= maxSize )
		{
			if(  rSize >= maxSize )
			{
				splitQuad( newParent, vpRow,
					cMin, cMid+1, rMin, rMid+1, maxSize ) ;
				splitQuad( newParent, vpRow,
					cMid, cMax, rMin, rMid+1, maxSize ) ;
				splitQuad( newParent, vpRow,
					cMin, cMid+1, rMid, rMax, maxSize ) ;
				splitQuad( newParent, vpRow,
					cMid, cMax, rMid, rMax, maxSize ) ;
			}
			else
			{
				splitQuad( newParent, vpRow,
					cMin, cMid+1, rMin, rMax, maxSize ) ;
				splitQuad( newParent, vpRow,
					cMid, cMax, rMin, rMax, maxSize ) ;
			}
		}
		else
		{
			splitQuad( newParent, vpRow,
				cMin, cMax, rMin, rMid+1, maxSize ) ;
			splitQuad( newParent, vpRow,
				cMin, cMax, rMid, rMax, maxSize ) ;
		}
	}
	else
	{
		SoSeparator *newParent = new SoSeparator ;
		parent->addChild( newParent ) ;
		newParent->renderCaching = SoSeparator::ON ;
		newParent->boundingBoxCaching = SoSeparator::ON ;
		newParent->renderCulling = SoSeparator::ON ;
		newParent->pickCulling = SoSeparator::ON ;

		SoIndexedTriangleStripSet *ts = new SoIndexedTriangleStripSet ;
		newParent->addChild( ts ) ;

		int32_t nIndices = ( rSize - 1 ) * ( 2 * cSize + 1 ) ;
		ts->coordIndex.setNum( nIndices ) ;
		int32_t *lval = ts->coordIndex.startEditing() ;
		int32_t n = 0 ;
		for( int32_t r = rMin ; r < rMax - 1 ; r++ )
		{
			int32_t cStart = r * vpRow ;
			for( int32_t c = cMin ; c < cMax ; c++ )
			{
				lval[n++] = cStart + c ;
				lval[n++] = cStart + c + vpRow ;
			}
			lval[n++] = -1 ;
		}
		assert( n == nIndices ) ;
		ts->coordIndex.finishEditing() ;
	}
}



//------------------------------------------------------------------------------
// Build the scene graph.
//------------------------------------------------------------------------------
void
GroundMesh::buildIt()
{
	int32_t vpr = verticesPerRow.getValue() ;
	int32_t vpc = verticesPerColumn.getValue() ;

	hints->ref() ;
	tex->ref() ;
	coords->ref() ;
	xform->ref() ;
	root->removeAllChildren() ;
	root->addChild( hints ) ;
	root->addChild( xform ) ;
	root->addChild( tex ) ;
	root->addChild( coords ) ;
	hints->unrefNoDelete() ;
	tex->unrefNoDelete() ;
	coords->unrefNoDelete() ;
	xform->unrefNoDelete() ;

	int32_t nPts = vpr * vpc ;
	tex->point.setNum( nPts ) ;
	coords->point.setNum( nPts ) ;
	if( nPts <= 1 )
	{
		tex->point.setValue( 0.0f, 0.0f ) ;
		coords->point.setValue( 0.0f, 0.0f, 0.0f ) ;
	}
	else
	{
		int32_t nZ = height.getNum() ;
		SbVec2f *t = tex->point.startEditing() ;
		SbVec3f *c = coords->point.startEditing() ;
		int32_t k = 0 ;
		deltaL = length.getValue() / (float)( vpr - 1 ) ;
		deltaW = width.getValue() / (float)( vpc - 1 ) ;
		float sFactor = repeatS.getValue() / width.getValue() ;
		float tFactor = repeatT.getValue() / length.getValue() ;
		for( int32_t i = 0 ; i < vpc ; i++ )
		{
			float x = (float)i * deltaW ;
			for( int32_t j = 0 ; j < vpr ; j++ )
			{
				float y = (float)j * deltaL ;
				c[k].setValue( x, y, height[k%nZ] );
				t[k].setValue( x * sFactor, y * tFactor ) ;
				k++ ;
			}
		}
		tex->point.finishEditing() ;
		coords->point.finishEditing() ;

		splitQuad( root, vpr, 0, vpr, 0, vpc, maxSize.getValue() ) ;
	}

/*
	static int num = 0 ;
	SbString s( "terrain_" ) ;
	s += SbString( num ) ;
	s += SbString( ".iv" ) ;
	Utilities::dumpIt( s.getString(), root ) ;
	num++ ;
*/
}



//------------------------------------------------------------------------------
// GetHeight.
//
// Locate the cell containing the point.
// Determine if the point is in the lower-left or upper-right triangle of
//   the cell.
// In non-dimensional, projected XY space, weight the 3 corners of the triangle
//   based on the areas of the 3 triangles created by the desired position
//   point and the corner points:
//
//	C +		actual = A * area(BPC) / area(ABC)
//	  |\		       + B * area(APC) / area(ABC)
//	  | \		       + C * area(APB) / area(ABC)
//	  |  \
//	  |   \
//	  | P. \
//	  |     \
//      A +------+ B
//------------------------------------------------------------------------------
float
GroundMesh::getHeight(
	float	w,
	float	l,
	SbVec3f	&weight,
	int32_t	&nA,
	int32_t	&nB,
	int32_t	&nC
	)
{
	float s = w / deltaW ;
	float t = l / deltaL ;

	int32_t nx = (int32_t)s ;
	int32_t ny = (int32_t)t ;

	s -= (float)nx ;
	t -= (float)ny ;

	int32_t vpRow = verticesPerRow.getValue() ;

	int32_t n = nx * vpRow + ny ;

	//
	// Upper right triangle of quad.
	//
	if( t + s > 1.0f )
	{
		nA = n + vpRow + 1 ;
		nB = n + vpRow ;
		nC = n + 1 ;
		weight.setValue( s + t - 1.0f, 1.0f - t, 1.0f - s ) ;
	}
	//
	// Lower left triangle of quad.
	//
	else
	{
		nA = n ;
		nB = n + 1 ;
		nC = n + vpRow ;
		weight.setValue( 1.0f - s - t, t, s ) ;
	}

	return( weight[0] * coords->point[nA][2] +
		weight[1] * coords->point[nB][2] +
		weight[2] * coords->point[nC][2] ) ;
}



//------------------------------------------------------------------------------
// Find the relative height of a point given X and Y on the shape, determine
// the triangle that it is on, and the interpolated normal at that point.
//------------------------------------------------------------------------------
void
GroundMesh::intersect(
	FollowTerrainAction	*fta,
	float			w,
	float			l
	)
{
	int32_t nA, nB, nC ;
	SbVec3f weight ;

	float zValue = getHeight( w, l, weight, nA, nB, nC ) ;

	// Calculate the normal.
	SbVec3f normal ;
	switch( SoNormalBindingElement::get( fta->getState() ) )
	{
		case SoNormalBindingElement::OVERALL :
		default :
			normal = vector[0] ;
			break ;

		case SoNormalBindingElement::PER_VERTEX :
			if( vector.getNum() == height.getNum() )
			{
				normal.setValue( weight, vector[nA], vector[nB],
						vector[nC] ) ;
				normal.normalize() ;
			}
			else
			{
				normal = vector[0] ;
			}
			break ;
		case SoNormalBindingElement::PER_FACE_INDEXED :
printf( "FIX PER_FACE_INDEXED terrain following!!!!\n" ) ;
			break ;
		case SoNormalBindingElement::PER_VERTEX_INDEXED :
			normal.setValue( weight, vector[nA], vector[nB],
					vector[nC] ) ;
			normal.normalize() ;
			break ;
	}

	// Adjust normal for rotation
	SbVec3f adjNorm( normal[0]*rotCos-normal[1]*rotSin,
			 normal[0]*rotSin+normal[1]*rotCos, normal[2] ) ;

	SbVec3f pt( fta->getSearchPoint() ) ;
	pt[2] = zValue + origin.getValue()[2] ;

	GroundPoint &gp = fta->addIntersection( pt, adjNorm ) ;
	// Save the resulting (transformed) points of the hit triangle.
	SbVec3f corner[3] ;
	const SbVec3f &orig = origin.getValue() ;
	corner[0].setValue(
		coords->point[nA][0]*rotCos-coords->point[nA][1]*rotSin+orig[0],
		coords->point[nA][0]*rotSin+coords->point[nA][1]*rotCos+orig[1],
		coords->point[nA][2]+orig[2] ) ;
	corner[1].setValue(
		coords->point[nB][0]*rotCos-coords->point[nB][1]*rotSin+orig[0],
		coords->point[nB][0]*rotSin+coords->point[nB][1]*rotCos+orig[1],
		coords->point[nB][2]+orig[2] ) ;
	corner[2].setValue(
		coords->point[nC][0]*rotCos-coords->point[nC][1]*rotSin+orig[0],
		coords->point[nC][0]*rotSin+coords->point[nC][1]*rotCos+orig[1],
		coords->point[nC][2]+orig[2] ) ;
	gp.setPoly( 3, corner ) ;
}



//------------------------------------------------------------------------------
// Check to see if a ray intersects the shape.
//------------------------------------------------------------------------------
SbBool
GroundMesh::checkObstruction( 
	SbVec3f	&localStart,
	SbVec3f	&localEnd,
	SbBool	findDetail
	)
{
	SbBool	hit = FALSE ;

	SbVec3f &delta = localEnd - localStart ;
	float t ;
	float maxDim[2] ;
	maxDim[0] = width.getValue() ;
	maxDim[1] = length.getValue() ;

	// Trim to X and Y edges.
	for( int i = 0 ; i < 2 ; i++ )
	{
		if( delta[i] != 0.0f )
		{
			if( localStart[i] < 0.0f )
			{
				t = -localStart[i] / delta[i] ;
				localStart.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
			else if( localStart[i] > maxDim[i] )
			{
				t = ( maxDim[i] - localStart[i] ) / delta[i] ;
				localStart.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
			if( localEnd[i] < 0.0f )
			{
				t = -localStart[i] / delta[i] ;
				localEnd.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
			else if( localEnd[i] > maxDim[i] )
			{
				t = ( maxDim[i] - localStart[i] ) / delta[i] ;
				localEnd.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
		}
	}

	// Now check to see on which side of the surface (above or below) the
	// endpoints are.
	SbVec3f	weight ;
	int32_t	nA, nB, nC ;
	float heightStart = getHeight( localStart[0], localStart[1], weight,
		nA, nB, nC ) ;
	float heightEnd = getHeight( localEnd[0], localEnd[1], weight,
		nA, nB, nC ) ;

	SbBool startAbove = ( heightStart < localStart[2] ) ;
	SbBool endAbove = ( heightEnd < localEnd[2] ) ;

	if( startAbove != endAbove && !findDetail )
	{
		return( TRUE ) ;
	}

	// Quick check if this is planar (ie, there is only one height)
	int32_t nZ = height.getNum() ;
	if( nZ == 1 && startAbove == endAbove )
	{
		return( FALSE ) ;
	}

	float	invDeltaW = 1.0f / deltaW ;
	float	invDeltaL = 1.0f / deltaL ;

	int vpr = verticesPerRow.getValue() ;

	// Check to see if line crosses X-constant lines.
	if( delta[0] != 0.0f )
	{
		int i0, i1, iDelta ;
		if( localStart[0] < localEnd[0] )
		{
			i0 = (int)ceilf( localStart[0] * invDeltaW ) ;
			i1 = (int)floorf( localEnd[0] * invDeltaW ) + 1 ;
			iDelta = 1 ;
		}
		else
		{
			i0 = (int)floorf( localStart[0] * invDeltaW ) ;
			i1 = (int)ceilf( localEnd[0] * invDeltaW ) - 1 ;
			iDelta = -1 ;
		}
		
		float k0 = delta[1] / delta[0] ;
		float k1 = localStart[1] - k0 * localStart[0] ;
		float m0 = delta[2] / delta[0] ;
		float m1 = localStart[2] - m0 * localStart[0] ;

		if( i0 != i1 )
		{
			for( i = i0 ; i != i1 ; i += iDelta )
			{
				float x = (float)i * deltaW ;
				float y = ( k0 * x + k1 ) * invDeltaL ;

				float y0 = floorf( y ) ;
				int j0 = (int)y0 ;
				int j1 = (int)ceilf( y ) ;

				float s = y - y0 ;
				int n0 = ( i * vpr + j0 ) % nZ ;
				int n1 = ( i * vpr + j1 ) % nZ ;
				float zGround = s * height[n1] +
						( 1.0f - s ) * height[n0] ;
				float zLine = m0 * x + m1 ;
				if( ( zLine > zGround ) != startAbove )
				{
					if( !findDetail )
					{
						return( TRUE ) ;
					}
					else
					{
						if( !adjustEnd( i - iDelta, i,
							j0, j1, localStart,
							localEnd ) )
						{
							localEnd.setValue( x,
							    y * deltaL, zLine ) ;
						}
						hit = TRUE ;
						delta = localEnd - localStart ;
						break ;
					}
				}
			}
		}
	}

	// Check to see if line crosses Y-constant lines.
	if( delta[1] != 0.0f )
	{
		int j0, j1, jDelta ;
		if( localStart[1] < localEnd[1] )
		{
			j0 = (int)ceilf( localStart[1] * invDeltaL ) ;
			j1 = (int)floorf( localEnd[1] * invDeltaL ) + 1 ;
			jDelta = 1 ;
		}
		else
		{
			j0 = (int)floorf( localStart[1] * invDeltaL ) ;
			j1 = (int)ceilf( localEnd[1] * invDeltaL ) - 1 ;
			jDelta = -1 ;
		}

		float k0 = delta[0] / delta[1] ;
		float k1 = localStart[0] - k0 * localStart[1] ;
		float m0 = delta[2] / delta[1] ;
		float m1 = localStart[2] - m0 * localStart[1] ;

		if( j0 != j1 )
		{
			for( int j = j0 ; j != j1 ; j += jDelta )
			{
				float y = (float)j * deltaL ;
				float x = ( k0 * y + k1 ) * invDeltaW ;

				float x0 = floorf( x ) ;
				int i0 = (int)x0 ;
				int i1 = (int)ceilf( x ) ;

				float s = x - x0 ;
				int n0 = ( i0 * vpr + j ) % nZ ;
				int n1 = ( i1 * vpr + j ) % nZ ;
				float zGround = s * height[n1] +
						( 1.0f - s ) * height[n0] ;
				float zLine = m0 * y + m1 ;
				if( ( zLine > zGround ) != startAbove )
				{
					if( !findDetail )
					{
						return( TRUE ) ;
					}
					else
					{
						if( !adjustEnd( i0, i1,
							j - jDelta, j,
							localStart, localEnd ) )
						{
							localEnd.setValue(
								x * deltaW,
								y, zLine ) ;
						}
						hit = TRUE ;
						delta = localEnd - localStart ;
						break ;
					}
				}
			}
		}
	}

	// Finally, check to see if any diagonal lines are crossed.
	if( delta[0] * deltaL != -delta[1] * deltaW )
	{
		float k0 = delta[1] / delta[0] ;
		float k1 = localStart[1] - k0 * localStart[0] ;
		float m0 = delta[2] / delta[0] ;
		float m1 = localStart[2] - m0 * localStart[0] ;

		//
		// Let l be a diagonal line such that i + j = constant = l
		//
		float lStart = ( localStart[0] * invDeltaW +
			localStart[1] * invDeltaL ) ;
		float lEnd = ( localEnd[0] * invDeltaW +
			localEnd[1] * invDeltaL ) ;

		int l0, l1, lDelta ;
		if( lStart < lEnd )
		{
			l0 = (int)ceilf( lStart ) ;
			l1 = (int)floorf( lEnd ) + 1 ;
			lDelta = 1 ;
		}
		else
		{
			l0 = (int)floorf( lStart ) ;
			l1 = (int)ceilf( lEnd ) - 1 ;
			lDelta = -1 ;
		}

		if( l0 != l1 )
		{
			float k2 = 1.0f / ( k0 + deltaL * invDeltaW ) ;
			for( int l = l0 ; l != l1 ; l += lDelta )
			{
				float x ;
				if( delta[0] == 0.0f )
				{
					x = localStart[0] * invDeltaW ;
				}
				else
				{
					x = ( (float)l * deltaL - k1 ) * k2 *
						invDeltaW ;
				}
				float s = floorf( x ) ;
				int i0 = (int)s ;
				int i1 = i0 + 1 ;
				s = x - s ;
				int n0 = ( i0 * vpr + l - i0 ) % nZ ;
				int n1 = ( i1 * vpr + l - i1 ) % nZ ;
				float zGround = s * height[n1] +
						( 1.0f - s ) * height[n0] ;
				x *= deltaW ;
				float zLine = m0 * x + m1 ;
				if( ( zLine > zGround ) != startAbove )
				{
					if( !findDetail )
					{
						return( TRUE ) ;
					}
					else
					{
						if( !adjustEnd( i0, i1,
							l - i0, l - i1,
							localStart, localEnd ) )
						{
							float y = k0 * x + k1 ;
							localEnd.setValue(
								x, y, zLine ) ;
						}
						hit = TRUE ;
						delta = localEnd - localStart ;
						break ;
					}
				}
				
			}
		}
	}
	//
	// If the following is true, it means we're looking for full detail
	// but didn't cross a line.  Therefore, assume it's a plane between
	// the 2 points and solve for where the line intersects the plane.
	//
	if( startAbove != endAbove && !hit )
	{
		heightStart -= localStart[2] ;
		heightEnd -= localEnd[2] ;
		float t = heightStart / ( heightStart - heightEnd ) ;
		localEnd = localStart + t * delta ;
		hit = TRUE ;
	}

	return( hit ) ;
}



//------------------------------------------------------------------------------
// Adjust the end of a ray by intersecting it with the quad given.
//------------------------------------------------------------------------------
SbBool
GroundMesh::adjustEnd(
	int		i0,
	int		i1,
	int		j0,
	int		j1,
	const SbVec3f	&localStart,
	SbVec3f		&localEnd
	)
{
	int ib ;
	int ie ;
	int jb ;
	int je ;

	if( i0 < i1 )
	{
		ib = i0 ;
		ie = i1 ;
	}
	else
	{
		ib = i1 ;
		ie = i0 ;
	}

	if( j0 < j1 )
	{
		jb = j0 ; 
		je = j1 ;
	}
	else
	{
		jb = j1 ;
		je = j0 ;
	}
	float xb = (float)ib * deltaW ;
	float xe = (float)ie * deltaW ;
	float yb = (float)jb * deltaL ;
	float ye = (float)je * deltaL ;
	int vpr = verticesPerRow.getValue() ;
	int32_t nZ = height.getNum() ;
	SbVec3f pbb( xb, yb, height[(ib * vpr + jb) % nZ] ) ;
	SbVec3f peb( xe, yb, height[(ie * vpr + jb) % nZ] ) ;
	SbVec3f pbe( xb, ye, height[(ib * vpr + je) % nZ] ) ;
	SbVec3f pee( xe, ye, height[(ie * vpr + je) % nZ] ) ;

	SbLine line( localStart, localEnd ) ;
	SbPlane plane( pbb, peb, pbe ) ;
	SbVec3f pt ;
	if( plane.intersect( line, pt ) && pt[0] >= xb && pt[1] >= yb &&
		pt[0] + pt[1] <= (float)( xb + ye ) )
	{
		// Make sure pt is within the endpoints.
		SbVec3f &a = pt - localStart ;
		SbVec3f &b = localEnd - localStart ;
		float d = a.dot( b ) ;
		if( d >= 0.0f && a.dot( a ) <= b.dot( b ) )
		{
			localEnd = pt ;
			return( TRUE ) ;
		}
	}

	plane = SbPlane( pee, peb, pbe ) ;
	if( plane.intersect( line, pt ) && pt[0] <= xe && pt[1] <= ye &&
		pt[0] + pt[1] >= (float)( xb + ye ) )
	{
		// Make sure pt is within the endpoints.
		SbVec3f &a = pt - localStart ;
		SbVec3f &b = localEnd - localStart ;
		float d = a.dot( b ) ;
		if( d >= 0.0f && a.dot( a ) <= b.dot( b ) )
		{
			localEnd = pt ;
			return( TRUE ) ;
		}
	}

	return( FALSE ) ;
}
