//------------------------------------------------------------------------------
// GroundShape abstract Inventor node.
//------------------------------------------------------------------------------

#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/elements/SoGLNormalElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <math.h>
#include <limits.h>
#include "GroundShape.h"
#include "FollowTerrainAction.h"
#include "LosAction.h"

SO_NODE_ABSTRACT_SOURCE( GroundShape ) ;

#define PRINT_VEC(s,v) \
	printf("%-20s: %7.4f %7.4f %7.4f\n",(s),(v)[0],(v)[1],(v)[2])

#define	XMIN_OUTER	( 1 << 0 )
#define	XMAX_OUTER	( 1 << 1 )
#define	YMIN_OUTER	( 1 << 2 )
#define	YMAX_OUTER	( 1 << 3 )

//------------------------------------------------------------------------------
// Initialize the GroundShape class.
//------------------------------------------------------------------------------
void
GroundShape::initClass()
{
	SO_NODE_INIT_ABSTRACT_CLASS( GroundShape, SoNode, "Node" ) ;

	SO_ENABLE( SoCallbackAction, SoGLNormalElement ) ;
	SO_ENABLE( SoGLRenderAction, SoGLNormalElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoGLNormalElement ) ;
	SO_ENABLE( SoPickAction, SoGLNormalElement ) ;
	SO_ENABLE( SoGetMatrixAction, SoGLNormalElement ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
GroundShape::GroundShape() :
	children( this )
{
	SO_NODE_CONSTRUCTOR( GroundShape ) ;
	SO_NODE_ADD_FIELD( origin, ( 0.0f, 0.0f, 0.0f ) ) ;
	SO_NODE_ADD_FIELD( rotation, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( width, ( 1.0f ) ) ;
	SO_NODE_ADD_FIELD( length, ( 1.0f ) ) ;
	SO_NODE_ADD_FIELD( repeatS, ( 1.0f ) ) ;
	SO_NODE_ADD_FIELD( repeatT, ( 1.0f ) ) ;
	SO_NODE_ADD_FIELD( height, ( 0.0f ) ) ;
	SO_NODE_ADD_FIELD( vector, ( 0.0f, 0.0f, 0.0f ) ) ;
	vector.deleteValues( 0 ) ; // Nuke bogus normal

	rotationChangedSensor = new SoFieldSensor( rotationChangedCB, this ) ;
	rotationChangedSensor->setPriority( 0 ) ;
	rotationChangedSensor->attach( &rotation ) ;

	widthChangedSensor = new SoFieldSensor( somethingChangedCB, this ) ;
	widthChangedSensor->setPriority( 0 ) ;
	widthChangedSensor->attach( &width ) ;

	lengthChangedSensor = new SoFieldSensor( somethingChangedCB, this ) ;
	lengthChangedSensor->setPriority( 0 ) ;
	lengthChangedSensor->attach( &length ) ;

	heightChangedSensor = new SoFieldSensor( somethingChangedCB, this ) ;
	heightChangedSensor->setPriority( 0 ) ;
	heightChangedSensor->attach( &height ) ;

	root = new SoSeparator ;
	root->ref() ;
	root->renderCaching = SoSeparator::OFF ;
	root->boundingBoxCaching = SoSeparator::ON ;
	root->renderCulling = SoSeparator::ON ;
	root->pickCulling = SoSeparator::ON ;

	hints = new SoShapeHints ;
	root->addChild( hints ) ;
	hints->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE ;
	hints->faceType = SoShapeHints::CONVEX ;
	hints->shapeType.setIgnored( TRUE ) ;

	xform = new SoTransform ;
	root->addChild( xform ) ;
	xform->translation.connectFrom( &origin ) ;

	tex = new SoTextureCoordinate2 ;
	root->addChild( tex ) ;

	coords = new SoCoordinate3 ;
	root->addChild( coords ) ;

	children.append( root ) ;

	deltaW = 1.0f ;
	deltaL = 1.0f ;
	rotCos = cosf( rotation.getValue() ) ;
	rotSin = sinf( rotation.getValue() ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GroundShape::~GroundShape()
{
	root->unref() ;
	delete rotationChangedSensor ;
	delete widthChangedSensor ;
	delete lengthChangedSensor ;
	delete heightChangedSensor ;
}



//------------------------------------------------------------------------------
// Read in the GroundShape node.
//------------------------------------------------------------------------------
SbBool
GroundShape::readInstance(
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoInput		*in,
	unsigned short	flags
#else
	SoInput	*in
#endif
	)
{
	SbBool status = TRUE ;

	// Detach sensors temporarily
	rotationChangedSensor->detach() ;
	widthChangedSensor->detach() ;
	lengthChangedSensor->detach() ;
	heightChangedSensor->detach() ;

	// Read field info as usual.
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	if( !SoFieldContainer::readInstance( in, flags ) )
#else
	if( !SoFieldContainer::readInstance( in ) )
#endif
	{
		status = FALSE ;
	}
	else
	{
		rotationChangedCB( this, NULL ) ;
		somethingChangedCB( this, NULL ) ;
	}

	// Reattach sensors
	rotationChangedSensor->attach( &rotation ) ;
	widthChangedSensor->attach( &width ) ;
	lengthChangedSensor->attach( &length ) ;
	heightChangedSensor->attach( &height ) ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Override the default value.
//------------------------------------------------------------------------------
SbBool
GroundShape::affectsState() const
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Field sensor callback.
//------------------------------------------------------------------------------
void
GroundShape::somethingChangedCB(
	void		*data,
	SoSensor	*
	)
{
	GroundShape *g = (GroundShape *)data ;

	g->buildIt() ;
}



//------------------------------------------------------------------------------
// Rotation field sensor callback.
//------------------------------------------------------------------------------
void
GroundShape::rotationChangedCB(
	void		*data,
	SoSensor	*
	)
{
	GroundShape *g = (GroundShape *)data ;

	g->rotCos = cosf( g->rotation.getValue() ) ;
	g->rotSin = sinf( g->rotation.getValue() ) ;
	g->xform->rotation.setValue( SbVec3f( 0.0f, 0.0f, 1.0f ),
					g->rotation.getValue() ) ;
}



//------------------------------------------------------------------------------
// Generic action handler.
//------------------------------------------------------------------------------
void
GroundShape::doAction(
	SoAction	*action
	)
{
	int numIndices ;
	const int *indices ;

	if( !vector.isIgnored() )
	{
		SoNormalElement::set( action->getState(), this,
			vector.getNum(), vector.getValues( 0 ) ) ;
	}

	if( action->getPathCode( numIndices, indices ) == SoAction::IN_PATH )
	{
		children.traverse( action, 0, indices[numIndices-1] ) ;
	}
	else
	{
		children.traverse( action ) ;
	}
}



//------------------------------------------------------------------------------
// GLRenderAction handler.
//------------------------------------------------------------------------------
void
GroundShape::GLRender(
	SoGLRenderAction *action
	)
{
	GroundShape::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetBoundBoxAction handler.
//------------------------------------------------------------------------------
void
GroundShape::getBoundingBox(
	SoGetBoundingBoxAction *action
	)
{
	GroundShape::doAction( action ) ;
}



//------------------------------------------------------------------------------
// CallbackAction handler.
//------------------------------------------------------------------------------
void
GroundShape::callback(
	SoCallbackAction *action
	)
{
	GroundShape::doAction( action ) ;
}



//------------------------------------------------------------------------------
// PickAction handler.
//------------------------------------------------------------------------------
void
GroundShape::pick(
	SoPickAction *action
	)
{
	GroundShape::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetMatrixAction handler.
//------------------------------------------------------------------------------
void
GroundShape::getMatrix(
	SoGetMatrixAction *action
	)
{
	GroundShape::doAction( action ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the root node.
//------------------------------------------------------------------------------
SoChildList *
GroundShape::getChildren() const
{
	return( (SoChildList *)&children ) ;
}



//------------------------------------------------------------------------------
// FollowTerrainAction static handler.
//------------------------------------------------------------------------------
void
GroundShape::followTerrainS(
	SoAction	*action,
	SoNode		*node
	)
{
	( (GroundShape *)node )->followTerrain( (FollowTerrainAction *)action );
}



//------------------------------------------------------------------------------
// FollowTerrainAction handler.
//------------------------------------------------------------------------------
void
GroundShape::followTerrain(
	FollowTerrainAction *fta
	)
{
	SbVec3f wVec( rotCos, rotSin, 0.0f ) ;
	SbVec3f lVec( -wVec[1], wVec[0], 0.0f ) ;

	SbVec3f &pt = fta->getSearchPoint() - origin.getValue() ;

	float w = wVec.dot( pt ) ;
	if( w < 0.0f || w > width.getValue() )
	{
		return ;
	}

	float l = lVec.dot( pt ) ;
	if( l < 0.0f || l > length.getValue() )
	{
		return ;
	}

	intersect( fta, w, l ) ;
}



//------------------------------------------------------------------------------
// LosAction static handler.
//------------------------------------------------------------------------------
void
GroundShape::obstructS(
	SoAction	*action,
	SoNode		*node
	)
{
	( (GroundShape *)node )->obstruct( (LosAction *)action ) ;
}



//------------------------------------------------------------------------------
// LosAction handler.
//------------------------------------------------------------------------------
void
GroundShape::obstruct(
	LosAction *la
	)
{
	SbVec3f &localStart = la->getStartPoint() - origin.getValue() ;
	float x = localStart[0] * rotCos + localStart[1] * rotSin ;
	float y = localStart[1] * rotCos - localStart[0] * rotSin ;
	localStart[0] = x ;
	localStart[1] = y ;

	SbVec3f &localEnd = la->getEndPoint() - origin.getValue() ;
	x = localEnd[0] * rotCos + localEnd[1] * rotSin ;
	y = localEnd[1] * rotCos - localEnd[0] * rotSin ;
	localEnd[0] = x ;
	localEnd[1] = y ;

	SbBool findDetail = ( la->getMode() == LosAction::FIND_CLOSEST ) ;

	//
	// Do a quick check to see if should be checked further.
	//
	int startStatus = 0 ;
	if( localStart[0] < 0.0f ) startStatus |= XMIN_OUTER ;
	if( localStart[1] < 0.0f ) startStatus |= YMIN_OUTER ;
	if( localStart[0] > width.getValue() ) startStatus |= XMAX_OUTER ;
	if( localStart[1] > length.getValue() ) startStatus |= YMAX_OUTER ;

	int endStatus = 0 ;
	if( localEnd[0] < 0.0f ) endStatus |= XMIN_OUTER ;
	if( localEnd[1] < 0.0f ) endStatus |= YMIN_OUTER ;
	if( localEnd[0] > width.getValue() ) endStatus |= XMAX_OUTER ;
	if( localEnd[1] > length.getValue() ) endStatus |= YMAX_OUTER ;

	if( ( startStatus & endStatus ) != 0 )
	{
		return ;
	}

	if( checkObstruction( localStart, localEnd, findDetail ) )
	{
		if( findDetail )
		{
			x = localEnd[0] * rotCos - localEnd[1] * rotSin ;
			y = localEnd[1] * rotCos + localEnd[0] * rotSin ;
			localEnd[0] = x ;
			localEnd[1] = y ;
			localEnd += origin.getValue() ;
			la->trimRay( localEnd ) ;
		}
		la->setObstructed() ;
	}
}



