//------------------------------------------------------------------------------
// GroundStrip Inventor node.
//------------------------------------------------------------------------------

#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <math.h>
#include <limits.h>
#include "GroundStrip.h"
#include "FollowTerrainAction.h"
#include "Utilities.h"

SO_NODE_SOURCE( GroundStrip ) ;

#define	XMIN_OUTER	( 1 << 0 )
#define	XMAX_OUTER	( 1 << 1 )
#define	YMIN_OUTER	( 1 << 2 )
#define	YMAX_OUTER	( 1 << 3 )

//------------------------------------------------------------------------------
// Initialize the GroundStrip class.
//------------------------------------------------------------------------------
void
GroundStrip::initClass()
{
	if( GroundShape::getClassTypeId() == SoType::badType() )
	{
		GroundShape::initClass() ;
	}
	SO_NODE_INIT_CLASS( GroundStrip, GroundShape, "GroundShape" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
GroundStrip::GroundStrip()
{
	SO_NODE_CONSTRUCTOR( GroundStrip ) ;

	root->addChild( new SoTriangleStripSet ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GroundStrip::~GroundStrip()
{
}



//------------------------------------------------------------------------------
// Build the scene graph.
//------------------------------------------------------------------------------
void
GroundStrip::buildIt()
{
	int32_t nZ = height.getNum() ;
	int32_t nPts = 2 * nZ ;
	coords->point.setNum( nPts ) ;
	tex->point.setNum( nPts ) ;
	if( nZ <= 1 )
	{
		coords->point.setValue( 0.0f, 0.0f, 0.0f ) ;
		tex->point.setValue( 0.0f, 0.0f ) ;
	}
	else
	{
		SbVec3f *c = coords->point.startEditing() ;
		SbVec2f *t = tex->point.startEditing() ;
		int32_t k = 0 ;
		deltaL = length.getValue() / (float)( nZ - 1 ) ;
		deltaW = width.getValue() ;
		float sFactor = repeatS.getValue() / width.getValue() ;
		float tFactor = repeatT.getValue() / length.getValue() ;
		for( int32_t i = 0 ; i < nZ ; i++ )
		{
			float y = (float)i * deltaL ;
			c[k  ].setValue( 0.0f, y, height[i] );
			c[k+1].setValue( deltaW, y, height[i] );
			t[k  ].setValue( 0.0f, y * tFactor ) ;
			t[k+1].setValue( deltaW * sFactor, y * tFactor ) ;
			k += 2 ;
		}
		coords->point.finishEditing() ;
		tex->point.finishEditing() ;
	}
}



//------------------------------------------------------------------------------
// Determine the Z location on the strip at a given X,Y location.
//------------------------------------------------------------------------------
float
GroundStrip::getHeight(
	float	l,
	SbVec2f	&weight,
	int32_t	&nA,
	int32_t	&nB
	)
{
	float	z ;
	weight[1] = l / deltaL ;

	nA = (int32_t)weight[1] ;

	weight[1] -= (float)nA ;
	weight[0] = 1.0f - weight[1] ;

	nB = nA + 1 ;
	SbVec3f actualPosition ;
	if( nB >= height.getNum() )
	{
		z = height[nA] ;
		nB = nA ;
		nA-- ;
		weight[0] = 0.0f ;
		weight[1] = 1.0f ;
	}
	else
	{
		z = weight[0] * height[nA] + weight[1] * height[nB] ;
	}

	return( z ) ;
}



//------------------------------------------------------------------------------
// Determine the Z location on the strip at a given X,Y location.
//------------------------------------------------------------------------------
void
GroundStrip::intersect(
	FollowTerrainAction	*fta,
	float			,
	float			l
	)
{
	SbVec2f	weight ;
	int32_t	nA, nB ;
	float zGround = getHeight( l, weight, nA, nB ) ;

	// Calculate the normal.
	nA *= 2 ;
	nB *= 2 ;
	SbVec3f normal ;
	switch( SoNormalBindingElement::get( fta->getState() ) )
	{
		case SoNormalBindingElement::OVERALL :
		default :
			normal = vector[0] ;
			break ;

		case SoNormalBindingElement::PER_VERTEX :
		case SoNormalBindingElement::PER_VERTEX_INDEXED :
			if( vector.getNum() == 2 * height.getNum() )
			{
				normal.setValue(
					weight[0] * vector[nA][0] +
						weight[1] * vector[nB][0],
					weight[0] * vector[nA][1] +
						weight[1] * vector[nB][1],
					weight[0] * vector[nA][2] +
						weight[1] * vector[nB][2] );
				normal.normalize() ;
			}
			else
			{
				normal = vector[0] ;
			}
			break ;
		case SoNormalBindingElement::PER_FACE :
		case SoNormalBindingElement::PER_FACE_INDEXED :
printf( "FIX PER_FACE terrain following!!!!\n" ) ;
			break ;
	}

	// Adjust normal for rotation
	SbVec3f adjNorm( normal[0]*rotCos-normal[1]*rotSin,
			 normal[0]*rotSin+normal[1]*rotCos, normal[2] ) ;

	SbVec3f pt( fta->getSearchPoint() ) ;
	pt[2] = zGround + origin.getValue()[2] ;

	GroundPoint &gp = fta->addIntersection( pt, adjNorm ) ;
	// Save the resulting (transformed) points of the hit triangle.
	SbVec3f corner[4] ;
	const SbVec3f &orig = origin.getValue() ;
	corner[0].setValue(
		coords->point[nA][0]*rotCos-coords->point[nA][1]*rotSin+orig[0],
		coords->point[nA][0]*rotSin+coords->point[nA][1]*rotCos+orig[1],
		coords->point[nA][2]+orig[2] ) ;
	corner[3].setValue(
		coords->point[nB][0]*rotCos-coords->point[nB][1]*rotSin+orig[0],
		coords->point[nB][0]*rotSin+coords->point[nB][1]*rotCos+orig[1],
		coords->point[nB][2]+orig[2] ) ;
	nA++ ;
	nB++ ;
	corner[1].setValue(
		coords->point[nA][0]*rotCos-coords->point[nA][1]*rotSin+orig[0],
		coords->point[nA][0]*rotSin+coords->point[nA][1]*rotCos+orig[1],
		coords->point[nA][2]+orig[2] ) ;
	corner[2].setValue(
		coords->point[nB][0]*rotCos-coords->point[nB][1]*rotSin+orig[0],
		coords->point[nB][0]*rotSin+coords->point[nB][1]*rotCos+orig[1],
		coords->point[nB][2]+orig[2] ) ;
	gp.setPoly( 4, corner ) ;
}



//------------------------------------------------------------------------------
// Check to see if a ray intersects the shape.
//------------------------------------------------------------------------------
SbBool
GroundStrip::checkObstruction( 
	SbVec3f	&localStart,
	SbVec3f	&localEnd,
	SbBool	findDetail
	)
{
	SbBool	hit = FALSE ;

	SbVec3f &delta = localEnd - localStart ;
	float t ;
	float maxDim[2] ;
	maxDim[0] = width.getValue() ;
	maxDim[1] = length.getValue() ;

	// Trim to X and Y edges.
	for( int i = 0 ; i < 2 ; i++ )
	{
		if( delta[i] != 0.0f )
		{
			if( localStart[i] < 0.0f )
			{
				t = -localStart[i] / delta[i] ;
				localStart.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
			else if( localStart[i] > maxDim[i] )
			{
				t = ( maxDim[i] - localStart[i] ) / delta[i] ;
				localStart.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
			if( localEnd[i] < 0.0f )
			{
				t = -localStart[i] / delta[i] ;
				localEnd.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
			else if( localEnd[i] > maxDim[i] )
			{
				t = ( maxDim[i] - localStart[i] ) / delta[i] ;
				localEnd.setValue(
						localStart[0] + t * delta[0],
						localStart[1] + t * delta[1],
						localStart[2] + t * delta[2] ) ;
				delta = localEnd - localStart ;
			}
		}
	}

	// Now check to see on which side of the surface (above or below) the
	// endpoints are.
	SbVec2f	weight ;
	int32_t	nA, nB ;
	float heightStart = getHeight( localStart[1], weight, nA, nB ) ;
	float heightEnd = getHeight( localEnd[1], weight, nA, nB ) ;

	SbBool startAbove = ( heightStart < localStart[2] ) ;
	SbBool endAbove = ( heightEnd < localEnd[2] ) ;

	if( startAbove != endAbove && !findDetail )
	{
		return( TRUE ) ;
	}

	// Quick check if this is planar (ie, there is only one height)
	int32_t nZ = height.getNum() ;
	if( nZ == 1 && startAbove == endAbove )
	{
		return( FALSE ) ;
	}

	float	invDeltaW = 1.0f / deltaW ;
	float	invDeltaL = 1.0f / deltaL ;

	// Check to see if line crosses Y-constant lines.
	if( delta[1] != 0.0f )
	{
		int j0, j1, jDelta ;
		if( localStart[1] < localEnd[1] )
		{
			j0 = (int)ceilf( localStart[1] * invDeltaL ) ;
			j1 = (int)floorf( localEnd[1] * invDeltaL ) + 1 ;
			jDelta = 1 ;
		}
		else
		{
			j0 = (int)floorf( localStart[1] * invDeltaL ) ;
			j1 = (int)ceilf( localEnd[1] * invDeltaL ) - 1 ;
			jDelta = -1 ;
		}

		float k0 = delta[0] / delta[1] ;
		float k1 = localStart[0] - k0 * localStart[1] ;
		float m0 = delta[2] / delta[1] ;
		float m1 = localStart[2] - m0 * localStart[1] ;

		if( j0 != j1 )
		{
			for( int j = j0 ; j != j1 ; j += jDelta )
			{
				float y = (float)j * deltaL ;
				float x = ( k0 * y + k1 ) * invDeltaW ;

				int n0 = ( j ) % nZ ;
				float zGround = height[n0] ;
				float zLine = m0 * y + m1 ;
				if( ( zLine > zGround ) != startAbove )
				{
					if( !findDetail )
					{
						return( TRUE ) ;
					}
					else
					{
						localEnd.setValue( x, y,
							zLine ) ;
						hit = TRUE ;
						delta = localEnd - localStart ;
						break ;
					}
				}
			}
		}
	}
	//
	// If the following is true, it means we're looking for full detail
	// but didn't cross a line.  Therefore, assume it's a plane between
	// the 2 points and solve for where the line intersects the plane.
	//
	if( startAbove != endAbove && !hit )
	{
		heightStart -= localStart[2] ;
		heightEnd -= localEnd[2] ;
		float t = heightStart / ( heightStart - heightEnd ) ;
		localEnd = localStart + t * delta ;
		hit = TRUE ;
	}

	return( hit ) ;
}



