//------------------------------------------------------------------------------
// Class for guided missiles.
//------------------------------------------------------------------------------

#include <Inventor/SbTime.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoSeparator.h>

#include "LosAction.h"
#include "GuidedMissile.h"
#include "Utilities.h"
#include "Constants.h"

#include "GuidedMissileGeom.h"

SO_NODE_SOURCE( GuidedMissile ) ;

SoSeparator		*GuidedMissile::geometry = NULL ;
SoSeparator		*GuidedMissile::icon = NULL ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
GuidedMissile::initClass()
{
	if( Missile::getClassTypeId() == SoType::badType() )
	{
		Missile::initClass() ;
	}
	SO_NODE_INIT_CLASS( GuidedMissile, Missile, "Missile" ) ;

	geometry = new SoSeparator ;
	geometry->ref() ;

	SoComplexity *comp = new SoComplexity ;
	geometry->addChild( comp ) ;
	comp->type = SoComplexity::OBJECT_SPACE ;
	comp->value = 0.2f ;

	SoCone *cone = new SoCone ;
	geometry->addChild( cone ) ;
	cone->height = Constants::MissileLength ;
	cone->bottomRadius = Constants::MissileRadius ;

	icon = Utilities::readAll( "icons/guided_missile.iv", FALSE ) ;
	if( icon == NULL )
	{
		SoInput input ;
		input.setBuffer( (void *)defaultGeom, sizeof( defaultGeom ) ) ;

		icon = SoDB::readAll( &input ) ;

		if( icon == NULL )
		{
			SoDebugError::post( "GuidedMissile", "can not create "
				"icon geometry" ) ;
			exit( 1 ) ;
		}
	}
	icon->ref() ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
GuidedMissile::GuidedMissile()
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Alternate constructor specifying the desired height.
//------------------------------------------------------------------------------
GuidedMissile::GuidedMissile(
	float	height
	) :
	Missile( height )
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common code for constructors.
//------------------------------------------------------------------------------
void
GuidedMissile::buildCommon()
{
	SO_NODE_CONSTRUCTOR( GuidedMissile ) ;
	SO_NODE_ADD_FIELD( maxPitchRate, ( 0.8f ) ) ;
	SO_NODE_ADD_FIELD( maxYawRate, ( 0.6f ) ) ;

	status = Missile::ACTIVE ;

	children.truncate( 0 ) ;
	children.append( geometry ) ;

	maxSpeed = maxSpeed * 0.5f ;

	buildTime = 10.0f ;

	damageAmount = 4.0f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
GuidedMissile::~GuidedMissile()
{
}



//------------------------------------------------------------------------------
// Move the missile and check for collisions.
//------------------------------------------------------------------------------
Missile::MissileStatus
GuidedMissile::update(
	float	mx,
	float	my,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( age > maxAge )
	{
		status = Missile::EXPIRED ;
		blowUp() ;
		return status ;
	}

	SbRotation rot = rotation.getValue() ;
	SbVec3f ahead ;
	Utilities::setForwardVector( rot, ahead ) ;

	if( -0.8f < ahead[2] && ahead[2] < 0.8f )
	{
		// Remove the roll.
		SbVec3f up ;
		SbVec3f right ;
		Utilities::setUpVector( rot, up ) ;
		Utilities::setRightVector( rot, right ) ;

		float x = up[2] ;
		float y = right[2] ;

		// If upside down, modify roll by 180 degrees.
		if( x < 0.0f )
		{
			x = -x ;
			y = -y ;
		}
		float roll = atan2f( y, x ) ;
		if( age < 0.5f )
		{
			roll *= 2.0f * age ;
		}
		rot = rot * SbRotation( ahead, roll ) ;
	}

	float factor = age / maxAge ;
	factor = ( 1.0f - factor * factor ) * deltaTime ;

	age += deltaTime ;

	SbVec3f	startPos( position.getValue() ) ;

	float yaw = maxYawRate.getValue() * mx * factor ;
	float pitch = maxPitchRate.getValue() * my * factor ;
	SbRotation turnPitch( SbVec3f( 1.0f, 0.0f, 0.0f ), pitch ) ;
	SbRotation turnYaw( SbVec3f( 0.0f, 0.0f, -1.0f ), yaw ) ;

	rot = turnPitch * turnYaw * rot ;

	float	forward = maxSpeed * deltaTime ;

	SbVec3f &curPos = startPos + forward * ahead ;

	los->setRay( startPos, curPos ) ;
	los->apply( scene ) ;

	if( los->isObstructed() )
	{
		blowUp() ;
		curPos = los->getEndPoint() ;
		status = Missile::STRUCK ;
	}

	position.setValue( curPos ) ;
	rotation.setValue( rot ) ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Return whether or not the missile is a guided type.
//------------------------------------------------------------------------------
SbBool
GuidedMissile::isGuided() const
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Return the missile type (this is a virtual method wrapper around the static
// GetClassTypeId).
//------------------------------------------------------------------------------
Missile::MissileType
GuidedMissile::getMissileType() const
{
	return( Missile::GUIDED_MISSILE ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the icon scene graph.
//------------------------------------------------------------------------------
SoNode *
GuidedMissile::getIconGraph()
{
	return( GuidedMissile::icon ) ;
}
