//------------------------------------------------------------------------------
// InterComm class.  Compresses and decompresses audio data.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <assert.h>
#include <bstring.h>
#include <limits.h>
#include "GameIO.h"
#include "AudioComp.h"
#include "SfxData.h"
#include "GameIO.h"
#include "GamePackets.h"
#include "InterComm.h"


//------------------------------------------------------------------------------
// Utility class to hold comm information.
//------------------------------------------------------------------------------
class CommHost {
    public :
	CommHost( HostAddr remoteId ) ;
	~CommHost() {}
	HostAddr	getAddress() const { return( hostAddr ) ; }
	void		setPacketData( void *newData, int len ) ;
	signed char	*getPacketData( int &len )
			    { len = dataLen ; dataLen = 0 ; return( data ) ; }
	SbBool		hasPacketData() const { return( dataLen > 0 ) ; }
	void		setLastBuffer( int num ) { lastBuffer = num ; }
	int		getLastBuffer() const { return( lastBuffer ) ; }

    private :
	CommHost() ;
	CommHost( const CommHost &host ) ;

	HostAddr	hostAddr ;
	int		dataLen ;
	int		lastBuffer ;
	signed char	data[CommPacket::COMP_DATA_SIZE] ;
} ;



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
InterComm::InterComm(
	GameIO *gameIO
	) :
	gio( gameIO ),
	outPacket( 0, 0 )
{
	comp = new AudioComp( AudioComp::IDVI8 ) ;
	rawSampleSize = comp->uncompressedBufSize( CommPacket::COMP_DATA_SIZE );

	readBuffer = 0 ;
	writeBuffer = 0 ;
	numBuffers = 0 ;
	incoming = new signed char[NUM_BUFFERS * rawSampleSize] ;
	numComm[0] = 0 ;
	numComm[1] = 0 ;

	//
	// First check to see if we can open up an input audio port.
	//
	if( !openAudioPort() )
	{
		enabled = 0 ;
		return ;
	}

	speaker = new SfxData( rawSampleSize, 20, AL_RATE_8000 ) ;
	mode = QUIET ;
	team = 0 ;
	enabled = 1 ;

	gio->addInputPacketType( COMM_PKT, sizeof( CommPacket ),
			(GamePacketHandler *)processCommPacket, this ) ;

	seqId = 0 ;
	sampsLeftToSend = 0 ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
InterComm::~InterComm()
{
	if( enabled )
	{
		ALcloseport( ap ) ;
		delete speaker ;
	}

	delete comp ;
	delete rawBuffer ;
	delete incoming ;
}



//------------------------------------------------------------------------------
// Open up the audio port.
//------------------------------------------------------------------------------
int
InterComm::openAudioPort()
{
	long		pvbuf[2] ;	// system long
	long		buflen = 2 ;	// system long
	ALconfig	audioPortConfig;

	rawBuffer = new signed char[4*rawSampleSize] ;
	if( rawBuffer == NULL || incoming == NULL || !Sfx::isEnabled() )
	{
		return( 0 ) ;
	}

	ALerrfunc originalErrorHandler = ALseterrorhandler( NULL ) ;

	pvbuf[0] = AL_INPUT_RATE ;
	pvbuf[1] = AL_RATE_8000 ;
	ALsetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;

	// configure and open audio port
	audioPortConfig = ALnewconfig();
	ALsetchannels( audioPortConfig, AL_MONO ) ;
	ALsetsampfmt( audioPortConfig, AL_SAMPFMT_TWOSCOMP ) ;
	ALsetwidth( audioPortConfig, AL_SAMPLE_8 ) ;
	ALsetqueuesize( audioPortConfig, 4 * rawSampleSize ) ;

	ap = ALopenport( "InterComm", "r", audioPortConfig ) ;

	ALfreeconfig( audioPortConfig ) ;

	ALseterrorhandler( originalErrorHandler ) ;

	return( ap != NULL ) ;
}



//------------------------------------------------------------------------------
// Set the team mode.
//------------------------------------------------------------------------------
void
InterComm::setTeam(
	int newTeam
	)
{
	team = newTeam ;

	if( mode == PRIVATE )
	{
		outPacket.audience = team ;
	}
}



//------------------------------------------------------------------------------
// Change the mode of the broadcast.
//------------------------------------------------------------------------------
void
InterComm::changeMode(
	Mode newMode
	)
{
	if( enabled && mode != newMode )
	{
		// Start broadcasting.  Throw away all data currently in the
		// input queue.
		if( mode == QUIET )
		{
			ALflush( ap, 2 * ALgetfilled( ap ) ) ;
		}
		// Stop broadcasting.
		else if( newMode == QUIET )
		{
			sampsLeftToSend = ALgetfilled( ap ) ;
			if( sampsLeftToSend == 0 )
			{
				outPacket.len = 0 ;
				outPacket.number = seqId++ ;
				gio->send( &outPacket, sizeof( CommPacket ) ) ;
			}
		}
		if( newMode == PUBLIC )
		{
			outPacket.audience = 0 ;
		}
		else if( newMode == PRIVATE )
		{
			outPacket.audience = team ;
		}
	}

	mode = newMode ;
}



//------------------------------------------------------------------------------
// Start processing.  Grab any input data, encode it, and send it.  Clear the
// mux buffer for any input data.
//------------------------------------------------------------------------------
void
InterComm::startCommProcessing()
{
	if( enabled == 0 )
	{
		return ;
	}

	//
	// First check to see if we need to send any data.
	//
	if( mode != QUIET && ALgetfilled( ap ) >= rawSampleSize )
	{
		sampsLeftToSend = rawSampleSize ;
	}

	if( sampsLeftToSend > 0 )
	{
		int amount = ( sampsLeftToSend < rawSampleSize ) ?
			sampsLeftToSend : rawSampleSize ;
		if( ALreadsamps( ap, rawBuffer, amount ) == 0 )
		{
			int n = comp->encode( rawBuffer, amount,
						outPacket.getDataPtr() ) ;
			for( int i = n ; i < CommPacket::COMP_DATA_SIZE ; i++ )
			{
				outPacket.getDataPtr()[i] = 0 ;
			}
			outPacket.len = n ;
			outPacket.number = seqId++ ;
			gio->send( &outPacket, sizeof( CommPacket ) ) ;

		}
		sampsLeftToSend -= amount ;

		if( sampsLeftToSend <= 0 && mode == QUIET )
		{
			outPacket.len = 0 ;
			outPacket.number = seqId++ ;
			gio->send( &outPacket, sizeof( CommPacket ) ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Static callback to handle an incoming comm packet.
//------------------------------------------------------------------------------
void
InterComm::processCommPacket(
	CommPacket	*pkt,
	void		*userData
	)
{
	( (InterComm *)userData )->processComm( pkt ) ;
}



//------------------------------------------------------------------------------
// Callback to handle an incoming comm packet.
//------------------------------------------------------------------------------
void
InterComm::processComm(
	CommPacket	*pkt
	)
{
	int hostNo = hostList.find( pkt->fromAddress ) ;
	CommHost *host = hostList[hostNo] ;

	if( !enabled || ( pkt->audience != 0 && pkt->audience != team ) )
	{
		return ;
	}
	int nData = pkt->getLength() ;

	int nSamps = 0 ;
	if( host->hasPacketData() )
	{
		int pLen ;
		signed char *ptr = host->getPacketData( pLen ) ;
		nSamps = comp->decode( ptr, pLen, rawBuffer ) ;
	}

	if( nData > 0 )
	{
		host->setPacketData( pkt->getDataPtr(), nData ) ;
	}

	if( nSamps > 0 )
	{
		addToOutputBuffer( nSamps, host ) ;
	}
}



//------------------------------------------------------------------------------
// Add decoded data to the output buffer.
//------------------------------------------------------------------------------
void
InterComm::addToOutputBuffer(
	int		nSamps,
	CommHost	*host
	)
{
	int rBuf ;
	int wBuf ;
	int lBuf = host->getLastBuffer() ;

	//
	// First, determine if last buffer written to has been dumped or not.
	//
	if( readBuffer > writeBuffer )
	{
		rBuf = readBuffer - NUM_BUFFERS ;
		lBuf -= NUM_BUFFERS ;
	}
	else
	{
		rBuf = readBuffer ;
	}

	if( rBuf <= lBuf && lBuf < writeBuffer )
	{
		wBuf = ( lBuf + 1 + NUM_BUFFERS ) % NUM_BUFFERS ;
	}
	else if( numBuffers < NUM_BUFFERS )
	{
		// wBuf = ( readBuffer + 1 ) % NUM_BUFFERS ;
		wBuf = readBuffer ;
		if( readBuffer == writeBuffer )
		{
			writeBuffer = ( writeBuffer + 1 ) % NUM_BUFFERS ;
			numComm[writeBuffer] = 0 ;
			numBuffers++ ;
		}
	}

	host->setLastBuffer( wBuf ) ;

	signed char *ptr = incoming + wBuf * rawSampleSize ;
	if( numComm[wBuf] > 0 )
	{
		int val ;
		for( int i = 0 ; i < nSamps ; i++ )
		{
			val = ptr[i] + rawBuffer[i] ;
			if( val > SCHAR_MAX )
			{
				val = SCHAR_MAX ;
			}
			else if( val < SCHAR_MIN )
			{
				val = SCHAR_MIN ;
			}
			ptr[i] = val ;
		}
	}
	else
	{
		for( int i = 0 ; i < nSamps ; i++ )
		{
			ptr[i] = rawBuffer[i] ;
		}
		while( i < rawSampleSize )
		{
			ptr[i++] = 0 ;
		}
	}
	numComm[wBuf]++ ;
}



//------------------------------------------------------------------------------
// Finish up comm processing by playing the collected incoming communications.
//------------------------------------------------------------------------------
void
InterComm::finishCommProcessing()
{
	if( enabled && numBuffers > 0 && speaker->getFilledElements() < 3 )
	{
		signed char *ptr = incoming + readBuffer * rawSampleSize ;
		speaker->playWait( (char *)ptr ) ;
		numComm[readBuffer] = 0 ;
		readBuffer = ( readBuffer + 1 ) % NUM_BUFFERS ;
		numBuffers-- ;
	}
}



//------------------------------------------------------------------------------
// CommHostList definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Truncate the CommHostList and delete the CommHost's.
//------------------------------------------------------------------------------
void
CommHostList::truncate(
	int	start
	)
{
	for( int i = start ; i < getLength() ; i++ )
	{
		delete (*this)[i] ;
	}

	SbPList::truncate( start ) ;
}



//------------------------------------------------------------------------------
// Find a CommHost within the CommHostList based on the host id.
//------------------------------------------------------------------------------
int
CommHostList::find(
	HostAddr remoteId
	)
{
	for( int i = 0 ; i < getLength() ; i++ )
	{
		if( (*this)[i]->getAddress() == remoteId )
		{
			return( i ) ;
		}
	}

	CommHost *ch = new CommHost( remoteId ) ;
	append( ch ) ;

	return( getLength() - 1 ) ;
}



//------------------------------------------------------------------------------
// Remove an entry from the CommHostList and delete it.
//------------------------------------------------------------------------------
void
CommHostList::remove(
	int	which
	)
{
	delete (*this)[which] ;
	SbPList::remove( which ) ;
}



//------------------------------------------------------------------------------
// CommHost definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
CommHost::CommHost(
	HostAddr remoteId
	)
{
	hostAddr = remoteId ;
	dataLen = 0 ;
	lastBuffer = 0 ;
}



//------------------------------------------------------------------------------
// Copy the data.
//------------------------------------------------------------------------------
void
CommHost::setPacketData(
	void	*newData,
	int	len
	)
{
	dataLen = len ;
	bcopy( (void *)newData, data, len ) ;

	if( len < sizeof( data ) )
	{
		bzero( data+len, sizeof( data ) - len ) ;
	}
}
