//------------------------------------------------------------------------------
// LosAction class.
//------------------------------------------------------------------------------

#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoComplexityElement.h>
#include <Inventor/elements/SoComplexityTypeElement.h>
#include <Inventor/elements/SoCoordinateElement.h>
#include <Inventor/elements/SoFontNameElement.h>
#include <Inventor/elements/SoFontSizeElement.h>
#include <Inventor/elements/SoProfileCoordinateElement.h>
#include <Inventor/elements/SoProfileElement.h>
#include <Inventor/elements/SoSwitchElement.h>
#include <Inventor/elements/SoUnitsElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/elements/SoViewingMatrixElement.h>
#include <Inventor/elements/SoViewportRegionElement.h>
#include <Inventor/elements/SoNormalBindingElement.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoGroup.h>
#include <math.h>
#include "LosAction.h"
#include "Obstacle.h"
#include "GroundShape.h"



SO_ACTION_SOURCE( LosAction ) ;


//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
LosAction::initClass()
{
	SO_ACTION_INIT_CLASS( LosAction, SoAction ) ;

	// Enable the elements that are involved.
	SO_ENABLE( LosAction, SoModelMatrixElement ) ;
	SO_ENABLE( LosAction, SoComplexityElement ) ;
	SO_ENABLE( LosAction, SoComplexityTypeElement ) ;
	SO_ENABLE( LosAction, SoCoordinateElement ) ;
	SO_ENABLE( LosAction, SoFontNameElement ) ;
	SO_ENABLE( LosAction, SoFontSizeElement ) ;
	SO_ENABLE( LosAction, SoProfileCoordinateElement ) ;
	SO_ENABLE( LosAction, SoProfileElement ) ;
	SO_ENABLE( LosAction, SoSwitchElement ) ;
	SO_ENABLE( LosAction, SoUnitsElement ) ;
	SO_ENABLE( LosAction, SoViewVolumeElement ) ;
	SO_ENABLE( LosAction, SoViewingMatrixElement ) ;
	SO_ENABLE( LosAction, SoViewportRegionElement ) ;
	SO_ENABLE( LosAction, SoNormalBindingElement ) ;

	// Register static methods fo the action.
	SO_ACTION_ADD_METHOD( SoNode, doNada ) ;

	SO_ACTION_ADD_METHOD( SoGroup, callDoAction ) ;
	SO_ACTION_ADD_METHOD( Obstacle, obstructS ) ;
	SO_ACTION_ADD_METHOD( GroundShape, GroundShape::obstructS ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
LosAction::LosAction() :
	self( NULL ),
	target( NULL ),
	startPt( 0.0f, 0.0f, 0.0f ),
	endPt( 0.0f, 0.0f, 0.0f )
{
	SO_ACTION_CONSTRUCTOR( LosAction ) ;

	obstructor = NULL ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
LosAction::~LosAction()
{
	if( obstructor != NULL )
	{
		obstructor->unref() ;
	}
}



//------------------------------------------------------------------------------
// Initiate the action.
//------------------------------------------------------------------------------
void
LosAction::beginTraversal(
	SoNode *node
	)
{
	obstructed = FALSE ;
	if( obstructor != NULL )
	{
		obstructor->unref() ;
		obstructor = NULL ;
	}
	traverse( node ) ;
}



//------------------------------------------------------------------------------
// Generic action.
//------------------------------------------------------------------------------
void
LosAction::callDoAction(
	SoAction	*action,
	SoNode		*node
	)
{
	node->doAction( action ) ;
}



//------------------------------------------------------------------------------
// Static method for ignoring the action.
//------------------------------------------------------------------------------
void
LosAction::doNada(
	SoAction *,
	SoNode	 *
	)
{
}



//------------------------------------------------------------------------------
// Set the start and endpoints of the ray.
//------------------------------------------------------------------------------
void
LosAction::setRay(
	const SbVec3f	&start,
	const SbVec3f	&end
	)
{
	startPt = start ;
	trimRay( end ) ;
}



//------------------------------------------------------------------------------
// Trivial rejection test based on the distance between the center and
// collision radii (TRUE means no hit).
//------------------------------------------------------------------------------
SbBool
LosAction::checkDistance(
	const SbVec3f	&center,
	float		radius
	)
{
	//
	// Check for trivial rejection based on distance between center of
	// object to the ray and the collision radius.
	//
	SbVec3f dVec( center[0]-startPt[0], center[1]-startPt[1],
			center[2]-startPt[2] ) ;
	float	f = dVec.dot( direction ) ;
	if( f < -radius || f > length + radius )
	{
		return( TRUE ) ;
	}

	SbVec3f norm( dVec[0] - f*direction[0], dVec[1] - f*direction[1],
			dVec[2] - f*direction[2] ) ;
	return( norm.dot( norm ) > ( radius * radius ) ) ;
}



//------------------------------------------------------------------------------
// Perform the collision check given the bounding box of the obstacle and its
// position and rotation.
//------------------------------------------------------------------------------
SbBool
LosAction::checkBoundingBox(
	const SbBox3f		&box,
	const SbRotation	&rotation,
	const SbVec3f		&center,
	float			radius
	)
{
	//
	// Check for trivial rejection based on distance between centers of
	// objects and sum of collision radii.
	//
	if( checkDistance( center, radius ) )
	{
		return FALSE ;
	}

	//
	// Transform the ray points from world points to local (bounding box)
	// points.
	//
	SbMatrix worldToLocalMat ;
	rotation.inverse().getValue( worldToLocalMat ) ;

	SbVec3f localStart ;
	SbVec3f localEnd ;

	worldToLocalMat.multVecMatrix( startPt - center, localStart ) ;
	worldToLocalMat.multVecMatrix( endPt - center, localEnd ) ;

	float ext[6] ;
	box.getBounds( ext[0], ext[2], ext[4], ext[1], ext[3], ext[5] ) ;

	// P = start + t * ( end - start )
	SbVec3f delta( localEnd[0] - localStart[0],
			localEnd[1] - localStart[1],
			localEnd[2] - localStart[2] ) ;
	float t ;
	SbVec3f p ;

	SbBool hitSomething = FALSE ;

	// Check the six faces to see if the ray intersects one.
	for( int n = 0 ; n < 6 ; n++ )
	{
		int i = n / 2 ;

		if( ( localStart[i] < ext[n] && ext[n] < localEnd[i] ) ||
			( localEnd[i] < ext[n] && ext[n] < localStart[i] ) )
		{
			int j = ( i + 1 ) % 3 ;
			int jLo = j + j ;
			int jHi = jLo + 1 ;
			t = ( ext[n] - localStart[i] ) / delta[i] ;
			p[j] = localStart[j] + t * delta[j] ;
			if( ext[jLo] <= p[j] && p[j] <= ext[jHi] )
			{
				int k = ( j + 1 ) % 3 ;
				int kLo = k + k ;
				int kHi = kLo + 1 ;
				p[k] = localStart[k] + t * delta[k] ;
				if( ext[kLo] <= p[k] && p[k] <= ext[kHi] )
				{
					setObstructed() ;
					if( mode == GROSS_CHECK )
					{
						return( TRUE ) ;
					}
					hitSomething = TRUE ;
					p[i] = ext[n] ;
					localEnd = p ;
				}
			}
		}
	}

	//
	// Got here only if no obstruction was found or if we want to trim
	// the ray.  In the latter case, transform localEnd to world space.
	//
	if( hitSomething )
	{
		SbMatrix localToWorldMat ;
		rotation.getValue( localToWorldMat ) ;
		localToWorldMat[3][0] = center[0] ;
		localToWorldMat[3][1] = center[1] ;
		localToWorldMat[3][2] = center[2] ;

		SbVec3f newEnd ;
		localToWorldMat.multVecMatrix( localEnd, newEnd ) ;
		trimRay( newEnd ) ;
	}

	return( hitSomething ) ;
}



//------------------------------------------------------------------------------
// Static method for all checking obstructions with all obstacles.
//------------------------------------------------------------------------------
void
LosAction::obstructS(
	SoAction *action,
	SoNode	*node
	)
{
	( (Obstacle *)node )->obstruct( (LosAction *)action ) ;
}



//------------------------------------------------------------------------------
// Set the self node.
//------------------------------------------------------------------------------
void
LosAction::setSelf(
	SoNode	*node
	)
{
	if( self != NULL )
	{
		self->unref() ;
	}
	self = node ;
	if( self != NULL )
	{
		self->ref() ;
	}
}



//------------------------------------------------------------------------------
// Set the target node.
//------------------------------------------------------------------------------
void
LosAction::setTarget(
	SoNode	*node
	)
{
	if( target != NULL )
	{
		target->unref() ;
	}
	target = node ;
	if( target != NULL )
	{
		target->ref() ;
	}
}



//------------------------------------------------------------------------------
// Adjust the end point of the ray.
//------------------------------------------------------------------------------
void
LosAction::trimRay(
	const SbVec3f &newEnd
	)
{
	endPt = newEnd ;

	direction = endPt - startPt ;
	length = direction.normalize() ;
}



//------------------------------------------------------------------------------
// Set the obstructed flag and the terminate flag if we're doing a GROSS_CHECK.
//------------------------------------------------------------------------------
void
LosAction::setObstructed()
{
	obstructed = TRUE ;
	if( mode == GROSS_CHECK )
	{
		setTerminated( TRUE ) ;
	}
	else
	{
		SoPath *newObs = getCurPath()->copy() ;
		newObs->ref() ;
		if( obstructor != NULL )
		{
			obstructor->unref() ;
		}
		obstructor = newObs ;
	}
}
