//------------------------------------------------------------------------------
// Mauler class.
//------------------------------------------------------------------------------

#include <Inventor/fields/SoMFString.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoInfo.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include "Missile.h"
#include "GuidedMissile.h"
#include "CruiseMissile.h"
#include "ZbufferMode.h"
#include "Mortar.h"
#include "Utilities.h"
#include "FollowTerrainAction.h"
#include "Mauler.h"


SO_NODE_SOURCE( Mauler ) ;



//------------------------------------------------------------------------------
// Initialize the Mauler class.
//------------------------------------------------------------------------------
void
Mauler::initClass()
{
	if( Tank::getClassTypeId() == SoType::badType() )
	{
		Tank::initClass() ;
	}
	SO_NODE_INIT_CLASS( Mauler, Tank, "Tank" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Mauler::Mauler() :
	Tank( "models/mauler.iv" )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
Mauler::Mauler(
	const char *name
	) :
	Tank( name )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
Mauler::createIt()
{
	SO_NODE_CONSTRUCTOR( Mauler ) ;
	setParams() ;
	addFirstPersonGraphics() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Mauler::~Mauler()
{
	rangeText->unref() ;
	targetRangeText->unref() ;
	rangeRot->unref() ;
}



//------------------------------------------------------------------------------
// Set up parameters.
//------------------------------------------------------------------------------
void
Mauler::setParams()
{
	maxSpeed = 1.25f ;
	maxTurnRate = 0.75f ;
	maxAcceleration = 1.0f ;
	maxDamage = 40.0f ;
	numBays = 3 ;

	numWeapons[WeaponControl::MAIN_POD] = 3 ;
	weaponList[WeaponControl::MAIN_POD].truncate( 0 ) ;
	weaponList[WeaponControl::MAIN_POD].append( Missile::MISSILE ) ;
	weaponList[WeaponControl::MAIN_POD].append( Missile::CRUISE_MISSILE ) ;
	weaponList[WeaponControl::MAIN_POD].append( Missile::GUIDED_MISSILE ) ;
	numWeapons[WeaponControl::AA_POD] = 1 ;
	weaponList[WeaponControl::AA_POD].truncate( 0 ) ;
	weaponList[WeaponControl::AA_POD].append( Missile::MORTAR ) ;

	modelLabel = "Mauler" ;

	points = 80 ;
}



//------------------------------------------------------------------------------
// Indicate that no con view is available.
//------------------------------------------------------------------------------
SbBool
Mauler::hasConView() const
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Move the conning tower view.
//------------------------------------------------------------------------------
float
Mauler::moveConView(
	float	mx,
	float	my,
	float	dt,
	SoNode	*scene
	)
{
	static const float	twoPi = 2.0f * (float)M_PI ;
	static const float	halfPi = 0.5f * (float)M_PI ;
	static const float	minIncl = 0.05f * (float)M_PI ;
	static const SbRotation	c( SbVec3f( 1.0f, 0.0f, 0.0f ), halfPi ) ;
	const float	azimRate = twoPi / 2.0f ;
	const float	inclRate = twoPi / 2.0f ;

	mx *= ( mx < 0.0f ) ? -mx : mx ;
	my *= ( my < 0.0f ) ? -my : my ;
	azim -= azimRate * mx * dt ;
	incl = inclRate * my ;

	if( azim > twoPi )
	{
		azim -= twoPi ;
	}
	else if( azim < 0.0f )
	{
		azim += twoPi ;
	}

	if( incl > halfPi )
	{
		incl = halfPi ;
	}
	else if( incl < minIncl )
	{
		incl = minIncl ;
	}

	SbRotation az( SbVec3f( 0.0f, 1.0f, 0.0f ), azim ) ;
	SbRotation in( SbVec3f( 1.0f, 0.0f, 0.0f ), incl ) ;

	launchDir = c.inverse() * in * az * c ;
	conView.setValue( az * c ) ;

	float energy = move( 0.0f, 0.0f, dt, scene ) ;

	radarRotation.setValue( 
		c.inverse() * conView.getValue() * rotation.getValue() ) ;

	// Get the height.
	float altitude ;
	fta->setSearchPoint( position.getValue() ) ;
	fta->apply( scene ) ;
	const GroundPointList &gpl = fta->getGroundPointList() ;
	if( gpl.getLength() > 0 )
	{
		altitude = -gpl[0]->getHeight() ;
	}
	else
	{
		altitude = 0.0f ;
	}

	float A = -0.125f * Constants::Gravity ; // -2 * Mortar::Grav / 2
	SbVec3f dir ;
	Utilities::setForwardVector( launchDir * rotation.getValue(), dir ) ;
	dir *= Mortar::muzzleVelocity ;
	float B = dir[2] ;
	float C = altitude ;
	float D = B * B - 2.0f * A * C ; // A has factor of 2 built in
	float t ;
	if( D > 0.0f )
	{
		D = sqrtf( D ) ;
		t = ( -B - D ) / A ; // A has factor of 2 built in
	}
	else if( D < 0.0f )
	{
		t = 0.0f ;
	}
	else
	{
		t = -B / A ; // A has factor of 2 built in
	}
	SbVec3f groundZero( dir[0] * t, dir[1] * t, -altitude ) ;

	char buf[10] ;
	sprintf( buf, "% 03d", (int)( groundZero.length() * 10.0f ) ) ;
	rangeText->string.setValue( buf ) ;

	rangeRot->angle = azim ;

	return( energy ) ;
}



//------------------------------------------------------------------------------
// Position a newly launched weapon.
//------------------------------------------------------------------------------
void
Mauler::launchWeapon(
	Missile				*weapon,
	WeaponControl::WeaponPod	pod
	)
{
	SbVec3f	up ;
	SbVec3f	ahead ;
	SbVec3f	right ;
	SbRotation rot = rotation.getValue() ;
	Utilities::setUpVector( rot, up ) ;
	Utilities::setForwardVector( rot, ahead ) ;
	Utilities::setRightVector( rot, right ) ;

	SbVec3f lp ;
	if( pod == WeaponControl::MAIN_POD )
	{
		lp = launchPoint + getCenterOffset() ;
	}
	else
	{
		lp = auxViewpoint + getCenterOffset() ;
		rot = launchDir * rot ;
	}
	SbVec3f pos = position.getValue() + lp[0] * right + lp[1] * ahead +
			lp[2] * up ;

	weapon->position.setValue( pos ) ;
	weapon->rotation.setValue( rot ) ;
	weapon->setVelocity( getVelocity() ) ;
}



//------------------------------------------------------------------------------
// Build the first person graphics.
//------------------------------------------------------------------------------
void
Mauler::addFirstPersonGraphics()
{
	// Main view graphics are nothing.
	fpGraphics->addChild( new SoInfo ) ;

	SoSeparator *auxBase = new SoSeparator ;
	fpGraphics->addChild( auxBase ) ;
	auxBase->renderCulling = SoSeparator::OFF ;
	auxBase->pickCulling = SoSeparator::OFF ;
	auxBase->boundingBoxCaching = SoSeparator::OFF ;
	auxBase->renderCaching = SoSeparator::OFF ;

	ZbufferMode *zb = new ZbufferMode ;
	auxBase->addChild( zb ) ;
	zb->value = ZbufferMode::ALWAYS ;

	SoTranslation *t = new SoTranslation ;
	auxBase->addChild( t ) ;
	t->translation.setValue( auxViewpoint ) ;

	rangeRot = new SoRotationXYZ ;
	rangeRot->ref() ;
	auxBase->addChild( rangeRot ) ;
	rangeRot->axis = SoRotationXYZ::Z ;
	rangeRot->angle = 0.0f ;

	t = new SoTranslation ;
	auxBase->addChild( t ) ;
	t->translation.setValue( SbVec3f( 0.0f, 1.1f, 0.0f ) ) ;

	SoGroup *common = new SoGroup ;
	auxBase->addChild( common ) ;

	SoLightModel *lm = new SoLightModel ;
	common->addChild( lm ) ;
	lm->model = SoLightModel::BASE_COLOR ;

	SoBaseColor *bc = new SoBaseColor ;
	common->addChild( bc ) ;
	bc->rgb.setValue( 1.0f, 0.0f, 0.0f ) ;

	SoFont *font = new SoFont ;
	common->addChild( font ) ;
	font->name.setValue( "Helvetica" ) ;
	font->size = 15.0f ;

	SoSeparator *rangeSep = new SoSeparator ;
	common->addChild( rangeSep ) ;
	rangeSep->renderCulling = SoSeparator::OFF ;
	rangeSep->pickCulling = SoSeparator::OFF ;
	rangeSep->boundingBoxCaching = SoSeparator::OFF ;
	rangeSep->renderCaching = SoSeparator::OFF ;

	t = new SoTranslation ;
	rangeSep->addChild( t ) ;
	t->translation.setValue( -0.30f, 0.0f, 0.32f ) ;

	rangeText = new SoText2 ;
	rangeText->ref() ;
	rangeSep->addChild( rangeText ) ;
	rangeText->justification = SoText2::LEFT ;
	rangeText->string.setValue( "000" ) ;

	t = new SoTranslation ;
	rangeSep->addChild( t ) ;
	t->translation.setValue( 0.60f, 0.0f, 0.0f ) ;

	targetRangeText = new SoText2 ;
	targetRangeText->ref() ;
	rangeSep->addChild( targetRangeText ) ;
	targetRangeText->justification = SoText2::RIGHT ;
	targetRangeText->string.setValue( "000" ) ;
}



//------------------------------------------------------------------------------
// Indicate the distance to the target.
//------------------------------------------------------------------------------
void
Mauler::setRange(
	float d
	)
{
	char buf[10] ;
	sprintf( buf, "% 03d", (int)( d * 10.0f ) ) ;
	targetRangeText->string.setValue( buf ) ;
}



//------------------------------------------------------------------------------
// Show vehicle-specific help text.
//------------------------------------------------------------------------------
void
Mauler::showHelp(
	SoMFString	*str,
	int		&n
	)
{
	str->set1Value( n++, "Right mouse - Toggle between Main and "
				"Mortar views" ) ;
	str->set1Value( n++, "" ) ;
	str->set1Value( n++, "MAIN VIEW" ) ;
	str->set1Value( n++, "Left mouse - fire next missile" ) ;
	str->set1Value( n++, "Middle mouse - change missile type to be "
				"loaded" ) ;
	str->set1Value( n++, "Mouse - control speed and direction" ) ;
	str->set1Value( n++, "" ) ;
	str->set1Value( n++, "MORTAR VIEW" ) ;
	str->set1Value( n++, "Left mouse - fire mortar" ) ;
	str->set1Value( n++, "Mouse - aim mortar" ) ;
	str->set1Value( n++, "Space - range finder" ) ;
	str->set1Value( n++, "Left number on overlay is estimated "
				"distance of" ) ;
	str->set1Value( n++, "weapon.  Right number is range finder "
				"distance." ) ;
}
