//------------------------------------------------------------------------------
// Base class for missiles.
//------------------------------------------------------------------------------

#include <Inventor/SbTime.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoBBoxModelMatrixElement.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoSeparator.h>

#include "FollowTerrainAction.h"
#include "LosAction.h"
#include "Missile.h"
#include "Utilities.h"
#include "Constants.h"

#include "MissileGeom.h"

SO_NODE_SOURCE( Missile ) ;

SoSeparator		*Missile::geometry = NULL ;
SoSeparator		*Missile::icon = NULL ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Missile::initClass()
{
	SO_NODE_INIT_CLASS( Missile, SoNode, "Node" ) ;
	SO_ENABLE( SoGetMatrixAction, SoModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoBBoxModelMatrixElement ) ;

	geometry = new SoSeparator ;
	geometry->ref() ;

	SoComplexity *comp = new SoComplexity ;
	geometry->addChild( comp ) ;
	comp->type = SoComplexity::OBJECT_SPACE ;
	comp->value = 0.2f ;

	SoCylinder *cyl = new SoCylinder ;
	geometry->addChild( cyl ) ;
	cyl->height = Constants::MissileLength ;
	cyl->radius = Constants::MissileRadius ;

	icon = Utilities::readAll( "icons/missile.iv", FALSE ) ;
	if( icon == NULL )
	{
		SoInput input ;
		input.setBuffer( (void *)defaultGeom, sizeof( defaultGeom ) ) ;

		icon = SoDB::readAll( &input ) ;

		if( icon == NULL )
		{
			SoDebugError::post( "Missile", "can not create icon "
				"geometry" ) ;
			exit( 1 ) ;
		}
	}
	icon->ref() ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Missile::Missile() :
	children( this )
{
	buildCommon() ;
	desiredHeight = -0.2f ;
}



//------------------------------------------------------------------------------
// Alternate constructor specifying the desired height.
//------------------------------------------------------------------------------
Missile::Missile(
	float	height
	) :
	children( this )
{
	buildCommon() ;
	desiredHeight = -height ;
}



//------------------------------------------------------------------------------
// Common code for constructors.
//------------------------------------------------------------------------------
void
Missile::buildCommon()
{
	SO_NODE_CONSTRUCTOR( Missile ) ;
	SO_NODE_ADD_FIELD( position, ( 0.0f, 0.0f, 0.0f ) ) ;
	SO_NODE_ADD_FIELD( rotation, ( SbVec3f( 0.0f, 0.0f, 1.0f ), 0.0f ) ) ;
	SO_NODE_ADD_FIELD( buildTime, ( 6.0f ) ) ;

	children.append( geometry ) ;

	los = new LosAction ;
	los->setMode( LosAction::FIND_CLOSEST ) ;
	los->setSelf( NULL ) ;
	los->setTarget( NULL ) ;

	age = 0.0f ;
	maxSpeed = 15.0f ;
	maxAge = 7.5f ;

	status = Missile::ACTIVE ;
	velocity.setValue( 0.0f, 0.0f, 0.0f ) ;

	damageAmount = 8.0f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Missile::~Missile()
{
	delete los ;
}



//------------------------------------------------------------------------------
// Main guts of the action handlers.
//------------------------------------------------------------------------------
void
Missile::actionGuts(
	SoState		*state,
	SoAction	*action
	)
{
	if( !position.isIgnored() && !position.isDefault() )
	{
		SoModelMatrixElement::translateBy( state, this,
						position.getValue() ) ;
	}

	if( !rotation.isIgnored() && !rotation.isDefault() )
	{
		SoModelMatrixElement::rotateBy( state, this,
						rotation.getValue() ) ;
	}

	int numIndices ;
	const int *indices ;

	if( action->getPathCode( numIndices, indices ) == SoAction::IN_PATH )
	{
		children.traverse( action, 0, indices[numIndices-1] ) ;
	}
	else
	{
		children.traverse( action ) ;
	}
}



//------------------------------------------------------------------------------
// Generic action handler.
//------------------------------------------------------------------------------
void
Missile::doAction(
	SoAction	*action
	)
{
	SoState *state = action->getState() ;

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SbMatrix saveMat = SoModelMatrixElement::pushMatrix( state ) ;

	actionGuts( state, action ) ;

	SoModelMatrixElement::popMatrix( state, saveMat ) ;
#else
	SbMatrix saveMat = SoModelMatrixElement::get( state ) ;

	actionGuts( state, action ) ;

	SoModelMatrixElement::set( state, this, saveMat ) ;
#endif
}



//------------------------------------------------------------------------------
// GLRenderAction handler.
//------------------------------------------------------------------------------
void
Missile::GLRender(
	SoGLRenderAction *action
	)
{
	Missile::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetBoundBoxAction handler.
//------------------------------------------------------------------------------
void
Missile::getBoundingBox(
	SoGetBoundingBoxAction *action
	)
{
	int		numIndices ;
	const int	*indices ;

	if( action->getPathCode( numIndices, indices ) == SoAction::OFF_PATH )
	{
		return ;
	}

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoState *state = action->getState() ;

	SbMatrix matrix, localMatrix ;
	SoBBoxModelMatrixElement::pushMatrix( state, matrix, localMatrix ) ;

	actionGuts( state, action ) ;

	SoBBoxModelMatrixElement::popMatrix( state, matrix, localMatrix ) ;
#else
	Missile::doAction( action ) ;
#endif
}



//------------------------------------------------------------------------------
// CallbackAction handler.
//------------------------------------------------------------------------------
void
Missile::callback(
	SoCallbackAction *action
	)
{
	Missile::doAction( action ) ;
}



//------------------------------------------------------------------------------
// PickAction handler.
//------------------------------------------------------------------------------
void
Missile::pick(
	SoPickAction *action
	)
{
	Missile::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetMatrixAction handler.
//------------------------------------------------------------------------------
void
Missile::getMatrix(
	SoGetMatrixAction *action
	)
{
	int		numIndices ;
	const int	*indices ;

	if( action->getPathCode( numIndices, indices ) != SoAction::IN_PATH )
	{
		return ;
	}

	SbMatrix &ctm = action->getMatrix() ;
	SbMatrix &inv = action->getInverse() ;
	SbMatrix m ;

	// Missile::doAction( action ) ;
	if( !position.isIgnored() && !position.isDefault() )
	{
		const SbVec3f &v = position.getValue() ;
		m.setTranslate( v ) ;
		ctm.multLeft( m ) ;
		m.setTranslate( -v ) ;
		inv.multRight( m ) ;
	}

	if( !rotation.isIgnored() && !rotation.isDefault() )
	{
		SbRotation r = rotation.getValue() ;
		r.getValue( m ) ;
		ctm.multLeft( m ) ;
		r.invert() ;
		r.getValue( m ) ;
		inv.multRight( m ) ;
	}
}



//------------------------------------------------------------------------------
// Return a pointer to the root node.
//------------------------------------------------------------------------------
SoChildList *
Missile::getChildren() const
{
	return( (SoChildList *)&children ) ;
}



//------------------------------------------------------------------------------
// Move the missile and check for collisions.
//------------------------------------------------------------------------------
Missile::MissileStatus
Missile::update(
	float	,
	float	,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( age > maxAge )
	{
		status = Missile::EXPIRED ;
		blowUp() ;
		return status ;
	}

	SbRotation	rot = rotation.getValue() ;

	// -Z vector
	SbVec3f ahead ;
	Utilities::setForwardVector( rot, ahead ) ;

	if( age == 0.0f )
	{
		velocity = maxSpeed * ahead ;
	}
	age += deltaTime ;

	SbVec3f	startPos( position.getValue() ) ;

	SbVec3f	&forward = velocity * deltaTime ;
	if( age > 5.0f )
	{
		velocity[2] -= 0.25f * Constants::Gravity * deltaTime ;
	}

	SbVec3f &curPos = startPos + forward ;

	los->setRay( startPos, curPos ) ;
	los->apply( scene ) ;

	if( los->isObstructed() )
	{
		blowUp() ;
		curPos = los->getEndPoint() ;
		status = Missile::STRUCK ;
	}

	position.setValue( curPos ) ;
	rotation.setValue( rot ) ;

	return( status ) ;
}



//------------------------------------------------------------------------------
// Detonate a missile.
//------------------------------------------------------------------------------
void
Missile::blowUp()
{
	age = maxAge + 1.0f ;
}



//------------------------------------------------------------------------------
// Return whether or not the missile is a guided type.
//------------------------------------------------------------------------------
SbBool
Missile::isGuided() const
{
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Return the missile type (this is a virtual method wrapper around the static
// GetClassTypeId).
//------------------------------------------------------------------------------
Missile::MissileType
Missile::getMissileType() const
{
	return( Missile::MISSILE ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the icon scene graph.
//------------------------------------------------------------------------------
SoNode *
Missile::getIconGraph()
{
	return( Missile::icon ) ;
}



