//------------------------------------------------------------------------------
// Class for mortars.
//------------------------------------------------------------------------------

#include <Inventor/SbTime.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoSeparator.h>

#include "LosAction.h"
#include "Mortar.h"
#include "Utilities.h"
#include "Constants.h"

#include "MortarGeom.h"

SO_NODE_SOURCE( Mortar ) ;

SoSeparator		*Mortar::geometry = NULL ;
SoSeparator		*Mortar::icon = NULL ;
const float		Mortar::muzzleVelocity = 7.0f ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Mortar::initClass()
{
	if( Bomb::getClassTypeId() == SoType::badType() )
	{
		Bomb::initClass() ;
	}
	SO_NODE_INIT_CLASS( Mortar, Bomb, "Bomb" ) ;

	geometry = new SoSeparator ;
	geometry->ref() ;

	SoComplexity *comp = new SoComplexity ;
	geometry->addChild( comp ) ;
	comp->type = SoComplexity::OBJECT_SPACE ;
	comp->value = 0.2f ;

	SoCylinder *cone = new SoCylinder ;
	geometry->addChild( cone ) ;
	cone->height = Constants::MissileLength ;
	cone->radius = Constants::MissileRadius ;

	icon = Utilities::readAll( "icons/mortar.iv", FALSE ) ;
	if( icon == NULL )
	{
		SoInput input ;
		input.setBuffer( (void *)defaultGeom, sizeof( defaultGeom ) ) ;

		icon = SoDB::readAll( &input ) ;

		if( icon == NULL )
		{
			SoDebugError::post( "Mortar", "can not create icon "
				"geometry" ) ;
			exit( 1 ) ;
		}
	}
	icon->ref() ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Mortar::Mortar()
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Alternate constructor specifying the desired height.
//------------------------------------------------------------------------------
Mortar::Mortar(
	float	height
	) :
	Bomb( height )
{
	buildCommon() ;
}



//------------------------------------------------------------------------------
// Common code for constructors.
//------------------------------------------------------------------------------
void
Mortar::buildCommon()
{
	SO_NODE_CONSTRUCTOR( Mortar ) ;

	children.truncate( 0 ) ;
	children.append( geometry ) ;

	los->setTarget( NULL ) ;

	velocity.setValue( 0.0f, 0.0f, 0.0f ) ;
	maxAge = 7.5f ;

	damageAmount = 10.0f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Mortar::~Mortar()
{
}



//------------------------------------------------------------------------------
// Move the missile and check for collisions.
//------------------------------------------------------------------------------
Missile::MissileStatus
Mortar::update(
	float	mx,
	float	my,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( age == 0.0f )
	{
		SbVec3f ahead ;
		Utilities::setForwardVector( rotation.getValue(), ahead ) ;
		velocity = muzzleVelocity * ahead ;
	}

	Missile::MissileStatus st = Bomb::update( mx, my, deltaTime, scene ) ;

	velocity[2] -= 0.025f * Constants::Gravity * deltaTime ;

	return( st ) ;
}



//------------------------------------------------------------------------------
// Return the missile type (this is a virtual method wrapper around the static
// GetClassTypeId).
//------------------------------------------------------------------------------
Missile::MissileType
Mortar::getMissileType() const
{
	return( Missile::MORTAR ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the icon scene graph.
//------------------------------------------------------------------------------
SoNode *
Mortar::getIconGraph()
{
	return( Mortar::icon ) ;
}
