//------------------------------------------------------------------------------
// NoPassGroup Inventor node.  Prohibits selected actions from traversing its
// children.
//------------------------------------------------------------------------------
//
// Copyright 1995, Silicon Graphics, Inc.
// All Rights Reserved.
//
// This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
// the contents of this file may not be disclosed to third parties, copied or
// duplicated in any form, in whole or in part, without the prior written
// permission of Silicon Graphics, Inc.
//
// RESTRICTED RIGHTS LEGEND:
// Use, duplication or disclosure by the Government is subject to restrictions
// as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
// and Computer Software clause at DFARS 252.227-7013, and/or in similar or
// successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
// rights reserved under the Copyright Laws of the United States.
//

#include <Inventor/actions/SoAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include "NoPassGroup.h"

#define	BANNED(x)	( ( banFlags.getValue() & (x) ) != 0 )

SO_NODE_SOURCE( NoPassGroup ) ;

//------------------------------------------------------------------------------
// Class initializer.
//------------------------------------------------------------------------------
void
NoPassGroup::initClass()
{
	SO_NODE_INIT_CLASS( NoPassGroup, SoGroup, "Group" ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
NoPassGroup::NoPassGroup() : SoGroup()
{
	commonConstructor() ;
}



//------------------------------------------------------------------------------
// Alternate constructor that takes the approximate number of children.
//------------------------------------------------------------------------------
NoPassGroup::NoPassGroup(
	int	numChildren
	) : SoGroup( numChildren )
{
	commonConstructor() ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
NoPassGroup::commonConstructor()
{
	SO_NODE_CONSTRUCTOR( NoPassGroup ) ;
	SO_NODE_ADD_FIELD( banFlags, (NONE) ) ;
	SO_NODE_ADD_FIELD( alternateRep, (NULL) ) ;
	SoGroup *group = new SoGroup ;
	group->ref() ;
	alternateRep = group ;

	// Set up static info for enum fields
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, NONE ) ;
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, RENDER ) ;
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, SEARCH ) ;
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, PICK ) ;
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, BOUNDINGBOX ) ;
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, MATRIX ) ;
	SO_NODE_DEFINE_ENUM_VALUE( ActionFlags, ALL ) ;

	// Set up info in enumerated type fields
	SO_NODE_SET_SF_ENUM_TYPE( banFlags, ActionFlags ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
NoPassGroup::~NoPassGroup()
{
	SoGroup *group = (SoGroup *)alternateRep.getValue() ;
	group->unref() ;
}



//------------------------------------------------------------------------------
// Affects state method.
//------------------------------------------------------------------------------
SbBool
NoPassGroup::affectsState() const
{
	// Could be smarter on this based on the way banFlags are set, but we'll
	// play it safe (and lazy) for now.
	return TRUE ;
}



//------------------------------------------------------------------------------
// GetBoundingBox action.
//------------------------------------------------------------------------------
void
NoPassGroup::getBoundingBox(
	SoGetBoundingBoxAction	*action
	)
{
	if( !BANNED( BOUNDINGBOX ) )
	{
		SoGroup::getBoundingBox( action ) ;
	}
}



//------------------------------------------------------------------------------
// GLRender action.
//------------------------------------------------------------------------------
void
NoPassGroup::GLRender(
	SoGLRenderAction	*action
	)
{
	if( !BANNED( RENDER ) )
	{
		SoGroup::GLRender( action ) ;
	}
}



//------------------------------------------------------------------------------
// Search action.
//------------------------------------------------------------------------------
void
NoPassGroup::search(
	SoSearchAction	*action
	)
{
	if( !BANNED( SEARCH ) )
	{
		SoGroup::search( action ) ;
	}
}



//------------------------------------------------------------------------------
// Pick action.
//------------------------------------------------------------------------------
void
NoPassGroup::pick(
	SoPickAction	*action
	)
{
	if( !BANNED( PICK ) )
	{
		SoGroup::pick( action ) ;
	}
}



//------------------------------------------------------------------------------
// Search action.
//------------------------------------------------------------------------------
void
NoPassGroup::getMatrix(
	SoGetMatrixAction	*action
	)
{
	if( !BANNED( MATRIX ) )
	{
		SoGroup::getMatrix( action ) ;
	}
}



//------------------------------------------------------------------------------
// Write action to handle alternate rep.
//------------------------------------------------------------------------------
void
NoPassGroup::write(
	SoWriteAction	*action
	)
{
	SoOutput *out = action->getOutput() ;
	if( out->getStage() == SoOutput::COUNT_REFS )
	{
		SoGroup *group = (SoGroup *)( alternateRep.getValue() ) ;
		group->removeAllChildren() ;
		for( int i = 0 ; i < getNumChildren() ; i++ )
		{
			group->addChild( getChild( i ) ) ;
		}
	}

	SoGroup::write( action ) ;

	if( out->getStage() == SoOutput::WRITE )
	{
		SoGroup *group = (SoGroup *)( alternateRep.getValue() ) ;
		group->removeAllChildren() ;
	}
}
