//------------------------------------------------------------------------------
// Abstract base class for obstacles.
//------------------------------------------------------------------------------

#include <Inventor/SoDB.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoBBoxModelMatrixElement.h>
#include <Inventor/elements/SoLocalBBoxMatrixElement.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoSeparator.h>
#ifdef VERBOSE
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#endif // VERBOSE

#include "Utilities.h"
#include "Obstacle.h"

#ifdef VERBOSE
SbBool	Obstacle::showBoundingBox = FALSE ;
#endif /* VERBOSE */


GeometryNodeList Obstacle::gList ;


SO_NODE_ABSTRACT_SOURCE( Obstacle ) ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
Obstacle::initClass()
{
	SO_NODE_INIT_ABSTRACT_CLASS( Obstacle, SoTransformation,
					"Transformation" ) ;
	SO_ENABLE( SoGetMatrixAction, SoModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoBBoxModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoLocalBBoxMatrixElement ) ;

#ifdef VERBOSE
	if( getenv( "PB_SHOW_BOUNDING_BOX" ) != NULL )
	{
		showBoundingBox = TRUE ;
	}
#endif /* VERBOSE */
}



//------------------------------------------------------------------------------
// Common constructor bits.
//------------------------------------------------------------------------------
void
Obstacle::createIt()
{
	SO_NODE_CONSTRUCTOR( Obstacle ) ;
	SO_NODE_ADD_FIELD( position, ( 0.0f, 0.0f, 0.0f ) ) ;
	SO_NODE_ADD_FIELD( rotation, ( SbVec3f( 0.0f, 0.0f, 1.0f ), 0.0f ) ) ;
	SO_NODE_ADD_FIELD( name, ( "<Undefined file>" ) ) ;

	root = new SoSeparator ;
	root->ref() ;
	root->renderCaching = SoSeparator::OFF ;
	root->boundingBoxCaching = SoSeparator::OFF ;
	root->renderCulling = SoSeparator::OFF ;
	root->pickCulling = SoSeparator::OFF ;
#ifdef VERBOSE
	root->setName( "ObstacleRoot" ) ;
#endif // VERBOSE

	children.append( root ) ;

	hasBoundingBox = FALSE ;
	dirtyCorners = TRUE ;

	nameSensor = new SoFieldSensor( nameChangedCB, this ) ;
	nameSensor->setPriority( 0 ) ;
	nameSensor->attach( &name ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Obstacle::Obstacle() :
	children( this ),
	centerOffset( 0.0f, 0.0f, 0.0f )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
Obstacle::Obstacle(
	const char *model
	) :
	children( this ),
	centerOffset( 0.0f, 0.0f, 0.0f )
{
	createIt() ;
	name.setValue( model ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Obstacle::~Obstacle()
{
	root->unref() ;
	delete nameSensor ;
}



//------------------------------------------------------------------------------
// Main guts of the action handlers.
//------------------------------------------------------------------------------
void
Obstacle::actionGuts(
	SoState		*state,
	SoAction	*action
	)
{
	if( !position.isIgnored() && !position.isDefault() )
	{
		SoModelMatrixElement::translateBy( state, this,
						position.getValue() ) ;
	}

	if( !rotation.isIgnored() && !rotation.isDefault() )
	{
		SoModelMatrixElement::rotateBy( state, this,
						rotation.getValue() ) ;
	}

	SoModelMatrixElement::translateBy( state, this,
						centerOffset.getValue() ) ;

	int numIndices ;
	const int *indices ;

	if( action->getPathCode( numIndices, indices ) == SoAction::IN_PATH )
	{
		children.traverse( action, 0, indices[numIndices-1] ) ;
	}
	else
	{
		children.traverse( action ) ;
	}
}



//------------------------------------------------------------------------------
// Generic action handler.
//------------------------------------------------------------------------------
void
Obstacle::doAction(
	SoAction	*action
	)
{
	SoState *state = action->getState() ;
	// state->push() ;

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SbMatrix saveMat = SoModelMatrixElement::pushMatrix( state ) ;

	Obstacle::actionGuts( state, action ) ;

	SoModelMatrixElement::popMatrix( state, saveMat ) ;
#else
	SbMatrix saveMat = SoModelMatrixElement::get( state ) ;

	Obstacle::actionGuts( state, action ) ;

	SoModelMatrixElement::set( action->getState(), this, saveMat ) ;
#endif

	// state->pop() ;
}



//------------------------------------------------------------------------------
// GLRenderAction handler.
//------------------------------------------------------------------------------
void
Obstacle::GLRender(
	SoGLRenderAction *action
	)
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	Obstacle::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetBoundBoxAction handler.
//------------------------------------------------------------------------------
void
Obstacle::getBoundingBox(
	SoGetBoundingBoxAction *action
	)
{
	int		numIndices ;
	const int	*indices ;

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	if( action->getPathCode( numIndices, indices ) == SoAction::OFF_PATH )
	{
		return ;
	}

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoState *state = action->getState() ;

	SbMatrix matrix, localMatrix ;
	SoBBoxModelMatrixElement::pushMatrix( state, matrix, localMatrix ) ;

	Obstacle::actionGuts( state, action ) ;

	SoBBoxModelMatrixElement::popMatrix( state, matrix, localMatrix ) ;
#else
	Obstacle::doAction( action ) ;
#endif
}



//------------------------------------------------------------------------------
// CallbackAction handler.
//------------------------------------------------------------------------------
void
Obstacle::callback(
	SoCallbackAction *action
	)
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	Obstacle::doAction( action ) ;
}



//------------------------------------------------------------------------------
// PickAction handler.
//------------------------------------------------------------------------------
void
Obstacle::pick(
	SoPickAction *action
	)
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	Obstacle::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetMatrixAction handler.
//------------------------------------------------------------------------------
void
Obstacle::getMatrix(
	SoGetMatrixAction *action
	)
{
	int		numIndices ;
	const int	*indices ;

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	if( action->getPathCode( numIndices, indices ) != SoAction::IN_PATH )
	{
		return ;
	}

	SbMatrix &ctm = action->getMatrix() ;
	SbMatrix &inv = action->getInverse() ;
	SbMatrix m ;

	if( !position.isIgnored() && !position.isDefault() )
	{
		const SbVec3f &v = position.getValue() ;
		m.setTranslate( v ) ;
		ctm.multLeft( m ) ;
		m.setTranslate( -v ) ;
		inv.multRight( m ) ;
	}

	if( !rotation.isIgnored() && !rotation.isDefault() )
	{
		SbRotation r = rotation.getValue() ;
		r.getValue( m ) ;
		ctm.multLeft( m ) ;
		r.invert() ;
		r.getValue( m ) ;
		inv.multRight( m ) ;
	}

	const SbVec3f &v = centerOffset.getValue() ;
	m.setTranslate( v ) ;
	ctm.multLeft( m ) ;
	m.setTranslate( -v ) ;
	inv.multRight( m ) ;
}



//------------------------------------------------------------------------------
// Return a pointer to the root node.
//------------------------------------------------------------------------------
SoChildList *
Obstacle::getChildren() const
{
	return( (SoChildList *)&children ) ;
}



//------------------------------------------------------------------------------
// Read in the Obstacle node.
//------------------------------------------------------------------------------
SbBool
Obstacle::readInstance(
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoInput		*in,
	unsigned short	flags
#else
	SoInput	*in
#endif
	)
{
	disableSensors() ;

	SbBool status = preRead() ;

	if( status )
	{
		// Read field info as usual.
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
		if( !SoFieldContainer::readInstance( in, flags ) )
#else
		if( !SoFieldContainer::readInstance( in ) )
#endif
		{
			status = FALSE ;
		}
		else
		{
			// status = readGeometry() ;
			nameChangedCB( (void *)this, NULL ) ;
			status = ( root->getNumChildren() > 0 ) ;
		}
	}

	status = postRead( status ) ;

	enableSensors() ;

	return( status ) ;
}



#ifdef VERBOSE
//------------------------------------------------------------------------------
// Debug.
//------------------------------------------------------------------------------
void
Obstacle::addBoundingBoxGeom()
{
	if( !showBoundingBox ) return ;

	int32_t indices[] = {
		0, 1, 2, 3, 0, -1, 4, 5, 6, 7, 4, -1,
		0, 4, -1, 1, 5, -1, 2, 6, -1, 3, 7, -1,
		8, 9, 10, 11, 8, -1,
		12, 13, 14, 15, 12, -1,
		16, 17, 18, 19, 16, -1 } ;
	SbVec3f	minPt = boundingBox.getMin() - centerOffset ;
	SbVec3f	maxPt = boundingBox.getMax() - centerOffset ;

	SbVec3f ctr( -centerOffset[0], -centerOffset[1], -centerOffset[2] ) ;

	SoSeparator *sep = new SoSeparator ;
	root->addChild( sep ) ;

	SoBaseColor *bc = new SoBaseColor ;
	sep->addChild( bc ) ;
	bc->rgb.setValue( 1.0f, 0.0f, 0.0f ) ;

	SoLightModel *lm = new SoLightModel ;
	sep->addChild( lm ) ;
	lm->model = SoLightModel::BASE_COLOR ;

	SoCoordinate3 *coord = new SoCoordinate3 ;
	sep->addChild( coord ) ;
	coord->point.setNum( 20 ) ;

	SbVec3f *c = coord->point.startEditing() ;
	c[0].setValue( minPt[0], minPt[1], minPt[2] ) ;
	c[1].setValue( minPt[0], maxPt[1], minPt[2] ) ;
	c[2].setValue( maxPt[0], maxPt[1], minPt[2] ) ;
	c[3].setValue( maxPt[0], minPt[1], minPt[2] ) ;
	c[4].setValue( minPt[0], minPt[1], maxPt[2] ) ;
	c[5].setValue( minPt[0], maxPt[1], maxPt[2] ) ;
	c[6].setValue( maxPt[0], maxPt[1], maxPt[2] ) ;
	c[7].setValue( maxPt[0], minPt[1], maxPt[2] ) ;

	c[8].setValue( minPt[0], minPt[1], ctr[2] ) ;
	c[9].setValue( minPt[0], maxPt[1], ctr[2] ) ;
	c[10].setValue( maxPt[0], maxPt[1], ctr[2] ) ;
	c[11].setValue( maxPt[0], minPt[1], ctr[2] ) ;
	c[12].setValue( minPt[0], ctr[1], minPt[2] ) ;
	c[13].setValue( minPt[0], ctr[1], maxPt[2] ) ;
	c[14].setValue( maxPt[0], ctr[1], maxPt[2] ) ;
	c[15].setValue( maxPt[0], ctr[1], minPt[2] ) ;
	c[16].setValue( ctr[0], minPt[1], minPt[2] ) ;
	c[17].setValue( ctr[0], minPt[1], maxPt[2] ) ;
	c[18].setValue( ctr[0], maxPt[1], maxPt[2] ) ;
	c[19].setValue( ctr[0], maxPt[1], minPt[2] ) ;
	coord->point.finishEditing() ;

	SoIndexedLineSet *ls = new SoIndexedLineSet ;
	sep->addChild( ls ) ;
	ls->coordIndex.setValues( 0, sizeof( indices ) / sizeof( indices[0] ),
				indices ) ;
}
#endif /* VERBOSE */



//------------------------------------------------------------------------------
// Center the bounding box by calculating the centerOffset value.
//------------------------------------------------------------------------------
void
Obstacle::centerBoundingBox()
{
	centerOffset = -boundingBox.getCenter() ;

	float x, y, z ;
	boundingBox.getSize( x, y, z ) ;
	x *= 0.5f ;
	y *= 0.5f ;
	z *= 0.5f ;
	collisionRadius = sqrtf( ( x * x + y * y + z * z ) ) ;

	boundingBox.setBounds( -x, -y, -z, x, y, z ) ;
#ifdef VERBOSE
	addBoundingBoxGeom() ;
#endif /* VERBOSE */
}



//------------------------------------------------------------------------------
// Fallback bounding function.
//------------------------------------------------------------------------------
void
Obstacle::boundItBasics()
{
	// Update the bounding box.
	if( root->getNumChildren() > 0 )
	{
		SoGetBoundingBoxAction *bba =
			new SoGetBoundingBoxAction( SbViewportRegion() ) ;
		bba->setInCameraSpace( FALSE ) ;
		bba->apply( root->getChild( 0 ) ) ;
		boundingBox = bba->getBoundingBox() ;
		centerBoundingBox() ;
		delete bba ;
		hasBoundingBox = TRUE ;
	}
}



//------------------------------------------------------------------------------
// Calculate and center the bounding box.
//------------------------------------------------------------------------------
void
Obstacle::boundIt()
{
	Obstacle::boundItBasics() ;
}



//------------------------------------------------------------------------------
// Return the center offset value.
//------------------------------------------------------------------------------
const SbVec3f &
Obstacle::getCenterOffset()
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	return( centerOffset ) ;
}



//------------------------------------------------------------------------------
// Read in the geometry.
//------------------------------------------------------------------------------
SbBool
Obstacle::readGeometry()
{
	SoNode *node = gList.getNode( name.getValue() ) ;

	root->removeAllChildren() ;

	if( node != NULL )
	{
		root->addChild( node ) ;
	}
	else
	{
		boundingBox.makeEmpty() ;
		centerOffset.setValue( 0.0f, 0.0f, 0.0f ) ;
	}
	hasBoundingBox = FALSE ;

	processGeometry() ;

	return ( node != NULL ) ;
}



//------------------------------------------------------------------------------
// Perform any additional steps on the geometry once it is read in (virtual).
//------------------------------------------------------------------------------
void
Obstacle::processGeometry()
{
}



//------------------------------------------------------------------------------
// GeometryNodeList class methods.
//------------------------------------------------------------------------------
void
GeometryNodeList::truncate(
	int	start
	)
{
	for( int i = start ; i < getLength() ; i++ )
	{
		if( (*this)[i] != NULL )
		{
			delete (*this)[i] ;
		}
	}

	SbPList::truncate( start ) ;
}



//------------------------------------------------------------------------------
// GeometryNode class constructor.
//------------------------------------------------------------------------------
GeometryNode::GeometryNode(
	const SbString &s,
	SoNode *n
	) :
	name( s ),
	node( n )
{
	if( node != NULL )
	{
		node->ref() ;
	}
}



//------------------------------------------------------------------------------
// GeometryNode class destructor.
//------------------------------------------------------------------------------
GeometryNode::~GeometryNode()
{
	if( node != NULL )
	{
		node->unref() ;
	}
}



//------------------------------------------------------------------------------
// GeometryNodeList class method to load or locate a previously loaded
// geometry scene graph.
//------------------------------------------------------------------------------
SoNode *
GeometryNodeList::getNode(
	const SbString &s
	)
{
	for( int i = 0 ; i < getLength() ; i++ )
	{
		GeometryNode *gn = (*this)[i] ;
		if( gn->name == s )
		{
			return( gn->node ) ;
		}
	}

	// File is has not been loaded.
	SoInput *in = Utilities::createInput() ;

	const char *filename = s.getString() ;

	if( !in->openFile( filename, TRUE ) )
	{
		SoReadError::post( in, "Can't open included file \"%s\" in "
			"Obstacle node", filename ) ;
		delete in ;
		return NULL ;
	}

	SoNode *node = SoDB::readAll( in ) ;

	if( node == NULL )
	{
		SoReadError::post( in, "Can't read included file \"%s\" in "
			"Obstacle node", filename ) ;
	}
	in->closeFile() ;

	delete in ;

	// Add to list even if no geometry was loaded (saves retrying to
	// open the file and spitting out error messages).
	append( new GeometryNode( s, node ) ) ;

	return( node ) ;
}



//------------------------------------------------------------------------------
// Name changed sensor callback.
//------------------------------------------------------------------------------
void
Obstacle::nameChangedCB(
	void		*data,
	SoSensor	*
	)
{
	Obstacle *obst = (Obstacle *)data ;

	obst->readGeometry() ;
	obst->hasBoundingBox = FALSE ;
}



//------------------------------------------------------------------------------
// PreRead function.
//------------------------------------------------------------------------------
SbBool
Obstacle::preRead()
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// PostRead function.
//------------------------------------------------------------------------------
SbBool
Obstacle::postRead(
	SbBool status
	)
{
	return( status ) ;
}



//------------------------------------------------------------------------------
// Detach the field sensors.
//------------------------------------------------------------------------------
void
Obstacle::disableSensors()
{
	nameSensor->detach() ;
}



//------------------------------------------------------------------------------
// Attach the field sensors.
//------------------------------------------------------------------------------
void
Obstacle::enableSensors()
{
	nameSensor->attach( &name ) ;
}



//------------------------------------------------------------------------------
// Indicate that the state is not changed outside the node.
//------------------------------------------------------------------------------
SbBool
Obstacle::affectsState() const
{
	return( FALSE ) ;
}
