//------------------------------------------------------------------------------
// Obstacle abstract base class.
//------------------------------------------------------------------------------
#ifndef _OBSTACLE_H_
#define _OBSTACLE_H_

#include <Inventor/SbBox.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFRotation.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/misc/SoChildList.h>
#include <Inventor/nodes/SoTransformation.h>
#include <Inventor/sensors/SoFieldSensor.h>

class SoSeparator ;
class CollisionAction ;
class LosAction ;
class SoState ;
class SoAction ;


//------------------------------------------------------------------------------
// GeometryNode helper class.
//------------------------------------------------------------------------------
class GeometryNode {
    public :
	SbString	name ;
	SoNode		*node ;
	GeometryNode( const SbString &s, SoNode *n ) ;
	~GeometryNode() ;
    private :
	GeometryNode() ;
	GeometryNode( GeometryNode &gn ) ;
	GeometryNode & operator = ( GeometryNode &gn ) ;
} ;



//------------------------------------------------------------------------------
// GeometryNodeList helper class.
//------------------------------------------------------------------------------
class GeometryNodeList : public SbPList {
    public :
	GeometryNodeList() : SbPList() {}
	~GeometryNodeList() { truncate( 0 ) ; }
	GeometryNode *operator []( int i ) const
	    { return ((GeometryNode *)((*(const SbPList *)this )[i]) ) ; }
	void append( GeometryNode *gn ) { SbPList::append( (void *)gn ) ; }
	void truncate( int start ) ;
	SoNode *getNode( const SbString &s ) ;
    private :
	GeometryNodeList( GeometryNodeList &gn ) ;
	GeometryNodeList & operator = ( GeometryNodeList &gn ) ;
} ;



class Obstacle : public SoTransformation {

	SO_NODE_ABSTRACT_HEADER( Obstacle ) ;

    public:
	static void 	initClass() ;

	SoSFVec3f	position ;
	SoSFRotation	rotation ;
	SoSFString	name ;
	const SbBox3f	&getBox() const { return( boundingBox ) ; }
	const SbVec3f	&getCenterOffset() ;
	static void purgeGeometry() { gList.truncate( 0 ) ; }

    SoEXTENDER public :
	virtual void	doAction( SoAction *action ) ;
	virtual void	getBoundingBox( SoGetBoundingBoxAction *action ) ;
	virtual void	GLRender( SoGLRenderAction *action ) ;
	virtual void	pick( SoPickAction *action ) ;
	virtual void	getMatrix( SoGetMatrixAction *action ) ;
	virtual void	callback( SoCallbackAction *action ) ;
	virtual void	collide( CollisionAction *action ) = 0 ;
	virtual void	obstruct( LosAction *action ) = 0 ;
	virtual SoChildList *getChildren() const ;
	virtual SbBool	affectsState() const ;

    protected:
	Obstacle() ;
	Obstacle( const char *s ) ;
	virtual	~Obstacle() ;
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	virtual SbBool	readInstance( SoInput *in, unsigned short flags ) ;
#else
	virtual SbBool	readInstance( SoInput *in ) ;
#endif
	const SbString	&getGeometryName() const { return name.getValue() ; }
	static GeometryNodeList	gList ;
	virtual void	processGeometry() ;
	virtual SbBool	preRead() ;
	virtual SbBool	postRead( SbBool status ) ;
	virtual void	disableSensors() ;
	virtual void	enableSensors() ;
	virtual void	boundIt() ;
	SbBool		readGeometry() ;

	SbBox3f		boundingBox ;
	SbVec3f		corner[2][2][2] ;
	SbVec3f		centerOffset ;
	float		collisionRadius ;
	SoSeparator	*root ;
	void		addBoundingBoxGeom() ;
	void		centerBoundingBox() ;
	void		boundItBasics() ;
	SbBool		hasBoundingBox ;
	SbBool		dirtyCorners ;
	void		actionGuts( SoState *state, SoAction *action ) ;

    private:

	SoChildList	children ;
	void		createIt() ;
	SoFieldSensor	*nameSensor ;
	static void	nameChangedCB( void *data, SoSensor *sensor ) ;

#ifdef VERBOSE
	static SbBool	showBoundingBox ;
#endif /* VERBOSE */

} ;
#endif /* !_OBSTACLE_H_ */
