//------------------------------------------------------------------------------
// Group class that only redraws its children if they have changed.
//------------------------------------------------------------------------------

#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include "RedrawGroup.h"


SO_NODE_SOURCE( RedrawGroup ) ;

int		RedrawGroup::redrawCount = 0 ;
const int	RedrawGroup::maxRedraws = 2 ;


//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
RedrawGroup::initClass()
{
	SO_NODE_INIT_CLASS( RedrawGroup, SoGroup, "Group" ) ;
	redrawCount = 0 ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
RedrawGroup::createIt()
{
	SO_NODE_CONSTRUCTOR( RedrawGroup ) ;

	sceneSensor = new SoNodeSensor( sceneChangedCB, this ) ;
	sceneSensor->attach( this ) ;
	redrawChildren = maxRedraws ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
RedrawGroup::RedrawGroup() :
	SoGroup()
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor that specifies the number of children to be added.
//------------------------------------------------------------------------------
RedrawGroup::RedrawGroup(
	int	nChildren
	) :
	SoGroup( nChildren )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
RedrawGroup::~RedrawGroup()
{
	delete sceneSensor ;
}



//------------------------------------------------------------------------------
// Handler for the GLRender action.
//------------------------------------------------------------------------------
void
RedrawGroup::GLRender(
	SoGLRenderAction	*action
	)
{
	//
	// If children have changed, draw them the requested number of times.
	//
	if( redrawChildren > 0 )
	{
		SoGroup::doAction( (SoAction *)action ) ;
		redrawChildren-- ;
	}
	//
	// If there is a global redraw in effect, redraw.
	//
	else if( redrawCount > 0 )
	{
		SoGroup::doAction( (SoAction *)action ) ;
	}
}



//------------------------------------------------------------------------------
// Callback function for when the children change.
//------------------------------------------------------------------------------
void
RedrawGroup::sceneChangedCB(
	void		*data,
	SoSensor	*
	)
{
	RedrawGroup *rg = (RedrawGroup *)data ;

	rg->redrawChildren = maxRedraws ;
}



//------------------------------------------------------------------------------
// Read in an instance of the node.
//------------------------------------------------------------------------------
SbBool
RedrawGroup::readInstance(
#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoInput		*in,
	unsigned short	flags
#else
	SoInput	*in
#endif
	)
{
	SbBool status = TRUE ;

	sceneSensor->detach() ;

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	status = SoGroup::readInstance( in, flags ) ;
#else
	status = SoGroup::readInstance( in ) ;
#endif
	
	sceneSensor->attach( this ) ;

	return( status ) ;
}
