//------------------------------------------------------------------------------
// Sfx -- audio sound effects class header.
//
// Chris Fouts - January 1995.
//
// $Id: Sfx.h,v 1.12 1995/07/23 00:39:03 fouts Exp fouts $
//------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////
//
// The Sfx initialization routine, Sfx::init, sets a signal handler on SIGCHLD
// using sigaction(2).  Using this library with any other type of signals is
// strongly not recommended.
//
////////////////////////////////////////////////////////////////////////////////
//
// Copyright 1995, Silicon Graphics, Inc.
// All Rights Reserved.
//
// This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
// the contents of this file may not be disclosed to third parties, copied or
// duplicated in any form, in whole or in part, without the prior written
// permission of Silicon Graphics, Inc.
//
// RESTRICTED RIGHTS LEGEND:
// Use, duplication or disclosure by the Government is subject to restrictions
// as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
// and Computer Software clause at DFARS 252.227-7013, and/or in similar or
// successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
// rights reserved under the Copyright Laws of the United States.
//

#ifndef	_SFX_H
#define	_SFX_H

#include <dmedia/audio.h>
#include <signal.h>
#include <ulocks.h>
#include <sys/ucontext.h>

class RingBuffer ;
class RingBufferInput ;
class RingBufferOutput ;
class Semaphore ;

class Sfx {
    public:
	enum		Flags {
				DISABLED = 0x01
				} ;
	enum		Volume {
				SOFTER,
				LOUDER
				} ;
	// Constructors
	Sfx( const char *fileName ) ;
	Sfx( const Sfx& s ) ;

	// Destructor
	virtual ~Sfx() ;

	// Start playing the sound.
	virtual void	play() ;

	// Stop playing the sound.
	virtual void	stop() ;

	// Play the sound in 'dt' seconds.
	void		delayedPlay( float dt ) ;

	// Return the length (in seconds) of the sound clip.
	float		duration() ;

	// Return the name of the audio file.
	const char	*getName() ;

	// Initialize the sound handler.  This must be called before any
	// other calls to the Sfx library.
	//
	//	dirName is a directory to search in for all relative sound
	//		filenames (full paths override this)
	//	qSize is the number of samples mixed per iteration.  This
	//		affects the granularity of when sounds start and
	//		stop
	static void	init( const char *dirName = ".",
				unsigned int qSize = 1000 ) ;

	// Kill the sound handler (should be called before exiting app).
	static void	quit() ;

	static int	isEnabled() { return( soundEnabled == 1 ) ; }
	// Change the master volume.
	static int	changeVolume( Volume direction ) ;

	// Set the master volume.
	static void	setVolume( int level ) ;

	// Start synchronization mode.  This will cause all sounds which
	// have their "play()" method called to start together when the
	// releaseSynch() method is called.
	static void	setSynch() ;
	static void	releaseSynch() ;

	// Reset the rate to read in new sounds at differing rates.  Call only
	// if all earlier sounds have been deleted.  Returns old rate.
	static int	resetRate() ;

	// Set the audio rate.  Returns previous rate.
	static int	setRate( int newRate ) ;

	// Disable audio, must be called before init().
	static void	disableSfx() ;

    protected:
	int			id ;		// Identifier
	int			flags ;		// Not used currently
	int			numChannels ;	// No. of audio channels (1,2)
	int			synch ;		// Synch flag
	long			numSamps ;	// Number of samples
	long			sampsToPlay ;	// No. of samples left to play
	long			sampsPlayed ;	// No. of samples played
	unsigned long		delay ;		// Delay time in samples
	char			*name ;		// File name
	int			*ref ;		// Reference counter
	short			*soundData ;	// Audio data
	void			loadAudioFile( const char *fileName ) ;
	void			signalHandler() ;
	virtual int		fillQueue( int *queue ) ;
	static long		origLeftGain ;
	static long		origRightGain ;
	static long		currLeftGain ;
	static long		currRightGain ;
	static long		origOutputRate ;
	static ALport		ap ;
	static char		*audioDir ;
	static int		soundEnabled ;
	static int		soundId ;
	static int		gainIndex ;
	static const int	MaxSampleValue ;
	static const int	MinSampleValue ;
	static int		queueSize ;
	static int		sampsPerSec ;
	static int		rate ;
	static pid_t		soundChild ;
	static int		endingOnPurpose ;
	static int		synchMode ;
	static RingBuffer	*rb ;
	static RingBufferInput	*in ;
	static RingBufferOutput	*out ;
	static void		openAudioPort() ;
	static void		soundDied( int, siginfo_t *, ucontext_t * ) ;
	static void		intrHandler( int, int, struct sigcontext * ) ;
	static void		soundHandler( void * ) ;
	static void		dieGracefully( int, int, struct sigcontext * ) ;
	static void		setRate() ;
	static void		setLevel( float, float ) ;
	static void		setGainIndex( long ) ;
	static void		checkVolume() ;
	static int		checkRate( int newRate ) ;
	static Semaphore	*deleteSem ;
	static usptr_t		*arena ;
	static char		*arenaName ;
	Sfx() ;

    } ;

#endif /* _SFX_H */
