//------------------------------------------------------------------------------
// Speedometer class.
//------------------------------------------------------------------------------

#include <Inventor/SoDB.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include "Utilities.h"
#include "Speedometer.h"



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Speedometer::Speedometer()
{
	SoInput	*in = Utilities::createInput() ;

	static const char *filename = "graphs/speed_display.iv" ;

	if( !in->openFile( filename, TRUE ) )
	{
		SoReadError::post( in, "Can't open %s", filename ) ;
		exit( -1 ) ;
	}

	display = SoDB::readAll( in ) ;
	in->closeFile() ;

	if( display == NULL )
	{
		SoReadError::post( in, "Can't read %s", filename ) ;
		exit( -1 ) ;
	}

	display->ref() ;

	SbString name( "SpeedLit" ) ;
	lit = (SoTriangleStripSet *)Utilities::getNamedNode(
				name.getString(),
				SoTriangleStripSet::getClassTypeId() ) ;
	if( lit == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"TriangleStripSet node named `%s'\n",
			filename, name.getString() ) ;
		exit( -1 ) ;
	}

	name = "SpeedUnlit" ;
	unlit = (SoTriangleStripSet *)Utilities::getNamedNode(
				name.getString(),
				SoTriangleStripSet::getClassTypeId() ) ;
	if( unlit == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"TriangleStripSet node named `%s'\n",
			filename, name.getString() ) ;
		exit( -1 ) ;
	}

	name = "SpeedColor" ;
	litColor = (SoBaseColor *)Utilities::getNamedNode(
				name.getString(),
				SoBaseColor::getClassTypeId() ) ;
	if( litColor == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"BaseColor node named `%s'\n", filename,
			name.getString() ) ;
		exit( -1 ) ;
	}

	delete in ;

	positive = TRUE ;
	speedLevel = -1 ;
	setSpeed( 0.0f ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Speedometer::~Speedometer()
{
	display->unref() ;
}



//------------------------------------------------------------------------------
// Set the speed level.
//------------------------------------------------------------------------------
void
Speedometer::setSpeed(
	float newSpeed
	)
{
	speed = newSpeed ;

	updateDisplay() ;
}



//------------------------------------------------------------------------------
// Update the speedometer display.
//------------------------------------------------------------------------------
void
Speedometer::updateDisplay()
{
	int		newLevel ;
	float		absSpeed = speed < 0.0f ? -speed : speed ;
	static int32_t	fours[10] = { 4, 4, 4, 4, 4, 4, 4, 4, 4, 4 } ;

	if( absSpeed > 1.0f )
	{
		absSpeed = 1.0f ;
		newLevel = 10 ;
	}
	else
	{
		newLevel = (int)( 10.0f * absSpeed + 0.5f ) ;
	}

	if( newLevel != speedLevel || ( ( speed < 0.0f ) == positive ) )
	{
		if( newLevel <= 0 )
		{
			newLevel = 0 ;
			lit->numVertices.setValue( 0 ) ;
			unlit->numVertices.setNum( 10 ) ;
			unlit->numVertices.setValues( 0, 10, fours ) ;
		}
		else if( newLevel >= 10 )
		{
			newLevel = 10 ;
			lit->numVertices.setNum( 10 ) ;
			lit->numVertices.setValues( 0, 10, fours ) ;
			unlit->numVertices.setValue( 0 ) ;
		}
		else
		{
			lit->numVertices.setNum( newLevel ) ;
			lit->numVertices.setValues( 0, newLevel, fours ) ;
			unlit->numVertices.setNum( 10 - newLevel ) ;
			unlit->numVertices.setValues( 0, 10 - newLevel,
							fours ) ;
		}
		unlit->startIndex.setValue( 4 * newLevel ) ;

		if( speed >= 0.0f )
		{
			litColor->rgb.setValue( SbColor( 0.0f, 0.5f, 0.0f ) );
		}
		else
		{
			litColor->rgb.setValue( SbColor( 0.5f, 0.0f, 0.0f ) );
		}

		speedLevel = newLevel ;
		positive = ( speed >= 0 ) ;
	}
}



