//------------------------------------------------------------------------------
// StaticObstacle class.
//------------------------------------------------------------------------------

#include <math.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include "StaticObstacle.h"
#include "CollisionAction.h"
#include "LosAction.h"
#include "Utilities.h"



SO_NODE_SOURCE( StaticObstacle ) ;

//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
StaticObstacle::initClass()
{
	if( Obstacle::getClassTypeId() == SoType::badType() )
	{
		Obstacle::initClass() ;
	}
	SO_NODE_INIT_CLASS( StaticObstacle, Obstacle, "Obstacle" ) ;
}



//------------------------------------------------------------------------------
// Common constructor steps.
//------------------------------------------------------------------------------
void
StaticObstacle::createIt()
{
	SO_NODE_CONSTRUCTOR( StaticObstacle ) ;

	highAspectRatio = FALSE ;

	positionChangedSensor = new SoFieldSensor( positionChangedCB, this ) ;
	positionChangedSensor->setPriority( 0 ) ;

	rotationChangedSensor = new SoFieldSensor( rotationChangedCB, this ) ;
	rotationChangedSensor->setPriority( 0 ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
StaticObstacle::StaticObstacle()
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
StaticObstacle::StaticObstacle(
	const char *name
	) : Obstacle( name )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
StaticObstacle::~StaticObstacle()
{
	delete positionChangedSensor ;
	delete rotationChangedSensor ;
}



//------------------------------------------------------------------------------
// Calculate and center the bounding box.
//------------------------------------------------------------------------------
void
StaticObstacle::boundIt()
{
	Obstacle::boundItBasics() ;
	calculateCorners() ;
}



/*
//------------------------------------------------------------------------------
// Read in the geometry.
//------------------------------------------------------------------------------
SbBool
StaticObstacle::readGeometry()
{
	SoNode *node = gList.getNode( name.getValue() ) ;

	root->removeAllChildren() ;

	if( node != NULL )
	{
		root->addChild( node ) ;
		hasBoundingBox = FALSE ;
	}
	else
	{
printf( "geometry is NULL\n" ) ;
		boundingBox.makeEmpty() ;
		centerOffset.setValue( 0.0f, 0.0f, 0.0f ) ;
		calculateCorners() ;
		hasBoundingBox = TRUE ;
	}

	return ( node != NULL ) ;
}
*/



//------------------------------------------------------------------------------
// Process the geometry after it has been read in.
//------------------------------------------------------------------------------
void
StaticObstacle::processGeometry()
{
	if( root->getNumChildren() < 1 )
	{
		calculateCorners() ;
	}
}



//------------------------------------------------------------------------------
// Calculate the corners in world space and update the collision box.
//------------------------------------------------------------------------------
void
StaticObstacle::calculateCorners()
{
	//
	// Set corners to the position field if the bounding box is empty.
	//
	if( boundingBox.isEmpty() )
	{
		corner[0][0][0] = position.getValue() ;
		corner[0][0][1] = position.getValue() ;
		corner[0][1][0] = position.getValue() ;
		corner[0][1][1] = position.getValue() ;
		corner[1][0][0] = position.getValue() ;
		corner[1][0][1] = position.getValue() ;
		corner[1][1][0] = position.getValue() ;
		corner[1][1][1] = position.getValue() ;
		centerOffset.setValue( 0.0f, 0.0f, 0.0f ) ;
		collisionRadius = 0.0f ;
	}
	else
	{
		SbMatrix m ;
		m.setTranslate( position.getValue() ) ;

		SbMatrix r ;
		rotation.getValue().getValue( r ) ;
		m.multLeft( r ) ;

		SbVec3f &min = boundingBox.getMin() ;
		SbVec3f &max = boundingBox.getMax() ;

		m.multVecMatrix( SbVec3f( min[0], min[1], min[2] ),
				corner[0][0][0] ) ;
		m.multVecMatrix( SbVec3f( min[0], min[1], max[2] ),
				corner[0][0][1] ) ;
		m.multVecMatrix( SbVec3f( min[0], max[1], min[2] ),
				corner[0][1][0] ) ;
		m.multVecMatrix( SbVec3f( min[0], max[1], max[2] ),
				corner[0][1][1] ) ;
		m.multVecMatrix( SbVec3f( max[0], min[1], min[2] ),
				corner[1][0][0] ) ;
		m.multVecMatrix( SbVec3f( max[0], min[1], max[2] ),
				corner[1][0][1] ) ;
		m.multVecMatrix( SbVec3f( max[0], max[1], min[2] ),
				corner[1][1][0] ) ;
		m.multVecMatrix( SbVec3f( max[0], max[1], max[2] ),
				corner[1][1][1] ) ;
	}
	cBox.setPoints( corner ) ;

	sideNormal[0] = cBox.getNormal( 0 ) ;
	halfSize[0] = 0.5f * sideNormal[0].normalize() ;
	sideNormal[1] = cBox.getNormal( 1 ) ;
	halfSize[1] = 0.5f * sideNormal[1].normalize() ;
	sideNormal[2] = cBox.getNormal( 2 ) ;
	halfSize[2] = 0.5f * sideNormal[2].normalize() ;

	float maxDim = halfSize[0] > halfSize[1] ?
		( halfSize[0] > halfSize[2] ? halfSize[0] : halfSize[2] ) :
		( halfSize[1] > halfSize[2] ? halfSize[1] : halfSize[2] ) ;
	float minDim = halfSize[0] < halfSize[1] ?
		( halfSize[0] < halfSize[2] ? halfSize[0] : halfSize[2] ) :
		( halfSize[1] < halfSize[2] ? halfSize[1] : halfSize[2] ) ;

	if( minDim == 0.0f || maxDim / minDim > 2.0f )
	{
		highAspectRatio = TRUE ;
	}
	else
	{
		highAspectRatio = FALSE ;
	}
}



//------------------------------------------------------------------------------
// CollisionAction handler.
//------------------------------------------------------------------------------
void
StaticObstacle::collide(
	CollisionAction *action
	)
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	SbVec3f location = position.getValue() ;
	if( highAspectRatio )
	{
		location -= action->getCenter() ;
		for( int i = 0 ; i < 3 ; i++ )
		{
			float d = fabs( location.dot( sideNormal[i] ) ) ;
			if( d > ( action->getRadius() + halfSize[i] ) )
			{
				return ;
			}
		}
	}
	// Trivial distance-based test using bounding sphere radii.
	else if( action->checkDistance( location, collisionRadius ) )
	{
		return ;
	}

		
	//
	// For each of the 8 corner points, transform to local collision
	// space and test if inside collision bounding box.
	//
	int n ;
	unsigned int pStatus[8] ;
	SbVec3f pt[8] ;
	float xMin, xMax, yMin, yMax, zMin, zMax ;
	action->getBounds( xMin, yMin, zMin, xMax, yMax, zMax ) ;
	for( n = 0 ; n < 8 ; n++ )
	{
		//
		// Clear status flags for the point.  If no point is inside,
		// the status flags are used to see which edges of the
		// obstacle's bounding box might intersect the collision box.
		//
		pStatus[n] = 0 ;
		action->xformWorldToLocalPt( cBox[n], pt[n] );
		if( pt[n][0] >= xMin )
			pStatus[n] |= CollisionAction::XMIN_INNER ;
		if( pt[n][0] <= xMax )
			pStatus[n] |= CollisionAction::XMAX_INNER ;
		if( pt[n][1] >= yMin )
			pStatus[n] |= CollisionAction::YMIN_INNER ;
		if( pt[n][1] <= yMax )
			pStatus[n] |= CollisionAction::YMAX_INNER ;
		if( pt[n][2] >= zMin )
			pStatus[n] |= CollisionAction::ZMIN_INNER ;
		if( pt[n][2] <= zMax )
			pStatus[n] |= CollisionAction::ZMAX_INNER ;
		//
		// If point was inside, save the point (and normal) and return.
		//
		if( pStatus[n] == CollisionAction::ALL_INNER )
		{
			int i = ( n & 0x04 ) >> 2 ;
			int j = ( n & 0x02 ) >> 1 ;
			int k = n & 0x01 ;

			// Object space normal
			SbVec3f oDir( i == 0 ? -1.0f : 1.0f,
				j == 0 ? -1.0f : 1.0f, k == 0 ? -1.0f : 1.0f ) ;

			action->savePointCollisionInfo( rotation.getValue(),
					oDir, TRUE, pt[n], cBox[n], FALSE ) ;
			return ;
		}
	}

	//
	// Didn't find a point inside, so check lines and then surfaces for
	// intersections.
	//
	action->detailedCheckCollision( pt, pStatus, rotation.getValue() ) ;
}



//------------------------------------------------------------------------------
// LosAction handler.
//------------------------------------------------------------------------------
void
StaticObstacle::obstruct(
	LosAction *action
	)
{
	if( action->isTarget( this ) )
	{
		return ;
	}

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	//
	// Check if line intersects bounding box.
	//
	action->checkBoundingBox( boundingBox, rotation.getValue(),
					position.getValue(), collisionRadius ) ;
}



//------------------------------------------------------------------------------
// Position changed sensor callback.
//------------------------------------------------------------------------------
void
StaticObstacle::positionChangedCB(
	void		*data,
	SoSensor	*
	)
{
	StaticObstacle *obst = (StaticObstacle *)data ;

	obst->calculateCorners() ;
}



//------------------------------------------------------------------------------
// Rotation changed sensor callback.
//------------------------------------------------------------------------------
void
StaticObstacle::rotationChangedCB(
	void		*data,
	SoSensor	*
	)
{
	StaticObstacle *obst = (StaticObstacle *)data ;

	obst->calculateCorners() ;
}



//------------------------------------------------------------------------------
// Detach the field sensors.
//------------------------------------------------------------------------------
void
StaticObstacle::disableSensors()
{
	Obstacle::disableSensors() ;
	positionChangedSensor->detach() ;
	rotationChangedSensor->detach() ;
}



//------------------------------------------------------------------------------
// Attach the field sensors.
//------------------------------------------------------------------------------
void
StaticObstacle::enableSensors()
{
	Obstacle::enableSensors() ;
	positionChangedSensor->attach( &position ) ;
	rotationChangedSensor->attach( &rotation ) ;
}



