//------------------------------------------------------------------------------
// Tank class.
//------------------------------------------------------------------------------

#include <Inventor/SbViewportRegion.h>
#include <Inventor/SoDB.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/fields/SoMFString.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/nodes/SoDrawStyle.h>

#include <assert.h>

#include "Utilities.h"
#include "FollowTerrainAction.h"
#include "CollisionAction.h"
#include "Tank.h"


#ifdef VERBOSE
static SbBool debugging = FALSE ;
static Utilities	*Util = NULL ;
#endif /* VERBOSE */

static const float	Gravity = Constants::Gravity ;

enum TankSide { LEFT_SIDE = 0, RIGHT_SIDE = 1,
		BACK_SIDE = 2, FRONT_SIDE = 3,
		BOTTOM_SIDE = 4, TOP_SIDE = 5
		} ;


//------------------------------------------------------------------------------
// Helper class.
//------------------------------------------------------------------------------
class GroundSpot {
    public :
	GroundSpot() {}
	~GroundSpot() {}
	unsigned int side ;
	SbVec3f point ;
	SbBool onGround ;
	float s ;
	float t ;
} ;



//------------------------------------------------------------------------------
// Helper class.
//------------------------------------------------------------------------------
class GroundPoly {
    public :
	GroundPoly( int nPts, const SbVec3f *pts ) ;
	~GroundPoly() {}
	SbBool intersect( const GroundSpot &a, const GroundSpot &b,
				GroundSpot &pt ) ;
    private :
	SbVec3f normal ;
	SbVec3f	point[4] ;
	int	numPts ;
	GroundPoly() ;
	GroundPoly( const GroundPoly &gt ) ;
} ;



//------------------------------------------------------------------------------
// Another helper class.
//------------------------------------------------------------------------------
class GroundLine {
    public :
	GroundLine() {}
	~GroundLine() {}
	GroundLine *nextLine ;
	int pIndex0 ;
	int pIndex1 ;
	SbBool isCorner0 ;
	SbBool isCorner1 ;
	int side ;
	void splitLine( GroundLine &gl, int newPoint ) ;
    private :
} ;




SO_NODE_SOURCE( Tank ) ;



//------------------------------------------------------------------------------
// Initialize the Tank class.
//------------------------------------------------------------------------------
void
Tank::initClass()
{
	if( Vehicle::getClassTypeId() == SoType::badType() )
	{
		Vehicle::initClass() ;
	}
	SO_NODE_INIT_CLASS( Tank, Vehicle, "Vehicle" ) ;

#ifdef VERBOSE
	Util = new Utilities( "DEBUG_TANK" ) ;
	debugging = ( Util->getDebugLevel() > 0 ) ;
#endif /* VERBOSE */
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
Tank::createIt()
{
	SO_NODE_CONSTRUCTOR( Tank ) ;

	reset() ;
	mass = 1.0f ;
	points = 50 ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Tank::Tank() :
	Vehicle( "models/tank.iv" )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
Tank::Tank(
	const char *s
	) :
	Vehicle( s )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Tank::~Tank()
{
}



//------------------------------------------------------------------------------
// Set up the bounding box for terrain following and collisions and calculate
// the moments of inertia.
//------------------------------------------------------------------------------
void
Tank::boundIt()
{
	Vehicle::boundIt() ;

	boundingBox.getSize( width, length, height ) ;
	// Assume (for now) that the tank is uniform density
	// throughout the bounding box region.
	float div = 1.0f / 12.0f ;
	Ix = ( length * length + height * height ) * div ;
	Iy = ( width * width + height * height ) * div ;
	Iz = ( width * width + length * length ) * div ;
	width *= 0.5f ;
	length *= 0.5f ;
	height *= 0.5f ;

#ifdef VERBOSE
	if( debugging )
	{
		Util->print( 9, "creating debug geometry\n" ) ;
		SoSeparator *dbSep = new SoSeparator ;
		root->addChild( dbSep ) ;

		SoResetTransform *rxf = new SoResetTransform ;
		dbSep->addChild( rxf ) ;
		rxf->whatToReset = SoResetTransform::TRANSFORM ;

		SoDrawStyle *ds = new SoDrawStyle ;
		dbSep->addChild( ds ) ;
		ds->lineWidth = 1 ;

		dbCoord = new SoCoordinate3 ;
		dbSep->addChild( dbCoord ) ;

		SoLightModel *lm = new SoLightModel ;
		dbSep->addChild( lm ) ;
		lm->model = SoLightModel::BASE_COLOR ;

		SoBaseColor *bc = new SoBaseColor ;
		dbSep->addChild( bc ) ;
		bc->rgb.setValue( 1.0f, 1.0f, 0.0f ) ;

		dbLine = new SoIndexedLineSet ;
		dbSep->addChild( dbLine ) ;
		dbLine->coordIndex.setNum( 0 ) ;

		dbFreeCoord = new SoCoordinate3 ;
		dbSep->addChild( dbFreeCoord ) ;

		bc = new SoBaseColor ;
		dbSep->addChild( bc ) ;
		bc->rgb.setValue( 0.0f, 1.0f, 1.0f ) ;

		SoIndexedLineSet *ls = new SoIndexedLineSet ;
		dbSep->addChild( ls ) ;
		ls->coordIndex.setNum( 5 ) ;
		ls->coordIndex.set1Value( 0, 0 ) ;
		ls->coordIndex.set1Value( 1, 1 ) ;
		ls->coordIndex.set1Value( 2, 2 ) ;
		ls->coordIndex.set1Value( 3, 3 ) ;
		ls->coordIndex.set1Value( 4, 0 ) ;
	}
#endif /* VERBOSE */
}



//------------------------------------------------------------------------------
// Utility routine.
//------------------------------------------------------------------------------
float
Tank::findTwist(
	const SbRotation	&r,
	const SbVec3f		&uAxis,
	const SbVec3f		&vAxis
	)
{
	float toler = 1e-6 ;
	SbMatrix rMat ;
	r.getValue( rMat ) ;
	SbVec3f result ;
	rMat.multDirMatrix( vAxis, result ) ;

	float x = result.dot( uAxis ) ;
	float y = result.dot( vAxis ) ;

	if( -toler < x && x < toler ) return( 0.0f ) ;

	float ang = atan2f( -x, y ) ;
	if( ang > (float)M_PI )
	{
		ang -= 2.0f * (float)M_PI ;
	}
	return( ang ) ; 
}



//------------------------------------------------------------------------------
// Helper routine.  Scale the rotation value of an SbRotation vector.  This
// is different from SbRotation::scaleAngle.
//------------------------------------------------------------------------------
void
scaleRotation(
	SbRotation	&r,
	float		factor
	)
{
	SbVec3f axis ;
	float	radians ;

	r.getValue( axis, radians ) ;

	if( radians > (float)M_PI )
	{
		r.setValue( axis, ( radians - 2.0f * (float)M_PI ) * factor ) ;
	}
	else
	{
		r.setValue( axis, radians * factor ) ;
	}
}



//------------------------------------------------------------------------------
// Correct the omega rotation for unwanted twist.
//------------------------------------------------------------------------------
void
Tank::calculateOmega(
	SbRotation	&newOrientation,
	SbRotation	&omega,
	float		desiredTwist
	)
{
	SbVec3f	uAxis ;
	SbVec3f vAxis ;
	SbVec3f wAxis ;

	switch( gpMode )
	{
		case BOTTOM_SIDE :
		case TOP_SIDE :
			uAxis.setValue( 1.0f, 0.0f, 0.0f ) ;
			vAxis.setValue( 0.0f, 1.0f, 0.0f ) ;
			wAxis.setValue( 0.0f, 0.0f, 1.0f ) ;
			break ;
		case LEFT_SIDE :
		case RIGHT_SIDE :
			uAxis.setValue( 0.0f, 1.0f, 0.0f ) ;
			vAxis.setValue( 0.0f, 0.0f, 1.0f ) ;
			wAxis.setValue( 1.0f, 0.0f, 0.0f ) ;
			break ;
		case FRONT_SIDE :
		case BACK_SIDE :
			uAxis.setValue( 0.0f, 0.0f, 1.0f ) ;
			vAxis.setValue( 1.0f, 0.0f, 0.0f ) ;
			wAxis.setValue( 0.0f, 1.0f, 0.0f ) ;
			break ;
	}

	//
	// Let
	//
	// R0 = orientation at time step N
	// R1 = orientation at time step N+1 = newOrientation
	// W = rotation (omega) in world space
	// Wl = rotation in local space (relative to tank so that twist is
	//	about the local twist axis which is normal to the ground
	//	plane of the tank)
	//
	// R1 = R0 * W  --> W = R0' * R1
	//
	// R1 = Wl * R0   --> Wl = R1 * R0'
	//
	// Cancel all twist in Wl --> Wn
	//
	// W = R0' * ( R1 ) = R0' * ( Wn * R0 )
	//

	SbRotation R0 = rotation.getValue() ;
	SbRotation &R0inv = R0.inverse() ;

	SbRotation &Wl = newOrientation * R0inv ;

	// Find twist by calculating the result of multiplying the Y vector
	// (ahead) by the rotation.  Use the XY-plane projection of the
	// resulting vector to determine the angle.  Since we only need the
	// X and Y result, let's use the direct equations:
	float twist = findTwist( Wl, uAxis, vAxis ) ;

	SbRotation anti( wAxis, -twist ) ;
	Wl *= anti ;

	// Wl now equals the non-twisting part of the rotation.  Let's smooth
	// it out for a better ride.
	static const float fudge[5] = { 1.0f, 0.925f, 0.8f, 0.6f, 0.2f } ;
	scaleRotation( Wl, fudge[onGround] ) ;

	// Now add back in the desired twist.
	Wl *= SbRotation( wAxis, desiredTwist ) ;

	//
	// R1 = Wl * R0
	//
	newOrientation = Wl * R0 ;
	//
	// W = R0' * ( R1 ) = R0' * ( Wn * R0 )
	//
	omega = R0inv * newOrientation ;
}



//------------------------------------------------------------------------------
// Move according to input values.
//
// Basic algorithm:
//   1)	Calculate position of tank as if in freefall.
//   2) Locate up to 8 points on the bottom of the tank (4 corner points plus
//	up to 1 point per edge) that intersect the ground -- this ASSUMES the
//	terrain is coarse enough so that no terrain triangle lies completely
//	within the tank.
//   3) Of the resulting points, calculate the energy required to move the
//	tank from the freefall position to the position dictated by each
//	group of 3 (miminum number required to define a plane).
//   4)	Choose the 3 points that give the minimum energy change.
//------------------------------------------------------------------------------
float
Tank::move(
	float	xInput,
	float	yInput,
	float	deltaTime,
	SoNode	*scene
	)
{
	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	int startOnGround = onGround ;

	float dInput = xInput * xInput + yInput * yInput ;
	if( dInput > 1.0f )
	{
		dInput = 1.0f / sqrt( dInput ) ;
		xInput *= dInput ;
		yInput *= dInput ;
	}
	xInput *= ( xInput > 0.0f ) ? xInput : -xInput ;
	yInput *= ( yInput > 0.0f ) ? yInput : -yInput ;

	float newRightEngine = yInput - xInput ;
	float newLeftEngine = yInput + xInput ;

	if( newRightEngine > rightEngine )
	{
		rightEngine += deltaTime ;
		if( rightEngine > newRightEngine )
		{
			rightEngine = newRightEngine ;
		}
	}
	else if( newRightEngine < rightEngine )
	{
		rightEngine -= deltaTime ;
		if( rightEngine < newRightEngine )
		{
			rightEngine = newRightEngine ;
		}
	}
	if( newLeftEngine > leftEngine )
	{
		leftEngine += deltaTime ;
		if( leftEngine > newLeftEngine )
		{
			leftEngine = newLeftEngine ;
		}
	}
	else if( newLeftEngine < leftEngine )
	{
		leftEngine -= deltaTime ;
		if( leftEngine < newLeftEngine )
		{
			leftEngine = newLeftEngine ;
		}
	}

	float twist = ( rightEngine + newRightEngine -
			leftEngine - newLeftEngine ) * 0.25f *
			maxTurnRate ;

#ifdef VERBOSE
	Util->print( 1, "\nTank::move (dt = %f)\n", deltaTime ) ;
#endif /* VERBOSE */
	int	i ;
	SbBool	terrainCollision ;
	float	ch = centerOffset[2] ;
	SbRotation startRot = rotation.getValue() ;
	float	invDeltaTime = 1.0f / deltaTime ;

	SbVec3f oldCenterPos( position.getValue() ) ;

	// Calculate turning due to omega (rotational velocity).
	SbRotation w( omegaAxis, omegaRad * deltaTime ) ;

	// Calculate the position of the tank without user input or environment
	// effects (ie, freefall).
	SbVec3f &freeCenterPos = oldCenterPos + velocity * deltaTime ;
	SbRotation &freeRot = startRot * w ;
	SbVec3f newCenterPos( freeCenterPos ) ;

	SbMatrix mat ;
	mat.setRotate( freeRot ) ;

	// -Z vector
	SbVec3f ahead( mat[1][0], mat[1][1], mat[1][2] ) ;
	// +Y vector
	SbVec3f up( mat[2][0], mat[2][1], mat[2][2] ) ;

	// Based on the current orientation, find the side that presents
	// the most space projected onto the XY plane.
	findBottomSide( mat ) ;

	SbVec3f	collisionNormal ;
	terrainCollision = followTerrain( scene, newCenterPos, mat,
					collisionNormal, deltaTime ) ;
#ifdef VERBOSE
	Util->printVec( 2, "newCenterPos (0)", newCenterPos ) ;
#endif /* VERBOSE */

	// SbRotation newOrientation( mat ) ;
	SbRotation newOrientation ;
	Utilities::buildRotation( newOrientation, mat ) ;

	ca->setLocation( newCenterPos, newOrientation ) ;
	ca->apply( scene ) ;

	const CollisionList &cList = ca->getCollisionList() ;

	// Handle a collision.
	if( cList.getLength() > 0 || terrainCollision )
	{
#ifdef VERBOSE
		Util->print( 2, "collision\n" ) ;
		Util->printRot( 2, "newOrientation", newOrientation ) ;
		Util->printRot( 2, "startRot", startRot ) ;
#endif /* VERBOSE */

		// Find vector of travel.
		SbVec3f &travel = newCenterPos - oldCenterPos ;
#ifdef VERBOSE
		Util->printVec( 2, "travel", travel ) ;
#endif /* VERBOSE */

		// Get world normal(s) direction off of obstacle(s).
		for( i = 0 ; i < cList.getLength() ; i++ )
		{
			collisionNormal += cList[i]->getWorldNormal() ;
		}
		collisionNormal.normalize() ;
#ifdef VERBOSE
		Util->printVec( 2, "collisionNormal", collisionNormal ) ;
#endif /* VERBOSE */

		// Find the amount of travel in the direction of the collision
		// normal.
		float tdn = travel.dot( collisionNormal ) ;
		if( tdn < 0.0f )
		{
			tdn = -tdn ;
		}
		if( tdn < 0.001f )
		{
			tdn = sqrtf( width * width + length * length +
						height * height ) ;
			travel = -0.1f * tdn * collisionNormal ;
		}
#ifdef VERBOSE
		Util->printVal( 2, "tdn", tdn ) ;
#endif /* VERBOSE */

		SbVec3f correction ;
		const float fudge = 1.3f ;
		correction = ( fudge * tdn ) * collisionNormal ;
#ifdef VERBOSE
		Util->printVec( 2, "correction", correction ) ;
#endif /* VERBOSE */

		// Get corrected position (no ground effect).
		SbVec3f modCenterPos = newCenterPos + correction ;
		mat.setRotate( startRot ) ;

		// Get ground effect on correction -> modCenterPos.
		followTerrain( scene, modCenterPos, mat, collisionNormal,
				deltaTime ) ;

		// Calculate the effective normal (normal plus ground effect)
		SbVec3f &effNormal = modCenterPos - newCenterPos ;
		effNormal.normalize() ;
#ifdef VERBOSE
		Util->printVec( 2, "effNormal", effNormal ) ;
		Util->printVec( 2, "modCenterPos", modCenterPos ) ;
#endif /* VERBOSE */
		tdn = travel.dot( effNormal ) ;
		if( tdn < 0.0f ) tdn = -tdn ;
#ifdef VERBOSE
		Util->printVal( 2, "tdn", tdn ) ;
#endif /* VERBOSE */

		float tdt = travel.dot( travel ) ;
		const float limitFactor = 0.075f ;
#ifdef VERBOSE
		Util->printVal( 2, "tdt", tdt ) ;
		Util->printVal( 2, "limitFactor * tdt", limitFactor * tdt ) ;
#endif /* VERBOSE */
		if( tdn * tdn > limitFactor * tdt )
		{
			tdn = tdt / tdn ;
#ifdef VERBOSE
			Util->printVal( 2, "tdn (a)", tdn ) ;
#endif /* VERBOSE */
		}
		else if( tdn > 0.0f )
		{
			float a = ( tdn * tdn / tdt ) / limitFactor ;
			tdn = ( a * tdt / tdn ) + ( 1.0f - a ) * tdn ;
#ifdef VERBOSE
			Util->printVal( 2, "tdn (b)", tdn ) ;
#endif /* VERBOSE */
		}
		correction = ( fudge * tdn ) * effNormal ;
		newCenterPos += correction ;
		
		if( followTerrain( scene, newCenterPos, mat,
				collisionNormal, deltaTime ) )
		{
			const SbVec3f &p = oldCenterPos ;
			position.setValue( p - ( 0.25f * deltaTime ) *
					velocity ) ;
#ifdef VERBOSE
			printf( "adjusting position bumps height\n" ) ;
			Util->printVec( 2, "new center position",
					position.getValue() ) ;
#endif /* VERBOSE */
			newOrientation = rotation.getValue() ;
			float vdn = velocity.dot( effNormal ) ;
			if( vdn < 0.0f )
			{
				velocity -= 1.5 * vdn * effNormal ;
			}
			else
			{
				velocity += 1.25 * vdn * effNormal ;
			}
#ifdef VERBOSE
			Util->printVec( 2, "velocity", velocity ) ;
#endif /* VERBOSE */
			if( startOnGround == 0 )
			{
				return( calculateEnergy( position.getValue(),
						newOrientation, freeCenterPos,
						freeRot, deltaTime ) ) ;
			}
			else if( gpMode != BOTTOM_SIDE && onGround > 0 &&
				omegaRad == 0.0f )
			{
				return( 1000.0f ) ;
			}
			else
			{
				return( 0.0f ) ;
			}
		}
		else
		{
			Utilities::buildRotation( newOrientation, mat ) ;
		}
		terrainCollision = TRUE ;
	}

	SbVec3f freeVel( velocity ) ;
	velocity = newCenterPos - oldCenterPos ;
	velocity *= ( invDeltaTime ) ;

	// Calculate omega.
	startRot.invert() ;
	SbRotation omega = startRot * newOrientation ;
#ifdef VERBOSE
	Util->printRot( 2, "omega (0)", omega ) ;
#endif /* VERBOSE */

	if( terrainCollision )
	{
		velocity *= 0.01f ;
		scaleRotation( omega, 0.01f ) ;
	}

	// Adjust velocity based on user input.
	if( onGround > 0 )
	{
		float desiredSpeed;
		float accel ;
		if( gpMode == BOTTOM_SIDE )
		{
			desiredSpeed = 0.5f * ( newLeftEngine + newRightEngine )
					* maxSpeed ;
			accel = maxAcceleration ;
		}
		else
		{
			desiredSpeed = 0.0f ;
			accel = 2.0f * Gravity ;
		}
		ahead.setValue( mat[1] ) ;
		float forwardSpeed = velocity.dot( gpAh ) ;
		float newSpeed = forwardSpeed ;
		if( forwardSpeed < desiredSpeed )
		{
			newSpeed += accel * deltaTime ;
			if( newSpeed > desiredSpeed )
			{
				newSpeed = desiredSpeed ;
			}
		}
		else if( newSpeed > desiredSpeed )
		{
			newSpeed -= accel * deltaTime ;
			if( newSpeed < desiredSpeed )
			{
				newSpeed = desiredSpeed ;
			}
		}
		velocity += ( newSpeed - forwardSpeed ) * gpAh ;

		// Cancel out side speed.
		SbVec3f right( mat[0] ) ;
		velocity -= velocity.dot( gpRt ) * gpRt ;

		float desiredTwist ;
		if( gpMode == BOTTOM_SIDE )
		{
			desiredTwist = twist * deltaTime ;
		}
		else
		{
			desiredTwist = 0.0f ;
		}
		calculateOmega( newOrientation, omega, desiredTwist ) ;

	}

	omega.getValue( omegaAxis, omegaRad ) ;
	if( omegaRad > (float)M_PI )
	{
		omegaRad -= 2.0f * M_PI ;
	}
	omegaRad *= invDeltaTime ;
	velocity[2] -= Gravity * deltaTime ;
	position.setValue( newCenterPos[0], newCenterPos[1], newCenterPos[2] ) ;
	rotation.setValue( newOrientation ) ;
	radarRotation.setValue( newOrientation ) ;
#ifdef VERBOSE
	Util->printVec( 2, "position", position.getValue() ) ;
	Util->printRot( 2, "rotation", rotation.getValue() ) ;
	Util->printVec( 2, "velocity", velocity ) ;
	Util->printVec( 2, "omegaAxis", omegaAxis ) ;
	Util->printVal( 2, "omegaRad", omegaRad ) ;
	Util->printVal( 2, "onGround", (int32_t)onGround ) ;
#endif /* VERBOSE */

	if( startOnGround == 0 )
	{
		return( calculateEnergy( newCenterPos, newOrientation,
					freeCenterPos, freeRot, deltaTime ) ) ;
	}
	else if( gpMode != BOTTOM_SIDE && onGround > 0 && omegaRad == 0.0f )
	{
		return( 1000.0f ) ;
	}
	else
	{
		return( 0.0f ) ;
	}
}



//------------------------------------------------------------------------------
// Locate point beneath specified point and adjust up vector.
//------------------------------------------------------------------------------
SbBool
Tank::followTerrain(
	SoNode		*terrainNode,
	SbVec3f		&freeCenterPos,	// Freefall position
	SbMatrix	&freeMat,	// Freefall orientation
	SbVec3f		&collisionNormal,
	float		deltaTime
	)
{
	int	i ;
	int	collisions = 0 ;

	collisionNormal.setValue( 0.0f, 0.0f, 0.0f ) ;

	// -Z vector
	SbVec3f ahead( freeMat[1][0], freeMat[1][1], freeMat[1][2] ) ;
	// +Y vector
	SbVec3f up( freeMat[2][0], freeMat[2][1], freeMat[2][2] ) ;
	// +X vector
	SbVec3f right( freeMat[0][0], freeMat[0][1], freeMat[0][2] ) ;

#ifdef VERBOSE
Util->print( 5, "followTerrain\n" ) ;
Util->printVec( 5, "right", right ) ;
Util->printVec( 5, "ahead", ahead ) ;
Util->printVec( 5, "up", up ) ;
Util->printVec( 5, "freeCenterPos", freeCenterPos ) ;
#endif /* VERBOSE */
	GroundSpot spot[12] ;
	spot[0].onGround = FALSE ;
	spot[0].side = ( 1 << 0 ) | ( 1 << 3 ) ;
	spot[0].s = -gpW ;
	spot[0].t = gpL ;
	spot[0].point.setValue(
		freeCenterPos[0]-gpW*gpRt[0]+gpL*gpAh[0]-gpH*gpUp[0],
		freeCenterPos[1]-gpW*gpRt[1]+gpL*gpAh[1]-gpH*gpUp[1],
		freeCenterPos[2]-gpW*gpRt[2]+gpL*gpAh[2]-gpH*gpUp[2] ) ;
	spot[1].onGround = FALSE ;
	spot[1].side = ( 1 << 1 ) | ( 1 << 0 ) ;
	spot[1].s = gpW ;
	spot[1].t = gpL ;
	spot[1].point.setValue(
		freeCenterPos[0]+gpW*gpRt[0]+gpL*gpAh[0]-gpH*gpUp[0],
		freeCenterPos[1]+gpW*gpRt[1]+gpL*gpAh[1]-gpH*gpUp[1],
		freeCenterPos[2]+gpW*gpRt[2]+gpL*gpAh[2]-gpH*gpUp[2] ) ;
	spot[2].onGround = FALSE ;
	spot[2].side = ( 1 << 2 ) | ( 1 << 1 ) ;
	spot[2].s = gpW ;
	spot[2].t = -gpL ;
	spot[2].point.setValue(
		freeCenterPos[0]+gpW*gpRt[0]-gpL*gpAh[0]-gpH*gpUp[0],
		freeCenterPos[1]+gpW*gpRt[1]-gpL*gpAh[1]-gpH*gpUp[1],
		freeCenterPos[2]+gpW*gpRt[2]-gpL*gpAh[2]-gpH*gpUp[2] ) ;
	spot[3].onGround = FALSE ;
	spot[3].side = ( 1 << 3 ) | ( 1 << 2 ) ;
	spot[3].s = -gpW ;
	spot[3].t = -gpL ;
	spot[3].point.setValue(
		freeCenterPos[0]-gpW*gpRt[0]-gpL*gpAh[0]-gpH*gpUp[0],
		freeCenterPos[1]-gpW*gpRt[1]-gpL*gpAh[1]-gpH*gpUp[1],
		freeCenterPos[2]-gpW*gpRt[2]-gpL*gpAh[2]-gpH*gpUp[2] ) ;

#ifdef VERBOSE
if( debugging )
{
	dbFreeCoord->point.set1Value( 0, spot[0].point ) ;
	dbFreeCoord->point.set1Value( 1, spot[1].point ) ;
	dbFreeCoord->point.set1Value( 2, spot[2].point ) ;
	dbFreeCoord->point.set1Value( 3, spot[3].point ) ;
}
#endif /* VERBOSE */

	GroundLine line[12] ;
	for( i = 0 ; i < 4 ; i++ )
	{
		line[i].pIndex0 = i ;
		line[i].pIndex1 = ( i + 1 ) % 4 ;
		line[i].isCorner0 = TRUE ;
		line[i].isCorner1 = TRUE ;
		line[i].side = i ;
		line[i].nextLine = ( i < 3 ) ? line + i + 1 : NULL ;
	}
	int nLines = 4 ;

	int nHits ;
	int	nPoints = 0 ;
	GroundPoly *gPoly[4] ;

	float dropHeight = position.getValue()[2] - freeCenterPos[2] ;
	float maxH ;
	if( gpMode == BOTTOM_SIDE )
	{
		maxH = 0.4f * 2.0f * gpH ;
	}
	else
	{
		maxH = 0.2f * 2.0f * height ;
	}
	maxH += dropHeight ;
	for( i = 0 ; i < 4 ; i++ )
	{
		gPoly[i] = NULL ;

		fta->setSearchPoint( spot[i].point ) ;
		fta->apply( terrainNode ) ;

		const GroundPointList &gpl = fta->getGroundPointList() ;

		nHits = gpl.getLength() ;
		if( nHits > 0 )
		{
			int n = 0 ;

			// Find the height that is accessible.
			while( ( n < nHits - 1 ) && gpl[n]->getHeight() > maxH )
			{
				n++ ;
			}

			float h = gpl[n]->getHeight() ;
			int nPolyPts ;
			const SbVec3f *poly = gpl[n]->getPoly( nPolyPts ) ;
			gPoly[i] = new GroundPoly( nPolyPts, poly ) ;

			float deltaH ;
			if( n > 0 )
			{
				deltaH = gpl[n-1]->getHeight() - h ;
				if( deltaH < 2.0f * gpH )
				{
					n-- ;
					h = 2.0f * gpH ;
				}
			}

			// If there is a collision...
			if( h > maxH || gpl[n]->getNormal()[2] < 0.71f )
			{
				collisions++ ;
				const SbVec3f &tN = gpl[n]->getNormal();
				const SbVec3f &tPt = gpl[n]->getPoint();
				float pdN = tN.dot( position.getValue() ) ;
#ifdef VERBOSE
				Util->printVec( 5, "tN", tN ) ;
				Util->printVec( 5, "tPt", tPt ) ;
				Util->printVal( 5, "old center pos.dot( tN )",
						pdN ) ;
				Util->printVal( 5, "tPt.dot( tN )",
						tPt.dot( tN ) ) ;
#endif /* VERBOSE */
				if( pdN > tPt.dot( tN ) )
				{
					collisionNormal += tN ;
				}
				else
				{
					collisionNormal -= tN ;
				}
			}
			else if( h > 0.0f )
			{
				spot[i].point = gpl[n]->getPoint() ;
				spot[i].onGround = TRUE ;
				nPoints++ ;
			}
		}
	}
#ifdef VERBOSE
Util->print( 5, "nPoints (on ground)= %d\n", nPoints ) ;
#endif /* VERBOSE */

	onGround = nPoints ;

	// Now have 4 corner points, some of which may be in the air.  For
	// those that are on the ground, see if the line connecting it and
	// the adjacent corner goes under the ground.  If it does, where the
	// boundary would be projected (in the XY plane) on the line.
	nPoints = 4 ;
	GroundLine *gl = &line[0] ;
	SbVec3f tmp ;
	while( gl != NULL )
	{
		int n0 = gl->pIndex0 ;
		int n1 = gl->pIndex1 ;

		// Point is on ground and is a corner point.  See if line goes
		// underground.
		if( gl->isCorner0 && gPoly[n0] &&
			gPoly[n0]->intersect( spot[n0], spot[n1],
						spot[nPoints] ) )
		{
			spot[nPoints].onGround = TRUE ;
			onGround++ ;
			gl->splitLine( line[nLines], nPoints ) ;
			nPoints++ ;
			nLines++ ;
		}
		else if( gl->isCorner1 && gPoly[n1] &&
			gPoly[n1]->intersect( spot[n1], spot[n0],
						spot[nPoints] ) )
		{
			spot[nPoints].onGround = TRUE ;
			onGround++ ;
			gl->splitLine( line[nLines], nPoints ) ;
			nPoints++ ;
			nLines++ ;
			// If we didn't do anything else, we would check the
			// last half of this line again which is redundant.
			// So let's skip it.
			gl = gl->nextLine ;
		}
		gl = gl->nextLine ;
	}

	for( i = 0 ; i < 4 ; i++ )
	{
		if( gPoly[i] )
		{
			delete gPoly[i] ;
		}
	}
#ifdef VERBOSE
Util->print( 5, "nPoints = %d\n", nPoints ) ;
Util->print( 5, "nLines = %d\n", nLines ) ;
if( debugging )
{
	for( int ii = 0 ; ii < nPoints ; ii++ )
	{
		Util->print( 5, "spot[%d]: %8.4f %8.4f %8.4f "
			"(%8.4f %8.4f)\n", ii,
			spot[ii].point[0], spot[ii].point[1], spot[ii].point[2],
			spot[ii].s, spot[ii].t ) ;
	}

	dbCoord->point.setNum( nPoints ) ;
	for( i = 0 ; i < nPoints ; i++ )
	{
		dbCoord->point.set1Value( i, spot[i].point ) ;
	}

	dbLine->coordIndex.setNum( nLines ) ;
	gl = &line[0] ;
	dbLine->coordIndex.set1Value( 0, gl->pIndex0 ) ;
	i = 1 ;
	while( gl != NULL && gl->pIndex1 != line[0].pIndex0 )
	{
		dbLine->coordIndex.set1Value( i++, gl->pIndex1 ) ;
		gl = gl->nextLine ;
	}
	dbLine->coordIndex.set1Value( i++, line[0].pIndex0 ) ;
}
#endif /* VERBOSE */

	// If not on ground, return without calculating energy positions
	// since there is nothing to change the freefall result.
	if( onGround == 0 )
	{
		return( collisions > 0 ) ;
	}
	else if( onGround > 4 )
	{
		onGround = 4 ;
	}

	// Now have up to 12 points (either ground points or freefall points).
	// Cycle through all the valid 3 point combinations and calculate the
	// resulting energy difference between the resulting position and the
	// freefall position.
	float minEnergy = 1e30 ;
	int iPoint, jPoint, kPoint ;

	gl = &line[0] ;
	while( gl != NULL )
	{
		i = gl->pIndex0 ;
		GroundLine *jGl = gl ;
		while( jGl != NULL && jGl->pIndex1 != line[0].pIndex0 )
		{
			if( jGl == gl || jGl->side != gl->side )
			{
				int j = jGl->pIndex1 ;
				GroundLine *kGl = jGl->nextLine ;

				while( kGl != NULL &&
					kGl->pIndex1 != line[0].pIndex0 )
				{
					int k = kGl->pIndex1 ;
					if( ( spot[i].side & spot[j].side &
						spot[k].side ) == 0 && k != i )
					{
						SbBool flip ;
						float e = findEnergy( spot,
							i, j, k, nPoints,
							freeCenterPos, freeMat,
							deltaTime, flip ) ;
						if( e < minEnergy )
						{
							iPoint = i ;
							if( flip )
							{
								jPoint = k ;
								kPoint = j ;
							}
							else
							{
								jPoint = j ;
								kPoint = k ;
							}
							minEnergy = e ;
						}
					}
					kGl = kGl->nextLine ;
				}
			}
			jGl = jGl->nextLine ;
		}
		gl = gl->nextLine ;
	}
#ifdef VERBOSE
	if( minEnergy == 1e30 )
	{
		printf( "MIN ENERGY NOT CALCULATED PROPERLY!!!!\n" ) ;
	}
#endif /* VERBOSE */
	findPosition( spot[iPoint], spot[jPoint], spot[kPoint], freeCenterPos,
			freeMat ) ;
#ifdef VERBOSE
Util->print( 4, "iPoint = %d, jPoint = %d, kPoint = %d\n", iPoint, jPoint, kPoint ) ;
Util->printVal( 4, "min energy", minEnergy ) ;
Util->printVec( 9, "freeCenterPos", freeCenterPos ) ;
#endif /* VERBOSE */

	return( collisions > 0 ) ;
}



//------------------------------------------------------------------------------
// Given 3 GroundSpots, find the resulting position and orientation.  Allow
// the center position to move only along the Z axis.
//------------------------------------------------------------------------------
void
Tank::findPosition(
	const GroundSpot	&a,
	const GroundSpot	&b,
	const GroundSpot	&c,
	SbVec3f			&center,
	SbMatrix		&mat
	)
{
#ifdef VERBOSE
Util->print( 6, "------findPosition\n" ) ;
Util->printVec( 6, "a", a.point ) ;
Util->printVec( 6, "b", b.point ) ;
Util->printVec( 6, "c", c.point ) ;
Util->printVec( 6, "center", center ) ;
#endif /* VERBOSE */
	// Given the s and t values of 3 non-linear points, calculate the
	// location of any other point P(s,t) with the following eqns.
	//
	//            C.		P = B + x * ( A - B ) + y * ( C - B )
	//            /			U = A - B
	//           /.....P		V = C - B
	//          /    ;
	//         /    ;		x * U + y * V = P - B = W
	//      B./____;____. A		Solve resulting simultaneous eqns.
	//
	//	d = 1 / ( ux * vy - uy * vx )
	//	x = ( wx * vy - wy * vx ) * d
	//	y = ( ux * wy - uy * wy ) * d
	//
	SbVec2f u( a.s - b.s, a.t - b.t ) ;
	SbVec2f v( c.s - b.s, c.t - b.t ) ;

	float d = u[0] * v[1] - u[1] * v[0] ;
	assert( d != 0.0f ) ;
	d = 1.0f / d ;

	// Calculate the right vector which is found by finding the point
	// one positive s unit from point B.
	//
	//	W = P - B = [ (1,0)+B ] - B = (1,0)
	//
	float x = v[1] * d ;		// x = ( 1 * vy - 0 * vx ) * d 
	float y = -u[1] * d ;		// y = ( 0 * wy - uy * 1 ) * d

	float z = -x - y ; // P - B = [ B + x * ( A - B ) + y * ( C - B ) ] - B

	SbVec3f right( x * a.point[0] + y * c.point[0] + z * b.point[0],
			x * a.point[1] + y * c.point[1] + z * b.point[1],
			x * a.point[2] + y * c.point[2] + z * b.point[2] ) ;
	right.normalize() ;

	// Calculate the forward vector which is found by finding the
	// point one positive t unit from point B.
	//
	// W = P - B = [ (0,1)+B ] - B = (0,1)
	//
	x = -v[0] * d ;			// x = ( 0 * vy - 1 * vx ) * d
	y = u[0] * d ;			// y = ( ux * 1 - uy * 0 ) * d

	z = -x - y ;	// P - B = [ B + x * ( A - B ) + y * ( C - B ) ] - B

	SbVec3f forward( x * a.point[0] + y * c.point[0] + z * b.point[0],
			x * a.point[1] + y * c.point[1] + z * b.point[1],
			x * a.point[2] + y * c.point[2] + z * b.point[2] ) ;
	forward.normalize() ;

	SbVec3f &up = right.cross( forward ) ;
	up.normalize() ;

	SbVec3f	tankRight ;
	SbVec3f	tankAhead ;
	SbVec3f	tankUp ;

	switch( gpMode )
	{
		case BOTTOM_SIDE :
			tankRight = right ;
			tankAhead = forward ;
			tankUp = up ;
			break ;

		case TOP_SIDE :
			tankRight = -right ;
			tankAhead = forward ;
			tankUp = -up ;
			break ;

		case RIGHT_SIDE :
			tankRight = -up ;
			tankAhead = -right ;
			tankUp = forward ;
			break ;

		case LEFT_SIDE :
			tankRight = up ;
			tankAhead = right ;
			tankUp = forward ;
			break ;

		case FRONT_SIDE :
			tankRight = forward ;
			tankAhead = -up ;
			tankUp = -right ;
			break ;

		case BACK_SIDE :
			tankRight = forward ;
			tankAhead = up ;
			tankUp = right ;
			break ;
	}

	mat[0][0] = tankRight[0] ;
	mat[0][1] = tankRight[1] ;
	mat[0][2] = tankRight[2] ;
	mat[0][3] = 0.0f ;

	mat[1][0] = tankAhead[0] ;
	mat[1][1] = tankAhead[1] ;
	mat[1][2] = tankAhead[2] ;
	mat[1][3] = 0.0f ;

	mat[2][0] = tankUp[0] ;
	mat[2][1] = tankUp[1] ;
	mat[2][2] = tankUp[2] ;
	mat[2][3] = 0.0f ;

	mat[3][0] = 0.0f ;
	mat[3][1] = 0.0f ;
	mat[3][2] = 0.0f ;
	mat[3][3] = 1.0f ;

	// Now that we have the rotation determined, find the location of
	// the center.  We know centerX and centerY.  We also know the plane
	// of orientation and the distance the center should be from the
	// plane.
	//
	// Plane: x * upx + y * upy + z * upz = d = up dot A
	//
	// x = center[0]
	// y = center[1]
	//
	// x * upx + y * upy + z * upz = d + | centerOffset |
	//
	// z = ( d + |centerOffset| - x * up[0] - y * up[1] ) / up[2]
	//
	// This fails only if up[2] == 0.  In this case, don't modify center.
	//

	d = up.dot( a.point ) + gpH ;
#ifdef VERBOSE
Util->printVal( 6, "d", d ) ;
#endif /* VERBOSE */

	if( up[2] != 0.0f )
	{
		center[2] = ( d - center[0]*up[0] - center[1]*up[1] ) / up[2] ;
#ifdef VERBOSE
Util->printVec( 6, "center", center ) ;
#endif /* VERBOSE */
	}
}



//------------------------------------------------------------------------------
// Calculate the energy difference between the calculated position and the
// freefall position.
//------------------------------------------------------------------------------
float
Tank::findEnergy(
	const GroundSpot	*spot,
	int			i,
	int			j,
	int			k,
	int			nPoints,
	const SbVec3f		&freeCenterPos,
	const SbMatrix		&freeMat,
	float			deltaTime,
	SbBool			&flip
	)
{
	// Make sure the 3 points -- I, J, and K -- are ordered correctly.
	SbVec2f aVec( spot[i].s - spot[j].s, spot[i].t - spot[j].t ) ;
	SbVec2f bVec( spot[k].s - spot[j].s, spot[k].t - spot[j].t ) ;
	if( aVec[0]*bVec[1] - aVec[1]*bVec[0] < 0.0f )
	{
		int l = j ;
		j = k ;
		k = l ;
		flip = TRUE ;
	}
	else
	{
		flip = FALSE ;
	}

	const GroundSpot &a = spot[i] ;
	const GroundSpot &b = spot[j] ;
	const GroundSpot &c = spot[k] ;

	//
	// Make sure this is a valid combination of points such that none
	// of the other points are significantly above the plane formed by
	// the 3 basis points.
	//
	SbVec3f &ab = a.point - b.point ;
	SbVec3f &cb = c.point - b.point ;
	SbVec3f &n = ab.cross( cb ) ;
	n.normalize() ;
	float d = n.dot( a.point ) ;
	for( int l = 0 ; l < nPoints ; l++ )
	{
		// If point is above the plane, kick out with an absurdly
		// high energy value.
		if( n.dot( spot[l].point ) - d > 0.0001f )
		{
#ifdef VERBOSE
			Util->print( 5, "Rejected  %d,%d,%d\n", i, j, k ) ;
#endif /* VERBOSE */
			return( 1.e20 ) ;
		}
	}

	SbVec3f		newCenterPos( freeCenterPos ) ;
	SbMatrix	newMat ;
	findPosition( a, b, c, newCenterPos, newMat ) ;

	SbVec3f	newVel = newCenterPos - freeCenterPos ;
	newVel *= ( 1.0f / deltaTime ) ;
	float ke = 0.5f * newVel.dot( newVel ) ;

	float pe = Gravity * ( newCenterPos[2] - freeCenterPos[2] ) ;

	// Find the rotational energy.
	// SbRotation freeRot( freeMat ) ;
	// SbRotation newRot( newMat ) ;
	SbRotation freeRot ;
	SbRotation newRot ;
	Utilities::buildRotation( freeRot, freeMat ) ;
	Utilities::buildRotation( newRot, newMat ) ;

	freeRot.invert() ;

	SbRotation w = newRot * freeRot ;
	SbVec3f axis ;
	float radians ;
	w.getValue( axis, radians ) ;

	SbVec3f	right( freeMat[0][0], freeMat[0][1], freeMat[0][2] ) ;
	SbVec3f	ahead( freeMat[1][0], freeMat[1][1], freeMat[1][2] ) ;
	SbVec3f	up( freeMat[2][0], freeMat[2][1], freeMat[2][2] ) ;

	float wdx = axis.dot( right ) ;
	float wdy = axis.dot( ahead ) ;
	float wdz = axis.dot( up ) ;

	float I = Ix * wdx * wdx + Iy * wdy * wdy + Iz * wdz * wdz ;
	radians /= deltaTime ;
	float re = 0.5f * I * radians * radians ;

#ifdef VERBOSE
Util->print( 5, "Energy at %d,%d,%d is %f (k:%.3f p:%.3f r%.3f)\n", i, j, k, ke + pe + re, ke, pe, re ) ;
#endif /* VERBOSE */
	return( ke + pe + re ) ;
}



//------------------------------------------------------------------------------
// Reset the tank.
//------------------------------------------------------------------------------
void
Tank::reset()
{
	velocity.setValue( 0.0f, 0.0f, 0.0f ) ;
	omegaAxis.setValue( 0.0f, 0.0f, 1.0f ) ;
	omegaRad = 0.0f ;
}



//------------------------------------------------------------------------------
// Create a GroundPoly structure.
//------------------------------------------------------------------------------
GroundPoly::GroundPoly(
	int		nPts,
	const SbVec3f	*pts
	)
{
	int	i ;

	for( i = 0 ; i < nPts ; i++ )
	{
		point[i] = pts[i] ;
	}
	numPts = nPts ;

	SbVec3f a ;
	SbVec3f b ;
	if( nPts == 3 )
	{
		a.setValue( pts[1][0]-pts[0][0], pts[1][1]-pts[0][1],
			pts[1][2]-pts[0][2] ) ;
		b.setValue( pts[2][0]-pts[0][0], pts[2][1]-pts[0][1],
			pts[2][2]-pts[0][2] ) ;
	}
	else if( nPts == 4 )
	{
		a.setValue( pts[3][0]-pts[1][0], pts[3][1]-pts[1][1],
			pts[3][2]-pts[1][2] ) ;
		b.setValue( pts[2][0]-pts[0][0], pts[2][1]-pts[0][1],
			pts[2][2]-pts[0][2] ) ;
	}

	normal = b.cross( a ) ;
	normal.normalize() ;
}



//------------------------------------------------------------------------------
// Utility function that intersects 2 lines in 2D space.
//------------------------------------------------------------------------------
SbBool
intersectLines(
	const SbVec2f	&a,
	const SbVec2f	&b,
	const SbVec2f	&q,
	const SbVec2f	&r,
	float		&s,
	float		&t
	)
{
	float xqr = q[0] - r[0] ;
	float yqr = q[1] - r[1] ;
	float xba = b[0] - a[0] ;
	float yba = b[1] - a[1] ;
	float xqa = q[0] - a[0] ;
	float yqa = q[1] - a[1] ;

	float D = xba * yqr - yba * xqr ;

	if( D == 0.0f ) return( FALSE ) ;

	s = xba * yqa - xqa * yba ;
	t = xqa * yqr - xqr * yqa ;

	if( D > 0.0f )
	{
		if( s < 0.0f || t < 0.0f || s > D || t > D )
		{
			return( FALSE ) ;
		}
	}
	else if( s > 0.0f || t > 0.0f || s < D || t < D )
	{
		return( FALSE ) ;
	}

	s /= D ;
	t /= D ;
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Intersect a line with the triangle.
//------------------------------------------------------------------------------
SbBool
GroundPoly::intersect(
	const GroundSpot	&A,
	const GroundSpot	&B,
	GroundSpot		&pt
	)
{
	static const float toler = 1.e-4f ;
	SbVec3f BA( B.point[0] - A.point[0], B.point[1] - A.point[1],
			B.point[2] - A.point[2] ) ;
	BA.normalize() ;

	if( BA.dot( normal ) > -0.001f )
	{
		return( FALSE ) ;
	}

	SbVec2f v[4] ;
	for( int i = 0 ; i < numPts ; i++ )
	{
		v[i].setValue( point[i][0], point[i][1] ) ;
	}

	SbVec2f a( A.point[0], A.point[1] ) ;
	SbVec2f b( B.point[0], B.point[1] ) ;

	float	s, t ;
	int	j ;

	for( i = 0 ; i < numPts ; i++ )
	{
		j = ( i + 1 ) % numPts ;
		if( intersectLines( a, b, v[i], v[j], s, t ) )
		{
			SbVec3f &l = point[j] - point[i] ;
			pt.point.setValue(
				point[i][0] + s * l[0], point[i][1] + s * l[1],
				point[i][2] + s * l[2] ) ;
			float origZ = A.point[2] +
				t * ( B.point[2] - A.point[2] ) ;
			if( origZ >= pt.point[2] ||
				pt.point.equals( A.point, toler ) ||
				pt.point.equals( B.point, toler ) )
			{
				return( FALSE ) ;
			}

			// Set s, t, and side values.
			SbVec2f r( B.s - A.s, B.t - A.t ) ;
			r.normalize() ;
			SbVec3f &pma = pt.point - A.point ;
			SbVec3f &pmb = pt.point - B.point ;
			float pmaL = pma.length() ;
			float pmbL = pmb.length() ;
			pt.s = 0.5f * ( A.s + pmaL * r[0] + B.s - pmbL * r[0] );
			pt.t = 0.5f * ( A.t + pmaL * r[1] + B.t - pmbL * r[1] );
			pt.side = A.side & B.side ;

			return( TRUE ) ;
		}
	}
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Smooth out the ride by blending the rotation matrix.
//------------------------------------------------------------------------------
void
Tank::smoothRide(
	SbMatrix &newMat
	)
{
	SbMatrix oldMat ;
	static const float fudge[5] = { 0.0f, 0.025f, 0.1f, 0.225f, 0.4f } ;
	rotation.getValue().getValue( oldMat ) ;

	SbVec3f oldUp( oldMat[2][0], oldMat[2][1], oldMat[2][2] ) ;
	SbVec3f newUp( newMat[2][0], newMat[2][1], newMat[2][2] ) ;

	SbVec3f ahead( newMat[1][0], newMat[1][1], newMat[1][2] ) ;
#ifdef VERBOSE
Util->print( 2, "onGround = %d\n", onGround ) ;
#endif /* VERBOSE */
	float a = fudge[onGround] ;
	float b = 1.0f - a ;
	SbVec3f up( a*oldUp[0]+b*newUp[0], a*oldUp[1]+b*newUp[1],
		a*oldUp[2]+b*newUp[2] ) ;
	up.normalize() ;
	SbVec3f &right = ahead.cross( up ) ;
	ahead = up.cross( right ) ;

	newMat[0][0] = right[0] ;
	newMat[0][1] = right[1] ;
	newMat[0][2] = right[2] ;

	newMat[1][0] = ahead[0] ;
	newMat[1][1] = ahead[1] ;
	newMat[1][2] = ahead[2] ;

	newMat[2][0] = up[0] ;
	newMat[2][1] = up[1] ;
	newMat[2][2] = up[2] ;
}



//------------------------------------------------------------------------------
// Find the side that projects the most area onto the XY plane.
//------------------------------------------------------------------------------
void
Tank::findBottomSide(
	const SbMatrix &mat
	)
{
	SbVec3f right( mat[0][0], mat[0][1], mat[0][2] ) ;
	SbVec3f ahead( mat[1][0], mat[1][1], mat[1][2] ) ;
	SbVec3f up( mat[2][0], mat[2][1], mat[2][2] ) ;

	float a[3] ;

	a[0] = ahead[0] * up[1] - ahead[1] * up[0] ;
	a[1] = up[0] * right[1] - up[1] * right[0] ;
	a[2] = right[0] * ahead[1] - right[1] * ahead[0] ;

	float maxArea = 0.0f ;

	for( int i = 0 ; i < 3 ; i++ )
	{
		if( a[i] > maxArea )
		{
			gpMode = 2 * i ;
			maxArea = a[i] ;
		}
		else if( -a[i] > maxArea )
		{
			gpMode = 2 * i + 1 ;
			maxArea = -a[i] ;
		}
	}

	switch( gpMode )
	{
		case BOTTOM_SIDE :
			gpW = width ;
			gpL = length ;
			gpH = height ;
			gpRt = right ;
			gpAh = ahead ;
			gpUp = up ;
			break ;

		case TOP_SIDE :
			gpW = width ;
			gpL = length ;
			gpH = height ;
			gpRt = -right ;
			gpAh = ahead ;
			gpUp = -up ;
			break ;

		case LEFT_SIDE :
			gpW = length ;
			gpL = height ;
			gpH = width ;
			gpRt = ahead ;
			gpAh = up ;
			gpUp = right ;
			break ;

		case RIGHT_SIDE :
			gpW = length ;
			gpL = height ;
			gpH = width ;
			gpRt = -ahead ;
			gpAh = up ;
			gpUp = -right ;
			break ;

		case BACK_SIDE :
			gpW = height ;
			gpL = width ;
			gpH = length ;
			gpRt = up ;
			gpAh = right ;
			gpUp = ahead ;
			break ;

		case FRONT_SIDE :
			gpW = height ;
			gpL = width ;
			gpH = length ;
			gpRt = -up ;
			gpAh = right ;
			gpUp = -ahead ;
			break ;
	}
}



//------------------------------------------------------------------------------
// Calculate the energy difference between the calculated position and the
// freefall position.
//------------------------------------------------------------------------------
float
Tank::calculateEnergy(
	const SbVec3f		&realCenterPos,
	const SbRotation	&realRot,
	const SbVec3f		&freeCenterPos,
	const SbRotation	&freeRot,
	float			deltaTime
	) const
{
	SbVec3f	newVel = realCenterPos - freeCenterPos ;
	newVel *= ( 1.0f / deltaTime ) ;
	float ke = 0.5f * newVel.dot( newVel ) ;

	float pe = Gravity * ( realCenterPos[2] - freeCenterPos[2] ) ;

	SbRotation w = realRot * freeRot.inverse() ;
	SbVec3f axis ;
	float radians ;
	w.getValue( axis, radians ) ;

	SbMatrix freeMat ;
	freeRot.getValue( freeMat ) ;

	SbVec3f	right( freeMat[0][0], freeMat[0][1], freeMat[0][2] ) ;
	SbVec3f	ahead( freeMat[1][0], freeMat[1][1], freeMat[1][2] ) ;
	SbVec3f	up( freeMat[2][0], freeMat[2][1], freeMat[2][2] ) ;

	float wdx = axis.dot( right ) ;
	float wdy = axis.dot( ahead ) ;
	float wdz = axis.dot( up ) ;

	float I = Ix * wdx * wdx + Iy * wdy * wdy + Iz * wdz * wdz ;
	radians /= deltaTime ;
	float re = 0.5f * I * radians * radians ;

	return( ke + pe + re ) ;
}



//------------------------------------------------------------------------------
// GroundLine routine to split the line and initialize the new one.
//------------------------------------------------------------------------------
void
GroundLine::splitLine(
	GroundLine	&newLine,
	int		newPtIndex
	)
{
	newLine.pIndex0 = newPtIndex ;
	newLine.pIndex1 = pIndex1 ;
	newLine.isCorner0 = FALSE ;
	newLine.isCorner1 = isCorner1 ;
	newLine.side = side ;
	newLine.nextLine = nextLine ;

	pIndex1 = newPtIndex ;
	isCorner1 = FALSE ;
	nextLine = &newLine ;
}



//------------------------------------------------------------------------------
// Show vehicle-specific help text.
//------------------------------------------------------------------------------
void
Tank::showHelp(
	SoMFString	*str,
	int		&n
	)
{
	str->set1Value( n++, "Right mouse - Toggle between Main and "
				"Anti-aircraft views" ) ;
	str->set1Value( n++, "" ) ;
	str->set1Value( n++, "MAIN VIEW" ) ;
	str->set1Value( n++, "Left mouse - fire next missile" ) ;
	str->set1Value( n++, "Middle mouse - change missile type to be "
				"loaded" ) ;
	str->set1Value( n++, "Mouse - control speed and direction" ) ;
	str->set1Value( n++, "" ) ;
	str->set1Value( n++, "AA VIEW" ) ;
	str->set1Value( n++, "Left mouse - fire cannon" ) ;
	str->set1Value( n++, "Mouse - aim turret" ) ;
}
