//------------------------------------------------------------------------------
// TransformSwitch node.
//------------------------------------------------------------------------------

#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoPickAction.h>
#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoBBoxModelMatrixElement.h>
#include "TransformSwitch.h"


SO_NODE_SOURCE( TransformSwitch ) ;



//------------------------------------------------------------------------------
// Initialize the class.
//------------------------------------------------------------------------------
void
TransformSwitch::initClass()
{
	SO_NODE_INIT_CLASS( TransformSwitch, SoSwitch, "Switch" ) ;
	SO_ENABLE( SoGetMatrixAction, SoModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoModelMatrixElement ) ;
	SO_ENABLE( SoGetBoundingBoxAction, SoBBoxModelMatrixElement ) ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
TransformSwitch::TransformSwitch()
{
	SO_NODE_CONSTRUCTOR( TransformSwitch ) ;
	SO_NODE_ADD_FIELD( position, ( 0.0f, 0.0f, 0.0f ) ) ;
	SO_NODE_ADD_FIELD( rotation, ( SbVec3f( 0.0f, 0.0f, 1.0f ), 0.0f ) ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
TransformSwitch::~TransformSwitch()
{
}



//------------------------------------------------------------------------------
// Main guts of the action handlers.
//------------------------------------------------------------------------------
void
TransformSwitch::actionGuts(
	SoState		*state,
	SoAction	*action
	)
{
	if( !position.isIgnored() && !position.isDefault() )
	{
		SoModelMatrixElement::translateBy( state, this,
						position.getValue() ) ;
	}

	if( !rotation.isIgnored() && !rotation.isDefault() )
	{
		SoModelMatrixElement::rotateBy( state, this,
						rotation.getValue() ) ;
	}

	SoSwitch::doAction( action ) ;
}



//------------------------------------------------------------------------------
// Generic action handler.
//------------------------------------------------------------------------------
void
TransformSwitch::doAction(
	SoAction	*action
	)
{
	SoState *state = action->getState() ;

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SbMatrix saveMat = SoModelMatrixElement::pushMatrix( state ) ;

	actionGuts( state, action ) ;

	SoModelMatrixElement::popMatrix( state, saveMat ) ;
#else
	SbMatrix saveMat = SoModelMatrixElement::get( state ) ;

	actionGuts( state, action ) ;

	SoModelMatrixElement::set( state, this, saveMat ) ;
#endif
}



//------------------------------------------------------------------------------
// GLRenderAction handler.
//------------------------------------------------------------------------------
void
TransformSwitch::GLRender(
	SoGLRenderAction *action
	)
{
	TransformSwitch::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetBoundBoxAction handler.
//------------------------------------------------------------------------------
void
TransformSwitch::getBoundingBox(
	SoGetBoundingBoxAction *action
	)
{
	int		numIndices ;
	const int	*indices ;

	if( action->getPathCode( numIndices, indices ) == SoAction::OFF_PATH )
	{
		return ;
	}

#if defined(SO_VERSION) && SO_VERSION > 1 && SO_VERSION_REVISION > 0
	SoState *state = action->getState() ;

	SbMatrix matrix, localMatrix ;
	SoBBoxModelMatrixElement::pushMatrix( state, matrix, localMatrix ) ;

	TransformSwitch::actionGuts( state, action ) ;

	SoBBoxModelMatrixElement::popMatrix( state, matrix, localMatrix ) ;
#else
	TransformSwitch::doAction( action ) ;
#endif
}



//------------------------------------------------------------------------------
// CallbackAction handler.
//------------------------------------------------------------------------------
void
TransformSwitch::callback(
	SoCallbackAction *action
	)
{
	TransformSwitch::doAction( action ) ;
}



//------------------------------------------------------------------------------
// PickAction handler.
//------------------------------------------------------------------------------
void
TransformSwitch::pick(
	SoPickAction *action
	)
{
	TransformSwitch::doAction( action ) ;
}



//------------------------------------------------------------------------------
// GetMatrixAction handler.
//------------------------------------------------------------------------------
void
TransformSwitch::getMatrix(
	SoGetMatrixAction *action
	)
{
	int		numIndices ;
	const int	*indices ;

	if( action->getPathCode( numIndices, indices ) != SoAction::IN_PATH )
	{
		return ;
	}

	SbMatrix &ctm = action->getMatrix() ;
	SbMatrix &inv = action->getInverse() ;
	SbMatrix m ;

	if( !position.isIgnored() && !position.isDefault() )
	{
		const SbVec3f &v = position.getValue() ;
		m.setTranslate( v ) ;
		ctm.multLeft( m ) ;
		m.setTranslate( -v ) ;
		inv.multRight( m ) ;
	}

	if( !rotation.isIgnored() && !rotation.isDefault() )
	{
		SbRotation r = rotation.getValue() ;
		r.getValue( m ) ;
		ctm.multLeft( m ) ;
		r.invert() ;
		r.getValue( m ) ;
		inv.multRight( m ) ;
	}
}



