//------------------------------------------------------------------------------
// Collection of utility routines.
//------------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <Inventor/SoDB.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoSeparator.h>

#include "Utilities.h"

SbString	Utilities::searchDirectory = "" ;

//------------------------------------------------------------------------------
// Initialize by setting the debug level from an environment variable.
//------------------------------------------------------------------------------
Utilities::Utilities(
	const char *envVar
	)
{
	char *value = getenv( envVar ) ;

	if( value != NULL )
	{
		debugLevel = atoi( value ) ;
	}
	else
	{
		debugLevel = 0 ;
	}
}



//------------------------------------------------------------------------------
// Print out a label and a 3D vector.
//------------------------------------------------------------------------------
void
Utilities::printVec(
	int		level,
	const char	*s,
	const SbVec3f	&v
	)
{
	print( level, "%-20s: %9.6f %9.6f %9.6f\n", s, v[0], v[1], v[2] ) ;
}



//------------------------------------------------------------------------------
// Print out a label and a 2D vector.
//------------------------------------------------------------------------------
void
Utilities::printVec(
	int		level,
	const char	*s,
	const SbVec2f	&v
	)
{
	print( level, "%-20s: %9.6f %9.6f\n", s, v[0], v[1] ) ;
}



//------------------------------------------------------------------------------
// Print out a label and a float.
//------------------------------------------------------------------------------
void
Utilities::printVal(
	int		level,
	const char	*s,
	float		f
	)
{
	print( level, "%-20s: %9.6f\n", s, f ) ;
}



//------------------------------------------------------------------------------
// Print out a label and a int32_t.
//------------------------------------------------------------------------------
void
Utilities::printVal(
	int		level,
	const char	*s,
	int32_t		l
	)
{
	print( level, "%-20s: %ld\n", s, l ) ;
}



//------------------------------------------------------------------------------
// Debug print statement.
//------------------------------------------------------------------------------
void
Utilities::print(
	int		level,
	const char	*fmt
	...
	)
{
	if( level <= debugLevel )
	{
		va_list	arg ;
		va_start( arg, fmt ) ;
		vprintf( fmt, arg ) ;
		va_end( arg ) ;
	}
}



//------------------------------------------------------------------------------
// Print out a label and a rotation axis and angle.
//------------------------------------------------------------------------------
void
Utilities::printRot(
	int			level,
	const char		*s,
	const SbRotation	&r
	)
{
	SbVec3f	axis ;
	float	radians ;

	r.getValue( axis, radians ) ;
	print( level, "%-20s: %9.6f %9.6f %9.6f, %9.6f\n", s, axis[0],
		axis[1], axis[2], radians ) ;
}



//------------------------------------------------------------------------------
// Debug utility function.
//------------------------------------------------------------------------------
void
Utilities::dumpIt(
	const char	*fileName,
	SoNode		*node
	)
{
	SoOutput out ;
	out.openFile( fileName ) ;
	SoWriteAction wa( &out ) ;
	wa.apply( node ) ;
	out.closeFile() ;
}



//------------------------------------------------------------------------------
// Get the forward vector of a rotation matrix.
//------------------------------------------------------------------------------
void
Utilities::setForwardVector(
	const SbRotation	&r,
	SbVec3f			&forward
	)
{
	float q0, q1, q2, q3 ;
	r.getValue( q0, q1, q2, q3 ) ;

	forward.setValue(
		2.0f * ( q0 * q1 - q2 * q3 ),
		1.0f - 2.0f * ( q2 * q2 + q0 * q0 ),
		2.0f * ( q1 * q2 + q0 * q3 )
		) ;
}



//------------------------------------------------------------------------------
// Get the right vector of a rotation matrix.
//------------------------------------------------------------------------------
void
Utilities::setRightVector(
	const SbRotation	&r,
	SbVec3f			&right
	)
{
	float q0, q1, q2, q3 ;
	r.getValue( q0, q1, q2, q3 ) ;

	right.setValue(
		1.0f - 2.0f * ( q1 * q1 + q2 * q2 ),
		2.0f * ( q0 * q1 + q2 * q3 ),
		2.0f * ( q2 * q0 - q1 * q3 )
		) ;
}



//------------------------------------------------------------------------------
// Get the up vector of a rotation matrix.
//------------------------------------------------------------------------------
void
Utilities::setUpVector(
	const SbRotation	&r,
	SbVec3f			&up
	)
{
	float q0, q1, q2, q3 ;
	r.getValue( q0, q1, q2, q3 ) ;

	up.setValue(
		2.0f * ( q2 * q0 + q1 * q3 ),
		2.0f * ( q1 * q2 - q0 * q3 ),
		1.0f - 2.0f * ( q1 * q1 + q0 * q0 )
		) ;
}



//------------------------------------------------------------------------------
// Locate a named node and check its type.
//------------------------------------------------------------------------------
SoNode *
Utilities::getNamedNode(
	const char	*name,
	SoType		type
	)
{
	SoNode *node = SoNode::getByName( name ) ;
	if( node != NULL && node->isOfType( type ) )
	{
		return( node ) ;
	}

	return( NULL ) ;
}



//------------------------------------------------------------------------------
// Read in a scene graph, aborting on failure if desired.
//------------------------------------------------------------------------------
SoSeparator *
Utilities::readAll(
	const char	*filename,
	SbBool		quit
	)
{
	SoInput	*in = createInput() ;

	if( !in->openFile( filename, TRUE ) )
	{
		if( quit )
		{
			SoReadError::post( in, "Can not open %s", filename ) ;
			exit( 1 ) ;
		}

		delete in ;
		return( NULL ) ;
	}

	SoSeparator *root = SoDB::readAll( in ) ;

	if( root == NULL && quit )
	{
		SoReadError::post( in, "Can not read %s", filename ) ;
		exit( 1 ) ;
	}

	delete in ;
	return( root ) ;
}



//------------------------------------------------------------------------------
// Create an SoInput object and add a search directory to it.
//------------------------------------------------------------------------------
SoInput *
Utilities::createInput()
{
	SoInput *in = new SoInput ;

	if( searchDirectory.getString() != NULL &&
		searchDirectory.getLength() > 0 )
	{
		in->addDirectoryFirst( searchDirectory.getString() ) ;
	}

	return( in ) ;
}


 
//------------------------------------------------------------------------------
// Build an SbRotation from an SbMatrix (doesn't spit out debug errors like
// SbRotation method does due to mismatches)>
//------------------------------------------------------------------------------
void
Utilities::buildRotation(
	SbRotation	&r,
	const SbMatrix	&m
	)
{
	int i, j, k ;
	float quat[4] ;

	// First, find largest diagonal in matrix:
	if( m[0][0] > m[1][1] )
	{
		if( m[0][0] > m[2][2] )
		{
			i = 0 ;
		}
		else
		{
			i = 2 ;
		}
	}
	else
	{
		if( m[1][1] > m[2][2] )
		{
			i = 1 ;
		}
		else
		{
			i = 2 ;
		}
	}
	if( m[0][0] + m[1][1] + m[2][2] > m[i][i] )
	{
		// Compute w first:
		quat[3] = sqrt( m[0][0]+m[1][1]+m[2][2]+m[3][3] ) * 0.5 ;

		// And compute other values:
		quat[0] = (m[1][2]-m[2][1])/(4*quat[3]);
		quat[1] = (m[2][0]-m[0][2])/(4*quat[3]);
		quat[2] = (m[0][1]-m[1][0])/(4*quat[3]);
	}
	else
	{
		// Compute x, y, or z first:
		j = (i+1)%3; k = (i+2)%3;

		// Compute first value:
		quat[i] = sqrt(m[i][i]-m[j][j]-m[k][k]+m[3][3]) * 0.5 ;

		// And the others:
		quat[j] = (m[i][j]+m[j][i])/(4*quat[i]);
		quat[k] = (m[i][k]+m[k][i])/(4*quat[i]);

		quat[3] = (m[j][k]-m[k][j])/(4*quat[i]);
	}

	r.setValue( quat ) ;
}
