//------------------------------------------------------------------------------
// Vehicle class to test FollowTerrainAction class.
//------------------------------------------------------------------------------

#include <Inventor/SbViewportRegion.h>
#include <Inventor/SoDB.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/fields/SoMFString.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoTransform.h>

#include "Cannon.h"
#include "Missile.h"
#include "CruiseMissile.h"
#include "GuidedMissile.h"
#include "FollowTerrainAction.h"
#include "CollisionAction.h"
#include "LosAction.h"
#include "Utilities.h"
#include "Vehicle.h"


static SbBool debugging = FALSE ;
#ifdef VERBOSE
#define PRINT if( debugging ) printf
#define PRINT_VEC(s,v) \
	PRINT("%-20s: %7.4f %7.4f %7.4f\n",(s),(v)[0],(v)[1],(v)[2])
#else
#define PRINT //
#define PRINT_VEC //
#endif /* VERBOSE */


SO_NODE_SOURCE( Vehicle ) ;



//------------------------------------------------------------------------------
// Initialize the Vehicle class.
//------------------------------------------------------------------------------
void
Vehicle::initClass()
{
	if( Obstacle::getClassTypeId() == SoType::badType() )
	{
		Obstacle::initClass() ;
	}
	SO_NODE_INIT_CLASS( Vehicle, Obstacle, "Obstacle" ) ;

	if( getenv( "DEBUG_VEHICLE" ) != NULL ) debugging = TRUE ;
}



//------------------------------------------------------------------------------
// Common constructor code.
//------------------------------------------------------------------------------
void
Vehicle::createIt()
{
	SO_NODE_CONSTRUCTOR( Vehicle ) ;
	SO_NODE_ADD_FIELD( conView, ( 0.0f, 0.0f, 1.0f, 0.0f ) ) ;
	SO_NODE_ADD_FIELD( radarRotation, ( 0.0f, 0.0f, 1.0f, 0.0f ) ) ;

	fta = new FollowTerrainAction ;
	ca = new CollisionAction ;
	ca->setSelf( this ) ;

	onGround = 1 ;

	leftEngine = 0.0f ;
	rightEngine = 0.0f ;

	setParams() ;

	resetConView() ;

	mainViewpoint.setValue( 0.0f, 0.11f, 0.155f ) ;
	auxViewpoint.setValue( 0.0f, 0.0f, 0.2f ) ;
	launchPoint.setValue( 0.0f, 0.1f, 0.0945f ) ;
	velocity.setValue( 0.0f, 0.0f, 0.0f ) ;

	fpGraphics = new SoSwitch ;
	fpGraphics->ref() ;
	fpGraphics->whichChild = -1 ;

	processGeometry() ;

	points = 0 ;
}



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
Vehicle::Vehicle()
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Alternate constructor.
//------------------------------------------------------------------------------
Vehicle::Vehicle(
	const char *s
	) :
	Obstacle( s )
{
	createIt() ;
}



//------------------------------------------------------------------------------
// Calculate and center the bounding box.
//------------------------------------------------------------------------------
void
Vehicle::boundIt()
{
	Obstacle::boundItBasics() ;
	float x, y ;
	boundingBox.getSize( x, y, heightClearance ) ;
	ca->setBoundingBox( boundingBox ) ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
Vehicle::~Vehicle()
{
	delete fta ;
	delete ca ;
	fpGraphics->unref() ;
}



//------------------------------------------------------------------------------
// Move according to input values.
//------------------------------------------------------------------------------
float
Vehicle::move(
	float	omega,
	float	accel,
	float	deltaTime,
	SoNode	*scene
	)
{
	int	i ;
	SbBool	terrainCollision ;
	float	ch = centerOffset[2] ;

	// Quickie model for now.
	speed = accel * maxSpeed ;
	float forward = speed * deltaTime ;

	float tAngZ = -maxTurnRate * deltaTime * omega ;
	SbRotation turnZ( SbVec3f( 0.0f, 0.0f, 1.0f ), tAngZ ) ;

	SbMatrix mat ;
	SbRotation rot = rotation.getValue() ;
	rot = turnZ * rot ;
	mat.setRotate( rot ) ;

	// -Z vector
	SbVec3f ahead( mat[1][0], mat[1][1], mat[1][2] ) ;
	// +Y vector
	SbVec3f up( mat[2][0], mat[2][1], mat[2][2] ) ;

	const SbVec3f &p = position.getValue() ;
	SbVec3f pos( p[0]+forward*ahead[0]+ch*up[0],
			p[1]+forward*ahead[1]+ch*up[1],
			p[2]+forward*ahead[2]+ch*up[2] ) ;

	SbVec3f	collisionNormal ;
	terrainCollision = followTerrain( scene, pos, mat,
					collisionNormal, deltaTime ) ;

	SbRotation newOrientation( mat ) ;

	up.setValue( mat[2][0], mat[2][1], mat[2][2] ) ;
	// Get current position
	SbVec3f newPos( pos[0]-ch*up[0], pos[1]-ch*up[1], pos[2]-ch*up[2] ) ;
	ca->setLocation( newPos, newOrientation ) ;
	ca->apply( scene ) ;

	const CollisionList &cList = ca->getCollisionList() ;
	if( cList.getLength() > 0 || terrainCollision )
	{

		// Find vector of travel.
		SbVec3f &travel = newPos - position.getValue() ;

		// Get world normal(s) direction off of obstacle(s).
		for( i = 0 ; i < cList.getLength() ; i++ )
		{
			collisionNormal += cList[i]->getWorldNormal() ;
		}
		collisionNormal.normalize() ;
		float tdn = travel.dot( collisionNormal ) ;
		if( tdn < 0.0f ) tdn = -tdn ;

		SbVec3f correction ;
		const float fudge = 1.3f ;
		correction = ( fudge * tdn ) * collisionNormal ;
		// Get corrected position (no ground effect).
		SbVec3f modPos = pos + correction ;
		mat.setRotate( rotation.getValue() ) ;
		// Get ground effect on correction -> modPos.
		followTerrain( scene, modPos, mat, collisionNormal,
				deltaTime ) ;

		// Calculate the effective normal (normal plus ground effect)
		SbVec3f &effNormal = modPos - pos ;
		effNormal.normalize() ;
		tdn = travel.dot( effNormal ) ;
		if( tdn < 0.0f ) tdn = -tdn ;

		float tdt = travel.dot( travel ) ;
		const float limitFactor = 0.075f ;
		if( tdn * tdn > limitFactor * tdt )
		{
			tdn = tdt / tdn ;
		}
		else if( tdn > 0.0f )
		{
			float a = ( tdn * tdn / tdt ) / limitFactor ;
			tdn = ( a * tdt / tdn ) + ( 1.0f - a ) * tdn ;
		}
		correction = ( fudge * tdn ) * effNormal ;
		pos += correction ;
		
		if( followTerrain( scene, pos, mat, collisionNormal,
				deltaTime ) )
		{
			printf( "adjusting position bumps height\n" ) ;
			return( 0.0f ) ;	
		}
		newOrientation.setValue( mat ) ;
	}

	position.setValue( pos[0]-ch*up[0], pos[1]-ch*up[1],
			   pos[2]-ch*up[2] ) ;
	rotation.setValue( newOrientation ) ;
	radarRotation.setValue( newOrientation ) ;

	return( 0.0f ) ;
}



//------------------------------------------------------------------------------
// Locate point beneath specified point and adjust up vector.
//------------------------------------------------------------------------------
SbBool
Vehicle::followTerrain(
	SoNode		*terrainNode,
	SbVec3f		&pos,
	SbMatrix	&mat,
	SbVec3f		&collisionNormal,
	float
	)
{
	int	collisions = 0 ;

	collisionNormal.setValue( 0.0f, 0.0f, 0.0f ) ;

	// -Z vector
	SbVec3f ahead( mat[1][0], mat[1][1], mat[1][2] ) ;
	// +Y vector
	SbVec3f up( mat[2][0], mat[2][1], mat[2][2] ) ;
	// +X vector
	SbVec3f right( mat[0][0], mat[0][1], mat[0][2] ) ;

	fta->setSearchPoint( pos ) ;
	fta->apply( terrainNode ) ;

	const GroundPointList &gpl = fta->getGroundPointList() ;

	int nHits = gpl.getLength() ;
	if( nHits > 0 )
	{
		int n = 0 ;
		while( ( n < nHits - 1 ) &&
			gpl[n]->getHeight() > 0.4f * heightClearance )
		{
			if( gpl[n]->getHeight() < heightClearance )
			{
				collisions++ ;
			}
			n++ ;
		}

		if( gpl[n]->getHeight() > 0.0f &&
			up.dot( gpl[n]->getNormal() ) < 0.7f )
		{
			collisions++ ;
		}

		if( collisions > 0 )
		{
			collisionNormal += position.getValue() - pos +
						centerOffset[2] * up ;
		}

		pos = gpl[n]->getPoint() ;
		// Average the current up with the new one to dampen out
		// the movement.
		up += gpl[n]->getNormal() ;
		up.normalize() ;

		right = ahead.cross( up ) ;
		ahead = up.cross( right ) ;
				
		mat[0][0] = right[0] ;
		mat[0][1] = right[1] ;
		mat[0][2] = right[2] ;

		mat[1][0] = ahead[0] ;
		mat[1][1] = ahead[1] ;
		mat[1][2] = ahead[2] ;

		mat[2][0] = up[0] ;
		mat[2][1] = up[1] ;
		mat[2][2] = up[2] ;
	}
	return( collisions > 0 ) ;
}



//------------------------------------------------------------------------------
// CollisionAction handler.
//------------------------------------------------------------------------------
void
Vehicle::collide(
	CollisionAction *action
	)
{
	if( action->isSelf( this ) )
	{
		return ;
	}

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

 	action->checkCollision( boundingBox, rotation.getValue(),
		position.getValue(), collisionRadius ) ;
}



//------------------------------------------------------------------------------
// LosAction handler.
//------------------------------------------------------------------------------
void
Vehicle::obstruct(
	LosAction *action
	)
{
	if( action->isTarget( this ) || action->isSelf( this ) )
	{
		return ;
	}

	if( !hasBoundingBox )
	{
		boundIt() ;
	}

	//
	// Check if line intersects bounding box.
	//
	action->checkBoundingBox( boundingBox, rotation.getValue(),
					position.getValue(), collisionRadius ) ;
}



//------------------------------------------------------------------------------
// Move the conning tower view.
//------------------------------------------------------------------------------
float
Vehicle::moveConView(
	float	mx,
	float	my,
	float	dt,
	SoNode	*scene
	)
{
	static const float	twoPi = 2.0f * (float)M_PI ;
	static const float	halfPi = 0.5f * (float)M_PI ;
	static const SbRotation	c( SbVec3f( 1.0f, 0.0f, 0.0f ), halfPi ) ;
	const float	azimRate = twoPi / 2.0f ;
	const float	inclRate = twoPi / 2.0f ;

	mx *= ( mx < 0.0f ) ? -mx : mx ;
	my *= ( my < 0.0f ) ? -my : my ;
	azim -= azimRate * mx * dt ;
	incl += inclRate * my * dt ;

	if( azim > twoPi )
	{
		azim -= twoPi ;
	}
	else if( azim < 0.0f )
	{
		azim += twoPi ;
	}

	if( incl > halfPi )
	{
		incl = halfPi ;
	}
	else if( incl < 0.0f )
	{
		incl = 0.0f ;
	}

	SbRotation az( SbVec3f( 0.0f, 1.0f, 0.0f ), azim ) ;
	SbRotation in( SbVec3f( 1.0f, 0.0f, 0.0f ), incl ) ;

	conView.setValue( in * az * c ) ;

	float energy = move( 0.0f, 0.0f, dt, scene ) ;

	radarRotation.setValue( 
		c.inverse() * conView.getValue() * rotation.getValue() ) ;

	return( energy ) ;
}



//------------------------------------------------------------------------------
// Return whether or not the vehicle has a con view.
//------------------------------------------------------------------------------
SbBool
Vehicle::hasConView() const
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Reset the con view.
//------------------------------------------------------------------------------
void
Vehicle::resetConView()
{
	azim = 0.0f ;
	incl = 0.0f ;
}



//------------------------------------------------------------------------------
// Return the number of weapon types contained in each pod.
//------------------------------------------------------------------------------
int
Vehicle::getNumWeapons(
	WeaponControl::WeaponPod pod
	)
{
	return( numWeapons[pod] ) ;
}



//------------------------------------------------------------------------------
// Return the type of weapon N in the given pod.
//------------------------------------------------------------------------------
Missile::MissileType
Vehicle::getWeaponType(
	int				n,
	WeaponControl::WeaponPod	pod
	)
{
	if( n < 0 || n >= weaponList[pod].getLength() )
	{
		return( Missile::BAD_TYPE ) ;
	}
	else
	{
		return( weaponList[pod][n] ) ;
	}
}



//------------------------------------------------------------------------------
// Set up vehicle parameters.
//------------------------------------------------------------------------------
void
Vehicle::setParams()
{
	maxSpeed = 2.0f ;
	maxTurnRate = 1.0f ;
	maxAcceleration = 2.0f ;
	maxDamage = 15.0f ;
	numBays = WeaponControl::MAX_BAYS ;

	numWeapons[WeaponControl::MAIN_POD] = 3 ;
	weaponList[WeaponControl::MAIN_POD].append( Missile::MISSILE );
	weaponList[WeaponControl::MAIN_POD].append( Missile::CRUISE_MISSILE ) ;
	weaponList[WeaponControl::MAIN_POD].append( Missile::GUIDED_MISSILE ) ;
	numWeapons[WeaponControl::AA_POD] = 1 ;
	weaponList[WeaponControl::AA_POD].append( Missile::CANNON ) ;

	bayPod = WeaponControl::MAIN_POD ;

	modelLabel = "Tank" ;
}



//------------------------------------------------------------------------------
// Handle a key press event (virtual).  Default is to do nothing.
//------------------------------------------------------------------------------
SbBool
Vehicle::handleKeyPress(
	KeySym
	)
{
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Position a newly launched weapon.
//------------------------------------------------------------------------------
void
Vehicle::launchWeapon(
	Missile				*weapon,
	WeaponControl::WeaponPod	pod
	)
{
	SbVec3f	up ;
	SbVec3f	ahead ;
	SbVec3f	right ;
	SbRotation rot = rotation.getValue() ;
	Utilities::setUpVector( rot, up ) ;
	Utilities::setForwardVector( rot, ahead ) ;
	Utilities::setRightVector( rot, right ) ;

	SbVec3f lp ;
	if( pod == WeaponControl::MAIN_POD )
	{
		lp = launchPoint + getCenterOffset() ;
	}
	else
	{
		lp = auxViewpoint + getCenterOffset() -
			SbVec3f( 0.0f, 0.0f, 0.075f ) ;
		rot = SbRotation( SbVec3f( -1.0f, 0.0f, 0.0f ), (float)M_PI_2 )
			* conView.getValue() * rot ;
	}
	SbVec3f pos = position.getValue() + lp[0] * right + lp[1] * ahead +
			lp[2] * up ;

	weapon->position.setValue( pos ) ;
	weapon->rotation.setValue( rot ) ;
	weapon->setVelocity( getVelocity() ) ;
}



//------------------------------------------------------------------------------
// Turn on the first person graphics.
//------------------------------------------------------------------------------
void
Vehicle::showFirstPersonGraphics(
	SbBool				show,
	WeaponControl::WeaponPod	pod
	)
{
	if( show && fpGraphics->getNumChildren() > 0 )
	{
		fpGraphics->whichChild = (int)pod ;
	}
	else
	{
		fpGraphics->whichChild = -1 ;
	}
}



//------------------------------------------------------------------------------
// Indicate if the vehicle has a range finder.
//------------------------------------------------------------------------------
SbBool
Vehicle::hasRangeFinder() const
{
	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Indicate if it is okay for the vehicle to fire under current conditions
// (neglecting if weapon is ready or not, that's checked elsewhere).
//------------------------------------------------------------------------------
SbBool
Vehicle::okayToFire(
	WeaponControl::WeaponPod
	)
{
	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Process the geometry just read in.
//------------------------------------------------------------------------------
void
Vehicle::processGeometry()
{
	SoNode *node ;
	if( root->getNumChildren() > 0 )
	{
		node = root->getChild( 0 ) ;

		// Search for view points.
		SoSearchAction search ;
		search.setName( "Viewpoints" ) ;
		search.setInterest( SoSearchAction::FIRST ) ;
		search.setSearchingAll( TRUE ) ;
		search.apply( node ) ;

		SoPath *path = search.getPath() ;
		SbBool okay = FALSE ;
		if( path != NULL )
		{
			SoNode *tail = path->getTail() ;
			if( tail->isOfType( SoCoordinate3::getClassTypeId() ) )
			{
				SoCoordinate3 *vp = (SoCoordinate3 *)tail ;
				if( vp->point.getNum() == 3 )
				{
					mainViewpoint = vp->point[0] ;
					auxViewpoint = vp->point[1] ;
					launchPoint = vp->point[2] ;
					okay = TRUE ;
				}
			}
		}

		if( !okay )
		{
			SoDebugError::post(
				getClassTypeId().getName().getString(),
				"\nCould not locate a Coordinate3 node named "
				"`Viewpoints' with 3 values\nin %s",
				name.getValue().getString() ) ;
		}
	}

	root->addChild( fpGraphics ) ;
}



//------------------------------------------------------------------------------
// Show vehicle-specific help text.
//------------------------------------------------------------------------------
void
Vehicle::showHelp(
	SoMFString	*str,
	int		&n
	)
{
	str->set1Value( n++, "Mouse - Control speed and direction" ) ;
}
