//------------------------------------------------------------------------------
// Class for weapon control.
//------------------------------------------------------------------------------

#include <assert.h>
#include <Inventor/SoDB.h>
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/errors/SoReadError.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include "ExplosionControl.h"
#include "Bomb.h"
#include "CruiseMissile.h"
#include "GuidedMissile.h"
#include "Missile.h"
#include "Mortar.h"
#include "Cannon.h"
#include "GuidedMissile.h"
#include "Utilities.h"
#include "Sfx.h"
#include "WeaponControl.h"


const HostAddr WeaponControl::DeadMissile = 0xffffffff ;
const HostAddr WeaponControl::ActiveMissile = 0 ;



//------------------------------------------------------------------------------
// Helper class.
//------------------------------------------------------------------------------
class WeaponBay {
    public :
	enum BayState {
		EMPTY = 0,
		LOADING = 1,
		READY = 2
		} ;
	WeaponBay() ;
	~WeaponBay() {}
	SbBool		fireBay() ;
	SbBool		updateBay( float deltaTime, SbBool okayToLoad,
				int nextWeaponTypeNo, float buildTime ) ;
	int		getWeaponTypeNo() const { return( weaponTypeNo ) ; }
	static void	resetSerialNumber() { nextSerialNumber = 0 ; }
	BayState	getBayState() { return( state ) ; }
	SbBool		isReady() { return( state == READY ) ; }
	void		reset() ;

    private :
	BayState	state ;
	float		stateAge ;
	int32_t		serialNo ;
	int		weaponTypeNo ;
	float		weaponBuildTime ;
	static int32_t	nextSerialNumber ;
} ;



//------------------------------------------------------------------------------
// Helper class.
//------------------------------------------------------------------------------
class BigGun {
    public :
	enum State {
		LOADING = 0,
		READY = 1
		} ;
	BigGun() ;
	~BigGun() {}
	SbBool		fire() ;
	SbBool		update( float deltaTime ) ;
	SbBool		isReady() { return( state == READY ) ; }
	void		reset() ;
	void		setLoadTime( float t ) { loadTime = t ; }

    private :
	State		state ;
	float		stateAge ;
	float		loadTime ;
} ;


int32_t		WeaponBay::nextSerialNumber = 0 ;



//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
WeaponControl::WeaponControl(
	int	numberOfBays
	) :
	numBays( numberOfBays )
{
	assert( numberOfBays >= 0 && numberOfBays < 4 ) ;

	weaponGroup = new SoGroup ;
	weaponGroup->ref() ;

	myWeaponGroup = new SoGroup ;
	weaponGroup->ref() ;

	statusColor[0].setValue( 0.1f, 0.1f, 0.1f ) ;
	statusColor[1].setValue( 0.2f, 0.2f, 0.0f ) ;
	statusColor[2].setValue( 0.0f, 0.2f, 0.0f ) ;
	statusColor[3].setValue( 0.0f, 0.0f, 0.0f ) ;

	createWeaponDisplay() ;

	mainWeapons = new SoGroup ;
	mainWeapons->ref() ;

	AAWeapons = new SoGroup ;
	AAWeapons->ref() ;

	WeaponBay::resetSerialNumber() ;
	bay = new WeaponBay[numBays] ;
	cannon = new BigGun ;

	reset() ;

	missileLoading = new Sfx( "missileLoading.aifc" ) ;
	missileReady = new Sfx( "missileReady.aifc" ) ;
	fireMissile = new Sfx( "fire.aifc" ) ;
	fireCannon = new Sfx( "fireCannon.aifc" ) ;
	cannonReady = new Sfx( "loadCannon.aifc" ) ;

	bayPod = MAIN_POD ;

	weaponClassType[Missile::MISSILE] = Missile::getClassTypeId() ;
	weaponClassType[Missile::CRUISE_MISSILE] =
					CruiseMissile::getClassTypeId() ;
	weaponClassType[Missile::GUIDED_MISSILE] =
					GuidedMissile::getClassTypeId() ;
	weaponClassType[Missile::MORTAR] = Mortar::getClassTypeId() ;
	weaponClassType[Missile::BOMB] = Bomb::getClassTypeId() ;
	weaponClassType[Missile::CANNON] = Cannon::getClassTypeId() ;
	weaponClassType[Missile::BAD_TYPE] = SoType::badType() ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
WeaponControl::~WeaponControl()
{
	delete bay ;
	delete cannon ;

	weaponGroup->unref() ;
	myWeaponGroup->unref() ;
	weaponDisplay->unref() ;
	mainWeapons->unref() ;
	AAWeapons->unref() ;

	delete missileLoading ;
	delete missileReady ;
	delete fireMissile ;
}



//------------------------------------------------------------------------------
// Reset the weapons.
//------------------------------------------------------------------------------
void
WeaponControl::reset()
{
	guidedMissile = NULL ;

	nextBay = 0 ;
	baySwitch->whichChild = nextBay ;

	nextMainWeapon = 0 ;
	nextAAWeapon = 0 ;
	nextLoadingBay = 0 ;
	nextBullet = 0 ;

	for( int i = 0 ; i < numBays ; i++ )
	{
		bay[i].reset() ;
		bayStatus[i]->rgb.setValue( statusColor[0] ) ;
		bayIcon[i]->removeAllChildren() ;
	}
	for( i = numBays ; i < MAX_BAYS ; i++ )
	{
		bayStatus[i]->rgb.setValue( statusColor[3] ) ;
		bayIcon[i]->removeAllChildren() ;
	}

	cannonStatus->rgb.setValue( statusColor[0] ) ;
	cannon->reset() ;
}



//------------------------------------------------------------------------------
// Add a weapon type to the arsenal.
//------------------------------------------------------------------------------
SbBool
WeaponControl::addWeaponType(
	Missile::MissileType	weaponClassTypeId,
	WeaponPod		view
	)
{
	SoGroup *group ;
	Missile *mis ;

	switch( view )
	{
		case MAIN_POD :
			group = mainWeapons ;
			break ;

		case AA_POD :
			group = AAWeapons ;
			break ;

		default :
			return( FALSE ) ;
	}

	for( int i = 0 ; i < group->getNumChildren() ; i++ )
	{
		mis = (Missile *)group->getChild( i ) ;
		if( mis->getMissileType() == weaponClassTypeId )
		{
			return( TRUE ) ;
		}
	}

	SoType wType = weaponClassType[weaponClassTypeId] ;

	if( !wType.canCreateInstance() )
	{
		SoDebugError::post( "WeaponControl::addWeaponType",
			"Can't create an instance of %s",
			wType.getName().getString() ) ;
		exit( -1 ) ;
	}

	mis = (Missile *)wType.createInstance() ;
	group->addChild( mis ) ;

	if( view == MAIN_POD && mainWeapons->getNumChildren() == 1 )
	{
		nextMainWeaponGroup->addChild( mis->getIconGraph() );
	}
	else if( view == AA_POD && AAWeapons->getNumChildren() == 1 )
	{
		nextAAWeaponGroup->addChild( mis->getIconGraph() );
	}

	if( group->getNumChildren() == 1 && view != bayPod )
	{
		float t = mis->getMaxAge() * 0.3f ;
		if( t < 0.5f )
		{
			t = 0.5f ;
		}
		cannon->setLoadTime( t ) ;
	}

	return( TRUE ) ;
}



//------------------------------------------------------------------------------
// Update the weapons and the weapon bays.
//------------------------------------------------------------------------------
void
WeaponControl::updateWeapons(
	float	mx,
	float	my,
	float	deltaTime,
	SbBool	updateBays,
	SoNode	*scene
	)
{
	int i ;

	//
	// Update the status on all of the weapon bays.
	//
	if( updateBays )
	{
		Missile *nextMis ;
		int nextWeapon ;
		if( bayPod == MAIN_POD )
		{
			nextWeapon = nextMainWeapon ;
			nextMis = (Missile *)
				mainWeapons->getChild( nextWeapon ) ;
		}
		else {
			nextWeapon = nextAAWeapon ;
			nextMis = (Missile *)
				AAWeapons->getChild( nextWeapon ) ;
		}
		float bt = nextMis->getBuildTime() ;

		for( i = 0 ; i < numBays ; i++ )
		{
			if( bay[i].updateBay( deltaTime,
				( i == nextLoadingBay ), nextWeapon, bt ) )
			{
				bayStatus[i]->rgb.setValue( statusColor[
						(int)bay[i].getBayState() ] ) ;
				if( bay[i].getBayState() == WeaponBay::LOADING )
				{
					bayIcon[i]->addChild(
						nextMis->getIconGraph() );
					missileLoading->play() ;
				}
				else if( bay[i].getBayState() ==
					WeaponBay::READY )
				{
					missileReady->play() ;
				}
			}
		}

		//
		// Check to see if we can start the loading procedure on
		// another bay.
		//
		if( bay[nextLoadingBay].getBayState() == WeaponBay::READY )
		{
			nextLoadingBay = ( nextLoadingBay + 1 ) % numBays ;
		}

		if( cannon->update( deltaTime ) )
		{
			cannonReady->play() ;
			cannonStatus->rgb.setValue( statusColor[2] ) ;
		}
	}

	int i0 = myWeaponGroup->getNumChildren() - 1 ;
	for( i = i0 ; i >= 0 ; i-- )
	{
		Missile *mis = (Missile *)myWeaponGroup->getChild( i ) ;

		Missile::MissileStatus result ;

		if( mis == guidedMissile )
		{
			result = mis->update( mx, my, deltaTime, scene ) ;
			if( result != Missile::ACTIVE )
			{
				guidedMissile = NULL ;
			}
		}
		else
		{
			result = mis->update( 0.0f, 0.0f, deltaTime, scene ) ;
		}

		switch( result )
		{
			case Missile::ACTIVE :
				break ;

			case Missile::STRUCK :
				if( missileStrikeHandler == NULL )
				{
					terminateMissile( mis, DeadMissile ) ;
				}
				else
				{
					SoNode *splat = mis->getObstructor() ?
						mis->getObstructor()->getTail():
						NULL ;
					HostAddr st = missileStrikeHandler(
						splat, mis, missileStrikeData );
					terminateMissile( mis, st ) ;
				}
				break ;

			case Missile::EXPIRED :
				terminateMissile( mis, DeadMissile ) ;
				break ;
		}
	}
}



//------------------------------------------------------------------------------
// Create a missile.
//------------------------------------------------------------------------------
Missile *
WeaponControl::createMissile(
	WeaponPod	currentPod,
	int		n
	)
{
	Missile::MissileType type ;

	if( currentPod == MAIN_POD )
	{
		type = ( (Missile *)mainWeapons->getChild( n ) )->
				getMissileType() ;
	}
	else
	{
		type = ( (Missile *)AAWeapons->getChild( n ) )->
				getMissileType() ;
	}
	return( (Missile *)weaponClassType[type].createInstance() ) ;
}



//------------------------------------------------------------------------------
// Launch a weapon.
//------------------------------------------------------------------------------
Missile *
WeaponControl::launchWeapon(
	WeaponPod	currentPod
	)
{
	if( guidedMissile != NULL )
	{
		return( NULL ) ;
	}

	if( currentPod == bayPod )
	{
		if( bay[nextBay].fireBay() )
		{
			Missile *ammo = createMissile( currentPod,
					bay[nextBay].getWeaponTypeNo() ) ;
			if( ammo->isOfType( Cannon::getClassTypeId() ) )
			{
				fireCannon->play() ;
			}
			else
			{
				fireMissile->play() ;
			}
			myWeaponGroup->addChild( ammo ) ;
			weaponGroup->addChild( ammo ) ;
			if( ammo->isGuided() )
			{
				guidedMissile = ammo ;
			}
			myMissile[nextBay].init( ammo ) ;
			bayStatus[nextBay]->rgb.setValue( statusColor[0] ) ;
			bayIcon[nextBay]->removeAllChildren() ;
			nextBay = ( nextBay + 1 ) % numBays ;
			baySwitch->whichChild = nextBay ;

			return( ammo ) ;
		}
	}
	else
	{
		if( cannon->fire() )
		{
			cannonStatus->rgb.setValue( statusColor[1] ) ;
			Missile *ammo = createMissile( currentPod, 0 ) ;
			if( ammo->isOfType( Cannon::getClassTypeId() ) )
			{
				fireCannon->play() ;
			}
			else
			{
				fireMissile->play() ;
			}
			myWeaponGroup->addChild( ammo ) ;
			weaponGroup->addChild( ammo ) ;
			if( ammo->isGuided() )
			{
				guidedMissile = ammo ;
			}
			myBullet[nextBullet].init( ammo ) ;
			nextBullet = ( nextBullet + 1 ) % MAX_BULLETS ;

			return( ammo ) ;
		}
	}

	return( NULL ) ;
}



//------------------------------------------------------------------------------
// Launch an automaton's missile.
//------------------------------------------------------------------------------
Missile *
WeaponControl::launchAutomatonMissile(
	SoType	type
	)
{
	Missile *ammo = (Missile *)type.createInstance() ;
	if( ammo )
	{
		weaponGroup->addChild( ammo ) ;
	}

	return( ammo ) ;
}



//------------------------------------------------------------------------------
// Launch another player's missile.
//------------------------------------------------------------------------------
Missile *
WeaponControl::launchPlayerMissile(
	Missile::MissileType	type
	)
{
	Missile *ammo = (Missile *)weaponClassType[type].createInstance() ;
	if( ammo )
	{
		weaponGroup->addChild( ammo ) ;
	}

	return( ammo ) ;
}



//------------------------------------------------------------------------------
// Change what weapon a bay is to load.
//------------------------------------------------------------------------------
SbBool
WeaponControl::changeWeapon(
	WeaponPod	currentPod
	)
{
	if( currentPod == bayPod )
	{
		if( bayPod == MAIN_POD )
		{
			nextMainWeapon = ( nextMainWeapon + 1 ) %
						mainWeapons->getNumChildren() ;
			Missile *nextMis = (Missile *)mainWeapons->getChild(
							nextMainWeapon ) ;
			nextMainWeaponGroup->removeAllChildren() ;
			nextMainWeaponGroup->addChild( nextMis->getIconGraph());
		}
		else
		{
			nextAAWeapon = ( nextAAWeapon + 1 ) %
						AAWeapons->getNumChildren() ;
			Missile *nextMis = (Missile *)AAWeapons->getChild(
							nextAAWeapon ) ;
			nextAAWeaponGroup->removeAllChildren() ;
			nextAAWeaponGroup->addChild( nextMis->getIconGraph() ) ;
		}
		return( TRUE ) ;
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Create the weapon display.
//------------------------------------------------------------------------------
void
WeaponControl::createWeaponDisplay()
{
	SoInput	*in = Utilities::createInput() ;
	static const char *filename = "graphs/weapon_display.iv" ;

	if( !in->openFile( filename, TRUE ) )
	{
		SoReadError::post( in, "Can't open %s", filename ) ;
		exit( -1 ) ;
	}

	weaponDisplay = SoDB::readAll( in ) ;

	if( weaponDisplay == NULL )
	{
		SoReadError::post( in, "Can't read %s", filename );
		exit( -1 ) ;
	}

	weaponDisplay->ref() ;

	baySwitch = (SoSwitch *)Utilities::getNamedNode( "BayIndicator",
						SoSwitch::getClassTypeId() ) ;
	if( baySwitch == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a Switch"
			" node named `BayIndicator'\n", filename ) ;
		exit( -1 ) ;
	}

	for( int i = 0 ; i < MAX_BAYS ; i++ )
	{
		SbString name( "BayStatus_" ) ;
		name += SbString( i ) ;
		bayStatus[i] = (SoBaseColor *)Utilities::getNamedNode(
			name.getString(), SoBaseColor::getClassTypeId() ) ;
		if( bayStatus[i] == NULL )
		{
			SoReadError::post( in, "File `%s' does not contain a "
				"BaseColor node named `%s'\n", filename,
				name.getString() ) ;
			exit( -1 ) ;
		}
		bayStatus[i]->rgb.setValue( statusColor[0] ) ;

		name = "BayIcon_" ;
		name += SbString( i ) ;
		bayIcon[i] = (SoSeparator *)Utilities::getNamedNode(
			name.getString(), SoSeparator::getClassTypeId() ) ;
		if( bayIcon[i] == NULL )
		{
			SoReadError::post( in, "File `%s' does not contain a "
				"Separator node named `%s'\n", filename,
				name.getString() ) ;
			exit( -1 ) ;
		}
	}

	SbString name( "NextMainWeaponIcon" ) ;
	nextMainWeaponGroup = (SoSeparator *)Utilities::getNamedNode(
			name.getString(), SoSeparator::getClassTypeId() ) ;
	if( nextMainWeaponGroup == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"Separator node named `%s'\n", filename,
			name.getString() ) ;
		exit( -1 ) ;
	}

	name = "NextAAWeaponIcon" ;
	nextAAWeaponGroup = (SoSeparator *)Utilities::getNamedNode(
			name.getString(), SoSeparator::getClassTypeId() ) ;
	if( nextAAWeaponGroup == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"Separator node named `%s'\n", filename,
			name.getString() ) ;
		exit( -1 ) ;
	}

	name = "AuxStatus" ;
	auxStatus = (SoBaseColor *)Utilities::getNamedNode(
		name.getString(), SoBaseColor::getClassTypeId() ) ;
	if( auxStatus == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"BaseColor node named `%s'\n", filename,
			name.getString() ) ;
		exit( -1 ) ;
	}
	auxStatus->rgb.setValue( 1.0f, 1.0f, 1.0f ) ;

	name = "MainStatus" ;
	mainStatus = (SoBaseColor *)Utilities::getNamedNode(
		name.getString(), SoBaseColor::getClassTypeId() ) ;
	if( mainStatus == NULL )
	{
		SoReadError::post( in, "File `%s' does not contain a "
			"BaseColor node named `%s'\n", filename,
			name.getString() ) ;
		exit( -1 ) ;
	}
	mainStatus->rgb.setValue( 1.0f, 1.0f, 1.0f ) ;

	cannonStatus = auxStatus ;

	in->closeFile() ;

	delete in ;
}



//------------------------------------------------------------------------------
// Indicate which pod controls the weapon bays.
//------------------------------------------------------------------------------
void
WeaponControl::setBayPod(
	WeaponPod pod
	)
{
	bayPod = pod ;

	if( pod == MAIN_POD )
	{
		cannonStatus = auxStatus ;
		mainStatus->rgb.setValue( 1.0f, 1.0f, 1.0f ) ;
	}
	else
	{
		cannonStatus = mainStatus ;
		auxStatus->rgb.setValue( 1.0f, 1.0f, 1.0f ) ;
	}
}



//------------------------------------------------------------------------------
// Terminate a missile.
//------------------------------------------------------------------------------
void
WeaponControl::terminateMissile(
	Missile		*mis,
	HostAddr	newStatus
	)
{
	int	n ;

	if( ( n = myWeaponGroup->findChild( mis ) ) != -1 )
	{
		if( mis == guidedMissile )
		{
			guidedMissile = NULL ;
		}

		myWeaponGroup->removeChild( n ) ;
	}

	terminateAnyonesMissile( mis ) ;

	for( n = 0 ; n < numBays ; n++ )
	{
		if( mis == myMissile[n].getMissile() )
		{
			myMissile[n].explode( newStatus ) ;
			break ;
		}
	}

	for( n = 0 ; n < MAX_BULLETS ; n++ )
	{
		if( mis == myBullet[n].getMissile() )
		{
			myBullet[n].explode( newStatus ) ;
			break ;
		}
	}
}



//------------------------------------------------------------------------------
// Terminate either a players or my own missile.
//------------------------------------------------------------------------------
void
WeaponControl::terminateAnyonesMissile(
	Missile *mis
	)
{
	if( explosionControl )
	{
		if( mis->isOfType( Cannon::getClassTypeId() ) )
		{
			explosionControl->addBulletExplosion(
				mis->position.getValue() ) ;
		}
		else
		{
			explosionControl->addExplosion(
				mis->position.getValue() ) ;
		}
	}

	int n ;
	if( ( n = weaponGroup->findChild( mis ) ) != -1 )
	{
		weaponGroup->removeChild( n ) ;
	}
	else
	{
		printf( "couldn't find missile in weaponGroup (mis = 0x%p)\n",
			mis ) ;
	}
}



//------------------------------------------------------------------------------
// Return true if the next weapon to launch is a guided type and is ready.
//------------------------------------------------------------------------------
SbBool
WeaponControl::isGuidedWeaponReady(
	WeaponPod	pod
	)
{
	if( pod == bayPod && bay[nextBay].isReady() )
	{
		int n = bay[nextBay].getWeaponTypeNo() ;
		Missile::MissileType type ;
		if( bayPod == MAIN_POD )
		{
			type = ( (Missile *)mainWeapons->getChild( n )
				)->getMissileType() ;
		}
		else
		{
			type = ( (Missile *)AAWeapons->getChild( n )
				)->getMissileType() ;
		}
		if( weaponClassType[type].isDerivedFrom(
			GuidedMissile::getClassTypeId() ) )
		{
			return( TRUE ) ;
		}
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Load the missile info into a packet for broadcasting.
//------------------------------------------------------------------------------
void
WeaponControl::loadMissileInfo(
	MissileInfo	*info,
	float		dt
	)
{
	static float zeros[4] = { 0.0f, 0.0f, 0.0f, 0.0f } ;

	for( int i = 0 ; i < MAX_BAYS ; i++ )
	{
		Missile *mis = myMissile[i].getMissile() ;
		if( myMissile[i].getStatus() == ActiveMissile )
		{
			info[i].setData( myMissile[i].getStatus(),
				mis->getMissileType(),
				mis->position.getValue().getValue(),
				mis->rotation.getValue().getValue() ) ;
		}
		else if( mis != NULL && myMissile[i].age( dt ) < 0.5f )
		{
			info[i].setData( myMissile[i].getStatus(),
				mis->getMissileType(),
				mis->position.getValue().getValue(),
				zeros ) ;
			info[i].setDamage( myMissile[i].getDamage() ) ;
		}
		else
		{
			info[i].setData( myMissile[i].getStatus(),
				Missile::BAD_TYPE, zeros, zeros ) ;
		}
	}

	for( i = 0 ; i < MAX_BULLETS ; i++ )
	{
		Missile *mis = myBullet[i].getMissile() ;
		if( myBullet[i].getStatus() == ActiveMissile )
		{
			info[MAX_BAYS+i].setData( myBullet[i].getStatus(),
				mis->getMissileType(),
				mis->position.getValue().getValue(),
				mis->rotation.getValue().getValue() ) ;
		}
		else if( mis != NULL && myBullet[i].age( dt ) < 0.5f )
		{
			info[MAX_BAYS+i].setData( myBullet[i].getStatus(),
				mis->getMissileType(),
				mis->position.getValue().getValue(), zeros ) ;
			info[MAX_BAYS+i].setDamage( myBullet[i].getDamage() ) ;
		}
		else
		{
			info[MAX_BAYS+i].setData( myBullet[i].getStatus(),
				Missile::BAD_TYPE, zeros, zeros ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Routines for the WeaponBay helper class.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
WeaponBay::WeaponBay()
{
	serialNo = nextSerialNumber++ ;
	reset() ;
}



//------------------------------------------------------------------------------
// Reset the weapon bay.
//------------------------------------------------------------------------------
void
WeaponBay::reset()
{
	state = EMPTY ;
	stateAge = 0.0f ;
}



//------------------------------------------------------------------------------
// Fire a weapon bay.
//------------------------------------------------------------------------------
SbBool
WeaponBay::fireBay()
{
	if( state == READY )
	{
		state = EMPTY ;
		stateAge = 0.0f ;
		return( TRUE ) ;
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Update the status on the bays.
//------------------------------------------------------------------------------
SbBool
WeaponBay::updateBay(
	float	deltaTime,
	SbBool	okayToLoad,
	int	nextWeaponTypeNo,
	float	buildTime
	)
{
	SbBool	changedState = FALSE ;

	switch( state )
	{
		case EMPTY :
			stateAge += deltaTime ;
			if( okayToLoad && stateAge >= 2.0f )
			{
				weaponTypeNo = nextWeaponTypeNo ;
				weaponBuildTime = buildTime ;
				state = LOADING ;
				changedState = TRUE ;
				stateAge -= 2.0f ;
				if( stateAge > deltaTime )
				{
					stateAge = deltaTime ;
				}
			}
			break ;

		case LOADING :
			stateAge += deltaTime ;
			if( stateAge >= weaponBuildTime )
			{
				state = READY ;
				changedState = TRUE ;
				stateAge -= weaponBuildTime ;
			}
			break ;

		case READY :
			break ;
	}

	return( changedState ) ;
}



//------------------------------------------------------------------------------
// Set the number of bays.
//------------------------------------------------------------------------------
void
WeaponControl::setNumberOfBays(
	int	newNumber
	)
{
	if( newNumber < 1 )
	{
		newNumber = 1 ;
	}
	else if( newNumber > MAX_BAYS )
	{
		newNumber = MAX_BAYS ;
	}

	if( newNumber != numBays )
	{
		delete bay ;
		numBays = newNumber ;
		bay = new WeaponBay[numBays] ;
	}

	reset() ;
}



//------------------------------------------------------------------------------
// Clear out the weapon types.
//------------------------------------------------------------------------------
void
WeaponControl::clearWeaponTypes()
{
	mainWeapons->removeAllChildren() ;
	AAWeapons->removeAllChildren() ;
	nextMainWeaponGroup->removeAllChildren() ;
	nextAAWeaponGroup->removeAllChildren() ;
}



//------------------------------------------------------------------------------
// Indicate whether or not a missile is a bomb.
//------------------------------------------------------------------------------
SbBool
WeaponControl::isBomb(
	Missile::MissileType	type
	) const
{
	return( 0 <= type && type < Missile::BAD_TYPE &&
		weaponClassType[type].isDerivedFrom(Bomb::getClassTypeId()) ) ;
}



//------------------------------------------------------------------------------
// Return the class string of a weapon type.
//------------------------------------------------------------------------------
const char *
WeaponControl::getWeaponName(
	Missile::MissileType	type
	) const
{
	return( ( 0 <= type && type < Missile::BAD_TYPE ) ?
			weaponClassType[type].getName().getString() :
			"Unknown Weapon" ) ;
}



//------------------------------------------------------------------------------
// MissilePack helper class definitions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
MissilePack::MissilePack()
{
	missile = NULL ;
	status = WeaponControl::DeadMissile ;
	damage = 0.0f ;
}



//------------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------------
MissilePack::~MissilePack()
{
	if( missile != NULL )
	{
		missile->unref() ;
		status = WeaponControl::DeadMissile ;
	}
}



//------------------------------------------------------------------------------
// Initialize.
//------------------------------------------------------------------------------
void
MissilePack::init(
	Missile	*m
	)
{
	m->ref() ;

	if( missile != NULL )
	{
		missile->unref() ;
	}

	missile = m ;
	status = WeaponControl::ActiveMissile ;
	damage = missile->getDamage() ;
}



//------------------------------------------------------------------------------
// Blow up the missile.
//------------------------------------------------------------------------------
void
MissilePack::explode(
	HostAddr newStatus
	)
{
	status = newStatus ;
	explosionAge = 0.0f ;
}



//------------------------------------------------------------------------------
// Routines for the Cannon helper class.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Constructor.
//------------------------------------------------------------------------------
BigGun::BigGun()
{
	reset() ;
}



//------------------------------------------------------------------------------
// Reset the cannon.
//------------------------------------------------------------------------------
void
BigGun::reset()
{
	state = LOADING ;
	stateAge = 0.0f ;
}



//------------------------------------------------------------------------------
// Fire the cannon.
//------------------------------------------------------------------------------
SbBool
BigGun::fire()
{
	if( state == READY )
	{
		reset() ;
		return( TRUE ) ;
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Update the status on the cannon.
//------------------------------------------------------------------------------
SbBool
BigGun::update(
	float	deltaTime
	)
{
	SbBool	changedState = FALSE ;

	switch( state )
	{
		case LOADING :
			stateAge += deltaTime ;
			if( stateAge >= loadTime )
			{
				state = READY ;
				changedState = TRUE ;
				stateAge = 0.0f ;
			}
			break ;

		case READY :
			break ;
	}

	return( changedState ) ;
}



