//------------------------------------------------------------------------------
// Generate normals for GroundShapes and GroundStrips.
//------------------------------------------------------------------------------
#include <stdio.h>
#include <Inventor/SoDB.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoSeparator.h>
#include "GroundShape.h"
#include "GroundStrip.h"
#include "GroundMesh.h"



static SoSeparator	*root ;

#define PRINT_VEC(s,v) \
	printf("%-20s: %7.4f %7.4f %7.4f\n",(s),(v)[0],(v)[1],(v)[2])

//------------------------------------------------------------------------------
// Usage routine.
//------------------------------------------------------------------------------
void
usage()
{
	fprintf( stderr, "Usage: gsnorm in.iv out.iv\n" ) ;
	exit( 1 ) ;
}



//------------------------------------------------------------------------------
// Generate the normals for a GroundStrip node at the end of a path.
//------------------------------------------------------------------------------
void
genStripNorms(
	SoPath *path
	)
{
	// Assume PER_VERTEX_INDEXED binding for now.
	GroundStrip *gs = (GroundStrip *)path->getTail() ;
	gs->ref() ;

	long vpr = gs->height.getNum() ;
	long numPoints = vpr * 2 ;

	if( gs->vector.getNum() == numPoints || vpr < 2 )
	{
		printf( "genStripNorms: %d vectors, %d height points\n",
			gs->vector.getNum(), vpr ) ;
		gs->unrefNoDelete() ;
		return ;
	}

	long j ;
	float deltaL = gs->length.getValue() / (float)( vpr - 1 ) ;

	long k = 0 ;
	gs->vector.setNum( numPoints ) ;
	SbVec3f *v = gs->vector.startEditing() ;
	float dr ;
	long jp, jm ;
	for( j = 0 ; j < vpr ; j++ )
	{
		if( j == 0 )
		{
			dr = deltaL ;
			jp = 1 ;
			jm = 0 ;
		}
		else if( j == vpr - 1 )
		{
			dr = deltaL ;
			jp = j ;
			jm = j - 1 ;
		}
		else
		{
			dr = 2.0f * deltaL ;
			jp = j + 1 ;
			jm = j - 1 ;
		}
		SbVec3f A( 1.0f, 0.0f, 0.0f ) ;
		SbVec3f B( 0.0f, dr, gs->height[jp] -
					gs->height[jm] ) ;
		v[k] = A.cross( B ) ;
		v[k].normalize() ;
		v[k+1] = v[k] ;
		k += 2 ;
	}
	gs->vector.finishEditing() ;
			
	gs->unrefNoDelete() ;
}



//------------------------------------------------------------------------------
// Generate the normals for a GroundShape node at the end of a path.
//------------------------------------------------------------------------------
void
genMeshNorms(
	SoPath *path
	)
{
	// Assume PER_VERTEX_INDEXED binding for now.
	GroundMesh *gs = (GroundMesh *)path->getTail() ;
	gs->ref() ;

	long vpr = gs->verticesPerRow.getValue() ;
	long vpc = gs->verticesPerColumn.getValue() ;
	long numPoints = vpr * vpc ;

	if( gs->vector.getNum() == numPoints || vpr < 2 || vpc < 2 )
	{
		gs->unrefNoDelete() ;
		return ;
	}

	float deltaL = gs->length.getValue() / (float)( vpr - 1 ) ;
	float deltaW = gs->width.getValue() / (float)( vpc - 1 ) ;

	long i ;
	long j ;
	long k = 0 ;
	long nZ = gs->height.getNum() ;
	gs->vector.setNum( numPoints ) ;
	SbVec3f *v = gs->vector.startEditing() ;
	float dc ;
	float dr ;
	long ip, im ;
	long jp, jm ;
	for( i = 0 ; i < vpc ; i++ )
	{
		if( i == 0 )
		{
			dc = deltaW ;
			ip = 1 ;
			im = 0 ;
		}
		else if( i == vpc - 1 )
		{
			dc = deltaW ;
			ip = i ;
			im = i - 1 ;
		}
		else
		{
			dc = 2.0f * deltaW ;
			ip = i + 1 ;
			im = i - 1 ;
		}
		for( j = 0 ; j < vpr ; j++ )
		{
			if( j == 0 )
			{
				dr = deltaL ;
				jp = 1 ;
				jm = 0 ;
			}
			else if( j == vpr - 1 )
			{
				dr = deltaL ;
				jp = j ;
				jm = j - 1 ;
			}
			else
			{
				dr = 2.0f * deltaL ;
				jp = j + 1 ;
				jm = j - 1 ;
			}
			SbVec3f A( dc, 0.0f, gs->height[(ip*vpr+j)%nZ] -
						gs->height[(im*vpr+j)%nZ] ) ;
			SbVec3f B( 0.0f, dr, gs->height[(i*vpr+jp)%nZ] -
						gs->height[(i*vpr+jm)%nZ] ) ;
			v[k] = A.cross( B ) ;
			v[k].normalize() ;
			k++ ;
		}
	}
	gs->vector.finishEditing() ;
			
	gs->unrefNoDelete() ;
}



//------------------------------------------------------------------------------
// Main routine.
//------------------------------------------------------------------------------
void
main(
	int	argc,
	char	**argv
	)
{
	int c ;
	extern char *optarg ;
	extern int optind ;
	while ( ( c = getopt( argc, argv, "" ) ) != EOF )
	{
		switch( c )
		{
			default :
				usage() ;
				break ;
		}
	}

	if( argc - 2 != optind )
	{
		usage() ;
	}

	SoDB::init() ;
	GroundShape::initClass() ;
	GroundStrip::initClass() ;
	GroundMesh::initClass() ;

	SoInput	in ;

	if( !in.openFile( argv[optind], TRUE ) )
	{
		fprintf( stderr, "%s: file not opened\n", argv[optind] ) ;
		exit( -1 ) ;
	}
	optind++ ;

	root = SoDB::readAll( &in ) ;
	in.closeFile() ;
	if( root == NULL )
	{
		fprintf( stderr, "%s: no data read\n", argv[optind] ) ;
		exit( 1 ) ;
	}
	root->ref() ;

	//
	// Search for ground strips.
	//
	SoSearchAction sa ;
	sa.setType( GroundStrip::getClassTypeId() ) ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setSearchingAll( TRUE ) ;
	sa.apply( root ) ;

	SoPathList &pathList = sa.getPaths() ;

	printf( "found %d GroundStrip nodes\n", pathList.getLength() ) ;
	for( int i = 0 ; i < pathList.getLength() ; i++ )
	{
		genStripNorms( pathList[i] ) ;
	}

	//
	// Search for ground shapes.
	//
	SoSearchAction sa2 ;
	sa2.setType( GroundMesh::getClassTypeId() ) ;
	sa2.setInterest( SoSearchAction::ALL ) ;
	sa2.setFind( SoSearchAction::TYPE ) ;
	sa2.setSearchingAll( TRUE ) ;
	sa2.apply( root ) ;

	pathList = sa2.getPaths() ;
	printf( "found %d GroundMesh nodes\n", pathList.getLength() ) ;
	for( i = 0 ; i < pathList.getLength() ; i++ )
	{
		genMeshNorms( pathList[i] ) ;
	}

	SoOutput out ;
	out.openFile( argv[optind] ) ;
	SoWriteAction wa( &out ) ;
	wa.apply( root ) ;
	out.closeFile() ;
}
