//------------------------------------------------------------------------------
// Main routine for Pointblank.
//------------------------------------------------------------------------------

#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/prctl.h>
#include <Inventor/SoDB.h>
#include <Inventor/SbString.h>
#include <Inventor/SbTime.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/fields/SoSFTime.h>
#include <Inventor/sensors/SoAlarmSensor.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/nodes/SoSeparator.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include "Automaton.h"
#include "Battery.h"
#include "DestructibleObstacle.h"
#include "Vehicle.h"
#include "Tank.h"
#include "Scout.h"
#include "Mauler.h"
#include "Mortar.h"
#include "Flyer.h"
#include "Goal.h"
#include "GroundMesh.h"
#include "GroundStrip.h"
#include "StaticObstacle.h"
#include "CollisionAction.h"
#include "FollowTerrainAction.h"
#include "LosAction.h"
#include "RotateRotation.h"
#include "NoPassGroup.h"
#include "Missile.h"
#include "Cannon.h"
#include "CruiseMissile.h"
#include "GuidedMissile.h"
#include "Bomb.h"
#include "Utilities.h"
#include "CockPit.h"
#include "RedrawGroup.h"
#include "ZbufferMode.h"
#include "RadarControl.h"
#include "Dashboard.h"
#include "WeaponControl.h"
#include "ExplosionControl.h"
#include "AnimationMaster.h"
#include "TransformSwitch.h"
#include "Debris.h"
#include "GameIO.h"
#include "NameList.h"
#include "GameControl.h"
#include "SceneryFile.h"
#include "GameScenario.h"
#include "Sfx.h"

#include "cursor_fg.bm"
#include "cursor_bg.bm"

static SbTime		lastTime ;
static SbTime		pauseTime ;
static SbTime		pauseStart ;
static SbBool		debugging = FALSE ;
static SbBool		showTiming = FALSE ;
static SbBool		pauseIt = FALSE ;
static SbBool		mouseLock = FALSE ;
static SbBool		stopNextFrame = FALSE ;
static SoSFTime		*myRealTime ;
static GameControl	*game ;
static GameIO		*gio ;
static CockPit		*cockPit ;
static const char	*serviceName = "sgi-pointblank" ;
static const char	*groupName = "224.0.1.2" ;
static const int	defaultPort = 5143 ;
static GameScenario	*scenario ;

#ifdef INDIZONE
static const char	*defaultDir = "/usr/demos/IndiZone/.data/Pointblank" ;
#else
static const char	*defaultDir = "data" ;
#endif // INDIZONE

extern int		errno ;

#define	PB_VERSION	3

//------------------------------------------------------------------------------
// Set the cursor.
//------------------------------------------------------------------------------
static void
setCursor(
	Widget w
	)
{
	XColor			fg ;
	XColor			bg ;
	Display *display = XtDisplay( w ) ;
    
	Pixmap shape = XCreatePixmapFromBitmapData( display, XtWindow( w ),
			(char *)cursor_fg_bits, cursor_fg_width,
			cursor_fg_height, 1, 0, 1 ) ;
	Pixmap mask = XCreatePixmapFromBitmapData( display, XtWindow( w ),
			(char *)cursor_bg_bits, cursor_bg_width,
			cursor_bg_height, 1, 0, 1 ) ;
	Colormap cmap = DefaultColormap( display, DefaultScreen( display ) ) ;
	XParseColor( display, cmap, "yellow", &fg ) ;
	XParseColor( display, cmap, "black", &bg ) ;
	Cursor cursor = XCreatePixmapCursor( display, shape, mask, &fg,
				&bg, cursor_fg_x_hot, cursor_fg_y_hot ) ;
	XDefineCursor( display, XtWindow( w ), cursor ) ;
	XFlush( display ) ;
}



//------------------------------------------------------------------------------
// Toggle the pause setting.
//------------------------------------------------------------------------------
void
togglePause()
{
	pauseIt = !pauseIt ;
	if( pauseIt )
	{
		pauseStart = SbTime::getTimeOfDay() ;
	}
	else
	{
		SbTime pauseLength =
			SbTime::getTimeOfDay() -
			pauseStart ;
		pauseTime += pauseLength ;
	}
	printf( "pauseIt set to %d\n", pauseIt ) ;
}



//------------------------------------------------------------------------------
// Quit the game after cleaning up.
//------------------------------------------------------------------------------
static void
quitGame()
{
	printf( "quitting the game\n" ) ;
	delete game ;
	delete gio ;
	delete cockPit ;
	Sfx::quit() ;
	exit( 0 ) ;
}



//------------------------------------------------------------------------------
// Idle sensor callback, update the display.
//------------------------------------------------------------------------------
void
frame(
	void		*,
	SoSensor	*sensor
	)
{
	static SbTime frameTime( 0.0 ) ;
	static int nFrames = 0 ;
	static const float minDeltaT = 0.075f ; // 0.05f ;

	( (SoIdleSensor *)sensor )->schedule() ;

	if( pauseIt )
	{
		if( cockPit->isVisible() )
		{
			cockPit->render() ;
		}
		myRealTime->setValue( SbTime::getTimeOfDay() ) ;
		return ;
	}

	SbTime deltaT = lastTime ;
	lastTime = SbTime::getTimeOfDay() - pauseTime ;

	deltaT = lastTime - deltaT ;
	if( debugging && deltaT.getValue() > 0.5 )
	{
		deltaT.setValue( minDeltaT - 0.001 ) ;
	}

	if( cockPit->isVisible() )
	{
		nFrames++ ;
		frameTime += deltaT ;

		float dT = (float)deltaT.getValue() ;

		if( dT > minDeltaT )
		{
			int steps = 1 + (int)( dT / minDeltaT ) ;
			float stepT = dT / (float)steps ;
			while( steps-- )
			{
				if( game->doOneFrame( stepT ) )
				{
					quitGame() ;
				}
			}
		}
		else
		{
			if( game->doOneFrame( dT ) )
			{
				quitGame() ;
			}
		}

		game->step( dT ) ;

		double ft = frameTime.getValue() ;
		if( showTiming && ft >= 1.0 )
		{
			printf( "%.2lf frames per second (%d, %f)\n",
				(double)nFrames / ft, nFrames, ft ) ;
			nFrames = 0 ;
			frameTime.setValue( 0.0 ) ;
		}
	}

	if( stopNextFrame )
	{
		togglePause() ;
	}

	myRealTime->setValue( SbTime::getTimeOfDay() ) ;
}



//------------------------------------------------------------------------------
// Window closed callback -> quit the game.
//------------------------------------------------------------------------------
static void
windowCloseCB(
	void		*,
	SoXtComponent	*
	)
{
	quitGame() ;
}



//------------------------------------------------------------------------------
// CockPit area event callback.
//------------------------------------------------------------------------------
SbBool
introEventCB(
	void		*,
	XAnyEvent	*event
	)
{
	KeySym ks ;
	SbBool handled = FALSE ;

	if( event->type == KeyRelease )
	{
		ks = XLookupKeysym( (XKeyEvent *)event, 0 ) ;
		if( ks == XK_Escape )
		{
			quitGame() ;
		}
	}
	return( handled ) ;
}



//------------------------------------------------------------------------------
// CockPit area event callback.
//------------------------------------------------------------------------------
SbBool
eventCB(
	void		*,
	XAnyEvent	*event
	)
{
	KeySym ks ;
	SbBool handled = FALSE ;

	if( event->type == MotionNotify && !mouseLock )
	{
		XMotionEvent *motion = (XMotionEvent *)event ;
		game->motionEvent( motion ) ;
		handled = TRUE ;
	}
	else if( event->type == KeyRelease )
	{
		ks = XLookupKeysym( (XKeyEvent *)event, 0 ) ;
		if( ks == XK_Escape )
		{
			quitGame() ;
		}
		else
		{
			handled = game->keyReleaseEvent( ks ) ;
		}
	}
	else if( event->type == ButtonRelease || event->type == ButtonPress )
	{
		game->buttonEvent( (XButtonEvent *)event ) ;
	}
	else if( event->type == KeyPress )
	{
		ks = XLookupKeysym( (XKeyEvent *)event, 0 ) ;
		switch( ks )
		{
#ifdef VERBOSE
			case XK_m :
				mouseLock = !mouseLock ;
				printf( "mouseLock set to %d\n", mouseLock ) ;
				handled = TRUE ;
				break ;

			case XK_f :
				if( ((XKeyEvent *)event)->state & Mod1Mask )
				{
					if( pauseIt )
					{
						togglePause() ;
					}
					stopNextFrame = TRUE ;
					handled = TRUE ;
				}
				break ;

			case XK_n :
				if( ((XKeyEvent *)event)->state & Mod1Mask )
				{
					if( pauseIt )
					{
						togglePause() ;
					}
					stopNextFrame = FALSE ;
					handled = TRUE ;
				}
				break ;
#endif // VERBOSE

			case XK_p :
			case XK_Pause :
				if( scenario->getGameMode() ==
					GameControl::SOLO )
				{
					togglePause() ;
					handled = TRUE ;
				}
				break ;

			case XK_F2 :
				showTiming = !showTiming ;
				break ;

			default :
				handled = game->keyPressEvent( ks ) ;
				break ;
		}
	}
	return( handled ) ;
}



//------------------------------------------------------------------------------
// Initialize all the new Inventor classes.
//------------------------------------------------------------------------------
static void
initAllClasses()
{
	GroundMesh::initClass() ;
	GroundStrip::initClass() ;
	FollowTerrainAction::initClass() ;
	Obstacle::initClass() ;
	StaticObstacle::initClass() ;
	Goal::initClass() ;
	CollisionAction::initClass() ;
	Vehicle::initClass() ;
	Tank::initClass() ;
	Scout::initClass() ;
	Flyer::initClass() ;
	LosAction::initClass() ;
	RotateRotation::initClass() ;
	Missile::initClass() ;
	CruiseMissile::initClass() ;
	GuidedMissile::initClass() ;
	Bomb::initClass() ;
	Cannon::initClass() ;
	RedrawGroup::initClass() ;
	ZbufferMode::initClass() ;
	AnimationMaster::initClass() ;
	TransformSwitch::initClass() ;
	Debris::initClass() ;
	NameList::initClass() ;
	Mauler::initClass() ;
	Mortar::initClass() ;
	NoPassGroup::initClass() ;
	SceneryFile::initClass() ;
	Automaton::initClass() ;
	Battery::initClass() ;
	DestructibleObstacle::initClass() ;
	SbString audioDir( defaultDir ) ;
	audioDir += "/audio" ;
	Sfx::init( audioDir.getString(), 1000 ) ;

	RedrawGroup::forceRedraw() ;
}



//------------------------------------------------------------------------------
// Usage routine.
//------------------------------------------------------------------------------
void
usage()
{
	fprintf( stderr, "Usage: pb [ [-l] | [-a] scenario ]\n" ) ;
	fprintf( stderr, "-a:  disable audio\n" ) ;
	fprintf( stderr, "-l:  list scenarios\n" ) ;
	exit( 1 ) ;
}



//------------------------------------------------------------------------------
// Main routine.
//------------------------------------------------------------------------------
void
main(
	int	argc,
	char	**argv
	)
{
	int c ;
	extern char *optarg ;
	extern int optind ;
	char selfName[L_cuserid] ;
	SbBool listGames = FALSE ;

	char *playName = getenv( "PB_NAME" ) ;
	if( playName == NULL )
	{
		cuserid( selfName ) ;
		playName = selfName ;
	}

#ifdef VERBOSE
	while ( ( c = getopt( argc, argv, "afdthl" ) ) != EOF )
#else
	while ( ( c = getopt( argc, argv, "al" ) ) != EOF )
#endif /* VERBOSE */
	{
		switch( c )
		{
			case 'd' :
				debugging = TRUE ;
				showTiming = FALSE ;
				break ;

			case 't' :
				showTiming = TRUE ;
				break ;

			case 'f' :
				stopNextFrame = TRUE ;
				break ;

			case 'a' :
				Sfx::disableSfx() ;
				printf( "audio disabled\n" ) ;
				break ;

			case 'l' :
				listGames = TRUE ;
				break ;

			case 'h' :
			default :
				usage() ;
				break ;
		}
	}

	GameScenario::setDataDir( defaultDir ) ;
	if( listGames )
	{
		GameScenario::listGames() ;
		exit( 0 ) ;
	}

	if( argc == optind )
	{
		scenario = new GameScenario( "game0" ) ;
	}
	else if( argc == optind + 1 )
	{
		scenario = new GameScenario( argv[optind] ) ;
	}
	else
	{
		usage() ;
	}

	if( prctl( PR_RESIDENT ) == -1 && errno != EPERM )
	{
		perror( "prctl( PR_RESIDENT )" ) ;
	}

	Widget topLevel = SoXt::init( "Pointblank", "Pointblank" ) ;

	int n = 0 ;
	Arg arg[6] ;

	XtSetArg( arg[n], XmNmaxAspectX, 5 ) ; n++ ;
	XtSetArg( arg[n], XmNmaxAspectY, 4 ) ; n++ ;
	XtSetArg( arg[n], XmNminAspectX, 5 ) ; n++ ;
	XtSetArg( arg[n], XmNminAspectY, 4 ) ; n++ ;
	XtSetArg( arg[n], XmNminWidth, 250 ) ; n++ ;
	XtSetArg( arg[n], XmNminHeight, 200 ) ; n++ ;
	XtSetValues( topLevel, arg, n ) ;

	SbName rt( "realTime" ) ;
	myRealTime = (SoSFTime *)SoDB::getGlobalField( rt ) ;

	//
	// Disable the real time interval interrupt since we update the
	// realTime global field once every frame.
	//
	SoDB::setRealTimeInterval( SbTime( 0.0 ) ) ;

	//
	// Create the CockPit display area.
	//
	cockPit = new CockPit( topLevel, "Cockpit" ) ;
	cockPit->setSize( SbVec2s( 640, 512 ) ) ;
	cockPit->setEventCallback( introEventCB, NULL ) ;
	cockPit->setTransparencyType( SoGLRenderAction::BLEND ) ;
	cockPit->setWindowCloseCallback( windowCloseCB ) ;
	cockPit->show() ;
	cockPit->setTitle( "Pointblank" ) ;

	SoXt::show( topLevel ) ;
	XFlush( XtDisplay( topLevel ) ) ;
	setCursor( cockPit->getWidget() ) ;

	cockPit->drainEvents() ;

	SbString introFile( defaultDir ) ;
	introFile += "/intro.iv" ;
	SoSeparator *intro = Utilities::readAll( introFile.getString(), FALSE );
	if( intro != NULL )
	{
		intro->ref() ;
		cockPit->addScene( intro, 0.0f, 1.0f, 0.0f, 1.0f, TRUE ) ;
	}
	else
	{
		printf( "Error reading or opening %s\n",
			introFile.getString() ) ;
	}
	cockPit->forceRedraw() ;
	cockPit->drainEvents() ;

	SbString dataFile( defaultDir ) ;
	if( cockPit->hasFastTextures() )
	{
		dataFile += "/tex" ;
	}
	else
	{
		dataFile += "/notex" ;
	}
	Utilities::setSearchDir( dataFile.getString() ) ;

	initAllClasses() ;
	cockPit->drainEvents() ;

	if( scenario->getGameMode() == GameControl::SOLO )
	{
		gio = new GameIO( NULL, NULL ) ;
	}
	else
	{
		gio = new GameIO( serviceName, groupName, NULL, defaultPort,
					0, 8 ) ;
	}
	GameIO::setMyName( playName ) ;
	gio->setIdentifier( scenario->getId(), PB_VERSION ) ;

	game = new GameControl( cockPit, gio, scenario ) ;

	SoIdleSensor *idle = new SoIdleSensor( frame, NULL ) ;
	idle->schedule() ;

	lastTime = SbTime::getTimeOfDay() ;
	pauseTime.zero() ;

	cockPit->setEventCallback( eventCB, NULL ) ;

	cockPit->forceRedraw() ;

	SoXt::mainLoop() ;
}
