//------------------------------------------------------------------------------
// Pbflat - remove all transforms in a Pointblank world file, adjusting
//    all Obstacles and GroundShapes accordingly.
//------------------------------------------------------------------------------

#include <stdio.h>
#include <Inventor/SbDict.h>
#include <Inventor/SoLists.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoInfo.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoTransformation.h>
#include "GroundShape.h"
#include "Obstacle.h"
#include "SceneryFile.h"
#include "Utilities.h"


static SbDict nodeDict ;



//------------------------------------------------------------------------------
// Check if a SceneryFile node in a path is transformed.
//------------------------------------------------------------------------------
void
checkSceneryFileTransform(
	SoPath		*path,
	SceneryFile	*sf
	)
{
	SoSearchAction sa ;
	sa.setNode( sf ) ;
	sa.setFind( SoSearchAction::NODE ) ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.apply( path ) ;

	SoPathList &paths = sa.getPaths() ;
	for( int i = 0 ; i < paths.getLength() ; i++ )
	{
		SoGetMatrixAction gm( SbViewportRegion( 1, 1 ) ) ;
		gm.apply( paths[i] ) ;
		SbMatrix &mat = gm.getMatrix() ;
		SbVec3f		translation ;
		SbRotation	rotation ;
		SbVec3f		scaleFactor ;
		SbRotation	scaleOrient ;
		mat.getTransform( translation, rotation, scaleFactor,
					scaleOrient ) ;
		SbVec3f axis ;
		float angle ;
		rotation.getValue( axis, angle ) ;
		if( translation != SbVec3f( 0.0f, 0.0f, 0.0f ) ||
			angle != 0.0f )
		{
			fprintf( stderr, "ERROR: SceneryFiles can not be "
				"transformed (%s)\n",
				sf->getFilename().getString() ) ;
			fprintf( stderr, "translation: %f %f %f\n",
				translation[0], translation[1],
				translation[2] ) ;
			fprintf( stderr, "rotation: %f %f %f  %f\n", axis[0],
				axis[1], axis[2], angle ) ;
			exit( 2 ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Check a path to see if it includes a SceneryFile.
//------------------------------------------------------------------------------
SceneryFile *
checkForSceneryFile(
	SoPath *path
	)
{
	for( int i = 0 ; i < path->getLength() ; i++ )
	{
		if( path->getNode( i )->isOfType(
					SceneryFile::getClassTypeId() ) )
		{
			return( (SceneryFile *)path->getNode( i ) ) ;
		}
	}

	return( NULL ) ;
}



//------------------------------------------------------------------------------
// Build a dictionary of all the nodes in the scene to determine multiple
// instances.
//------------------------------------------------------------------------------
void
buildDictionary(
	SoNode *root
	)
{
	void *value ;
	//
	// If node is already in the dictionary, increment the value.
	//
	if( nodeDict.find( root->getNodeId(), value ) )
	{
		//
		// Overwrite the entry with a new value.
		//
		value = (void *)( (int)value + 1 ) ;
	}
	//
	// Otherwise, set the count to 1.
	//
	else
	{
		value = (void *)1 ;
	}
	nodeDict.enter( root->getNodeId(), value ) ;

	if( root->isOfType( SoGroup::getClassTypeId() ) )
	{
		SoGroup *g = (SoGroup *)root ;
		for( int i = 0 ; i < g->getNumChildren() ; i++ )
		{
			buildDictionary( g->getChild( i ) ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Check to see if any node on the path is referenced twice.  If so, make a
// copy of it and replace it with the copy.
//------------------------------------------------------------------------------
SbBool
checkForDuplicate(
	SoPath *path
	)
{
	void		*value ;
	SceneryFile	*scenery = NULL ;

	for( int i = 1 ; i < path->getLength() ; i++ )
	{
		SoNode *n = path->getNode( i ) ;
		if( !nodeDict.find( n->getNodeId(), value ) )
		{
			fprintf( stderr, "FATAL ERROR: node (\"%s\") not "
				"traversed previously\n",
				n->getName().getString() ) ;
			exit( 1 ) ;
		}
		if( (int)value > 1 )
		{
			if( scenery )
			{
				fprintf( stderr, "WARNING:  found multiple "
					"instances of nodes in SceneryFile "
					"\"%s\".\n\tProcess this file "
					"separately for proper behavior in "
					"Pointblank.\n",
					scenery->getFilename().getString() ) ;
				break ;
			}
			SoNode *parent = path->getNode( i - 1 ) ;
			if( parent->isOfType( SoGroup::getClassTypeId() ) )
			{
				SoNode *newOne = n->copy() ;
				SbString s( n->getName().getString() ) ;
				s += "_" ;
				s += (int)value ;
				newOne->setName( s ) ;

				int k = path->getIndexFromTail( i - 1 ) ;
				( (SoGroup *)parent )->replaceChild( n,
								newOne ) ;
				printf( "replaced duplicate %s\n",
					n->getName().getString() ) ;
				return( TRUE ) ;
			}
			else
			{
				fprintf( stderr, "Found duplicate but could "
					"not replace!!!\n" ) ;
			}
		}
		//
		// Check if path includes a SceneryFile node.
		//
		if( scenery == NULL &&
			n->isOfType( SceneryFile::getClassTypeId() ) )
		{
			scenery = (SceneryFile *)n ;
		}
	}

	return( FALSE ) ;
}



//------------------------------------------------------------------------------
// Replace all transform nodes with Info nodes.
//------------------------------------------------------------------------------
void
removeTransforms(
	SoPath *path
	)
{
	int j = 1 ;

	SceneryFile *sf = checkForSceneryFile( path ) ;
	if( sf != NULL )
	{
		fprintf( stderr, "WARNING: SceneryFile %s contains "
			"transformations that\nwill not be removed.  "
			"Run this program on them separately.\n",
			sf->getFilename().getString() ) ;
		return ;
	}

	while( j < path->getLength() )
	{
		SoNode *parent = path->getNodeFromTail( j ) ;

		if( parent && parent->isOfType( SoGroup::getClassTypeId() ) )
		{
			SoGroup *g = (SoGroup *)parent ;
			if( parent->isOfType( SoSwitch::getClassTypeId() ) )
			{
				g->replaceChild( path->getTail(), new SoInfo ) ;
				return ;
			}
			else
			{
				g->removeChild( path->getNodeFromTail( j-1 ) ) ;
				if( g->getNumChildren() > 0 )
				{
					return ;
				}
			}
		}
		else if( j == 1 )
		{
			fprintf( stderr, "ERROR replacing SoTransformation node"
				" (\"%s\")\n",
				path->getTail()->getName().getString() ) ;
			exit( 3 ) ;
		}
	}
}



//------------------------------------------------------------------------------
// Find the transformation along the given path and apply it to the GroundShape
// at the tail of the path.
//------------------------------------------------------------------------------
void
transformGroundShape(
	SoPath *path
	)
{
	SoGetMatrixAction gm( SbViewportRegion( 1, 1 ) ) ;

	gm.apply( path ) ;

	SbMatrix &mat = gm.getMatrix() ;

	GroundShape *gs = (GroundShape *)path->getTail() ;

	SbMatrix gsMat ;
	SbVec3f ones( 1.0f, 1.0f, 1.0f ) ;
	SbVec3f zAxis( 0.0f, 0.0f, 1.0f ) ;
	SbRotation gsRot( zAxis, gs->rotation.getValue() ) ;
	gsMat.setTransform( gs->origin.getValue(), gsRot, ones ) ;
	mat.multLeft( gsMat ) ;

	SbVec3f		translation ;
	SbRotation	rotation ;
	SbVec3f		scaleFactor ;
	SbRotation	scaleOrient ;
	mat.getTransform( translation, rotation, scaleFactor, scaleOrient ) ;

	SbVec3f &diff = scaleFactor - ones ;
	if( !scaleFactor.equals( ones, 1e-6 ) )
	{
		fprintf( stderr, "ERROR: scale factor for node (\"%s\") is"
			"(%g,%g,%g)\n", gs->getName().getString(),
			scaleFactor[0], scaleFactor[1], scaleFactor[2] ) ;
		fprintf( stderr, "\tOnly scale factors of (1,1,1) are "
			"allowed\n" ) ;
		exit( 2 ) ;
	}

	SbVec3f axis ;
	float angle ;
	rotation.getValue( axis, angle ) ;

	if( !axis.equals( zAxis, 1e-06 ) )
	{
		fprintf( stderr, "ERROR: rotation for node (\"%s\") is"
			"not about the Z axis\n", gs->getName().getString() ) ;
		exit( 2 ) ;
	}

	SceneryFile *sf = checkForSceneryFile( path ) ;
	if( sf != NULL )
	{
		checkSceneryFileTransform( path, sf ) ;
		return ;
	}

	gs->origin.setValue( translation ) ;
	gs->rotation.setValue( angle ) ;
}



//------------------------------------------------------------------------------
// Find the transformation along the given path and apply it to the Obstacle
// at the tail of the path.
//------------------------------------------------------------------------------
void
transformObstacle(
	SoPath *path
	)
{
	SoGetMatrixAction gm( SbViewportRegion( 1, 1 ) ) ;

	gm.apply( path ) ;

	SbMatrix &mat = gm.getMatrix() ;

	Obstacle *obst = (Obstacle *)path->getTail() ;

	SbMatrix obMat ;
	SbVec3f ones( 1.0f, 1.0f, 1.0f ) ;
	obMat.setTransform( obst->position.getValue(),
				obst->rotation.getValue(), ones ) ;
	mat.multLeft( obMat ) ;

	SbVec3f		translation ;
	SbRotation	rotation ;
	SbVec3f		scaleFactor ;
	SbRotation	scaleOrient ;
	mat.getTransform( translation, rotation, scaleFactor, scaleOrient ) ;

	if( !scaleFactor.equals( ones, 1e-6 ) )
	{
		fprintf( stderr, "ERROR: scale factor for node (\"%s\") is"
			"(%g,%g,%g)\n", obst->getName().getString(),
			scaleFactor[0], scaleFactor[1], scaleFactor[2] ) ;
		fprintf( stderr, "\tOnly scale factors of (1,1,1) are "
			"allowed\n" ) ;
		exit( 2 ) ;
	}

	SceneryFile *sf = checkForSceneryFile( path ) ;
	if( sf != NULL )
	{
		checkSceneryFileTransform( path, sf ) ;
		return ;
	}

	obst->position.setValue( translation ) ;
	obst->rotation.setValue( rotation ) ;
}



//------------------------------------------------------------------------------
// Remove groups that have an Obstacle as their only child.
//------------------------------------------------------------------------------
void
removeIsolations(
	SoPath *path
	)
{
	int i = 1 ;
	Obstacle *obst = (Obstacle *)path->getTail() ;

	if( checkForSceneryFile( path ) )
	{
		return ;
	}

	SoGroup *target = NULL ;
	SoGroup *targetParent = NULL ;

	while( path->getLength() > i + 2 )
	{
		SoNode	*node ;
		node = path->getNodeFromTail( i ) ;
		if( !node->isOfType( SoGroup::getClassTypeId() ) )
		{
			break ;
		}
		SoGroup *parent = (SoGroup *)node ;

		if( parent->getNumChildren() > 1 )
		{
			break ;
		}

		node = path->getNodeFromTail( i + 1 ) ;
		if( !node->isOfType( SoGroup::getClassTypeId() ) )
		{
			break ;
		}
		targetParent = (SoGroup *)node ;
		target = parent ;
		i++ ;
	}

	if( target )
	{
		targetParent->replaceChild( target, obst ) ;
	}
}



//------------------------------------------------------------------------------
// Print the usage and exit
//------------------------------------------------------------------------------
void
usage()
{
	fprintf( stderr, "Usage: pbflat infile outfile\n" ) ;
	exit( 1 ) ;
}



//------------------------------------------------------------------------------
// Main routine.
//------------------------------------------------------------------------------
void
main(
	int	argc,
	char	**argv
	)
{
	int c ;
	int i ;
	extern char *optarg ;
	extern int optind ;
	while ( ( c = getopt( argc, argv, "" ) ) != EOF )
	{
		switch( c )
		{
			default :
				usage() ;
				break ;
		}
	}

	if( argc - 2 != optind )
	{
		usage() ;
	}

	SoDB::init() ;
	GroundShape::initClass() ;
	Obstacle::initClass() ;
	SceneryFile::initClass() ;

	SoInput	in ;

	if( !in.openFile( argv[optind], TRUE ) )
	{
		perror( argv[optind] ) ;
		exit( 1 ) ;
	}
	optind++ ;

	SoSeparator *root = new SoSeparator ;

	SoNode *n ;
	SbBool readOk = TRUE ;
	while( readOk )
	{
		if( SoDB::read( &in, n ) )
		{
			if( n != NULL )
			{
				root->addChild( n ) ;
			}
			else
			{
				readOk = FALSE ;
			}
		}
		else
		{
			exit( 1 ) ;
		}
	}

	in.closeFile() ;
	if( root == NULL || root->getNumChildren() < 1 )
	{
		fprintf( stderr, "%s: no data read\n", argv[optind] ) ;
		exit( 1 ) ;
	}
	root->ref() ;

	SoSearchAction sa ;

	//
	// Phase I: find all Obstacle nodes and make sure they are not
	// traversed twice.
	//
	printf( "Phase I:  check for duplicate Obstacles and GroundShapes\n" ) ;
	SbBool foundCopies = TRUE ;
	sa.setType( Obstacle::getClassTypeId() ) ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setSearchingAll( TRUE ) ;
	while( foundCopies )
	{
		foundCopies = FALSE ;
		nodeDict.clear() ;
		buildDictionary( root ) ;
		sa.apply( root ) ;
		SoPathList &paths = sa.getPaths() ;

		for( i = 0 ; !foundCopies && i < paths.getLength() ; i++ )
		{
			if( checkForDuplicate( paths[i] ) )
			{
				foundCopies = TRUE ;
			}
		}
	}

	//
	// Phase Ib: find all GroundShape nodes and make sure they are not
	// traversed twice.
	//
	foundCopies = TRUE ;
	sa.setType( GroundShape::getClassTypeId() ) ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setSearchingAll( TRUE ) ;
	while( foundCopies )
	{
		foundCopies = FALSE ;
		nodeDict.clear() ;
		buildDictionary( root ) ;
		sa.apply( root ) ;
		SoPathList &paths = sa.getPaths() ;

		for( i = 0 ; !foundCopies && i < paths.getLength() ; i++ )
		{
			if( checkForDuplicate( paths[i] ) )
			{
				foundCopies = TRUE ;
			}
		}
	}

	//
	// Phase II: find all Obstacle nodes and update their positions
	// and rotations.
	//
	printf( "Phase II:  transforming nodes\n" ) ;
	sa.setType( Obstacle::getClassTypeId() ) ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setSearchingAll( TRUE ) ;
	sa.apply( root ) ;

	SoPathList &pathList = sa.getPaths() ;

	printf( "found %d Obstacle nodes\n", pathList.getLength() ) ;
	for( i = 0 ; i < pathList.getLength() ; i++ )
	{
		transformObstacle( pathList[i] ) ;
	}

	//
	// Phase IIb: find all GroundShape nodes and update their positions
	// and rotations.
	//
	sa.setType( GroundShape::getClassTypeId() ) ;
	sa.setInterest( SoSearchAction::ALL ) ;
	sa.setFind( SoSearchAction::TYPE ) ;
	sa.setSearchingAll( TRUE ) ;
	sa.apply( root ) ;

	SoPathList &gsPathList = sa.getPaths() ;

	printf( "found %d GroundShape nodes\n", gsPathList.getLength() ) ;
	for( i = 0 ; i < pathList.getLength() ; i++ )
	{
		transformGroundShape( gsPathList[i] ) ;
	}

	//
	// Phase III: remove all transform nodes.
	//
	printf( "Phase II:  removing transformations\n" ) ;
	sa.setType( SoTransformation::getClassTypeId() ) ;
	sa.apply( root ) ;
	printf( "found %d SoTransformation nodes\n", pathList.getLength() ) ;
	SoPathList &xfPathList = sa.getPaths() ;
	for( i = 0 ; i < xfPathList.getLength() ; i++ )
	{
		if( !xfPathList[i]->getTail()->isOfType(
			Obstacle::getClassTypeId() ) )
		{
			removeTransforms( xfPathList[i] ) ;
		}
	}

	//
	// Phase IV: search for all Separators with only one child and which
	// is an Obstacle
	//
	printf( "Phase IV:  removing isolated separators\n" ) ;
	sa.setType( Obstacle::getClassTypeId() ) ;
	sa.apply( root ) ;
	SoPathList &isPathList = sa.getPaths() ;
	for( i = 0 ; i < isPathList.getLength() ; i++ )
	{
		removeIsolations( isPathList[i] ) ;
	}

	SoOutput out ;
	if( !out.openFile( argv[optind] ) )
	{
		perror( argv[optind] ) ;
		exit( 1 ) ;
	}
	SoWriteAction wa( &out ) ;
	for( i = 0 ; i < root->getNumChildren() ; i++ )
	{
		wa.apply( root->getChild( i ) ) ;
	}
	out.closeFile() ;

	exit( 0 ) ;
}
