#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/SbTime.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/sensors/SoAlarmSensor.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCamera.h>

#include "GroundMesh.h"
#include "GroundStrip.h"
#include "Obstacle.h"
#include "StaticObstacle.h"
#include "Goal.h"
#include "Vehicle.h"
#include "Tank.h"
#include "Scout.h"
#include "Flyer.h"
#include "RotateRotation.h"
#include "Missile.h"
#include "CruiseMissile.h"
#include "GuidedMissile.h"
#include "Bomb.h"
#include "Cannon.h"
#include "RedrawGroup.h"
#include "ZbufferMode.h"
#include "AnimationMaster.h"
#include "TransformSwitch.h"
#include "Debris.h"
#include "NameList.h"
#include "Mauler.h"
#include "Mortar.h"
#include "NoPassGroup.h"
#include "SceneryFile.h"
#include "Automaton.h"
#include "Battery.h"
#include "DestructibleObstacle.h"
#include "Utilities.h"

static SbTime lastTime ;
static SbTime lastOverlayTime ;
static SbString fileName ;
static SbString olFileName ;
static SoSeparator *root ;
static SoSeparator *olRoot ;
static SoXtExaminerViewer *myViewer ;

static void
readOverlayFile()
{
printf( "readOverlayFile %s\n", olFileName.getString() ) ;
   SoInput input ;
   lastOverlayTime.setToTimeOfDay() ;

   if( olRoot )
   {
 	olRoot->unref() ;
   }

   if( !input.openFile( olFileName.getString() ) )
   {
	perror( olFileName.getString() ) ;
	return ;
   }

   olRoot = SoDB::readAll( &input ) ;
   if( olRoot == NULL )
   {
	fprintf( stderr, "problem reading %s\n", olFileName.getString() ) ;
	return ;
   }
   else
   {
	olRoot->ref() ;
	myViewer->setOverlaySceneGraph(olRoot);
   }
   input.closeFile() ;
}



static void
readFile()
{
   SoInput input ;
   lastTime.setToTimeOfDay() ;
   SbBool setView = FALSE ;
   SoCamera *camera ;
   SoType camType ;

   if( root )
   {
 	root->unref() ;
	setView = TRUE ;
	camera = myViewer->getCamera() ;
	camType = myViewer->getCameraType() ;
	camera->ref() ;
   }

   if( !input.openFile( fileName.getString() ) )
   {
	perror( fileName.getString() ) ;
	return ;
   }

   Obstacle::purgeGeometry() ;
   root = SoDB::readAll( &input ) ;
   if( root == NULL )
   {
	fprintf( stderr, "problem reading %s\n", fileName.getString() ) ;
	if( setView )
	{
		camera->unref() ;
	}
	return ;
   }
   else
   {
	root->ref() ;
	myViewer->setSceneGraph(root);

	if( setView )
	{
		myViewer->setCameraType( camType ) ;
		SoCamera *newCam = myViewer->getCamera() ;
		newCam->viewportMapping = camera->viewportMapping.getValue();
		newCam->position.setValue( camera->position.getValue() ) ;
		newCam->orientation.setValue( camera->orientation.getValue() ) ;
		newCam->aspectRatio.setValue( camera->aspectRatio.getValue() ) ;
		newCam->nearDistance.setValue(
					camera->nearDistance.getValue() ) ;
		newCam->farDistance.setValue(
					camera->farDistance.getValue() ) ;
		newCam->focalDistance.setValue(
					camera->focalDistance.getValue() ) ;
		camera->unref() ;
	}

   }
   input.closeFile() ;
}



static void
checkFile(
    void *,
    SoSensor *sensor
    )
{
    SoAlarmSensor *as = (SoAlarmSensor *)sensor ;

    as->setTimeFromNow( SbTime( 1.0 ) ) ;
    as->schedule() ;

    struct stat buf ;
    if( stat( fileName.getString(), &buf ) < 0 )
    {
	perror( fileName.getString() ) ;
	return ;
    }
    else
    {
	if( (double)buf.st_mtime > lastTime.getValue() )
	{
		readFile() ;
	}
    }

    if( olFileName.getLength() > 0 )
    {
	if( stat( olFileName.getString(), &buf ) < 0 )
	{
	    perror( olFileName.getString() ) ;
	    return ;
	}
	else
	{
	    if( (double)buf.st_mtime > lastOverlayTime.getValue() )
	    {
		    readOverlayFile() ;
	    }
	}
    }
}



void
main(int argc, char **argv)
{
   Widget myWindow = SoXt::init(argv[0]);
   GroundMesh::initClass() ;
   GroundStrip::initClass() ;
   Obstacle::initClass() ;
   StaticObstacle::initClass() ;
   Goal::initClass() ;
   Vehicle::initClass() ;
   Tank::initClass() ;
   Scout::initClass() ;
   Flyer::initClass() ;
   RotateRotation::initClass() ;
   Missile::initClass() ;
   CruiseMissile::initClass() ;
   GuidedMissile::initClass() ;
   Bomb::initClass() ;
   Cannon::initClass() ;
   RedrawGroup::initClass() ;
   ZbufferMode::initClass() ;
   AnimationMaster::initClass() ;
   TransformSwitch::initClass() ;
   Debris::initClass() ;
   NameList::initClass() ;
   Mauler::initClass() ;
   Mortar::initClass() ;
   NoPassGroup::initClass() ;
   SceneryFile::initClass() ;
   Automaton::initClass() ;
   Battery::initClass() ;
   DestructibleObstacle::initClass() ;
   RedrawGroup::forceRedraw() ;

   if (myWindow == NULL) exit(1);

   if( argc != 2 && argc != 3 )
   {
	fprintf( stderr, "Usage: %s file [overlay_file] \n", argv[0] ) ;
	exit( 1 ) ;
   }

   fileName = argv[1] ;

   root = NULL ;

   // Set up viewer:
   myViewer = new SoXtExaminerViewer(myWindow);
   myViewer->setTitle( fileName.getString() );
   myViewer->show();
   myViewer->setBackgroundColor( SbColor( 0.2f, 0.2f, 0.4f ) ) ;
   myViewer->setTransparencyType( SoGLRenderAction::BLEND ) ;
   myViewer->setDrawStyle( SoXtViewer::INTERACTIVE,
		SoXtViewer::VIEW_SAME_AS_STILL ) ;

   readFile() ;

   if( argc == 3 )
   {
	olFileName = argv[2] ;
	readOverlayFile() ;
	SbColor omapColors[3] ;
	omapColors[0].setValue( 1.0f, 1.0f, 0.0f ) ;
	omapColors[1].setValue( 1.0f, 0.0f, 0.0f ) ;
	omapColors[2].setValue( 0.0f, 0.0f, 0.0f ) ;
	myViewer->setOverlayColorMap( 1, 3, omapColors ) ;
   }
   else
   {
	olFileName.makeEmpty() ;
   }

   SoAlarmSensor *sensor = new SoAlarmSensor( checkFile, NULL ) ;
   sensor->setTimeFromNow( SbTime( 1.0 ) ) ;
   sensor->schedule() ;

   SoXt::show(myWindow);
   SoXt::mainLoop();
}

